#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/argocd_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/certificate_utils.sh

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Get additional CA certificate"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  --outpath                           Stores additional CA certificates in outpath directory"
  echo "  -h|--help                           Display help" 
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;
    
    --outpath)
      if [ -n "$2" ]; then
        CERT_OUTPATH="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      exit 1
    ;;
    esac
  done
}

function fetch_additional_ca_certs() {
  local namespace="uipath"
  local secret_name="additional-ca-certificates"
  local cert_name="additionalca.crt"
  if secret=$(kubectl get secret ${secret_name} --namespace ${namespace} -o json 2> /dev/null);
  then
    cert=$(echo "${secret}" | jq -r ".data.\"$cert_name\"" | base64 -d)

    [[ -z "${CERT_OUTPATH}" ]] && echo "${cert}"
    [[ -n "${CERT_OUTPATH}" ]] && mkdir -p "${CERT_OUTPATH}" && echo "${cert}" > "${CERT_OUTPATH}" && info "Certificates are stored in ${CERT_OUTPATH}"
  else
    error "Could not fetch Additional CA Certificates"
  fi
}

function main() {
  parse_long_args "$@"
  
  fetch_additional_ca_certs
}

main "$@"
