#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/Fabric_Installer/bin:/usr/local/bin

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/k8sutils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/certificate_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cluster_utils.sh

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Update additional CA certificates. It will either add additional certificate or replace all"
  echo "the previously configured CA certificates with the new CA certificates"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  --ca-cert-file                      Additional Certificate Authority bundle to"
  echo "                                        configure in cluster trust store"
  echo "  -r|--replace                        Replace all the previously configured CA certificates"
  echo "  -h|--help                           Display help" 
  echo
  echo "***************************************************************************************"
  echo
}

REPLACE=0

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    --ca-cert-file)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        CERT_CA_FILE_PATH=$2
        CERT_CA_FILE_PATH="${CERT_CA_FILE_PATH/#\~/$HOME}"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
    ;;

    -r|--replace)
      shift
      REPLACE=1
    ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      exit 1
    ;;
    esac
  done

  if [[ ! -f ${CERT_CA_FILE_PATH} ]]; then 
    display_usage 
    error "Certificate Authority bundle \"${CERT_CA_FILE_PATH}\" not found"
  fi

  echo
}

function update_additional_ca_certificates() {
  local namespace="uipath"
  local cert_file="additionalca.crt"
  local secret_name="additional-ca-certificates"
  
  declare -a error
  response=$(validate_pem_format "${CERT_CA_FILE_PATH}")

  [[ -n ${response} ]] && error+=("${response}") && failure_output "Validating Certificate..... failed" "${error[@]}"

  info "Validating additional ca certificates... DONE"
  info "Configuring additional ca certificates... IN PROGRESS"

  tmp_dir=$(mktemp -d certificates.XXXXXXX)
  tmp_file="additionalca.crt"

  if [[ ${REPLACE} == 0 ]]; 
  then
    secret=$(kubectl get secret ${secret_name} \
            --namespace ${namespace} \
            -o json \
            --ignore-not-found)

    if [[ -n ${secret} ]]; then
      echo "${secret}" | jq -r ".data.\"$cert_file\"" | base64 -d | cat > "${tmp_dir}"/"${tmp_file}"
    fi
  fi

  cat "${CERT_CA_FILE_PATH}" >> "${tmp_dir}"/"${tmp_file}"

  kubectl delete secret "${secret_name}" \
    --namespace ${namespace} >/dev/null 2>&1 || true

  kubectl create secret generic "${secret_name}" \
    --from-file="${cert_file}"="${tmp_dir}"/"${tmp_file}" \
    --namespace "${namespace}" \
    --dry-run=client -o yaml | kubectl apply -f - >/dev/null 2>&1

  rm -rf "${tmp_dir}" 

  profile=$(get_cluster_config_default "profile" "ha")

  if [[ "$profile" == "ha"  ]]; 
  then
    kubectl rollout restart deployment --namespace "${namespace}" >/dev/null
    wait_for_all_pods_in_namespace_ready "${namespace}" >/dev/null \
      || error "Certificate update failed - timeout waiting for all pods to restart and ready in namespace $namespace"
  else
    scale_down_all_deployment "$namespace" \
      || error "Certificate update failed - Scalling down all deployment in namespace $namespace failed"
    scale_up_all_deployment "$namespace" "1" \
      || error "Certificate update failed - Scalling up all deployment in namespace $namespace failed" 
  fi

  info "Configuring additional ca certificates... DONE"
}

function check_confirmation() {
  local input
  echo -e "\n"
  warn "Please be aware this operation will update additional ca certificate for your cluster"
  warn "If you wish to continue, type 'yes' and hit Enter for confirmation."
  read -r input </dev/tty
  if [[ $input == 'yes' ]] || [[ $input == "'yes'" ]]
  then
    info "Proceeding to update CA certificate"
  else
    info "Unexpected user input received: ${input}, exiting CA certificate update operation"
    exit 2
  fi
  echo -e "\n"
}

function main() {
  parse_long_args "$@"
  check_confirmation
  update_additional_ca_certificates
}

main "$@"
exit 0