#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export DOT_FOLDER=~/.uipath

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/argocd_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cluster_utils.sh

ARGOCD_CONFIG_FILE="${MODULES_DIR}"/argocd_config.json

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Get or generate cluster_config.json file"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  -i|--input                   [Optional] Path to the old cluster_config.json file."
  echo "                                  If not provided then, you may have to manually add few paramaters"                     
  echo "  -o|--outpath                 [Optional] Path when the file should be stored."
  echo "                                  If not provided then file will be store in the current directory"
  echo "  -h|--help                    Display help" 
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;
    
    -o|--outpath)
      if [ -n "$2" ]; then
        OUTPATH="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;
    
    -i|--input)
      if [ -n "$2" ]; then
        INFILE="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      exit 1
    ;;
    esac
  done

  if [[ -z ${OUTPATH} ]]; then 
    OUTPATH=$(mktemp -p "${DOT_FOLDER}" -t cluster_config.XXXXXXXXXX.json)
    info "Outpath is not provided, config file will be generated in: $OUTPATH"
  fi
  if [[ ! -f ${INFILE} ]]; then 
    warn "Old or existing cluster_config.json file is not provided, you may have to edit the file manually with the missing values"
  else 
    jq '.' "${INFILE}" &> /dev/null || error "Input file ${INFILE} is not a valid json and cannot be parsed"
  fi

  echo
}

function get_docker_registry_password() {
  local config=$1

  username=$(cat < /etc/rancher/rke2/registries.yaml | grep "username" | tail -n1 | awk '{print $2}')
  password=$(cat < /etc/rancher/rke2/registries.yaml | grep "password" | tail -n1 | awk '{print $2}')

  [[ -n $username ]] && config=$(jq ". |= .+ {\"infra.docker_registry.username\":\"$username\"}"  <<< "$config")
  [[ -n $password ]] && config=$(jq ". |= .+ {\"infra.docker_registry.password\":\"$password\"}"  <<< "$config")

  echo "$config"
}

function get_fixed_rke_address() {
  local config=$1
  fixed_rke_address=$(get_cluster_config "fixed_rke_address")
  [[ -n $fixed_rke_address ]] && config=$(jq ". |= .+ {\"fixed_rke_address\":\"$fixed_rke_address\"}"  <<< "$config")
  echo "$config"
}

function get_is_multinode() {
  local config=$1
  multinode=$(get_cluster_config "multinode")
  [[ -n $multinode ]] && config=$(jq ". |= .+ {\"multinode\":\"$multinode\"}"  <<< "$config")
  echo "$config"
}

function get_gpu_support() {
  local config=$1
  gpu_support=$(get_cluster_config "gpu_support")
  [[ -n $gpu_support ]] && config=$(jq ". |= .+ {\"gpu_support\":\"$gpu_support\"}"  <<< "$config")
  echo "$config" 
}

function get_rke_token() {
  local config=$1

  token=$(get_cluster_config "rke_token")
  [[ -z $token ]] && token=$(cat < /etc/rancher/rke2/config.yaml | grep "token" | tail -n1 | awk '{print $2}')
  [[ -n $token ]] && config=$(jq ". |= .+ {\"rke_token\":\"$token\"}"  <<< "$config")
  echo "$config" 
}

function update_proxy_field() {
  local config=$1

  no_proxy=$(jq -r '."proxy.no_proxy"' <<< "$config")
  fqdn=$(jq -r .fqdn <<< "$config")
  value=$(echo "$no_proxy" \
    | sed "s/\\\\,$fqdn\\\\,127.0.0.0\/8\\\\,localhost\\\\,kubernetes\\\\,.local\\\\,.svc\\\\,.cluster\\\\,rook-ceph-rgw-rook-ceph.rook-ceph.svc.cluster.local//g" \
    | sed 's/\\,/,/g')
    
  config=$(jq ". += { \"proxy.no_proxy\": \"$value\" }" <<< "$config")

  echo "$config"
}

function get_cluster_config_json_file() {
  argocd_app="uipath"
  argocd_config_map=$(cat "${ARGOCD_CONFIG_FILE}")
  app_params=$(argocd app get $argocd_app --output json | jq .spec.source.helm.parameters) 
  
  config="{}"
  while IFS=$'\t' read -r key value; 
  do 
    val=$value
    param="$(echo "$argocd_config_map" | jq ".[] | select(.argocd_param == \"$key\").config_param")"
    is_encoded="$(echo "$argocd_config_map" | jq ".[] | select(.argocd_param == \"$key\").is_encoded")"
    if [[ $is_encoded == '"true"' ]]; then 
      val=$(base64 -d <<< "$value" 2> /dev/null)
    fi
    [[ -n $param && -n $val && $val != "PLACEHOLDER" ]] && config=$(jq ". |= .+ {$param:\"$val\"}"  <<< "$config")
  done <<< "$(echo "$app_params" | jq -r '.[] | [.name, .value] | @tsv')"

  config=$(get_docker_registry_password "$config")
  config=$(get_fixed_rke_address "$config")
  config=$(get_gpu_support "$config")
  config=$(get_rke_token "$config")
  config=$(get_is_multinode "$config")
  config=$(update_proxy_field "$config")

  jq '
    to_entries
    | map(.key |= (split(".") | map(if test("^\\[[0-9]+\\]$"; "") then .[1:-1] | tonumber else . end)))
    | reduce .[] as $x (null; setpath($x.key; $x.value))
  ' <<< "$config" > "$OUTPATH"

  #shellcheck disable=SC2005
  [[ -f $INFILE ]] && echo "$(jq -s '.[0] * .[1]' "${INFILE}" "${OUTPATH}")" > "${OUTPATH}"

  info "Cluster configuration is stored in file: $OUTPATH"
}

function main() {
  parse_long_args "$@"
  
  get_cluster_config_json_file
}

main "$@"
