#!/bin/bash
set -e

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Add System (Host) Administrator"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  -u|--username                       Host admin username (required)"
  echo "  -e|--email                          Host admin email (optional, except if required by"
  echo "                                        external IDP"
  echo "  -p|--password                       Host admin password (required for basic auth)"
  echo "  -h|--help                           Display help"
  echo 
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;
    
    -u | --username)
      if [ -n "$2" ]; then
        HOST_ADMIN_USERNAME="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;
    
    -e | --email)
      if [ -n "$2" ]; then
        HOST_ADMIN_EMAIL="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;
    
    -p | --password)
      if [ -n "$2" ]; then
        HOST_ADMIN_PASSWORD="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      exit 1
      ;;
    esac
  done
}

function create_data_migrator_job() {
  if [[ -z "${HOST_ADMIN_USERNAME}" ]]; then
    error "Username required"
    exit 1
  fi

  IdentityApiImageName=$(kubectl -n uipath get deployment identity-service-api -o jsonpath="{.spec.template.spec.containers[0].image}" | sed 's/identity-service-api/identity-service-datamigrator/')

  set +e
  kubectl create secret generic -n uipath identity-add-host-admin-secret \
    --from-literal=HOST_ADMIN_USERNAME="${HOST_ADMIN_USERNAME}" \
    --from-literal=HOST_ADMIN_EMAIL="${HOST_ADMIN_EMAIL}" \
    --from-literal=HOST_ADMIN_PASSWORD="${HOST_ADMIN_PASSWORD}"
  sed "s|--DataMigratorImageTag--|${IdentityApiImageName}|" "${SCRIPT_DIR}"/addhostadmin.yaml | kubectl create -f -

  n=0
  until [ "$n" -ge 15 ]
  do
    PodName=$(kubectl get pod -n uipath | awk '/^identity-service-add-host-admin-[^\s]+/ {print $1}')
    echo Created pod "${PodName}"
    kubectl wait --for=condition=ready pod "${PodName}" --timeout=30s -n uipath && break
    n=$((n+1))
    sleep 2
  done

  echo
  echo
  kubectl logs "${PodName}" --follow -n uipath --pod-running-timeout=20s --ignore-errors=true
  echo
  echo
  kubectl delete -f "${SCRIPT_DIR}/addhostadmin.yaml"
  kubectl delete secret identity-add-host-admin-secret -n uipath
  set -e
}

function main() {
  parse_long_args "$@"
  
  create_data_migrator_job
}

main "$@"
