#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR

export MODULES_DIR
export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/Fabric_Installer/bin:/usr/local/bin

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Manage Identity Configuration"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [command]"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Available Commands:"
  echo "  token-cert                          Manage identity token signing certificates"
  echo "  saml2-cert                          Manage identity SAML2 certificates"
  echo "  kerberos-auth                       Manage kerberos auth settings"
  echo "  windows-auth                        Manage windows authentication settings"
  echo "  add-host-admin                      Add a service administrator to the host partition"
  echo "  enable-basic-auth                   Enable basic authentication for an organization"
  echo
  echo "Flags:"
  echo "  -h|--help                           Display help"
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    COMPONENT=$1
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    token-cert)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}" 
      configure_token_cert "$@";
      ;;

    saml2-cert)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_saml2_config "$@"
      ;;

    kerberos-auth)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_kerberos_auth "$@"
      ;;

    windows-auth)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_win_auth "$@"
      ;;

    add-host-admin)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      add_host_admin "$@"
      ;;

    enable-basic-auth)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      enable_basic_auth "$@"
      ;;

    -*|*) # unsupported flags
      display_usage
      exit 1
      ;;
    esac
  done

  display_usage
}

function configure_token_cert() {
  if "${SCRIPT_DIR}"/signing-cert/signing-cert.sh "$@"; then
    exit 0
  fi
    exit 1
}

function configure_saml2_config {
  if "${SCRIPT_DIR}"/saml2-cert/saml2-cert.sh "$@"; then
    exit 0
  fi
    exit 1
}

function configure_kerberos_auth() {
  if "${SCRIPT_DIR}"/kerberos-auth/kerberos.sh "$@"; then
    exit 0
  fi
    exit 1
}

function configure_win_auth() {
  if "${SCRIPT_DIR}"/windows-auth/main.sh "$@"; then
    exit 0
  fi
    exit 1
}

function add_host_admin() {
  if "${SCRIPT_DIR}"/add-host-admin/main.sh "$@"; then
    exit 0
  fi
    exit 1
}

function enable_basic_auth() {
  if "${SCRIPT_DIR}"/enable-basic-auth/main.sh "$@"; then
    exit 0
  fi
    exit 1
}

function main() {
  parse_long_args "$@"
  exit 0
}

main "$@"

