#! /bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/argocd_utils.sh

function update_common_kerberos() {
    argocd app set uipath -p "global.kerberosAuthConfig.adDomain=${AD_DOMAIN}" \
                          -p "global.kerberosAuthConfig.adUserName=${DEFAULT_AD_USERNAME}" \
                          -p "global.kerberosAuthConfig.enabled=${IS_KERBEROS_ENABLED}" \
                          -p "global.kerberosAuthConfig.userKeytab=${KEYTAB}" \
                          -p "global.kerberosAuthConfig.ticketLifetimeInHours=${TICKET_LIFETIME}"

    argocd app sync uipath
}

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Update Global Kerberos Auth Settings"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  -d|--debug            Run this script in debug mode"
  echo "  --enabled             (Required) Input: true/false; Enable/Disable Sql Integrated Auth"
  echo "  --adDomain            (Required) Enter the Active Directory Domain"
  echo "  --username            (Required) Enter the default Ad username"
  echo "  --keytab              (Required) Enter the keytab file string in base64 encoding"
  echo "  --lifetime            (Required) Enter the ticket lifetime in hour (8-168)"
  echo "  -h|--help             Display help" 
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    --enabled)
      if [ "$2" == "true" ] || [ "$2" == "false" ]; then
        IS_KERBEROS_ENABLED=$2
        shift 2
      else
        display_usage
        error "Wrong input for --enabled flag"
        exit 1
      fi
      ;;

    --adDomain)
      AD_DOMAIN=$2
      shift 2
      ;;

    --username)
      DEFAULT_AD_USERNAME=$2
      shift 2
      ;;

    --keytab)
      KEYTAB=$2
      shift 2
      ;;

    --lifetime)
      if [ "$2" -lt 8 ] || [ "$2" -gt 168 ]; then
        display_usage
        error "Ticket lifetime does not follow requirement"
      fi
      TICKET_LIFETIME=$2
      shift 2
      ;;

    -*|*) # unsupported flags
      display_usage
      ;;
    esac
  done
}

IS_KERBEROS_ENABLED="PLACEHOLDER"
AD_DOMAIN="PLACEHOLDER"
DEFAULT_AD_USERNAME="PLACEHOLDER"
KEYTAB="PLACEHOLDER"
TICKET_LIFETIME=8

function main() {
    if [ $# -lt 10 ]; then
	    display_usage
	    error "Missing command flags"
    fi

    parse_long_args "$@"

    echo "Updating kerberos auth....."
    if update_common_kerberos; then
        echo "Success!"
        echo "If you want to enable sql integrated auth for a service, please update the sql connection string."
        exit 0
    else 
        echo "Failed!"
        echo "Please provide valid kerberos auth configuration values."
        exit 1
    fi
}


main "$@"
exit 0
