#! /bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
   error "Could not determine script path"
fi
BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/argocd_utils.sh

function update_service_group() {
    argocd app set uipath -p "global.userInputs.$SERVICE_GROUP.kerberosAuthConfig.adUserName=$USERNAME" \
                          -p "global.userInputs.$SERVICE_GROUP.kerberosAuthConfig.userKeytab=$KEYTAB"

    argocd app sync uipath
}

function validate_service_group(){

    if [ "$SERVICE_GROUP" != "orchestrator" ] &&
        [ "$SERVICE_GROUP" != "platform" ] &&
        [ "$SERVICE_GROUP" != "discoverygroup" ] &&
        [ "$SERVICE_GROUP" != "testmanager" ] &&
        [ "$SERVICE_GROUP" != "automationops" ] &&
        [ "$SERVICE_GROUP" != "aicenter" ] &&
        [ "$SERVICE_GROUP" != "documentunderstanding" ] &&
        [ "$SERVICE_GROUP" != "insights" ]; then
          display_usage
          error "Invalid service group"
    fi
} 

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Update Service's Kerberos Auth Setting"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  -d|--debug        Run this script in debug mode"
  echo "  --sg              (Required) Name of the service group (Available services below)"
  echo "  --username        (Required) Enter username of AD"
  echo "  --keytab          (Required) Enter keytab file in base64 string format"
  echo "  -h|--help         Display help" 
  echo
  echo "Available Services (Case sensitive):"
  echo "   - orchestrator"
  echo "   - platform"
  echo "   - discoverygroup"
  echo "   - testmanager"
  echo "   - automationops"
  echo "   - aicenter"
  echo "   - documentunderstanding"
  echo "   - insights"
  echo "***************************************************************************************"
  echo
}


function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    --sg)
      SERVICE_GROUP=$2
      if validate_service_group; then
        shift 2
      fi
      ;;

    --username)
      USERNAME=$2
      shift 2
      ;;

    --keytab)
      KEYTAB=$2
      shift 2
      ;;

    -*|*) # unsupported flags
      display_usage
      ;;
    esac
  done
}

SERVICE_GROUP="PLACEHOLDER"
USERNAME="PLACEHOLDER"
KEYTAB="PLACEHOLDER"

function main() {
    if [ $# -lt 6 ] 
    then
      display_usage
      error "Missing command flags"
      exit 1
    fi

    parse_long_args "$@"  

    echo "Updating kerberos config for ${SERVICE_GROUP} service group....."
    if update_service_group; then
        echo "Success!"
        echo "If you want to enable sql integrated auth for the ${SERVICE_GROUP} service goup, please enable the sql connection string  (global.userInputs.${SERVICE_GROUP}.sqlConnectionString)"
        exit 0
    else 
        echo "Failed!"
        echo "Please provide valid kerberos auth configuration values."
        exit 1
    fi
}


main "$@"
exit 0
