#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR
export DOT_FOLDER=~/.uipath

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/Fabric_Installer/bin:/usr/local/bin

#shellcheck disable=SC1090,SC1091
{
  source "${MODULES_DIR}"/utils.sh
  source "${MODULES_DIR}"/argocd_utils.sh
}

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Switch secondary certificate as a primary certificate"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  --help|-h                           Display help"
  echo 
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      echo "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      ;;
    esac
  done
  
  echo
}

function rotate_saml_certificate() {
  
  local cert_name_1="global.userInputs.certificate.identity.saml.currentServiceCert"
  local pass_1="global.userInputs.certificate.identity.saml.currentServiceCertPassword"
  local cert_name_2="global.userInputs.certificate.identity.saml.futureServiceCert"
  local pass_2="global.userInputs.certificate.identity.saml.futureServiceCertPassword"

  local primary_cert primary_pass secondary_cert secondary_pass argo_params
  
  argo_params=$(argocd app get uipath --show-params -o json | jq -er  '.spec.source.helm.parameters')

  set +e
  # don't stop the script execution in case of errors when attempting to load current cert / current certPassword
  # the cert might be null for a fresh cluster and the request will return a non-zero result in jq 
  primary_cert=$( (echo "${argo_params}" | jq -er '.[] | select(.name == "'${cert_name_1}'") | .value') || \
    (echo "Q0VSVElGSUNBVEVfUExBQ0VIT0xERVI="))
  primary_pass=$( (echo "${argo_params}" | jq -er '.[] | select(.name == "'${pass_1}'") | .value') || \
    (echo "UExBQ0VIT0xERVI="))
  
  # restore errexit flag as future certificate is mandatory 
  set -e

  secondary_cert=$(echo "${argo_params}" | jq -er '.[] | select(.name == "'${cert_name_2}'") | .value')
  secondary_pass=$(echo "${argo_params}" | jq -er '.[] | select(.name == "'${pass_2}'") | .value')

  argocd app set uipath -p "${cert_name_1}=${secondary_cert}"
  argocd app set uipath -p "${pass_1}=${secondary_pass}"
  argocd app set uipath -p "${cert_name_2}=${primary_cert}"
  argocd app set uipath -p "${pass_2}=${primary_pass}"
  argocd app sync uipath
}

function main() {
  parse_long_args "$@"
  rotate_saml_certificate
  echo -e "SAML2 certificate is rotated successfully\n"
}

main "$@"
exit 0