#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR
export DOT_FOLDER=~/.uipath

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/Fabric_Installer/bin:/usr/local/bin

#shellcheck disable=SC1090,SC1091
{
  source "${MODULES_DIR}"/utils.sh
  source "${MODULES_DIR}"/certificate_utils.sh
  source "${MODULES_DIR}"/argocd_utils.sh
}

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Update secondary SAML2 service certificate in preparation for certificate upgrade"
  echo "This will not replace the primary certificate for SAML2"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  --cert-file-path                    Path to the PKCS#12 archive (.pfx)"
  echo "                                        certificate file path"
  echo "  --password                          Password decrypt certificate [OPTIONAL]"
  echo "  --help|-h                           Display help"
  echo 
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    --cert-file-path)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        CERT_FILE_PATH=$2
        CERT_FILE_PATH="${CERT_FILE_PATH/#\~/$HOME}"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;
    --password)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        PASSWORD="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      ;;
    esac
  done

  if [[ ! -f ${CERT_FILE_PATH} ]]; then
    display_usage 
    error "Certificate \"${CERT_FILE_PATH}\" not found"
  fi
  if [[ -z ${PASSWORD} ]]; then
    echo "Please enter password to decrypt certificate:"
    read -sr PASSWORD
  fi
}

function update_saml2_cert() {
  
  local cert_file=$1
  local password=$2

  local cert_file_base64 password_base64
  cert_file_base64=$(base64 "${cert_file}" -w0)
  password_base64=$( echo "${password}" | base64)
  
  argocd app set uipath -p "global.userInputs.certificate.identity.saml.futureServiceCert=${cert_file_base64}"
  argocd app set uipath -p "global.userInputs.certificate.identity.saml.futureServiceCertPassword=${password_base64}"
  argocd app sync uipath
}

function check_confirmation() {
  local input
  echo -e "\n"
  warn "Please be aware this operation will update SAML2 certificate for your cluster"
  warn "If you wish to continue, type 'yes' and hit Enter for confirmation."
  read -r input </dev/tty
  if [[ $input == 'yes' ]] || [[ $input == "'yes'" ]]
  then
    info "Proceeding to update SAML2 certificate"
  else
    info "Unexpected user input received: ${input}, exiting SAML2 certificate"
    exit 2
  fi
  echo -e "\n"
}

function main() {
  parse_long_args "$@"
  check_confirmation
  validate_signing_certificate "${CERT_FILE_PATH}" "${PASSWORD}"
  update_saml2_cert "${CERT_FILE_PATH}" "${PASSWORD}"
  echo -e "\nSAML2 certificate is updated successfully\n"
}

main "$@"
exit 0