#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/Fabric_Installer/bin:/usr/local/bin

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Get identity token signing certificates"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  --outpath                           Stores identity-token-cert certificates in"
  echo "                                        given path for --outpath flag"
  echo "  -h|--help                           Display help" 
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    --outpath)
      if [ -n "$2" ]; then
        CERT_OUTPATH="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      ;;
    esac
  done
}

function check_confirmation() {
  local input
  echo -e "\n"
  warn "Please be aware this operation will print Token signing certificate password on console"
  warn "If you wish to continue, type 'yes' and hit Enter for confirmation."
  read -r input </dev/tty
  if [[ $input == 'yes' ]] || [[ $input == "'yes'" ]]
  then
    info "Proceeding to fetch Token signing certificate password"
  else
    info "Unexpected user input received: ${input}, exiting Token signing certificate"
    exit 2
  fi
  echo -e "\n"
}

function fetch_token_signing_certs() {
  local namespace="uipath"
  local secret_name="identity-token-signing-certificate"
  local identity_cert="IdentityServer1.pfx"
  local password="password1"
  if secret=$(kubectl get secret ${secret_name} --namespace ${namespace} -o json 2> /dev/null);
  then
    [ -n "${CERT_OUTPATH}" ] && mkdir -p "${CERT_OUTPATH}" 

    local password_decoded
    password_decoded=$(echo "${secret}" | jq -r ".data.\"$password\"" | base64 -d)
    if [ -n "${CERT_OUTPATH}" ]; then 
      echo "${secret}" | jq -r ".data.\"$identity_cert\"" | base64 -d | cat > "${CERT_OUTPATH}"/"${identity_cert}"
      echo "${password_decoded}" | cat > "${CERT_OUTPATH}"/"${password}"
      
      echo -e "\nToken Signing certificate is stored at ${CERT_OUTPATH} \n"
    else
      local tmp_dir
      tmp_dir=$(mktemp -d certificates.XXXXXXX)

      echo -e "\n\nToken Signing certificate is printed below\n\n"
      echo "${secret}" | jq -r ".data.\"$identity_cert\"" | base64 -d | cat > "${tmp_dir}"/"${identity_cert}"
      openssl pkcs12 -in "${tmp_dir}"/"${identity_cert}" -nodes --passin pass:"${password_decoded}"

      check_confirmation
      echo -e "\nPassword is printed below\n\n"
      echo "${password_decoded}"
      echo -e "\n"

      rm -rf "$tmp_dir"
    fi
  else
    error "Could not fetch token signing certs"
  fi
}

function main() {
  parse_long_args "$@"
  
  fetch_token_signing_certs

}

main "$@"
exit 0