#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR
export DOT_FOLDER=~/.uipath

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/Fabric_Installer/bin:/usr/local/bin

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Switch secondary certificate as a primary certificate"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  --help|-h                           Display help"
  echo 
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      echo "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      ;;
    esac
  done
  
  echo
}

function rotate_token_certificate() {
  local namespace="uipath"
  local secret_name="identity-token-signing-certificate"
  local cert_name_1="IdentityServer1.pfx"
  local pass_1="password1"
  local cert_name_2="IdentityServer2.pfx"
  local pass_2="password2"

  local primary_cert primary_pass secondary_cert secondary_pass
  
  if secret=$(kubectl get secret ${secret_name} --namespace ${namespace} -o json 2> /dev/null);
  then
    [[ -d "${DOT_FOLDER}" ]] || mkdir -p "${DOT_FOLDER}"
  
    echo "${secret}" | jq -r ".data.\"$cert_name_1\"" | base64 -d > "${DOT_FOLDER}/$cert_name_1"
    primary_pass=$(echo "${secret}" | jq -r ".data.$pass_1" | base64 -d)
    primary_cert="${DOT_FOLDER}/$cert_name_1"
    primary_cert=${primary_cert/#\~/$HOME} #allow paths relative to the home directory

    echo "${secret}" | jq -r ".data.\"$cert_name_2\"" | base64 -d > "${DOT_FOLDER}/$cert_name_2"
    secondary_pass=$(echo "${secret}" | jq -r ".data.$pass_2" | base64 -d)
    secondary_cert="${DOT_FOLDER}/$cert_name_2"
    secondary_cert=${secondary_cert/#\~/$HOME} #allow paths relative to the home directory


    # now make secondary cert as 1st cert, and primary cert as 2nd cert
    kubectl create secret generic "${secret_name}" \
    --from-file="${cert_name_1}"="${secondary_cert}" \
    --from-literal="${pass_1}"="${secondary_pass}" \
    --from-file="${cert_name_2}"="${primary_cert}" \
    --from-literal="${pass_2}"="${primary_pass}" \
    --namespace "${namespace}" \
    --dry-run=client -o yaml | kubectl apply -f - >/dev/null 2>&1
  else
    error "Identity token signing certificate is not found"
    echo
  fi
}

function main() {
  parse_long_args "$@"
  
  rotate_token_certificate
  echo -e "Token signing certificate is rotated successfully\n"
}

main "$@"
exit 0