#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR
export DOT_FOLDER=~/.uipath

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/Fabric_Installer/bin:/usr/local/bin

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/certificate_utils.sh

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Update secondary certificate for signing the authentication token"
  echo "However this will not replace the primary certificate to sign token"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  --cert-file-path                    Path to the PKCS#12 archive (.pfx)"
  echo "                                        certificate file path"
  echo "  --password                          Password decrypt certificate [OPTIONAL]"
  echo "  --help|-h                           Display help"
  echo 
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    --cert-file-path)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        CERT_FILE_PATH=$2
        CERT_FILE_PATH="${CERT_FILE_PATH/#\~/$HOME}"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;
    --password)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        PASSWORD="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      ;;
    esac
  done

  if [[ ! -f ${CERT_FILE_PATH} ]]; then
    display_usage 
    error "Certificate \"${CERT_FILE_PATH}\" not found"
  fi
  if [[ -z ${PASSWORD} ]]; then
    echo "Please enter password to decrypt certificate:"
    read -sr PASSWORD
  fi
}

function update_token_sigining_cert() {
  local namespace="uipath"
  local secret_name="identity-token-signing-certificate"
  local cert_name_1="IdentityServer1.pfx"
  local pass_1="password1"
  local cert_name_2="IdentityServer2.pfx"
  local pass_2="password2"

  local original_cert=$1
  local original_pass=$2
  local cert_file=$1
  local password=$2
  
  if secret=$(kubectl get secret ${secret_name} --namespace ${namespace} -o json 2> /dev/null);
  then
    [[ -d "${DOT_FOLDER}" ]] || mkdir -p "${DOT_FOLDER}"
  
    echo "${secret}" | jq -r ".data.\"$cert_name_1\"" | base64 -d > "${DOT_FOLDER}/$cert_name_1"
    original_pass=$(echo "${secret}" | jq -r ".data.$pass_1" | base64 -d)
    original_cert="${DOT_FOLDER}/$cert_name_1"
    original_cert=${original_cert/#\~/$HOME} #allow paths relative to the home directory
  fi

  kubectl create secret generic "${secret_name}" \
    --from-file="${cert_name_1}"="${original_cert}" \
    --from-literal="${pass_1}"="${original_pass}" \
    --from-file="${cert_name_2}"="${cert_file}" \
    --from-literal="${pass_2}"="${password}" \
    --namespace "${namespace}" \
    --dry-run=client -o yaml | kubectl apply -f - >/dev/null 2>&1 || error "Failed to update Token signing certificate"
}

function check_confirmation() {
  local input
  echo -e "\n"
  warn "Please be aware this operation will update Token signing certificate for your cluster"
  warn "If you wish to continue, type 'yes' and hit Enter for confirmation."
  read -r input </dev/tty
  if [[ $input == 'yes' ]] || [[ $input == "'yes'" ]]
  then
    info "Proceeding to update Token signing certificate"
  else
    info "Unexpected user input received: ${input}, exiting Token signing certificate"
    exit 2
  fi
  echo -e "\n"
}

function main() {
  parse_long_args "$@"
  check_confirmation
  validate_signing_certificate "${CERT_FILE_PATH}" "${PASSWORD}"
  update_token_sigining_cert "${CERT_FILE_PATH}" "${PASSWORD}"
  echo -e "\nToken signing certificate is updated successfully\n"
}

main "$@"
exit 0