#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR
export DOT_FOLDER=~/.uipath

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/Fabric_Installer/bin:/usr/local/bin

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/k8sutils.sh
#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/scripts/rotate-secrets.sh

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Rotate Cert"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  --help|-h                           Display help"
  echo 
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      echo "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      ;;
    esac
  done
  
  echo
}


function create_service_account() {
  kubectl apply -f "${SCRIPT_DIR}"/manifests/rbac.yaml || echo "Unable to create service account"
}

function delete_service_account() {
  kubectl delete -f "${SCRIPT_DIR}"/manifests/rbac.yaml || echo "Unable to delete service account"
}

function renew_certificate() {
  local status="notready"
  kubectl delete -f "${SCRIPT_DIR}"/manifests/renew-cert-job.yaml || echo ""
  kubectl apply -f "${SCRIPT_DIR}"/manifests/renew-cert-job.yaml
  info "Renewing the certificate"
  kubectl -n uipath-infra wait --for=condition=complete --timeout=90s job/renew-certificate-mongodb && status="ready"
  if [ "$status" = "notready" ]; then
    info "Certificate renewal not successful. Please retry or contact UiPath support."
    exit 1
  fi
}

function restart_mongodb_components() {
  local namespace="mongodb"
  local statefulset_list=(mongodb-replica-set)

  for statefulset_name in "${statefulset_list[@]}"; do
    replica_count=$(kubectl get statefulset "$statefulset_name" -n "${namespace}" -o=jsonpath='{.status.replicas}')
    scale_down_statefulset "${statefulset_name}" "${namespace}"

    kubectl -n "${namespace}" scale statefulsets "${statefulset_name}" --replicas="$replica_count"
    kubectl -n "$namespace" rollout status statefulset "$statefulset_name" -w --timeout=600s

    # shellcheck disable=SC2181
    if [[ "$?" -ne 0 ]]; 
    then
      info "$statefulset_name deployment failed in namespace $namespace. Exiting !!!"
      exit 1
    fi
  done
}

function restart_apps_service() {
  local namespace="uipath"
  local deployment_list=(apps-server apps-wsserver)

  for deployment_name in "${deployment_list[@]}"; do
    replica_count=$(kubectl get deploy "$deployment_name" -n "${namespace}" -o=jsonpath='{.spec.replicas}')
    scale_down_deployment "${deployment_name}" "${namespace}"
    
    #handle corner cases
    [ -z "$replica_count" ] && replica_count=1
    [[ $replica_count -eq 0 ]] && replica_count=1

    kubectl -n "${namespace}" scale deploy "${deployment_name}" --replicas="$replica_count"
    kubectl -n "$namespace" rollout status deploy "$deployment_name" -w --timeout=600s
    
    # shellcheck disable=SC2181
    if [[ "$?" -ne 0 ]]; 
    then
      info "$deployment_name deployment failed in namespace $namespace. Exiting !!!"
      exit 1
    fi
  done
}

function update_additional_secrets() {
  #extract updated pem file name
  newPemFileName=$(kubectl -n mongodb get secret mongodb-replica-set-server-certificate-key -o json | jq -r '.data'| jq -r keys[0])
  echo "New pem file name ${newPemFileName}"

  #extract stale pem file name
  oldPemFile=$(kubectl -n mongodb get secret mongodb-replica-set-config -o json | jq -r '.data."cluster-config.json"' | base64 -d | jq -r '.processes[0].args2_6.net.tls.certificateKeyFile')
  oldPemFileName=$(basename "$oldPemFile")
  echo "Stale pem file name ${oldPemFileName}"

  if [[ "$oldPemFileName" != "$newPemFileName" ]]; then
    echo "Pem file entries do not match. replacing"

    #extract replica set secret  cluser config json to file
    kubectl -n mongodb get secret mongodb-replica-set-config -o json | jq -r '.data."cluster-config.json"' | base64 -d > /tmp/clusterConfig.json

    #replace old pem file name with new in the json file
    sed -i -e "s@$oldPemFileName@$newPemFileName@g" /tmp/clusterConfig.json

    #encode the json
    encodedUpdatedClusterConfig=$(jq -r '. | @base64 | "\(.)"' /tmp/clusterConfig.json)

    #patch replica set secret with updated cluster config
    kubectl -n mongodb patch secret mongodb-replica-set-config --type='json' -p='[{"op" : "replace" ,"path" : "/data/cluster-config.json" ,"value" : "'"$encodedUpdatedClusterConfig"'"}]'

  else
    echo "Pem file entries match; not updating"
  fi
}

function rotate_mongodb_certificate(){
  create_service_account
  renew_certificate
  rotate_secrets
  update_additional_secrets
  delete_service_account
  info "Cert Renewal successful and all MongoDB components are up and running"
}

function main() {
  parse_long_args "$@"
  
  rotate_mongodb_certificate
  info "MongoDB certificates rotated successfully\n"
}

main "$@"
exit 0