#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/argocd_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/certificate_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/k8sutils.sh

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Get cluster tls certificates"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  --outpath                           Stores tls certificates in given path for --outpath flag"
  echo "  -h|--help                           Display help" 
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;
    
    --outpath)
      if [ -n "$2" ]; then
        CERT_OUTPATH="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      exit 1
    ;;
    esac
  done
}

function fetch_tls_certs() {
  
  local ca_crt_file="ca.crt"
  local tls_crt_file="tls.crt"
  local tls_key_file="tls.key"
  local secret
  secret=$(get_tls_certs)

  [ -n "${CERT_OUTPATH}" ] && mkdir -p "${CERT_OUTPATH}"
  
  [ -n "${CERT_OUTPATH}" ] && echo -e "TLS Certificate Authority Bundle is stored at ${CERT_OUTPATH}/${ca_crt_file}\n" || echo -e "\n\nTLS Certificate Authority Bundle is printed below\n\n"
  #shellcheck disable=SC2015
  echo "${secret}" | jq -r ".data.\"$ca_crt_file\"" | base64 -d | ( [ -n "${CERT_OUTPATH}" ] && cat > "${CERT_OUTPATH}"/"${ca_crt_file}" || cat )
  
  [ -n "${CERT_OUTPATH}" ] && echo -e "TLS Public Certificate is stored at ${CERT_OUTPATH}/${tls_crt_file}\n" || echo -e "\n\nTLS Public Certificate is printed below\n\n"
  #shellcheck disable=SC2015
  echo "${secret}" | jq -r ".data.\"$tls_crt_file\"" | base64 -d | ( [ -n "${CERT_OUTPATH}" ] && cat > "${CERT_OUTPATH}"/"${tls_crt_file}" || cat )
  
  [ -n "${CERT_OUTPATH}" ] && echo -e "TLS Certificate Private Key is stored at ${CERT_OUTPATH}/${tls_key_file}\n" || echo -e "\n\nTLS Certificate Private Key is printed below\n\n"
  #shellcheck disable=SC2015
  echo "${secret}" | jq -r ".data.\"$tls_key_file\"" | base64 -d | ( [ -n "${CERT_OUTPATH}" ] && cat > "${CERT_OUTPATH}"/"${tls_key_file}" || cat )
}

function main() {
  parse_long_args "$@"
  
  fetch_tls_certs
}

main "$@"
