#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/Fabric_Installer/bin:/usr/local/bin

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/k8sutils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/argocd_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cluster_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/certificate_utils.sh

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Update cluster tls and server certificate"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  --ca-cert-file                      Certificate Authority bundle to"
  echo "                                        sign the tls certificate"
  echo "  --tls-cert-file                     Public tls certificate for FQDN"
  echo "  --tls-key-file                      Private key for certificate provided"
  echo "                                        with --tls-cert-file"
  echo "  -h|--help                           Display help" 
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;
    --ca-cert-file)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        CERT_CA_FILE_PATH=$2
        CERT_CA_FILE_PATH="${CERT_CA_FILE_PATH/#\~/$HOME}"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
    ;;
    --tls-cert-file)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        CERT_TLS_FILE_PATH=$2
        CERT_TLS_FILE_PATH="${CERT_TLS_FILE_PATH/#\~/$HOME}"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
    ;;
    --tls-key-file)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        CERT_KEY_FILE_PATH=$2
        CERT_KEY_FILE_PATH="${CERT_KEY_FILE_PATH/#\~/$HOME}"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
    ;;
    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      exit 1
    ;;
    esac
  done

  if [[ ! -f ${CERT_CA_FILE_PATH} ]]; then 
    display_usage 
    error "Certificate Authority bundle \"${CERT_CA_FILE_PATH}\" not found"
  fi
  if [[ ! -f ${CERT_TLS_FILE_PATH} ]]; then 
    display_usage 
    error "Public tls certificate \"${CERT_TLS_FILE_PATH}\" not found"
  fi
  if [[ ! -f ${CERT_KEY_FILE_PATH} ]]; then 
    display_usage 
    error "Private key for certificate \"${CERT_KEY_FILE_PATH}\" not found"
  fi

  echo
}

function update_docker_registry_certs() {
  local tls_secret_name="docker-registry-tls"
  local namespace="docker-registry"
  local deployment="docker-registry"
  
  kubectl create secret generic ${tls_secret_name} \
    --namespace ${namespace} \
    --from-file=tls.crt="${CERT_TLS_FILE_PATH}" \
    --from-file=tls.key="${CERT_KEY_FILE_PATH}" \
    --dry-run=client -o yaml | kubectl apply -f -
  
  old_replica_count=$(kubectl get deploy ${deployment} -n ${namespace} -o=jsonpath="{.spec.replicas}") || old_replica_count=1
  # This statement is just to be on safe side, if somehow kubectl command returns 0 as old_replica_count then it will reset to 1
  [[ "${old_replica_count}" == "0" ]] && old_replica_count=1
  
  trap '' SIGINT
  scale_down_deployment "${deployment}" "${namespace}"
  scale_up_deployment "${deployment}" "${namespace}" "${old_replica_count}"
  trap - SIGINT

  if wait_for_all_pods_in_namespace_ready "${namespace}"; then
    info "Docker Registry update successfully"
  else
    error "Docker Registry update failed"
  fi
}

function update_istio_certificate() {
  local tls_secret_name="istio-ingressgateway-certs"
  local namespace="istio-system"
  local label='config-discovery=yes'
  local service_namespace="uipath"
  
  kubectl create secret generic ${tls_secret_name} \
    --namespace ${namespace} \
    --from-file=ca.crt="${CERT_CA_FILE_PATH}" \
    --from-file=tls.crt="${CERT_TLS_FILE_PATH}" \
    --from-file=tls.key="${CERT_KEY_FILE_PATH}" \
    --dry-run=client -o yaml | kubectl apply -f -

  kubectl label secret ${tls_secret_name} "${label}" \
    --namespace ${namespace} --overwrite=true

  kubectl create secret generic ${tls_secret_name} \
    --namespace ${service_namespace} \
    --from-file=ca.crt="${CERT_CA_FILE_PATH}" \
    --from-file=tls.crt="${CERT_TLS_FILE_PATH}" \
    --from-file=tls.key="${CERT_KEY_FILE_PATH}" \
    --dry-run=client -o yaml | kubectl apply -f -
}

function update_server_certificate() {
  local service_namespace="uipath"
  cert_validation_status=$(validate_server_certificate "${CERT_CA_FILE_PATH}" "${CERT_TLS_FILE_PATH}" "${CERT_KEY_FILE_PATH}" "${CLUSTER_FQDN}" "${ARGO_FQDN}" "${OBJECTSTORE_FQDN}" "${REGISTRY_FQDN}" "${MONITORING_FQDN}" "${INSIGHTS_FQDN}")
  [[ "${cert_validation_status}" == "success" ]] && info "Validating Server Certificate... Done"

  info "TLS Server Certificate updation is in process, may take upto 30mins \n" 
  info "Configuring TLS ca certificates... IN PROGRESS"

  if [[ "${INSTALL_TYPE}" == "offline" ]]; then
    update_docker_registry_certs >/dev/null
    info "Docker Registry Update... DONE"
    update_argocd_trusted_certs "${REGISTRY_FQDN}" >/dev/null
    info "argocd update done..."
  fi

  update_istio_certificate >/dev/null
  info "Istio Routing Update... DONE"
  
  if [[ "$PROFILE" == "ha" ]]; 
  then
    info "Now restarting pods in uipath namespace"
    kubectl rollout restart deployment --namespace "${service_namespace}" >/dev/null
    info "Waiting for pods to be ready"
    wait_for_all_pods_in_namespace_ready "${service_namespace}" \
      || error "Certificate update failed - timeout waiting for all pods to restart and ready in namespace $service_namespace"
  else
    scale_down_all_deployment "$service_namespace" \
      || error "Certificate update failed - Scalling down all deployment in namespace $service_namespace failed"
    scale_up_all_deployment "$service_namespace" "1" \
      || error "Certificate update failed - Scalling up all deployment in namespace $service_namespace failed" 
  fi
  
  info "Certificate update successfully..."
  exit 0
}

function set_fqdn() {
  CLUSTER_FQDN=$(get_cluster_config "CLUSTER_FQDN")
  #shellcheck disable=SC2015
  [ -z "${CLUSTER_FQDN}" ] && error "Cluster fqdn is not found in cluster" || true

  ARGO_FQDN=$(get_cluster_config "ARGOCD_FQDN")
  OBJECTSTORE_FQDN=$(get_cluster_config "OBJECTSTORE_FQDN")
  REGISTRY_FQDN=$(get_cluster_config "REGISTRY_FQDN")
  MONITORING_FQDN=$(get_cluster_config "MONITORING_FQDN")
  INSIGHTS_FQDN=$(get_cluster_config "INSIGHTS_FQDN")
  INSTALL_TYPE=$(get_cluster_config "INSTALL_TYPE")

  PROFILE=$(get_cluster_config_default "PROFILE" "ha")
}

function check_confirmation() {
  local input
  echo -e "\n"
  warn "Please be aware this operation will update TLS certificate for your cluster"
  warn "In eval profile, this can cause downtime for the time certificate is getting updated in cluster"
  warn "If you wish to continue, type 'yes' and hit Enter for confirmation."
  read -r input </dev/tty
  if [[ $input == 'yes' ]] || [[ $input == "'yes'" ]]
  then
    info "Proceeding to update tls cert"
  else
    info "Unexpected user input received: ${input}, exiting TLS cert update operation"
    exit 2
  fi
  echo -e "\n"
}

function main() {
  parse_long_args "$@"
  check_confirmation
  set_fqdn
  update_server_certificate
}

main "$@"
exit 0