#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

readonly SCRIPT_DIR
readonly MODULES_DIR

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/k8sutils.sh
#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/disaster_recovery_utils.sh
#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/longhorn_helper.sh

export DISABLE_BACKUP=${DISABLE_BACKUP:-false}
export ONE_TIME_MANUAL_BACKUP=${ONE_TIME_MANUAL_BACKUP:-false}

# kube config file
KUBECONFIG="/etc/rancher/rke2/rke2.yaml"
MOUNTED_BACKUP_DIR="/mnt/backup"
BASE_BACKUP_DIR=${MOUNTED_BACKUP_DIR}/backup
export MERGEDINPUTFILE="${SCRIPT_DIR}/merged_infra_input.json"
export NODE_ROLE=${NODE_ROLE:-none}
export MOUNTED_BACKUP_DIR=${MOUNTED_BACKUP_DIR}

########
# Parse input json file to set the config values
# Globals:
#   INFILE
# Outputs:
#   Sets global config variables
########
function parse_input_file() {
  set -e
  LONGHORN_MOUNT_PATH="$(get_mandatory_json_value ".infra.longhorn.mount")"
  ETCD_BACKUP_PATH=$(get_mandatory_json_value ".backup.etcdBackupPath")
  BACKUP_INTERVAL=$(get_mandatory_json_value ".backup_interval")
  NFS_ENDPOINT="$(get_mandatory_json_value ".backup.nfs.endpoint")"
  NFS_MOUNT_PATH="$(get_mandatory_json_value ".backup.nfs.mountpath")"
  STORAGE_CLASS="$(get_mandatory_json_value ".storage_class")"
  STORAGE_CLASS_SINGLE_REPLICA="$(get_mandatory_json_value ".storage_class_single_replica")"
  PROXY_ENABLED="$(get_optional_json_value ".proxy.enabled")"
  HTTP_PROXY="$(get_optional_json_value ".proxy.http_proxy")"

  export PROXY_ENABLED="${PROXY_ENABLED}"
  export HTTP_PROXY="${HTTP_PROXY}"
  export STORAGE_CLASS="${STORAGE_CLASS}"
  export STORAGE_CLASS_SINGLE_REPLICA="${STORAGE_CLASS_SINGLE_REPLICA}"
  set +e
}


########
# Validates NFS Server Connection & Access Permissions
# Input:
# ENDPOINT - endpoint of nfs server
# MOUNT_PATH - directory exposed by nfs server for backup
########
function validate_nfs_server(){
  local endpoint=$1
  local mount_path=$2

  validate_nfs_server_connection "${endpoint}" "${mount_path}"
  # validate access permissions
  validate_nfs_access_permissions "${endpoint}" "${mount_path}"
}


########
# Mount NFS Server exposed mounted path to local directory
# Input:
# ENDPOINT - endpoint of nfs server
# MOUNT_PATH - directory exposed by nfs server for backup
########
function mount_nfs_server_backup(){
  local endpoint=$1
  local mount_path=$2

  # mount nfs server to client
  mkdir -p ${MOUNTED_BACKUP_DIR}

  local mount_result
  mount_result=$(mount -t nfs "${endpoint}":"${mount_path}" "${MOUNTED_BACKUP_DIR}" 2>&1) || true
  if { [ -n "${mount_result}" ] || [ "${mount_result}" == " " ] ; }; then
    error "Mount Path not exposed for nfs server ${endpoint}"
  fi

  # create backup folder for cluster
  mkdir -p "${BASE_BACKUP_DIR}"/"${ETCD_BACKUP_PATH}"

  # mount NFS Server mapping to fstab for handling reboot
  if ! grep -q "${endpoint}:${mount_path}" /etc/fstab; then
    echo "${endpoint}:${mount_path}    ${MOUNTED_BACKUP_DIR}    nfs    rw    0    0" >> /etc/fstab
    info "Updated fstab with nfs mount successfully."
  fi

  info "Backup mounted at ${BASE_BACKUP_DIR}/${ETCD_BACKUP_PATH} in nfs server"
}


########
# Create Backup Job for copying etcd snapshots & config files to backupstore
########
function create_backupjob(){
  # create a job to copy backup files to mounted volume
  sed -e "s|BACKUP_MOUNT_PLACEHOLDER|${BASE_BACKUP_DIR}/${ETCD_BACKUP_PATH}|g" -e "s|NODE_PLACEHOLDER|${SETUP_HOSTNAME}|g" -e "s|LONGHORN_MOUNT_PLACEHOLDER|${LONGHORN_MOUNT_PATH}|g" -e "s|NODE_ROLE_PLACEHOLDER|${NODE_ROLE}|g" "${SCRIPT_DIR}"/backupjob.sh > backupjob

  cp backupjob /usr/bin/backupjob
  chmod +x /usr/bin/backupjob
  crontab -l > backupcron

  # clean previous entry
  sed -i '/backupjob/d' backupcron

  echo "*/${BACKUP_INTERVAL} * * * *  /usr/bin/backupjob" >> backupcron
  crontab backupcron
  rm backupcron
  rm backupjob

  info "Setup CronJob to take backup at every ${BACKUP_INTERVAL}min interval."
}

########
# Enable Backup for Longhorn volumes with nfs backup target
########
function enable_longhorn_nfs_backup() {
  local endpoint=$1
  local mount_path=$2

  # update settings for backup
  update_default_settings "default-longhorn-static-storage-class" "${STORAGE_CLASS}"
  update_default_settings "backup-target" "nfs://${endpoint}:${mount_path}"

  info "Update longhorn-default-settings to enable nfs longhorn volume backup"
}

function setup_backup() {
  local endpoint=$1
  local mount_path=$2

  # validate nfs server connection & access permissions
  validate_nfs_server "${endpoint}" "${mount_path}"

  # mount backup path to NFS Server
  mount_nfs_server_backup "${endpoint}" "${mount_path}"

  # create cron job for taking backup at given interval
  create_backupjob

  # enable volume backup
  if [[ "${NODE_ROLE}" == "none" ]]; then
    enable_longhorn_nfs_backup "${endpoint}" "${mount_path}"
    enable_pvc_backup
  fi

  # run backup for the first time manually
  /usr/bin/backupjob || error "Failed to run backup job"

  info "Backup Successfully enabled for node ${SETUP_HOSTNAME} at ${endpoint}:${mount_path}"
}

function validate_backup() {
  local backup_location="${BASE_BACKUP_DIR}/${ETCD_BACKUP_PATH}"
  local maxtry=30
  local try=0
  local success=0
  while (( try != maxtry )) ; do
    if [ -z "$(ls -A "${backup_location}")" ]; then
      try=$((try+1))
      info "waiting for the backup to be generated...${try}/${maxtry}" &&  sleep 5;
    else
      success=1
      break
    fi
  done

  if [ ${success} -eq 0 ]; then
    error "Failed to backup in the given time interval."
  fi
}

function run_manual_backup() {
  # Setup Backup
  setup_backup "${NFS_ENDPOINT}" "${NFS_MOUNT_PATH}"
  # validate if backup resources are getting copied to backupstore
  validate_backup
  # Disable Backup
  disable_backup "${NODE_ROLE}" "${ONE_TIME_MANUAL_BACKUP}"
}

#MAIN
function main() {
  parse_args "$@"
  merge_default_values
  parse_input_file

  export KUBECONFIG="${KUBECONFIG}"
  export PATH=${PATH}:/var/lib/rancher/rke2/bin

  # Check the current NODE_ROLE and reject the backup if it's opposite of the existing role
  if [[ ${NODE_ROLE} == "agent" ]];then
    [[ "$(systemctl is-enabled rke2-server)" == "enabled" ]] && error "The node is part of cluster as server, can't enable backup for agent"
  else
    [[ "$(systemctl is-enabled rke2-agent)" == "enabled" ]] && error "The node is part of cluster as Agent, can't enable backup for Server."
  fi

  # validate if cluster is ready, maxtry 60
  validate_cluster_ready 10 "${NODE_ROLE}"

  # validate if hostname exits in Node list
  validate_hostname

  info "Running Backup for HostName: ${SETUP_HOSTNAME}"

  # Set Longhorn Url
  set_longhorn_url "${NODE_ROLE}"
  validate_longhorn_reachable

  # if ONE_TIME_MANUAL_BACKUP selected, backup will take etcd backup, volume backup and then disable the backup as well
  if [[ "${ONE_TIME_MANUAL_BACKUP}" == "true" ]]; then
    run_manual_backup "${NFS_ENDPOINT}" "${NFS_MOUNT_PATH}" "${NODE_ROLE}"
  else
    if [[ "${DISABLE_BACKUP}" == "true"  ]]; then
      # Disable Backup
      disable_backup "${NODE_ROLE}"
    else
      # Setup Backup
      setup_backup "${NFS_ENDPOINT}" "${NFS_MOUNT_PATH}"
      # validate if backup resources are getting copied to backupstore
      validate_backup
    fi
  fi
}

main "$@"
