#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

# shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh
# shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/k8sutils.sh
# shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/longhorn_helper.sh

########
# Validate HostName
########
function validate_hostname() {
  local HOSTNAME

  HOSTNAME=$(hostname)
  export SETUP_HOSTNAME="${HOSTNAME}"
}

########
# Disable Backup
########
function disable_backup(){
  local node_role=$1
  local one_time_manual_backup=${2:-false}

  if [[ "${node_role}" == "none" ]]; then
    validate_manual_backup_complete "${one_time_manual_backup}"
  fi
  info "Disable ETCD backup in the setup if any"
  crontab -l > backupcron
  sed -i '/backupjob/d' backupcron
  crontab backupcron
  rm backupcron
  rm -rf /usr/bin/backupjob

  info "Disable Longhorn volume Backup"
  if [[ "${node_role}" == "none" ]]; then
    # shellcheck disable=SC2119
    disable_pvc_backup
    # reset settings for backup
    update_default_settings "backup-target" "" "ignore_failure"
  fi

  # unmount backup path
  umount "${MOUNTED_BACKUP_DIR}" || true
  info "Disabled Backup Successfully."
}

function create_recurring_job() {
  cron_job_expression="${1}"

  cat <<EOF | kubectl apply -f -
apiVersion: longhorn.io/v1beta1
kind: RecurringJob
metadata:
  name: uipath-backup
  namespace: longhorn-system
spec:
  concurrency: 4
  cron: "${cron_job_expression}"
  groups:
  - uipath-backup
  retain: 1
  task: backup
EOF

  # validate pvc cronjob for backup created
  local try=0
  local maxtry=10
  local exists=""
  while (( try < maxtry ));
  do
    exists=$(kubectl -n longhorn-system get cronjobs uipath-backup) || true
    if [[ -n "${exists}" ]]; then
      info "recurring cronjob uipath-backup created successfully"
      break;
    fi

    warn "recurring cronjob uipath-backup not created yet, ${try}/${maxtry}..."; sleep 5
    try=$((try + 1))
  done

  if [[ -z "${exists}" ]];then
    error "failed to create recurring job for uipath-backup"
  fi
}

function delete_recurring_job() {
  backup_job_name="${1}"

  # validate pvc cronjob for backup created
  local try=0
  local maxtry=10
  local exists=""
  while (( try < maxtry ));
  do
    # delete recurring job
    kubectl -n longhorn-system delete recurringjob "${backup_job_name}" --force || true
    kubectl -n longhorn-system delete cronjobs "${backup_job_name}" --force || true

    exists=$(kubectl -n longhorn-system get cronjobs "${backup_job_name}") || true
    if [[ -z "${exists}" ]]; then
      info "recurring cronjob ${backup_job_name} deleted successfully"
      break;
    fi

    warn "recurring cronjob ${backup_job_name} not deleted yet, ${try}/${maxtry}..."; sleep 5
    try=$((try + 1))
    exists=""
  done

  if [[ -n "${exists}" ]];then
    error "failed to delete recurring job for ${backup_job_name}"
  fi
}

function enable_pvc_backup() {
  # setup recurringJob for backupStorageClass
  local cronexpression="*/${BACKUP_INTERVAL} * * * *"
  create_recurring_job "${cronexpression}"
  kubectl -n longhorn-system delete job/uipath-backup-manual || true
  kubectl -n longhorn-system create job --from=cronjob/uipath-backup "uipath-backup-manual"
}

function validate_manual_backup_complete() {
  local one_time_manual_backup=${1}

  if [[ "${one_time_manual_backup}" == "true" ]]; then
    # wait for jobs completion max timeout 1hr
    success=$(kubectl -n longhorn-system wait --for=condition=complete job/uipath-backup-manual --timeout=3600s)
    if [[ "${success}" != *"condition met"* ]]; then
      error "Failed to complete manual backup."
    fi
  fi

  local label="recurring-job-group.longhorn.io/uipath-backup=enabled"
  local volume_list
  volume_list=$(kubectl -n longhorn-system get volumes.longhorn.io -l "${label}" -o json | jq -r ".items[].metadata.name")
  local failed=0
  for volume_name in ${volume_list}; do
    get_backup_list_by_pvname "${volume_name}"
    BACKUP_PATH=$(echo "${PV_BACKUP_PATH}"| jq -r ".url")
    unset PV_BACKUP_PATH
    if [[ -z "${BACKUP_PATH}" ]]; then
      error_without_exit "no backup taken for pvc ${volume_name}, restore pvc will not work."
      failed=1
    fi
    sleep 5
  done

  if [[ "${failed}" -eq 1 ]]; then
    if [[ "${one_time_manual_backup}" == "true" ]]; then
      error "failed to take all backups, please enable backup again to make sure backup is taken."
    fi
    error_without_exit "please enable backup again to make sure backup is taken."
  fi
}

function disable_pvc_backup() {
  delete_recurring_job "uipath-backup"
}

function validate_nfs_server_connection() {
  local endpoint=$1
  local mount_path=$2

  local connection
  local proxy
  proxy=""
  if [[ "${PROXY_ENABLED}" == "true" ]] ; then
      http_proxy_url=${HTTP_PROXY/http:\/\//}
      proxy="--proxy ${http_proxy_url}"
  fi
  # shellcheck disable=SC2086
  # validate connection
  connection=$(nc ${proxy} -vz "${endpoint}" 2049 2>&1)
  if [[ ${connection} != *"Connected"* ]]; then
    error "Failed to connect to NFS Server at ${endpoint}"
  fi

  info "Validated NFS Server connection successfully."
}

function mount_local_path() {
  local endpoint=$1
  local mount_path=$2
  local local_path=$3

  local mount_result
  # mount nfs server to client
  mount_result=$(mount -t nfs "${endpoint}":"${mount_path}" "${local_path}" 2>&1) || true
  if { [ -n "${mount_result}" ] || [ "${mount_result}" == " " ] ; }; then
    error "Mount Path not exposed for nfs server ${endpoint}, please check nfs server mount path"
  else
    info "Mounted nfs server at ${local_path} successfully."
  fi
}

########
# Validates NFS Server Access Permissions
# Input:
# ENDPOINT - endpoint of nfs server
# MOUNT_PATH - directory exposed by nfs server for backup
########
function validate_nfs_access_permissions(){
  local endpoint=$1
  local mount_path=$2

  local temp_dir="/mnt/tmp_dir"
  local temp_dup_dir="/mnt/tmp_dup_dir"
  local validation_file="test.txt"
  local validation_data

  validation_data=$(tr -dc A-Za-z0-9 </dev/urandom | head -c 13 ; echo '')
  mkdir -p ${temp_dir}
  mount_local_path "${endpoint}" "${mount_path}" ${temp_dir}
  # create a file and add validation data to it
  echo "${validation_data}" > ${temp_dir}/${validation_file} || true
  if [ -f ${temp_dir}/${validation_file} ]; then
    info "Created test file at ${validation_file}."
  else
    error "File not available... check nfs server permissions."
  fi

  # unmount and remove the directory
  umount ${temp_dir} || true
  rm -rf ${temp_dir} || true

  # validate a newly mounted directory with nfs volume contains the file
  mkdir -p ${temp_dup_dir}
  mount_local_path "${endpoint}" "${mount_path}" ${temp_dup_dir}

  local success=0
  local found
  local data
  found=$(find ${temp_dup_dir} -maxdepth 1 -name "${validation_file}"| wc -l)
  if [[ ${found} -eq 1 ]]; then
    data=$(< ${temp_dup_dir}/${validation_file})
    if [[ "${data}" = "${validation_data}" ]]; then
      info "NFS Server Connection/Permission Validated successfully"
      success=1
    fi
  fi

  if [[ ${success} -eq 0 ]]; then
    error "Failed to validate NFS Permissions."
  fi

  # cleanup
  rm -rf ${temp_dup_dir}/${validation_file:?} || true
  umount ${temp_dup_dir} || true
  rm -rf ${temp_dup_dir} || true
}

function clean_pod_with_io_error() {
  local namespace=$1
  local pod_list
  pod_list=$(kubectl get pods -n "${namespace}" -o json|jq -r ".items[].metadata.name")
  for pod in ${pod_list}; do
    local container_list
    container_list=$(kubectl -n "${namespace}" get pods "${pod}" -o jsonpath='{.spec.containers[*].name}')
    for container in ${container_list}; do
      local corrupt_container_log
      corrupt_container_log=$(kubectl -n "${namespace}" logs "${pod}" -c "${container}" | grep -i "Input/output error") || true
      if [[ -n "${corrupt_container_log}"  ]]; then
        info "Cleaning corrupt pod ${pod} under namespace ${namespace}."
        kubectl -n "${namespace}" delete pod "${pod}" --timeout=60s || true
      fi
    done
  done
}
