#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

# shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh
# shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/longhornutils.sh

function update_cron_expression() {
  local pv_name=$1
  local recurring_job_expression=$2
  if [[ "${recurring_job_expression}" != "null" ]]; then
      # patch cronexpression to volumes
      kubectl -n longhorn-system patch volumes.longhorn.io "${pv_name}" --type json -p="[{\"op\": \"replace\", \"path\": \"/spec/recurringJobs\", \"value\": ${recurring_job_expression}}]"
      info "Update cronexpression for volume ${pv_name} successfully."
  else
    info "Skipping cronexpression update for volume: ${pv_name}"
  fi
}

########
# Update Default Settings of longhorn
# Input:
# KEY - longhorn default settings key to be updated
# VALUE - longhorn default settings value to be updated
# IGNORE_FAILURE - throw exception if IGNORE_FAILURE is not passed
########
function update_default_settings(){
  local key=$1
  local value=$2
  local ignore_failure=""
  local response
  local success=0
  local url="${LONGHORN_URL}/v1/settings/${key}"

  if [ $# -gt 2 ]; then
    ignore_failure=$3
  fi
  response=$(curl -X PUT -H 'Content-Type: application/json;charset=UTF-8' "${url}" --data-raw "{\"name\":\"${key}\",\"type\":\"setting\",\"value\":\"${value}\"}")

  if { [ -n "${response}" ] && [ "${response}" != " " ]; }; then
    updated_value=$(echo "${response}"| jq -r ".value")
    if [ "${updated_value}" = "${value}"  ]; then
      info "Updated longhorn default settings ${key} with ${value} successfully."
      success=1
    fi
  fi

  if [ ${success} -eq 0 ]; then
    local message="Failed to configure ${key} in longhorn default settings ${response}"
    if [ -n "${ignore_failure}" ]; then
      warn "${message}"
      return
    fi
    error "${message}"
  fi
}

########
# Validates longhorn backup endpoint is reachable
########
function validate_longhorn_backup_reachable(){
  local try=0
  local maxtry=30
  local success=0

  while (( try != maxtry )) ; do
    local backup_resp=""
    local resp_message=""
    backup_resp=$(curl "${LONGHORN_URL}/v1/backupvolumes?")
    info "BackupVolumes response: ${backup_resp}"
    if { [ -n "${backup_resp}" ] && [ "${backup_resp}" != " " ]; }; then
      resp_message=$(echo "${backup_resp}"| jq -r ".data[]")
      if { [ -n "${resp_message}" ] && [ "${resp_message}" != " " ]; }; then
        info "Longhorn BackupStore is Accessible"
        success=1
        break;
      fi
      info "waiting for longhorn frontend backup to comeup ...${try}/${maxtry}";
    fi
    try=$((try+1))
    sleep 5
  done

  if [ ${success} -eq 0 ]; then
    error "Longhorn BackupStore not accessible, error: ${backup_resp}."
  fi
}

########
# Validates PV backup availability in backup server
# Input:
# PV_NAME - persistent volume name
########
function validate_pv_backup_available(){
  local pv_name=$1
  local validate_pv_backup_available_resp=0
  local backup_resp=""
  local resp_status_code=""
  local resp_message=""

  local try=0
  local maxtry=3
  while (( try != maxtry )) ; do
    backup_resp=$(curl "${LONGHORN_URL}/v1/backupvolumes/${pv_name}?") || true
    info "Backup Response: ${backup_resp}"
    if { [ -n "${backup_resp}" ] && [ "${backup_resp}" != " " ]; }; then
      resp_status_code=$(echo "${backup_resp}"| jq -c ".status")
      resp_message=$(echo "${backup_resp}"| jq -c ".message")
      if [[ -n "${resp_status_code}" && "${resp_status_code}" != " " && "${resp_status_code}" != "null" && (( resp_status_code -gt 200 )) ]] ; then
        validate_pv_backup_available_resp=0
      else
        resp_message=$(echo "${backup_resp}"| jq -c ".messages.error")
        if { [ -z "${resp_message}" ] || [ "${resp_message}" = "null" ]; }; then
          info "PVC Backup is available for ${pv_name}"
          # shellcheck disable=SC2034
          validate_pv_backup_available_resp=1
          break;
        fi
      fi
    fi
    try=$((try+1))
    sleep 10
  done
  export IS_BACKUP_AVAILABLE="${validate_pv_backup_available_resp}"
}

########
# Get backup of given PV from backup server
# Input:
# PV_NAME - persistent volume name
########
function get_backup_list_by_pvname(){
  local pv_name=$1
  local get_backup_list_by_pvname_resp=""
  local backup_list_response=""

  local try=0
  local maxtry=3
  while (( try != maxtry )) ; do

    backup_list_response=$(curl "${LONGHORN_URL}/v1/backupvolumes/${pv_name}?action=backupList" -X 'POST' -H 'Content-Length: 0' -H 'Accept: application/json')
    if [[ -n "${backup_list_response}" && "${backup_list_response}" != " " && -n $( echo "${backup_list_response}"|jq ".data" )  && -n $( echo "${backup_list_response}"|jq ".data[]" ) ]]; then
      info "Backup List Response: ${backup_list_response}"
      # pick first backup data with url non empty
      # shellcheck disable=SC2034
      get_backup_list_by_pvname_resp=$(echo "${backup_list_response}"|jq -c '[.data[]|select(.url | . != null and . != "")][0]')
      break;
    fi
    try=$((try+1))
    sleep 10
  done
  export PV_BACKUP_PATH="${get_backup_list_by_pvname_resp}"
}

########
# Restarts Longhorn, KubeProxy and RKE-Server Service
########
function restart_longhorn_and_rke_services() {
  local LONGHORN_MANAGER_POD_LIST
  local LONGHORN_CSI_ATTACHER_POD_LIST
  local PROXY_PID
  local KUBE_PROXY_POD_ID

  # restart longhorn manager
  LONGHORN_MANAGER_POD_LIST=$(kubectl -n longhorn-system get pods |grep ^longhorn-manager|awk '{print $1}') || true
  for LONGHORN_MANAGER in ${LONGHORN_MANAGER_POD_LIST};
  do
    kubectl -n longhorn-system delete pod "${LONGHORN_MANAGER}" --grace-period=0 --force &
    sleep 15
  done

  LONGHORN_CSI_ATTACHER_POD_LIST=$(kubectl -n longhorn-system get pods |grep ^csi-attacher|awk '{print $1}') || true
  for CSI_ATTACHER in ${LONGHORN_CSI_ATTACHER_POD_LIST};
  do
    kubectl -n longhorn-system delete pod "${CSI_ATTACHER}" --grace-period=0 --force &
    sleep 15
  done

  # restart kube-proxy
  PROXY_PID=$(pgrep kube-proxy)
  if [ -n "${PROXY_PID}" ]; then
    info "Restarting kube-proxy..."
    kill -9 "${PROXY_PID}" || true
  else
    KUBE_PROXY_POD_ID=$(kubectl -n kube-system get pods |grep ^kube-proxy|awk '{print $1}') || true
    kubectl -n kube-system delete pod "${KUBE_PROXY_POD_ID}" --grace-period=0 --force &
    sleep 15
  fi
  sleep 10

  # stop rke2 server
  stop_rke2_server
  sleep 5
  # start rke2 server
  start_rke2_server
}

function progress_bar {
  local progress=$1
  local replica=$2
  local done=${progress}
  local left=$((100-done))

  fill=$(printf "%${done}s")
  empty=$(printf "%${left}s")

  printf "\rRestoreStatus ${replica} : [${fill// /\#}${empty// /-}]${progress}%%"
  echo ""
}

function restore_progress() {
  local url=$1
  local data
  data=$(curl -s "${url}"|jq -c ".restoreStatus[]|{progress: .progress, replica: .replica}")|| true
  if [[ -n "${data}" && "${data}" != "null" && "${data}" != " " ]]; then
    for ele in ${data}; do
      local progress
      local replica
      progress=$(echo "${ele}"| jq -r ".progress")
      replica=$(echo "${ele}"| jq -r ".replica")
      progress_bar "${progress}" "${replica}" || true
    done
  fi
}

function validate_longhorn_pods_health() {
  imagepullbackoffcount=$(kubectl get pods -n longhorn-system | grep -c "CrashLoopBackOff")
  if [[ "${imagepullbackoffcount}" -gt 0 ]]; then
    kubectl get pods -n longhorn-system || true
    error "Few pods in CrashLoopBackOff, plese fix those before running volume restore"
  fi
}

function check_if_all_volumes_backed_up() {
  LONGHORN_URL=$(kubectl -n longhorn-system get svc longhorn-backend -o jsonpath="{.spec.clusterIP}"):9500
  local label="recurring-job-group.longhorn.io/uipath-backup=enabled"
  local volume_list
  volume_list=$(kubectl -n longhorn-system get volumes.longhorn.io -l "${label}" -o json | jq -r ".items[].metadata.name")
  for volume_name in ${volume_list}; do
    get_backup_list_by_pvname "${volume_name}" &>/dev/null
    BACKUP_PATH=$(echo "${PV_BACKUP_PATH}"| jq -r ".url")
    unset PV_BACKUP_PATH
    if [[ -z "${BACKUP_PATH}" ]]; then
      warn "No backup taken for PV ${volume_name}"
    else
      info "Backup present for PV ${volume_name}"
    fi
    sleep 5
  done

  local no_backup_volumes
  label="recurring-job-group.longhorn.io/uipath-backup!=enabled"
  no_backup_volumes=$(kubectl -n longhorn-system get volumes.longhorn.io -l "${label}" -o name  | cut -d '/' -f2)
  backup_disabled_volumes=()
  for vol in ${no_backup_volumes};
  do
    storage_class=$(kubectl get pv "${vol}" -o jsonpath="{.spec.storageClassName}" )
    [[ "${storage_class}" == "longhorn" ]] || backup_disabled_volumes+=("${vol}")
  done
  if [[ "${#backup_disabled_volumes[@]}" == 0 ]];
  then
    info "All the volumes have backup enabled"
  else
    warn "Following volumes do not have backup enabled"
    echo "${backup_disabled_volumes[@]}"
  fi
}