#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

INFRA_DIR="${SCRIPT_DIR}/../"

readonly SCRIPT_DIR
readonly MODULES_DIR
readonly INFRA_DIR

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh
#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/longhorn_helper.sh
#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/disaster_recovery_utils.sh
#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/volume_restore.sh

# kube config file
RESTORE_DIR="/mnt/restore"
export MERGEDINPUTFILE="${SCRIPT_DIR}/merged_infra_input.json"
export NODE_ROLE=${NODE_ROLE:-none}
export RESTORE_VOLUME=${RESTORE_VOLUME:-false}

########
# Parse input json file to set the config values
# Globals:
#   INFILE
# Outputs:
#   Sets global config variables
########
function parse_input_file() {
  set -e
  ETCD_RESTORE_PATH="$(get_mandatory_json_value ".restore.etcdRestorePath")"
  NFS_ENDPOINT="$(get_mandatory_json_value ".restore.nfs.endpoint")"
  NFS_MOUNT_PATH="$(get_mandatory_json_value ".restore.nfs.mountpath")"
  ARGOCD_VERSION="$(get_mandatory_json_value ".fabric.argocd_version")"
  FQDN="$(get_mandatory_json_value ".fqdn")"
  ARGOCD_FQDN="$(get_mandatory_json_value ".subdomain.argo_cd").${FQDN}"
  STORAGE_CLASS="$(get_mandatory_json_value ".storage_class")"
  STORAGE_CLASS_SINGLE_REPLICA="$(get_mandatory_json_value ".storage_class_single_replica")"
  PROXY_ENABLED="$(get_optional_json_value ".proxy.enabled")"
  HTTP_PROXY="$(get_optional_json_value ".proxy.http_proxy")"

  export ARGOCD_VERSION="${ARGOCD_VERSION}"
  export FQDN="${FQDN}"
  export STORAGE_CLASS="${STORAGE_CLASS}"
  export ARGOCD_FQDN="${ARGOCD_FQDN}"
  export PROXY_ENABLED="${PROXY_ENABLED}"
  export HTTP_PROXY="${HTTP_PROXY}"
  export STORAGE_CLASS="${STORAGE_CLASS}"
  export STORAGE_CLASS_SINGLE_REPLICA="${STORAGE_CLASS_SINGLE_REPLICA}"
  set +e
}

########
# Copy backup files
########
function copy_backup_files(){

  info "Copying files to rke2 locations"

  local local_restore_path
  local_restore_path="${RESTORE_DIR}/${ETCD_RESTORE_PATH}/${SETUP_HOSTNAME}"

  if [[ "${NODE_ROLE}" != "agent" ]]; then
    # copy encryption key & longhorn disk files
    mkdir -p /var/lib/rancher/rke2/server/
    cp -r "${local_restore_path}"/cred /var/lib/rancher/rke2/server/
    cp -r "${local_restore_path}"/tls /var/lib/rancher/rke2/server/
  fi

  mkdir -p /etc/rancher/node
  cp -r "${local_restore_path}"/node/* /etc/rancher/node/.

  # find all longhorn disk config longhorn-disk.cfg
  longhorn_disk_paths=$(find "${local_restore_path}"/longhorn_disk_metadata -name longhorn-disk.cfg) || true
  for longhorn_disk_path in ${longhorn_disk_paths}; do
    longhorn_disk_dir="$(dirname "${longhorn_disk_path}")"
    longhorn_disk_dir="${longhorn_disk_dir/"$local_restore_path/longhorn_disk_metadata"/""}"
    mkdir -p "${longhorn_disk_dir}"
    mkdir -p "${longhorn_disk_dir}"/replicas
    cp "${longhorn_disk_path}" "${longhorn_disk_dir}" || true
  done

  info "Copy backup files completed."
}

function restart_pods_with_servicetoken() {
  local secret_list
  info "Cleaning up pods with service token..."
  secret_list=$(kubectl get secrets -o json -A|jq -c ".items[]|select(.type==\"kubernetes.io/service-account-token\")| {name: .metadata.name, namespace: .metadata.namespace}")
  for secret in ${secret_list}; do
    local secret_name
    local secret_namespace
    local pod_list
    secret_name=$(echo "${secret}"|jq -r ".name")
    secret_namespace=$(echo "${secret}"|jq -r ".namespace")
    pod_list=$(kubectl get pods -n "${secret_namespace}" -o json |jq -c ".items[]| {name: .metadata.name, secretname: .spec | select( has (\"volumes\") ).volumes[]| select( has (\"secret\") ).secret| select( .secretName == \"${secret_name}\" )|.secretName}")

    kubectl delete secret "${secret_name}" -n "${secret_namespace}"

    echo "Pod List attached to secret: ${secret_name}: ${pod_list}"
    echo "kubectl get pods -n \"${secret_namespace}\" -o json |jq -c \".items[]| {name: .metadata.name, secretname: .spec | select( has (\"volumes\") ).volumes[]| select( has (\"secret\") ).secret| select( .secretName == \"${secret_name}\" )|.secretName}\""
    for pod in ${pod_list}; do
      local pod_name
      pod_name=$(echo "${pod}"|jq -r ".name")
      kubectl delete pod "${pod_name}" -n "${secret_namespace}" --force
    done
  done
  info "Successfully cleanup for pods."
}

########
# Restore ETCD Snapshots
########
function restore_etcd_snapshot(){
  info "Restoring ETCD for Hostname ${ETCD_RESTORE_PATH} started..."

  # copy rke2 files
  copy_backup_files

  # install rke2
  "${INFRA_DIR}"/infra-installer.sh -i "${INFILE}" -o "${INFRA_DIR}/restore_infra_output.json"

  # stop rke2-server before etcd restore
  stop_rke2_server

  # create dir if not exist
  mkdir -p /var/lib/rancher/rke2/server/db || true
  mkdir -p /var/lib/kubelet || true
  # restore mount path
  mount -a || true

  local local_etcd_restorepath
  local_etcd_restorepath=${RESTORE_DIR}/${ETCD_RESTORE_PATH}/${SETUP_HOSTNAME}

  local latest_snapshot
  local restore_file
  # shellcheck disable=SC2012
  latest_snapshot=$(ls -t1 "${local_etcd_restorepath}/snapshots"|  head -n 1)

  restore_file=${local_etcd_restorepath}/snapshots/${latest_snapshot}
  info "Restore ETCD with snapshot ${latest_snapshot} from ${restore_file} started..."

  if [ ! -f "${restore_file}" ]; then
    error "Restore file not found at ${restore_file}"
  fi

  rke2 server --cluster-reset --cluster-reset-restore-path="${restore_file}"
  rke2-killall.sh
  # verify kubelet killed
  kill_kubelet_containerd
  # restore mount path
  mount -a || true

  start_rke2_server

  sleep 60

  validate_cluster_ready 50 "${NODE_ROLE}"

  disable_pvc_backup

  containercreatingcount=-1
  try=0
  maxtry=60
  while(( try < maxtry )); do
    # kubectl get pods -A | grep -v "Running"| grep -v "Complete"| grep -v "Terminating"; sleep 10
    try=$(( try + 1 ))
    containercreatingcount=$(kubectl get pods -A | grep -c "ContainerCreating")
    info "Total pods in ContainerCreating ${containercreatingcount}, Waiting for all pods to come up... ${try}/${maxtry}"; sleep 10
    if [[ "${containercreatingcount}" -le 2 ]]; then
      break
    fi
  done

  set_longhorn_url "${NODE_ROLE}"
  validate_longhorn_reachable ; sleep 120

  # reset settings for backup
  update_default_settings "backup-target" "" "ignore_failure"
  
  info "Restoring ETCD for Hostname ${SETUP_HOSTNAME} successfull."
}

########
# Validate NFS Server Connection and Mount Path
########
function validate_nfs_server_connection_and_mountpath() {
  local endpoint=$1
  local mount_path=$2

  validate_nfs_server_connection "${endpoint}" "${mount_path}"

  # mount nfs server to client
  mkdir -p ${RESTORE_DIR}

  # backup location is <mount_path>/backup for etcd and data backups
  mount_result=$(mount -t nfs "${endpoint}:${mount_path}/backup" ${RESTORE_DIR} 2>&1)
  if { [ -n "${mount_result}" ] || [ "${mount_result}" == " " ] ; }; then
    error "Mount Path not exposed for nfs server ${endpoint}"
  fi

  local local_etcd_restorepath
  local_etcd_restorepath=${RESTORE_DIR}/${ETCD_RESTORE_PATH}/${SETUP_HOSTNAME}

  # validate if etcd backup exists for restore hostname exists
  if ! { [ -d "${local_etcd_restorepath}" ] ; }; then
    error "Backup for hostname ${SETUP_HOSTNAME} not available in backup store."
  fi

  if [[ "${NODE_ROLE}" != "agent" ]]; then
    # validate all backup folders exits
    if { ! [ -d "${local_etcd_restorepath}/snapshots" ] || [ -z "$(ls -A "${local_etcd_restorepath}"/snapshots/*etcd-snapshot*)" ] ; }; then
      error "ETCD Snapshot Backup for hostname ${SETUP_HOSTNAME} not available in backup store ${local_etcd_restorepath}/snapshots."
    fi

    if { ! [ -d "${local_etcd_restorepath}/cred"  ] || [ -z "$(ls -A "${local_etcd_restorepath}"/cred)" ]; }; then
      error "Cred files for ETCD Snapshot Backup for hostname ${SETUP_HOSTNAME} not available in backup store."
    fi

    if { ! [ -d "${local_etcd_restorepath}/tls"  ] || [ -z "$(ls -A "${local_etcd_restorepath}"/tls)" ]; }; then
      error "TLS Files for ETCD Snapshot Backup for hostname ${SETUP_HOSTNAME} not available in backup store."
    fi
  fi

  info "NFS Server connection & mountpath successfully validated."
}

########
# Restore Cluster with NFS Server
########
function restore_cluster_with_nfs_server(){
  local endpoint=$1
  local mount_path=$2

  info "Start Restoring Cluster for hostname ${SETUP_HOSTNAME}"
  # validate connection with nfs server and mount path
  validate_nfs_server_connection_and_mountpath "${endpoint}" "${mount_path}"

  if [[ "${NODE_ROLE}" == "none" ]]; then
    if [[ "${RESTORE_VOLUME}" == "true"  ]]; then
      # validate and restore PVC's and Deployments
      restore_pvcs_and_dependent_deployments
    else
      # restore etcd snapshot
      restore_etcd_snapshot
    fi
  else
    # copy backup files
    copy_backup_files

    # install rke2
    "${INFRA_DIR}"/infra-installer.sh -i "${INFILE}" -o "${INFRA_DIR}/restore_infra_output.json"
  fi
}

#MAIN
function main() {
  parse_args "$@"
  merge_default_values
  parse_input_file

  export ETCD_RESTORE="true"

  validate_hostname

  # restore cluster
  restore_cluster_with_nfs_server "${NFS_ENDPOINT}" "${NFS_MOUNT_PATH}"

  # unmount restore path
  umount "${RESTORE_DIR}" || true
  info "Cluster ${HOSTNAME} successfully restored..."
}

main "$@"
