#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/longhorn_helper.sh
#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/disaster_recovery_utils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/argocd_utils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/k8sutils.sh
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

########
# Delete PVC Resources
# Input:
# PVC_NAME
# PVC_NAMESPACE
# PV_NAME
# VOLUME_ATTACHMENT_NAME
########
function delete_pvc_resources(){
  local pv_name=$1
  local volumeattachment_name=$2

  info "deleting pv & volumes forcefully"
  delete_pv_forcefully "${pv_name}"
  sleep 2
  delete_longhornvolume_forcefully "${pv_name}"
  sleep 2
  if [ -n "${volumeattachment_name}" ]; then
    info "deleting volumeattachments forcefully"
    delete_volumeattachment_forcefully "${volumeattachment_name}"
    sleep 5
  fi
}

########
# Store PVC Resources
# Input:
# PVC_NAME
# PVC_NAMESPACE
# PV_NAME
# VOLUME_ATTACHMENT_NAME
########
function store_pvc_resources(){
  local pvc_name=$1
  local namespace=$2
  local pv_name=$3
  local volumeattachment_name=$4

  if [[ -n "${volumeattachment_name}" && "${volumeattachment_name}" != " " ]]; then
    result=$(kubectl get volumeattachments "${volumeattachment_name}" -o json > "${LOCAL_RESTORE_PATH}/${volumeattachment_name}".json) || true
    info "${result}"
  fi

  kubectl get pv "${pv_name}" -o json > "${LOCAL_RESTORE_PATH}/${pv_name}"-pv.json
  kubectl -n longhorn-system get volumes.longhorn.io "${pv_name}" -o json > "${LOCAL_RESTORE_PATH}/${pv_name}"-volume.json
  kubectl get pvc "${pvc_name}" -n "${namespace}" -o json > "${LOCAL_RESTORE_PATH}/${pvc_name}"-pvc.json
}

function wait_pvc_bound() {
  local namespace=$1
  local pvc_name=$2

  try=0
  maxtry=30
  while (( try < maxtry )); do
    local status=""
    status=$(kubectl -n "${namespace}" get pvc "${pvc_name}" -o json| jq -r ".status | select( has(\"phase\")).phase")
    if [[ "${status}"  == "Bound" ]]; then
      info "PVC ${pvc_name} Bouned successfully."
      break;
    fi
    info "waiting for PVC to bound...${try}/${maxtry}"; sleep 30
    try=$((try+1))
  done
}

########
# Store PVC Resources
# Input:
# PVC_NAME
# PVC_NAMESPACE
# PV_NAME
# VOLUME_ATTACHMENT_NAME
########
function create_pvc_resources(){
  local pvc_name=$1
  local namespace=$2
  local pv_name=$3
  local volumeattachment_name=$4
  local pvc_uid=""

  kubectl create -f "${LOCAL_RESTORE_PATH}/${pv_name}"-volume.json
  kubectl create -f "${LOCAL_RESTORE_PATH}/${pv_name}"-pv.json

  pvc_uid=$(kubectl -n "${namespace}" get pvc "${pvc_name}" -o json| jq -r ".metadata.uid")

  # update pv with pvc uid
  kubectl patch pv "${pv_name}" --type json -p "[{\"op\": \"add\", \"path\": \"/spec/claimRef/uid\", \"value\": \"${pvc_uid}\"}]"

  wait_pvc_bound "${namespace}" "${pvc_name}"
  info "${result}"
}

function create_volume() {
  local pv_name=$1
  local backup_path=$2
  local response
  create_volume_status="succeed"
  info "Creating Volume with PV: ${pv_name} and BackupPath: ${backup_path}"
  local accessmode
  local replicacount
  accessmode=$(< "${LOCAL_RESTORE_PATH}/${pv_name}-volume.json" jq -r ".spec.accessMode")
  replicacount=$(< "${LOCAL_RESTORE_PATH}/${pv_name}-volume.json" jq -r ".spec.numberOfReplicas")

  # create volume from backup pv
  response=$(curl "${LONGHORN_URL}/v1/volumes" -H 'Accept: application/json' -H 'Content-Type: application/json;charset=UTF-8' --data-raw "{\"name\":\"${pv_name}\", \"accessMode\":\"${accessmode}\", \"fromBackup\": \"${backup_path}\", \"numberOfReplicas\": ${replicacount}}")

  sleep 5

  if [[ -n "${response}" && "${response}" != "null" && "${response}" != " " && -n $(echo "${response}"|jq -r ".id") ]]; then
  # wait for volume to be detached , max wait 4hr
    local try=0
    local maxtry=480
    local success=0
    while (( try < maxtry ));do
      status=$(kubectl -n longhorn-system get volumes.longhorn.io "${pv_name}"  -o json |jq -r ".status.state") || true
      info "volume ${pv_name} status: ${status}"
      if [[ "${status}" == "detached" ]]; then
        # update label
        kubectl -n longhorn-system label volumes.longhorn.io/"${pv_name}" recurring-job-group.longhorn.io/uipath-backup=enabled
        success=1
        info "Volume ready to use"
        break
      fi
      info "waiting for volume to be ready to use${try}/${maxtry}..."; sleep 30
      restore_status_url="${LONGHORN_URL}/v1/volumes/${pv_name}?action=snapshotList"
      restore_progress "${restore_status_url}"
      try=$(( try + 1 ))
    done

    if [ "${success}" -eq 0 ]; then
      create_volume_status="failed"
      error_without_exit "${pv_name} Volume is not ready to use with status ${status}"
    fi
  else
    create_volume_status="failed"
    kubectl create -f "${LOCAL_RESTORE_PATH}/${pv_name}"-volume.json || true
    error_without_exit "${pv_name} Volume creation failed ${response} "
  fi
  echo "${create_volume_status}"
}

function restore_withoutbackupvolume() {
  local pvc_name=$1
  local namespace=$2
  local pv_name=$3
  local volumeattachment_name=$4
  store_pvc_resources "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}"
  sleep 5
  delete_pvc_resources "${pv_name}" "${volumeattachment_name}"
  sleep 5
  create_pvc_resources "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}"
  sleep 5
}

function restore_with_backupvolume() {
  local pvc_name=$1
  local namespace=$2
  local pv_name=$3
  local volumeattachment_name=$4
  local backup_path=$5
  create_volume_status="succeed"
  store_pvc_resources "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}"
  sleep 5
  delete_pvc_resources "${pv_name}" "${volumeattachment_name}"
  sleep 5
  create_volume "${pv_name}" "${backup_path}"
  sleep 10

  kubectl create -f "${LOCAL_RESTORE_PATH}/${pv_name}"-pv.json
  if [[ -n "${create_volume_status}" && "${create_volume_status}" != "succeed" ]]; then
    error_without_exit "Backup volume restore failed for pvc ${pvc_name} in namespace ${namespace}"
    restore_pvc_status="failed"
  else 
    local pvc_uid=""
    pvc_uid=$(kubectl -n "${namespace}" get pvc "${pvc_name}" -o json| jq -r ".metadata.uid")

    # update pv with pvc uid
    kubectl patch pv "${pv_name}" --type json -p "[{\"op\": \"add\", \"path\": \"/spec/claimRef/uid\", \"value\": \"${pvc_uid}\"}]"

    wait_pvc_bound "${namespace}" "${pvc_name}"
    info "${result}"
  fi
  echo ${restore_pvc_status}
}

function restore_pvc(){
  local pvc_name=$1
  local namespace=$2
  local pv_name=$3
  local volumeattachment_name=$4
  local backup_path=$5
  local backup_pvc=${6:-true}
  restore_pvc_status="succeed"
  if [[ -n "${backup_path}" && "${backup_pvc}" == "true" ]]; then
    restore_with_backupvolume "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}" "${backup_path}"
  else
    restore_withoutbackupvolume "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}"
  fi
  echo ${restore_pvc_status}
  # attach_volume "${pv_name}"
}

function attach_volume() {
  local pv_name=$1
  local node_name

  # attach volume to node
  node_name=$(curl "${LONGHORN_URL}/v1/nodes?"|jq -r ".data[0].name")
  curl "${LONGHORN_URL}/v1/volumes/${pv_name}?action=attach" --data-raw "{\"hostId\":\"${node_name}\"}"
}

function get_pvc_resources() {

  get_pvc_resources_resp=""

  local PVC_NAME=$1
  local PVC_NAMESPACE=$2

  # PV have one to one mapping with PVC
  PV_NAME=$(kubectl -n "${PVC_NAMESPACE}" get pvc "${PVC_NAME}" -o json|jq -r ".spec.volumeName")

  VOLUME_ATTACHMENT_LIST=$(kubectl get volumeattachments -n "${PVC_NAMESPACE}" -o=json|jq -c ".items[]|\
{name: .metadata.name, pvClaimName:.spec.source| select( has (\"persistentVolumeName\")).persistentVolumeName}")

  VOLUME_ATTACHMENT_NAME=""
  for VOLUME_ATTACHMENT in ${VOLUME_ATTACHMENT_LIST};
  do
    PV_CLAIM_NAME=$(echo "${VOLUME_ATTACHMENT}"|jq -r ".pvClaimName")
    if [[ "${PV_NAME}" = "${PV_CLAIM_NAME}" ]]; then
      VOLUME_ATTACHMENT_NAME=$(echo "${VOLUME_ATTACHMENT}"|jq -r ".name")
      break;
    fi
  done

  BACKUP_PATH=""
  validate_pv_backup_available "${PV_NAME}"
  local is_backup_available=${IS_BACKUP_AVAILABLE:- 0}
  unset IS_BACKUP_AVAILABLE

  if [ "${is_backup_available}" -eq 1 ]; then
    info "Backup is available for PVC ${PVC_NAME}"

    get_backup_list_by_pvname "${PV_NAME}"
    BACKUP_PATH=$(echo "${PV_BACKUP_PATH}"| jq -r ".url")
    unset PV_BACKUP_PATH
  fi

  get_pvc_resources_resp="{\"pvc_name\": \"${PVC_NAME}\", \"pv_name\": \"${PV_NAME}\", \"volumeattachment_name\": \"${VOLUME_ATTACHMENT_NAME}\", \"backup_path\": \"${BACKUP_PATH}\"}"
  echo "${get_pvc_resources_resp}"
}

function restore_pvc_attached_to_deployments() {
  local namespace="${1}"
  local deployment_list
  # list of all deployment using PVC
  deployment_list=$(kubectl get deployments -n "${namespace}" -o=json | jq -c ".items[] |\
{name: .metadata.name, namespace: .metadata.namespace, replicas: .spec.replicas, pvcList:.spec.template.spec |\
[select( has (\"volumes\") ).volumes[] |\
select( has (\"persistentVolumeClaim\") ).persistentVolumeClaim.claimName]|select(length > 0) }")

  for deployment in ${deployment_list};
  do
    local pvc_restore_failed
    pvc_restore_failed=""
    deployment_name=$(echo "${deployment}"| jq -r ".name")
    pvc_list=$(echo "${deployment}"| jq -r ".pvcList[]")
    namespace=$(echo "${deployment}"| jq -r ".namespace")
    replica=$(echo "${deployment}" | jq -r ".replicas")
    restore_pvc_status=""
    scale_down_deployment "${deployment_name}" "${namespace}"

    for pvc_name in ${pvc_list};
    do
      local pvc_exist
      pvc_exist=$(kubectl -n "${namespace}" get pvc "${pvc_name}") || true
      if [[ -z "${pvc_exist}" || "${pvc_exist}" == " " ]]; then
        warn "PVC not available for the deployment ${deployment}, skipping restore."
        continue;
      fi

      # get pv, volumeattachments for pvc
      get_pvc_resources_resp=""
      get_pvc_resources "${pvc_name}" "${namespace}"

      local pv_name
      local volumeattachment_name
      local backup_path

      pv_name=$(echo "${get_pvc_resources_resp}"| jq -r ".pv_name")
      volumeattachment_name=$(echo "${get_pvc_resources_resp}"| jq -r ".volumeattachment_name")
      backup_path=$(echo "${get_pvc_resources_resp}"| jq -r ".backup_path")

      if [[ -z "${backup_path}" || "${backup_path}" == " " || "${backup_path}" == "null" ]]; then
        error_without_exit "pvc backup path for deployment ${deployment_name} in namespace ${namespace} is empty"
        pvc_restore_failed="error"
        break;
      fi
      restore_pvc_status="succeed"
      restore_pvc "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}" "${backup_path}"

      if [[ -n "${restore_pvc_status}" && "${restore_pvc_status}" != "succeed" ]]; then
        pvc_restore_failed="error"
        break;
      fi
    done

    sleep 10
    scale_up_deployment "${deployment_name}" "${namespace}" "${replica}"
    sleep 5

    if [[ -n "${pvc_restore_failed}" && "${pvc_restore_failed}" == "error" ]]; then
        error "Failed to restore pvc for deployment ${deployment_name} in namespace ${namespace}."
    fi
    
  done
}

function restore_pvc_attached_to_statefulsets() {
  local namespace
  local backup_pvc
  namespace="${1}"
  backup_pvc=${2:-true}
  local statefulset_list

  # list of all statefulset using PVC
  statefulset_list=$(kubectl get statefulset -n "${namespace}" -o=json | jq -c ".items[] |\
{name: .metadata.name, namespace: .metadata.namespace, replicas: .spec.replicas, claimName:.spec |\
select( has (\"volumeClaimTemplates\") ).volumeClaimTemplates[].metadata.name, ownerReferences: .metadata.ownerReferences }")

  for statefulset in ${statefulset_list};
  do
    local statefulset_name
    local claimtemplate_name
    local namespace
    local replica
    local ownerReferences
    local pvc_restore_failed
    pvc_restore_failed=""
    restore_pvc_status=""
    statefulset_name=$(echo "${statefulset}"| jq -r ".name")
    claimtemplate_name=$(echo "${statefulset}"| jq -r ".claimName")
    namespace=$(echo "${statefulset}"| jq -r ".namespace")
    replica=$(echo "${statefulset}" | jq -r ".replicas")
    ownerReferences=$(echo "${statefulset}" | jq -c ".ownerReferences")

    info "Scaling down Statefulset ${statefulset_name} with ${replica} under namespace ${namespace}"
    scale_down_statefulset "${statefulset_name}" "${namespace}" "${ownerReferences}"

    local pvc_prefix
    pvc_prefix="${claimtemplate_name}-${statefulset_name}"
    for((i=0;i<"${replica}";i++))
    do
      local pvc_name
      pvc_name="${pvc_prefix}-${i}"

      pvc_exist=$(kubectl -n "${namespace}" get pvc "${pvc_name}") || true
      if [[ -z "${pvc_exist}" || "${pvc_exist}" == " " ]]; then
        warn "PVC not available for the statefulset ${statefulset_name}, skipping restore."
        continue;
      fi

      # get pv, volumeattachments for pvc
      get_pvc_resources_resp=""
      get_pvc_resources "${pvc_name}" "${namespace}"

      local pv_name
      local volumeattachment_name
      local backup_path
      pv_name=$(echo "${get_pvc_resources_resp}"| jq -r ".pv_name")
      volumeattachment_name=$(echo "${get_pvc_resources_resp}"| jq -r ".volumeattachment_name")
      backup_path=$(echo "${get_pvc_resources_resp}"| jq -r ".backup_path")
      if [[ -z "${backup_path}" || "${backup_path}" == " " || "${backup_path}" == "null" ]]; then
        pvc_restore_failed="error"
        break;
      fi

      restore_pvc_status="succeed"
      restore_pvc "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}" "${backup_path}" "${backup_pvc}"
      if [[ -n "${restore_pvc_status}" && "${restore_pvc_status}" != "succeed" ]]; then
        pvc_restore_failed="error"
        break;
      fi
    done

    sleep 10
    scale_up_statefulset "${statefulset_name}" "${namespace}" "${replica}" "${ownerReferences}"
    sleep 5

    if [[ -n "${pvc_restore_failed}" && "${pvc_restore_failed}" == "error" ]]; then
        error "Failed to restore pvc for Statefulset ${statefulset_name} in namespace ${namespace}."
    fi    
  done
}

# rook-ceph disaster-recovery https://github.com/rook/rook/blob/master/Documentation/ceph-disaster-recovery.md
function restore_rook_ceph_cluster() {

  local namespace="rook-ceph"
  # scale down operator deployment
  scale_down_deployment "rook-ceph-operator" "${namespace}"

  local deployment_list
  # delete all deployment except "rook-ceph-operator" & "rook-ceph-tools"
  deployment_list=$(kubectl -n "${namespace}" get deploy -o json| jq -r ".items[].metadata.name")
  for deployment in ${deployment_list};
  do
    if [[ "${deployment}" != *"rook-ceph-operator"* && "${deployment}" != *"rook-ceph-tools"* ]]; then
      kubectl -n "${namespace}" delete deploy "${deployment}"
    fi
  done

  local pvc_list
  local rookceph_restore_failed
  pvc_list=$(kubectl -n "${namespace}" get pvc -o json|jq -r ".items[].metadata.name")

  rookceph_restore_failed="succeed"
  restore_pvc_status="succeed"

  for pvc_name in ${pvc_list}; do
    # get pv, volumeattachments for pvc
    get_pvc_resources_resp=""
    get_pvc_resources "${pvc_name}" "${namespace}"

    local pv_name
    local volumeattachment_name
    local backup_path

    pv_name=$(echo "${get_pvc_resources_resp}"| jq -r ".pv_name")
    volumeattachment_name=$(echo "${get_pvc_resources_resp}"| jq -r ".volumeattachment_name")
    backup_path=$(echo "${get_pvc_resources_resp}"| jq -r ".backup_path")
    
    if [[ -z "${backup_path}" || "${backup_path}" == " " || "${backup_path}" == "null" ]]; then
      error_without_exit "rook ceph pvc backup path in namespace ${namespace} is empty"
      rookceph_restore_failed="error"
      break;
    fi

    restore_pvc "${pvc_name}" "${namespace}" "${pv_name}" "${volumeattachment_name}" "${backup_path}"

    if [[ -n "${restore_pvc_status}" && "${restore_pvc_status}" != "succeed" ]]; then
      rookceph_restore_failed="error"
      break;
    fi
  done

  # clean osd prepare pods
  kubectl -n rook-ceph delete pods -l "app=rook-ceph-osd-prepare" --force

  scale_up_deployment "rook-ceph-operator" "${namespace}" "1"

  if [[ -n "${rookceph_restore_failed}" &&  "${rookceph_restore_failed}" != "succeed" ]]; then
    error "rook ceph restore failed"
  fi

  local try=0
  local maxtry=30
  local success=0
  while (( try < maxtry )); do
    result=$(kubectl -n "${namespace}" get pod -l app=rook-ceph-osd-prepare) || true
    if [[ -n "${result}" && "${result}" != " " ]]; then
      success=1
      break;
    fi
    info "waiting for rook-ceph-osd-prepare...${try}/${maxtry}"; sleep 30
    try=$(( try + 1 ))
  done

  local try=0
  local maxtry=30
  local success=0
  while (( try < maxtry )); do
    local result
    result=$(kubectl -n rook-ceph get deploy -l "app=rook-ceph-osd") || true
    # clean any pod with corrupt io error
    clean_pod_with_io_error "rook-ceph"
    if [[ -n "${result}" && "${result}" != " " ]]; then
      success=1
      break;
    fi
    info "waiting for deployment rook-ceph-osd...${try}/${maxtry}"; sleep 30
    try=$(( try + 1 ))
  done

  # clean osd pods
  kubectl -n rook-ceph delete pods -l "app=rook-ceph-osd" --force

  local deploy_list
  deploy_list=$(kubectl -n rook-ceph get deploy -l "app=rook-ceph-osd" -o json| jq -r ".items[].metadata.name")

  for deployment_name in ${deploy_list}; do
    scale_up_deployment "${deployment_name}" "${namespace}" "1"
  done
}

function restore_fabric_component() {
  restore_rook_ceph_cluster
  # exclude namespace for component which are handled explicitly
  local namespace_list
  namespace_list=$(kubectl get namespace -o json|jq -r ".items[].metadata| \
select( .name != \"docker-registry\" \
and .name != \"uipath\" \
and .name != \"rook-ceph\" \
and .name != \"redis-system\").name")

  for namespace in ${namespace_list}; do
    restore_pvc_attached_to_deployments "${namespace}"
    restore_pvc_attached_to_statefulsets "${namespace}"
  done
}

function restore_services_component() {
  namespace_list="uipath"
  for namespace in ${namespace_list}; do
    restore_pvc_attached_to_deployments "${namespace}"
    restore_pvc_attached_to_statefulsets "${namespace}"
  done
  # clean any pod with corrupt io error
  clean_pod_with_io_error "uipath"

  # rollout deployments in uipath namespace
  kubectl rollout restart deployment --namespace "uipath"
  if wait_for_all_pods_in_namespace_ready "uipath"; then
    info "Rollback successful"
  else
    error "Rollback of deployments failed."
  fi
}

function restore_infra_component() {
  namespace_list="docker-registry"
  for namespace in ${namespace_list}; do
    restore_pvc_attached_to_deployments "${namespace}"
    restore_pvc_attached_to_statefulsets "${namespace}"
  done
}

########
# Restore PVC & Deployments
########
function restore_pvcs_and_dependent_deployments() {

  set_longhorn_url
  validate_longhorn_reachable; sleep 120
  validate_longhorn_pods_health

  # update settings for backup
  update_default_settings "default-longhorn-static-storage-class" "${STORAGE_CLASS}"
  update_default_settings "backup-target" "nfs://${NFS_ENDPOINT}:${NFS_MOUNT_PATH}"

  # validate longhorn backup reachable
  validate_longhorn_backup_reachable; sleep 120

  # Save restored data in case of failures
  LOCAL_RESTORE_PATH="restoredata"
  mkdir -p ${LOCAL_RESTORE_PATH}
  export LOCAL_RESTORE_PATH="${LOCAL_RESTORE_PATH}"

  restore_infra_component

  # wait for docker registry to come up
  if [[ "${INSTALL_TYPE}" == "offline" ]]; then
    info "Waiting for all pods to pick images from private docker registry..."
    imagepullbackoffcount=-1
    try=0
    maxtry=120
    while(( try < maxtry )); do
      # kubectl get pods -A | grep -v "Running"| grep -v "Complete"| grep -v "Terminating"; sleep 10
      try=$(( try + 1 ))
      imagepullbackoffcount=$(kubectl get pods -A | grep -c "ImagePullBackOff")
      info "Total pods in ImagePullBackOff ${imagepullbackoffcount}, Waiting for all pods to come up... ${try}/${maxtry}"; sleep 30
      if [[ "${imagepullbackoffcount}" -lt 1 ]]; then
        break
      fi
    done

    if [[ "${imagepullbackoffcount}" -ge 1 ]]; then
      error "Failed to restore images for services"
    fi
  fi

  # check argocd is up
  try=0
  maxtry=10
  ready=0
  while(( try < maxtry )); do
    is_argocd_up
    ready="$?"
    if (( ready == 1 )) ; then
      break;
    fi
    info "Waiting for argocd to come up...  ${try}/${maxtry}"; sleep 20
    try=$(( try + 1 ))
  done

  if (( ready == 0 )); then
    error "Failed to start argocd."
  fi

  get_fabric_install_paths "${SCRIPT_DIR}"
  get_argocd_install_params
  login_argocd "${ARGOCD_NODEPORT_HTTP}" "${ARGOCD_FQDN}" "${ARGOCD_PASSWORD_SECRET_NAME}" "${ARGOCD_NAMESPACE}" "argocd"

  restore_fabric_component
  restore_services_component

  rm -rf ${LOCAL_RESTORE_PATH}
  info "PVC and Deployment successfully restored..."
}
