#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

YELLOW='\033[0;33m'
RED='\033[0;31m'
NC='\033[0m' # No Color

function error() {
  echo -e "${RED}[ERROR][$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*${NC}\n" >&2
  exit 1
}

function info() {
  echo "[INFO] [$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*"
}

function warn() {
  echo "${YELLOW}[WARN] [$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*${NC}" >&2
}


########
# Validates whether the argument is empty or not
########
function check_if_empty() {
    local arg_name=$1
    local arg_val=$2

    if [ -z "$arg_val" ]; then
        error "$arg_name argument is missing"
    fi
}


########
# Wait till the rollout is successfull
# Arguments:
#   namespace
#   object_type (deployment, statefulset etc)
########
function wait_till_rollout() {
    local namespace=$1
    local object_type=$2

    check_if_empty namespace "$namespace"
    check_if_empty object_type "$object_type"

    deployments=$(kubectl -n "$namespace" get "$object_type" -o name)
    for i in $deployments; 
    do 
    kubectl -n "$namespace" rollout status "$i" -w --timeout=600s; 
    # shellcheck disable=SC2181
    if [[ "$?" -ne 0 ]]; 
    then
        echo "$i deployment failed in namespace $namespace. Exiting !!!"
        exit 1
    fi
    done
}

########
# Wait till the pods of an particular app are all in Ready status
# Arguments:
#   namespace
#   app_label
########
function wait_for_pod_ready() {
    local namespace=$1
    local app_label=$2

    check_if_empty namespace "$namespace"
    check_if_empty app_label "$app_label"

    for pod in $(kubectl -n "$namespace" get pods -l app="$app_label" -o name); do
        local try=0
        local maxtry=120
        while [[ $(kubectl -n "$namespace" get "$pod" -o 'jsonpath={..status.conditions[?(@.type=="Ready")].status}') != "True" ]]; do
            if [ $try == $maxtry ]; then
                error "Pod: $pod under namespace: $namespace is not in Ready status"
            fi
            echo "Waiting for app: $app_label to be available" && sleep 5;
            try=$((try+1))
        done
    done
}

########
# Label configMap/secret to enable secret/configmap copy to other namespaces
# Arguments:
#   namespace
#   object_type
#   object_name
########
function label_object() {
    local namespace=$1
    local object_type=$2
    local object_name=$3

    check_if_empty namespace "$namespace"
    check_if_empty object_type "$object_type"
    check_if_empty object_name "$object_name"

    kubectl -n "$namespace" label "$object_type" "$object_name" config-discovery=yes 2>/dev/null || true
}


########
# Wait till an resource exists
########
function wait_till_resource_exists() {
    local namespace=$1
    local object_type=$2
    local object_name=$3

    check_if_empty namespace "$namespace"
    check_if_empty object_type "$object_type"
    check_if_empty object_name "$object_name"

    local try=0
    local maxtry=120

    while ! [ "$(kubectl -n "$namespace" get "$object_type" "$object_name" --ignore-not-found -o name | wc -l)" -gt 0 ] ; do
        if [ $try == $maxtry ]; then
        error "Unable to find objectName $object_name of type $object_type under the namespace $namespace"
        fi
        
        info "Waiting for objectName $object_name of type $object_type under the namespace $namespace to become available"
        sleep 5
        try=$((try+1))
    done
}

########
# urlencode <string>
########
function urlencode() {
    local LANG=C
    local length="${#1}"
    for (( i = 0; i < length; i++ )); do
        local c="${1:i:1}"
        case $c in
            [a-zA-Z0-9.~_-]) printf "%s" "$c" ;;
            *) printf '%%%02X' "'$c" ;;
        esac
    done
}

########
# generate_password
########
#shellcheck disable=SC2005
function generate_password() {
    chars='@#$%&_+'
    random_string=$(date +%s | sha256sum | base64 | head -c 14 ; echo)
    random_special_char=${chars:$((RANDOM % ${#chars})):1}
    random_number=$((1 + "$RANDOM" % 9))
    echo "$random_string$random_special_char$random_number"
}

# Set s3cmd context passing following params
# 1 Access key of the user
# 2 Secret key of the user
function setS3CMDContext() {
    OBJECT_GATEWAY_INTERNAL_HOST=$(kubectl -n rook-ceph get services/rook-ceph-rgw-rook-ceph -o jsonpath="{.spec.clusterIP}")
    OBJECT_GATEWAY_INTERNAL_PORT=$(kubectl -n rook-ceph get services/rook-ceph-rgw-rook-ceph -o jsonpath="{.spec.ports[0].port}")
    AWS_ACCESS_KEY=$1
    AWS_SECRET_KEY=$2
    
    # Reference https://rook.io/docs/rook/v1.5/ceph-object.html#consume-the-object-storage
    export AWS_HOST=$OBJECT_GATEWAY_INTERNAL_HOST
    export AWS_ENDPOINT=$OBJECT_GATEWAY_INTERNAL_HOST:$OBJECT_GATEWAY_INTERNAL_PORT
    export AWS_ACCESS_KEY_ID=$AWS_ACCESS_KEY
    export AWS_SECRET_ACCESS_KEY=$AWS_SECRET_KEY
}
