#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

dependencies() {
  local targets=("$@")
  local allFound=true
  for ((i=0; i<${#targets[@]}; i++)); do
    local target=${targets[$i]}
    if [ "$(which "$target")" == "" ]; then
      allFound=false
      echo Not found: "$target"
    fi
  done
  if [ "$allFound" == "false" ]; then
    echo "Please install missing dependencies."
    exit 2
  fi
}

create_ds() {
  kubectl create -n longhorn-system -f ./longhorn/tests/longhorn_test_ds.yaml
}

cleanup() {
  echo "cleaning up..."
  kubectl delete -n longhorn-system -f ./longhorn/tests/longhorn_test_ds.yaml
  echo "clean up complete"
}

wait_ds_ready() {
  while true; do
    local ds
    local numberReady
    local desiredNumberScheduled
    ds=$(kubectl -n longhorn-system get ds/longhorn-environment-check -o json --ignore-not-found)
    numberReady=$(echo "$ds" | jq .status.numberReady)
    desiredNumberScheduled=$(echo "$ds" | jq .status.desiredNumberScheduled)

    if [ "$desiredNumberScheduled" == "$numberReady" ] && [ "$desiredNumberScheduled" != "0" ]; then
      echo "all pods ready ($numberReady/$desiredNumberScheduled)"
      return
    fi

    echo "waiting for pods to become ready ($numberReady/$desiredNumberScheduled)"
    sleep 3
  done
}

validate_ds() {
  local allSupported=true
  local pods
  pods=$(kubectl -n longhorn-system -l app=longhorn-environment-check get po -o json)

  for ((i=0; i<1; i++)); do
    local pod
    local nodeName
    local mountPropagation

    pod=$(echo "$pods" | jq .items[$i])
    nodeName=$(echo "$pod" | jq -r .spec.nodeName)
    mountPropagation=$(echo "$pod" | jq -r '.spec.containers[0].volumeMounts[] | select(.name=="mountpoint") | .mountPropagation')
    if [ "$mountPropagation" != "Bidirectional" ]; then
      allSupported=false
      echo "node $nodeName: MountPropagation DISABLED"
    fi
  done

  if [ "$allSupported" != "true" ]; then
    echo
    echo "  MountPropagation is disabled on at least one node."
    echo "  As a result, CSI driver and Base image cannot be supported."
    echo
    exit 1
  else
    echo -e "\n  MountPropagation is enabled!\n"
  fi
}

main() {
  dependencies kubectl jq mktemp
  trap cleanup EXIT
  create_ds
  wait_ds_ready
  validate_ds
  exit 0
}

main "$@"