local location_service_handler = {}

local json_parser = require ("var.lib.istio.data.json_parser")
local request_handler = require ("var.lib.istio.data.request_handler")

-- Call Location Service to get org and/or tenant location
function location_service_handler.get_location(request_handle, org, tenant, service_type)
    local base_path = "/api/requestrouting/location/"
    local path = base_path..org.."/"..service_type
    if tenant ~= nil then
        path = base_path..org.."/"..tenant.."/"..service_type
    end
    request_handle:logInfo("Location Service call path: "..path)

    local location_response_headers, location_response_body = request_handle:httpCall(
    "outbound|80||platform-location-service.uipath.svc.cluster.local",
    {
        [":method"] = "GET",
        [":path"] = path,
        [":authority"] = "ingressgateway"
    },
    "",
    5000)

    local location_response_code = tonumber(location_response_headers[":status"])
    request_handle:logInfo("Location response status: "..location_response_code)
    request_handle:logInfo("Location response body: "..location_response_body)

    if location_response_code ~= 200 then
        local error_message = "Location request failed with status code: "..location_response_code..", response body: "..location_response_body
        request_handle:logErr(error_message)
        
        local error_redirect_path = "/portal_"
        if location_response_code == 404 then
            error_redirect_path = error_redirect_path.."/unregistered?serviceType="..service_type.."&organizationName="..org
            if tenant ~= nil then
                error_redirect_path = error_redirect_path.."&tenantName="..tenant
            end
        else
            -- Redirect to an error page with messages in contract wtih FE [OR-44416]
            error_redirect_path = error_redirect_path.."/genericerror?errorCode=9003"
        end

        request_handler.redirect_request(request_handle, error_redirect_path)
        return
    end

    local location_response_dto = json_parser.parse_json(location_response_body)

    local organization_id = location_response_dto.organizationId
    local organization_name_in_location = location_response_dto.organizationName
    local tenant_id = location_response_dto.tenantId
    local tenant_name_in_location = location_response_dto.tenantName

    return organization_id, organization_name_in_location, tenant_id, tenant_name_in_location
end

-- Get standardUrlEnabled from Servicedefinition
function location_service_handler.check_standard_url(request_handle, service_type)
    local base_path = "/api/serviceregistry/servicedefinition/"
    local path = base_path..service_type
    request_handle:logInfo("ServiceDefinition call path: "..path)

    local service_definition_response_headers, service_definition_response_body = request_handle:httpCall(
    "outbound|80||platform-location-service.uipath.svc.cluster.local",
    {
        [":method"] = "GET",
        [":path"] = path,
        [":authority"] = "ingressgateway"
    },
    "",
    5000)

    local service_definition_response_code = tonumber(service_definition_response_headers[":status"])
    request_handle:logInfo("ServiceDefinition response status: "..service_definition_response_code)
    request_handle:logInfo("ServiceDefinition response body: "..service_definition_response_body)

    if service_definition_response_code ~= 200 then
        local error_message = "ServiceDefinition request failed with status code: "..service_definition_response_code..", response body: "..service_definition_response_body
        request_handle:logErr(error_message)

        -- Redirect to an error page with messages in contract wtih FE [OR-44416]
        local error_redirect_path = "/portal_/genericerror?errorCode="
        if service_definition_response_code == 404 then
            error_redirect_path = error_redirect_path.."9001"
        else
            error_redirect_path = error_redirect_path.."9002"
        end

        request_handler.redirect_request(request_handle, error_redirect_path)
        return
    end

    local service_definition_response_dto = json_parser.parse_json(service_definition_response_body)

    local service_metadata = service_definition_response_dto.serviceMetadata
    local is_standard_url_enabled = service_metadata.standardUrlEnabled

    return is_standard_url_enabled
end

return location_service_handler
