local request_handler = {}

-- Redirect request to a given path
function request_handler.redirect_request(request_handle, redirect_path)
    request_handle:logErr("Redirecting to ".. redirect_path)
    request_handle:respond(
        {
            [":status"] = "302",
            ["Location"] = redirect_path
        },
        "302 Moved Temporarily"
    )
end

-- Match and extract parameters from standard request path
-- Including org name or guid, and/or tenant name or guid, and service_type, and/or additional path segment
-- Reference: https://www.lua.org/manual/5.3/manual.html#6.4.1
-- %a: represents all letters, A-Z, a-z
-- %w: represents all alphanumeric characters, A-Z, a-z, 0-9
-- Note that there should not be any underscore at the end of org or tenant name
-- Tenant level regex is trying to match /{org}/{tenant}/{}_/
-- Org level regex is trying to match /{org}/{}_/
function request_handler.extract_standard_params(request_handle, path)
    local org, tenant, service_type, path_segment = nil, nil, nil, ""
    local path_segment_index = string.find(path, "_/")
    if path_segment_index == nil then
        path_segment_index = string.find(path, "_%?")
    end
    if path_segment_index ~= nil then
        path_segment = string.sub(path, path_segment_index+1)
        request_handle:logInfo("Extracted path segment: "..path_segment)
        path = string.sub(path, 0, path_segment_index)
        request_handle:logInfo("Extracted path: "..path)
    end
    local tenant_regex = "^/([%w-_]+[%w])/([%w-_]+[%w])/([%a]+)_$"
    local org_regex = "^/([%w-_]+[%w])/([%a]+)_$"
    org, tenant, service_type = string.match(path, tenant_regex)
    if org == nil then
        org, service_type = string.match(path, org_regex)
        if org == nil then
            request_handle:logInfo("Cannot find organizaton in the standard request path")
            return nil, nil, nil, path_segment
        else
            request_handle:logInfo("Detect org/serviceType_ route: ["..org.."/"..service_type.."_]")
            return org, nil, service_type, path_segment
        end
    else
        request_handle:logInfo("Detect org/tenant/serviceType_ route: ["..org.."/"..tenant.."/"..service_type.."_]")
        return org, tenant, service_type, path_segment
    end
end

-- Match and extract parameters from default orchestrator request path
-- Including org name or guid, and tenant name or guid, and/or additional path segment
-- Reference: https://www.lua.org/manual/5.3/manual.html#6.4.1
-- %w: represents all alphanumeric characters, A-Z, a-z, 0-9
-- .: (a dot) represents all characters
-- Note that there should not be any underscore at the end of org or tenant name
-- Regex is trying to match /{org}/{tenant}/{} or /{org}/{tenant}
function request_handler.extract_orch_params(request_handle, path)
    local org, tenant, path_segment = nil, nil, ""
    local regex = "^/([%w-_]+[%w])/([%w-_]+[%w])/(.*)$"
    org, tenant, path_segment = string.match(path, regex)
    if path_segment == nil then
        regex = "^/([%w-_]+[%w])/([%w-_]+[%w])$"
        org, tenant = string.match(path, regex)
        if org == nil then
            request_handle:logInfo("Cannot find organizaton in the default orchestrator request path")
            return nil, nil, nil
        else
            request_handle:logInfo("Detect org/tenant route: ["..org.."/"..tenant.."]")
            return org, tenant, ""
        end
    else
        request_handle:logInfo("Detect org/tenant/segment route: ["..org.."/"..tenant.."/"..path_segment.."]")
        return org, tenant, "/"..path_segment
    end
end

-- Populate org and/or tenant headers
function request_handler.populate_headers(request_handle, organization_id, organization_name_in_location, include_tenant_headers, tenant_id, tenant_name_in_location)
    request_handle:logInfo("Replacing/adding org id to headers: "..organization_id)
    request_handle:headers():replace("AccountID", organization_id)
    request_handle:headers():replace("X-UiPath-Internal-AccountId", organization_id)

    request_handle:logInfo("Replacing/adding org name to headers: "..organization_name_in_location)
    request_handle:headers():replace("X-UiPath-Internal-AccountName", organization_name_in_location)
    
    if include_tenant_headers then
        request_handle:logInfo("Replacing/adding tenant id to headers: "..tenant_id)
        request_handle:headers():replace("TenantID", tenant_id)
        request_handle:headers():replace("X-UiPath-Internal-TenantId", tenant_id)

        request_handle:logInfo("Replacing/adding tenant name to headers: "..tenant_name_in_location)
        request_handle:headers():replace("X-UiPath-Internal-TenantName", tenant_name_in_location)
    end
end

-- Construct outgoing path
function request_handler.construct_outgoing_path(request_handle, is_standard_url_enabled, org, tenant, service_type, path_segment)
    local forwarding_path = nil

    if is_standard_url_enabled == false then
        -- Remove org and tenant name/id and servicetype_ for non-standard url
        -- Outgoing path: /{path_segment}
        forwarding_path = path_segment
    else
        -- Outgoing path: /{org}/{servicetype}_/{path_segment}
        forwarding_path = "/"..org.."/"..service_type.."_"..path_segment

        if tenant ~= nil then
        -- Outgoing path: /{org}/{tenant}/{servicetype}_/{path_segment}
        forwarding_path = "/"..org.."/"..tenant.."/"..service_type.."_"..path_segment
        end
    end

    request_handle:logInfo("Request path is being replaced with: "..forwarding_path)
    request_handle:headers():replace(":path", forwarding_path)
end

return request_handler
