local luaunit = require("luaunit")

local ls_handler = require("src.location_service_handler")
local mock_request_handle = require("mock_request_handle")

function test_get_location()
    local request_handle = mock_request_handle:new(nil)
    request_handle:setUpstreamHeaders({ [":status"] = "200" })
    local org_id = "org_id"
    local upstream_body = "{\"organizationId\":\""..org_id.."\"}"
    request_handle:setUpstreamBody(upstream_body)

    local res_org_id = ls_handler.get_location(request_handle, "org", nil, "aifabic")

    luaunit.assertEquals(res_org_id, org_id)
end

function test_get_location_with_tenant()
    local request_handle = mock_request_handle:new(nil)
    request_handle:setUpstreamHeaders({ [":status"] = "200" })
    local org_id, org_name, tenant_id = "org_id", "org_name", "tenant_id"
    local upstream_body = "{\"organizationId\":\""..org_id.."\",\"organizationName\":\""..org_name.."\",\"tenantId\":\""..tenant_id.."\"}"
    request_handle:setUpstreamBody(upstream_body)

    local res_org_id, res_org_name, res_tenant_id = ls_handler.get_location(request_handle, "org", "tenant", "aifabic")

    luaunit.assertEquals(res_org_id, org_id)
    luaunit.assertEquals(res_org_name, org_name)
    luaunit.assertEquals(res_tenant_id, tenant_id)
end

local function assert_redirect(request_handle, redirect_path)
    luaunit.assertEquals(request_handle:getResponseHeaders():get(":status"), "302")
    luaunit.assertEquals(request_handle:getResponseHeaders():get("Location"), redirect_path)
    luaunit.assertEquals(request_handle:getResponseBody(), "302 Moved Temporarily")
end

function test_get_location_redirect_404()
    local request_handle = mock_request_handle:new(nil)
    request_handle:setUpstreamHeaders({ [":status"] = "404" })
    local org, svc = "org", "aifabic"

    ls_handler.get_location(request_handle, org, nil, svc)

    local redirect_path = "/portal_/unregistered?serviceType="..svc.."&organizationName="..org
    assert_redirect(request_handle, redirect_path)
end

function test_get_location_redirect_404_with_tenant()
    local request_handle = mock_request_handle:new(nil)
    request_handle:setUpstreamHeaders({ [":status"] = "404" })
    local org, tenant, svc = "org", "tenant", "aifabic"

    ls_handler.get_location(request_handle, org, tenant, svc)

    local redirect_path = "/portal_/unregistered?serviceType="..svc.."&organizationName="..org.."&tenantName="..tenant
    assert_redirect(request_handle, redirect_path)
end

function test_get_location_redirect_with_error_code()
    local request_handle = mock_request_handle:new(nil)
    request_handle:setUpstreamHeaders({ [":status"] = "500" })

    ls_handler.get_location(request_handle, "org", nil, "apps")

    local redirect_path = "/portal_/genericerror?errorCode=9003"
    assert_redirect(request_handle, redirect_path)
end

function test_check_standard_url()
    local request_handle = mock_request_handle:new(nil)
    request_handle:setUpstreamHeaders({ [":status"] = "200" })
    local upstream_body = "{\"serviceMetadata\":{\"standardUrlEnabled\":false}}"
    request_handle:setUpstreamBody(upstream_body)

    local res_standardUrlEnabled = ls_handler.check_standard_url(request_handle, "aifabic")

    luaunit.assertEvalToFalse(res_standardUrlEnabled)
end

function test_check_standard_url_redirect_404()
    local request_handle = mock_request_handle:new(nil)
    request_handle:setUpstreamHeaders({ [":status"] = "404" })

    ls_handler.check_standard_url(request_handle, "aifabic")

    local redirect_path = "/portal_/genericerror?errorCode=9001"
    assert_redirect(request_handle, redirect_path)
end

function test_check_standard_url_redirect_with_error_code()
    local request_handle = mock_request_handle:new(nil)
    request_handle:setUpstreamHeaders({ [":status"] = "500" })

    ls_handler.check_standard_url(request_handle, "aifabic")

    local redirect_path = "/portal_/genericerror?errorCode=9002"
    assert_redirect(request_handle, redirect_path)
end

luaunit.LuaUnit:run()
