local luaunit = require("luaunit")

local request_handler = require("src.request_handler")
local mock_request_handle = require("mock_request_handle")

local function assert_redirect(request_handle, redirect_path)
    luaunit.assertEquals(request_handle:getResponseHeaders():get(":status"), "302")
    luaunit.assertEquals(request_handle:getResponseHeaders():get("Location"), redirect_path)
    luaunit.assertEquals(request_handle:getResponseBody(), "302 Moved Temporarily")
end

function test_redirect_request()
    local redirect_path = "/redirect"
    local request_handle = mock_request_handle:new(nil)

    request_handler.redirect_request(request_handle, redirect_path)

    assert_redirect(request_handle, redirect_path)
end

-- Request URL: /org/portal_
function test_extract_standard_params_for_org_empty_seg()
    local org, svc = "org", "portal"
    local path = "/"..org.."/"..svc.."_"
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_tenant, nil)
    luaunit.assertEquals(res_svc, svc)
    luaunit.assertEquals(res_seg, "")
end

-- Request URL: /org/portal_/api1/get
function test_extract_standard_params_for_org()
    local org, svc, seg = "org", "portal", "/api1/get"
    local path = "/"..org.."/"..svc.."_"..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_tenant, nil)
    luaunit.assertEquals(res_svc, svc)
    luaunit.assertEquals(res_seg, seg)
end

-- Request URL: /14DC909F-FD89-4AD2-A31F-08E3C0DBE38B/tenant/aifabric_/api1/get
function test_extract_standard_params_for_org_guid()
    local org, tenant, svc, seg = "14DC909F-FD89-4AD2-A31F-08E3C0DBE38B", "tenant", "aifabric", "/api1/get"
    local path = "/"..org.."/"..tenant.."/"..svc.."_"..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_tenant, tenant)
    luaunit.assertEquals(res_svc, svc)
    luaunit.assertEquals(res_seg, seg)
end

-- Request URL: /org_name/tenant/aifabric_/api1/get
function test_extract_standard_params_with_underscore_in_middle_org_name()
    local org, tenant, svc, seg = "org_name", "tenant", "aifabric", "/api1/get"
    local path = "/"..org.."/"..tenant.."/"..svc.."_"..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_tenant, tenant)
    luaunit.assertEquals(res_svc, svc)
    luaunit.assertEquals(res_seg, seg)
end

-- Request URL: /org/tenant/aifabric_
function test_extract_standard_params_for_tenant_empty_seg()
    local org, tenant, svc = "org", "tenant", "aifabric"
    local path = "/"..org.."/"..tenant.."/"..svc.."_"
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_tenant, tenant)
    luaunit.assertEquals(res_svc, svc)
    luaunit.assertEquals(res_seg, "")
end

-- Request URL: /org/tenant/aifabric_/api1/get
function test_extract_standard_params_for_tenant()
    local org, tenant, svc, seg = "org", "tenant", "aifabric", "/api1/get"
    local path = "/"..org.."/"..tenant.."/"..svc.."_"..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_tenant, tenant)
    luaunit.assertEquals(res_svc, svc)
    luaunit.assertEquals(res_seg, seg)
end

-- Request URL: /org/14DC909F-FD89-4AD2-A31F-08E3C0DBE38B/aifabric_/api1/get
function test_extract_standard_params_for_tenant_guid()
    local org, tenant, svc, seg = "org", "14DC909F-FD89-4AD2-A31F-08E3C0DBE38B", "aifabric", "/api1/get"
    local path = "/"..org.."/"..tenant.."/"..svc.."_"..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_tenant, tenant)
    luaunit.assertEquals(res_svc, svc)
    luaunit.assertEquals(res_seg, seg)
end

-- Request URL: /org/tenant_name/aifabric_/api1/get
function test_extract_standard_params_with_underscore_in_middle_tenant_name()
    local org, tenant, svc, seg = "org", "tenant_name", "aifabric", "/api1/get"
    local path = "/"..org.."/"..tenant.."/"..svc.."_"..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_tenant, tenant)
    luaunit.assertEquals(res_svc, svc)
    luaunit.assertEquals(res_seg, seg)
end

-- Request URL: /org/tenant/apps_?api1/get
function test_extract_standard_params_with_question_mark_separator()
    local org, tenant, svc, seg = "org", "tenant", "apps", "?api1/get"
    local path = "/"..org.."/"..tenant.."/"..svc.."_"..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_tenant, tenant)
    luaunit.assertEquals(res_svc, svc)
    luaunit.assertEquals(res_seg, seg)
end

-- Request URL: /org/apps_?api1/get
function test_extract_standard_params_with_question_mark_separator_org()
    local org, svc, seg = "org", "apps", "?api1/get"
    local path = "/"..org.."/"..svc.."_"..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_svc, svc)
    luaunit.assertEquals(res_seg, seg)
end

-- Request URL: /api/internal/app_state
function test_extract_standard_params_with_non_standard_url_should_return_nil()
    local org, tenant, svc, seg = "api", "internal", "app", "state"
    local path = "/"..org.."/"..tenant.."/"..svc.."_"..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, nil)
    luaunit.assertEquals(res_tenant, nil)
    luaunit.assertEquals(res_svc, nil)
    luaunit.assertEquals(res_seg, "")
end

-- Request URL: /org//tenant/aifabric_/api1/get
function test_extract_standard_params_with_non_alphanumeric_character_at_end_of_org_name_should_return_nil()
    local org, tenant, svc, seg = "org/", "tenant", "aifabric", "/api1/get"
    local path = "/"..org.."/"..tenant.."/"..svc.."_"..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, nil)
    luaunit.assertEquals(res_tenant, nil)
    luaunit.assertEquals(res_svc, nil)
    luaunit.assertNotNil(res_seg, seg)
end

-- Request URL: /org/tenant//aifabric_/api1/get
function test_extract_standard_params_with_non_alphanumeric_character_at_end_of_tenant_name_should_return_nil()
    local org, tenant, svc, seg = "org", "tenant/", "aifabric", "/api1/get"
    local path = "/"..org.."/"..tenant.."/"..svc.."_"..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_svc, res_seg = request_handler.extract_standard_params(request_handle, path)

    luaunit.assertEquals(res_org, nil)
    luaunit.assertEquals(res_tenant, nil)
    luaunit.assertEquals(res_svc, nil)
    luaunit.assertEquals(res_seg, seg)
end

-- Request URL: /org/tenant
function test_extract_orch_params()
    local org, tenant = "org", "tenant"
    local path = "/"..org.."/"..tenant
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_seg = request_handler.extract_orch_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_tenant, tenant)
    luaunit.assertEquals(res_seg, "")
end

-- Request URL: /org/tenant/api1/get
function test_extract_orch_params_with_path_seg()
    local org, tenant, seg = "org", "tenant", "/api1/get"
    local path = "/"..org.."/"..tenant..seg
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_seg = request_handler.extract_orch_params(request_handle, path)

    luaunit.assertEquals(res_org, org)
    luaunit.assertEquals(res_tenant, tenant)
    luaunit.assertEquals(res_seg, seg)
end

-- Request URL: /org
function test_extract_orch_params_with_non_standard_url_should_return_nil()
    local org = "org"
    local path = "/"..org
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_seg = request_handler.extract_orch_params(request_handle, path)

    luaunit.assertEquals(res_org, nil)
    luaunit.assertEquals(res_tenant, nil)
    luaunit.assertEquals(res_seg, nil)
end

-- Request URL: /org//tenant/
function test_extract_orch_params_with_non_alphanumeric_character_at_end_of_org_name_should_return_nil()
    local org, tenant = "org/", "tenant"
    local path = "/"..org.."/"..tenant
    local request_handle = mock_request_handle:new(nil)

    local res_org, res_tenant, res_seg = request_handler.extract_orch_params(request_handle, path)

    luaunit.assertEquals(res_org, nil)
    luaunit.assertEquals(res_tenant, nil)
    luaunit.assertEquals(res_seg, nil)
end

function test_populate_headers()
    local org_id, org_name, tenant_id, tenant_name = "org_id", "org_name", "tenant_id", "tenant_name"
    local request_handle = mock_request_handle:new(nil)

    request_handler.populate_headers(request_handle, org_id, org_name, true, tenant_id, tenant_name)

    luaunit.assertEquals(request_handle:headers():get("AccountID"), org_id)
    luaunit.assertEquals(request_handle:headers():get("X-UiPath-Internal-AccountId"), org_id)
    luaunit.assertEquals(request_handle:headers():get("X-UiPath-Internal-AccountName"), org_name)
    luaunit.assertEquals(request_handle:headers():get("TenantID"), tenant_id)
    luaunit.assertEquals(request_handle:headers():get("X-UiPath-Internal-TenantId"), tenant_id)
    luaunit.assertEquals(request_handle:headers():get("X-UiPath-Internal-TenantName"), tenant_name)
end

function test_construct_outgoing_path_standard_url_enabled_false()
    local seg = "/api1/get"
    local request_handle = mock_request_handle:new(nil)

    request_handler.construct_outgoing_path(request_handle, false, nil, nil, nil, seg)
    
    luaunit.assertEquals(request_handle:headers():get(":path"), seg)
end

function test_construct_outgoing_path_standard_url_enabled_true_with_org()
    local org, svc, seg = "org", "aifabric", "/api1/get"
    local request_handle = mock_request_handle:new(nil)

    request_handler.construct_outgoing_path(request_handle, true, org, nil, svc, seg)

    luaunit.assertEquals(request_handle:headers():get(":path"), ("/"..org.."/"..svc.."_"..seg))
end

function test_construct_outgoing_path_standard_url_enabled_true_with_tenant()
    local org, tenant, svc, seg = "org", "tenant", "aifabric", "/api1/get"
    local request_handle = mock_request_handle:new(nil)

    request_handler.construct_outgoing_path(request_handle, true, org, tenant, svc, seg)

    luaunit.assertEquals(request_handle:headers():get(":path"), ("/"..org.."/"..tenant.."/"..svc.."_"..seg))
end

luaunit.LuaUnit:run()
