#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cluster_utils.sh

function get_argocd_install_params() {
  readonly ARGOCD_NAMESPACE="argocd"
  readonly ARGOCD_NODEPORT_HTTP="$(get_mandatory_json_value ".cluster_port")"
  readonly ARGOCD_BIN_FILENAME="argocd"
  readonly ARGOCD_PASSWORD_SECRET_NAME="argocd-admin-password"
  readonly ARGO_DIR="${VENDOR_DIR}/argocd"
  readonly ARGO_INSTALL_DIR="${ARGO_DIR}/install"

  export ARGOCD_NAMESPACE
  export ARGOCD_NODEPORT_HTTP
  export ARGOCD_BIN_FILENAME
  export ARGOCD_PASSWORD_SECRET_NAME
  export ARGO_DIR
  export ARGO_INSTALL_DIR
}

########
# Install/Update ArgoCD
# Globals:
#   ARGOCD_NAMESPACE, VENDOR_DIR, ARGOCD_NODEPORT_HTTP
########
function install_or_update_argocd() {
    local try=0
    local maxtry=30

    info "Creating ${ARGOCD_NAMESPACE} namespace"
    kubectl create namespace "${ARGOCD_NAMESPACE}" --dry-run=client -o yaml | kubectl apply -f -
    info "Installing ArgoCD installation manifest"

    if kubectl -n "${ARGOCD_NAMESPACE}" get deploy argocd-dex-server &> /dev/null;
    then
      # We do not use argocd-dex-server deployment, so deleting it. 
      # It won't get installed in fresh installations
      kubectl -n "${ARGOCD_NAMESPACE}" delete deploy argocd-dex-server
    fi

    mkdir -p "${ARGO_INSTALL_DIR}"

    if [ "${PROFILE}" == "ha" ]; then
      # shellcheck disable=SC2038
      find "${ARGO_DIR}" -maxdepth 1 -type f ! -name install.yaml | xargs cp -t "${ARGO_INSTALL_DIR}"
      cp -r "${ARGO_DIR}/ha/"* "${ARGO_INSTALL_DIR}/."
      if [ "${ZONE_RESILIENCE}" == "true" ]; then
        ## The backslash before cp command is to ignore alias
        \cp -r "${ARGO_DIR}/zone-resilience/"* "${ARGO_INSTALL_DIR}/."
      fi
    else
      # shellcheck disable=SC2038
      find "${ARGO_DIR}" -maxdepth 1 -type f | xargs cp -t "${ARGO_INSTALL_DIR}"
    fi

    find "${ARGO_INSTALL_DIR}" -maxdepth 1 -type f -print0 | xargs -0 sed -i  "s/##DOCKER_REGISTRY_URL##/${DOCKER_REGISTRY_URL}/g"

    kubectl apply -k "${ARGO_INSTALL_DIR}" -n "${ARGOCD_NAMESPACE}"

    if [[ -d "${ARGO_INSTALL_DIR}" ]]; then
      rm -rf "${ARGO_INSTALL_DIR}"
    fi

    # patch proxy configuration if proxy is enabled
    if [[ "${PROXY_ENABLED}" == "true" ]] ; then
      patch_proxy_configuration
    fi

    ARGOCD_PORT="${ARGOCD_NODEPORT_HTTP}"
    export ARGOCD_PORT

    info "Checking if argocd deployment is ready"
    local status="notready"
    while [[ ${status} == "notready" ]] && (( try != maxtry )); do
        try=$((try+1))
        pod_statuses=$(kubectl get pods \
            -l app.kubernetes.io/name=argocd-server \
            -n "${ARGOCD_NAMESPACE}" \
            -o 'jsonpath={..status.conditions[?(@.type=="Ready")].status}' \
            || true) # allow the retry loop to continue in case of kubeapi transient failures

        for pod_status in ${pod_statuses}; do
            if [[ "${pod_status}" == "True" ]];then
                status="ready"
                break;
            fi
        done

        if [ "${status}" == "ready" ]; then
            info "Successfully installed Argocd."
            break;
        fi

        info "still waiting...${try}/${maxtry}" && sleep 15;
    done
    [[ ${status} == "ready" ]] || error "Argocd deployment failed"
}

function is_argocd_up() {
  local argocd_svc
  local argocd_ip
  local argocd_port

  # get argocd server ip & port
  argocd_svc=$(kubectl -n argocd get svc argocd-server -o=json | jq -c '.spec| {ip: .clusterIP, port: .ports[0].port}')
  argocd_ip=$(echo "${argocd_svc}"| jq -r ".ip")
  argocd_port=$(echo "${argocd_svc}"| jq -r ".port")

  local argocd_url
  local argocd_result
  argocd_url="http://${argocd_ip}:${argocd_port}"

  echo "${argocd_url}"
  argocd_result=$(curl "${argocd_url}/api/version") || true
  if [[ -z "${argocd_result}" || "${argocd_result}" != *"Version"* ]]; then
    return 0
  fi

  return 1
}

function login_argocd_using_cluster_config() {
  local argocd_fqdn
  argocd_fqdn=$(get_cluster_config "ARGOCD_FQDN")
  local argocd_nodeport_http
  argocd_nodeport_http=$(get_cluster_config_default "CLUSTER_PORT" 443)
  
  login_argocd "${argocd_nodeport_http}" "${argocd_fqdn}" "argocd-admin-password" "argocd" "argocd" >/dev/null
}

function login_argocd() {
    local argocd_nodeport_http=$1
    local argocd_hostname=$2
    local argocd_password_secret_name=$3
    local argocd_namespace=$4
    local argocd_binary=$5

    ARGOCD_PORT="${argocd_nodeport_http}"
    export ARGOCD_PORT

    local ARGOCD_ENDPOINT="${argocd_hostname}:${ARGOCD_PORT}"
    export ARGOCD_ENDPOINT

    local argocd_password

    argocd_password=$(kubectl get secrets/"${argocd_password_secret_name}" -n "${argocd_namespace}" --template \{\{.data.password\}\} | base64 -d)

    ARGO_PLAINTEXT=${ARGO_PLAINTEXT:-false}
    if [[ "${ARGO_PLAINTEXT}" == "true" ]];
    then
      "${argocd_binary}" login \
      "${ARGOCD_ENDPOINT}" --plaintext \
      --grpc-web --grpc-web-root-path "/" \
      --username admin --password "${argocd_password}" \
      --insecure || error "Failed to log in"
    else
      "${argocd_binary}" login \
      "${ARGOCD_ENDPOINT}" \
      --grpc-web --grpc-web-root-path "/" \
      --username admin --password "${argocd_password}" \
      --insecure  || error "Failed to log in"
    fi
    info "Login successful!"
}

########
# Download the ArgoCD cli tool
# Globals:
#   BINARY_DIR, ARGOCD_BIN_FILENAME, ARGOCD_VERSION
########
function download_argocd() {
    local argocd_binary=$1
    local argocd_version=$2

    if [ -f "${argocd_binary}" ]; then
        rm -f "${argocd_binary}"
    fi
    info "Downloading Argocd CLI tool"
    wget "https://github.com/argoproj/argo-cd/releases/download/${argocd_version}/argocd-linux-amd64" -O "${argocd_binary}" -q || error "Failed to download ArgoCD"

    chmod +x "${argocd_binary}"
}

function sync_with_wait() {
    local service_name=$1
    local argocd_binary=$2
    local apps_labels=${3:-""}
    local presync_jobs_label=${4:-""}
    local namespace=${5:-uipath}
    MISSING_DEGRADED_APP_LIST=""

    info "Syncing ${service_name}..."
    "${argocd_binary}" app terminate-op "${service_name}" >/dev/null 2>&1 || true
    "${argocd_binary}" app wait "${service_name}" --operation >/dev/null 2>&1 || true # Wait for last terminate operation to finish
    "${argocd_binary}" app sync "${service_name}" --prune >/dev/null 2>&1 || true

    if [[ ${presync_jobs_label} != "" ]];
    then
      # Fail if presync validations failed
      local try=0
      local maxtry=3
      local status="notready"
      while [[ ${status} == "notready" ]]  && (( try != maxtry )); do
          try=$((try+1))
          local jobs=""
          info "Looking for presync checks...${try}/${maxtry}"
          jobs=$(kubectl -n "${namespace}" get job -l "${presync_jobs_label}" \
            -o jsonpath='{.items[*].metadata.name}' \
            --ignore-not-found)
          if [[ ${jobs} != "" ]]; then
              info "Checking the presync validation jobs: ${jobs}...${try}/${maxtry}"
              if kubectl wait -n "${namespace}" --for=condition=complete \
              --timeout=120s job -l "${presync_jobs_label}" 2>/dev/null; then
                status="ready"
              else
                status="failed"
              fi
          elif (( try ==  maxtry )); then
              status="notfound"
          fi
          sleep 10
      done
      if [[ ${status} != "notfound" ]];
      then
        kubectl -n "${namespace}" logs -l "${presync_jobs_label}" --tail=-1 || true
        [[ ${status} == "failed" ]] && error "Presync checks failed"
        [[ ${status} == "ready" ]] && info "Presync checks succeeded"
      else
        info "No presync validation jobs found"
      fi
    fi

    # Wait for services to become healthy
    status="notready"
    declare -a wait_args
    local try=0
    local maxtry=1
    local wait_args
    local max_wait_time=2400
    local start_timestamp
    local current_timestamp

    start_timestamp="$(date +%s)"
    while (( try <= maxtry )); do
      current_timestamp="$(date +%s)"
      wait_time=$(( max_wait_time - (current_timestamp - start_timestamp) ))
      wait_args+=(--timeout "${wait_time}")
      wait_args+=(--health)

      [[ ${apps_labels} != "" ]] && wait_args+=(-l "${apps_labels}")
      info "Waiting for ${service_name} app to become healthy...try ${try}/${maxtry}"
          "${argocd_binary}" app wait "${service_name}"\
              "${wait_args[@]}" >/dev/null 2>&1 \
          && status=ready && break
      sleep 20
      try=$((try+1))
    done
    
    [[ ${status} == "notready" ]] && get_apps_with_missing_or_degraded_state "${argocd_binary}"
    if [[ -n "${MISSING_DEGRADED_APP_LIST}" ]]; then
      info "################## Failed Component ##################"
      get_argocd_error_events "${argocd_binary}" "${MISSING_DEGRADED_APP_LIST}"
      error "Failed to sync ${service_name}"
    fi
}

function wait_for_sync_and_health() {
  local app_name=$2
  local argocd_binary=$1
  local wait_args
  local status="notready"
  wait_args+=(--timeout 2400)
  wait_args+=(--health)
  wait_args+=(--sync)
  info "Waiting for ${app_name} app to become healthy.."
  "${argocd_binary}" app wait "${app_name}"\
      "${wait_args[@]}" >/dev/null 2>&1 \
  && status=ready

  [[ ${status} == "ready" ]] && return 0

  return 1
}

function get_apps_with_missing_or_degraded_state() {
  local argocd_binary_path=$1
  local app_downgraded_or_missing_list
  app_downgraded_or_missing_list=$("${argocd_binary_path}" app list -o json | jq -c ".[]|{name: .metadata.name, status: .status.health| select((.status == \"Missing\") or (.status == \"Degraded\")).status}")
  export MISSING_DEGRADED_APP_LIST="${app_downgraded_or_missing_list}"
}

function get_argocd_error_events() {
    local argocd_binary_path=$1
    local app_downgraded_or_missing_list=$2
    for apps in ${app_downgraded_or_missing_list}; do
      app_name=$(echo "${apps}"| jq -r ".name")
      status=$(echo "${apps}"| jq -r ".status")
      info "====================================================="
      info "Application ${app_name} went into status ${status}"
      "${argocd_binary_path}" app get "${app_name}"
      info "====================================================="
      info "run 'kubectl get events -n namespace' to get more info"
    done
}

function update_argocd_trusted_certs() {
  local fqdn=$1
  local argocd_server_pod
  local docker_registry_namespace="docker-registry"
  local docker_registry_secret="docker-registry-tls"

  if ! docker_registry_secret_json=$(kubectl get secret "${docker_registry_secret}" --namespace "${docker_registry_namespace}" -o json); then
    error "docker-registry tls certificate not found"
  fi

  tls_crt=$(echo "${docker_registry_secret_json}" | jq -r '.data."tls.crt"' | base64 -d)

  if argocd cert list --cert-type https | grep "${fqdn}"; then
    argocd cert rm --cert-type https "${fqdn}"
  fi

  info "Adding offline docker-registry cert to argocd"
  echo "$tls_crt" | argocd cert add-tls "${fqdn}" || error "Failed to add docker-registry-tls certificate to argocd"

  argocd cert list --cert-type https \
      | grep -q "${fqdn}" \
      || error "Can't find added certificate for ${fqdn}"

  local try=0
  local maxtry=30
  local status="notready"
  while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
          try=$((try+1))
          argocd_server_pod=$(kubectl get po -l app.kubernetes.io/name=argocd-server -A -o json | jq -r '.items[0].metadata.name') || true
          kubectl exec -n argocd "${argocd_server_pod}" -- \
              bash -c "ls /app/config/tls" \
              | grep -q "${fqdn}" \
              && status="ready"
          info "Checking if the cert has been mounted to pods...${try}/${maxtry}" &&  sleep 10;
  done
  [[ ${status} == "ready" ]] || error "Added cert was not mounted to the argocd pods"
}

# Get argo app status or fail after a few retries
function get_argo_app_status() {
  local app="$1"
  local app_status=""
  local try=0
  local maxtry=200

  local status="notready"
  while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
          try=$((try+1))
          argocd app get "${app}" -o json &>/dev/null || { status="notready"; sleep 6; continue; }
          status="ready"
          app_status="$(argocd app get "${app}" -o json 2>/dev/null | jq -r '.status.health.status // ""')"
  done

  [[ ${status} == "notready" ]] && error "Failed to get status for argocd app: ${app}"

  echo -n "${app_status}"
}
