#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

function generate_v3_file() {
    local v3_file=$1
    local fqdn=$2
    cat <<EOF > "${v3_file}"
authorityKeyIdentifier=keyid,issuer
basicConstraints=CA:FALSE
keyUsage = digitalSignature, nonRepudiation, keyEncipherment, dataEncipherment
extendedKeyUsage = serverAuth
subjectAltName = @alt_names
[alt_names]
DNS.1 = ${fqdn}
DNS.2 = *.${fqdn}
EOF
}


function create_self_issued_certificate(){
    local fqdn=$1
    local root_ca_cert_file="$2"
    local root_ca_key_file="$2.key"
    local tls_cert_file="$3"
    local tls_key_file="$4"
    local root_ca_subj="/C=US/ST=NY/O=UiPath, Inc./CN=UiPath AS Root CA"
    local tls_cert_subj="/C=US/ST=NY/O=UiPath, Inc./CN=UiPath Automation Suite"
    local key_length=2048
    local ca_expiry=365
    local tls_expiry=90

    local pass
    pass=$(tr </dev/urandom -dc _A-Z-a-z-0-9 | head -c 16)

    CERT_DIR="$(dirname "${root_ca_cert_file}")"
    [[ -d ${CERT_DIR} ]] || mkdir -p "${CERT_DIR}"

    local tls_csr_file="${CERT_DIR}/server.csr"
    local v3_file="${CERT_DIR}/v3.txt"

    # Generate certificate authority key
    openssl genrsa -des3 -passout pass:"$pass" $key_length > "$root_ca_key_file" 2> /dev/null
    
    # Generate certificate authority bundle
    openssl req -x509 -new -nodes -key "$root_ca_key_file" -sha256 -days $ca_expiry  -out "$root_ca_cert_file" -subj "$root_ca_subj" -passin pass:"$pass" 2> /dev/null

    # Generate v3 file to add SAN and serverAuth for extendedKeyUsage (iOS support)
    generate_v3_file "$v3_file" "$fqdn" 

    # Generate certificate signing request and tls private key
    openssl req -new -nodes -out "$tls_csr_file" -newkey rsa:$key_length -keyout "$tls_key_file" -subj "$tls_cert_subj" 2> /dev/null

    # Generate tls certificate
    openssl x509 -req -in "$tls_csr_file" -CA "$root_ca_cert_file" -CAkey "$root_ca_key_file" -CAcreateserial -out "$tls_cert_file" -days $tls_expiry -sha256 -extfile "$v3_file" -passin pass:"$pass" 2> /dev/null

    info "TLS certificate generated..."
}

function create_token_signing_certificate(){
    local cert_file="$1"
    local key_file="$2"
    local out_path="$3"
    local pass="$4"

    if ! openssl pkcs12 -export -out "$out_path" -inkey "$key_file" -in "$cert_file" --passout pass:"$pass" 2> /dev/null; then
        error "Unable to create token signing certificate"
    fi

    info "Token signing certificate generated..."
}

function validate_certificate_key_match() {
  local cert_file=$1
  local cert_key=$2
  
  cert_checksum=$(openssl x509 -noout -modulus -in "${cert_file}" 2> /dev/null | openssl md5)
  key_checksum=$(openssl rsa -noout -modulus -in "${cert_key}" 2> /dev/null | openssl md5)

  if [[ "${cert_checksum}" == "${key_checksum}" ]]; then 
    echo ""
  else
    echo "Certificate and Key checksum is mismatched."
  fi  
}

########
# $1 should be tls.crt file
# $2 should be ca.crt file
########
function validate_tls_cert_issuer_and_ca_subject_hash_match() {
  local tls_cert_file=$1
  local ca_cert_file=$2
  local tmp_dir
  tmp_dir=$(mktemp -d certificates.XXXXXXX)
  
  tls_issuer_hash=$(openssl x509 -in "${tls_cert_file}" -noout --issuer_hash 2> /dev/null ) 
  #shellcheck disable=SC2002
  cat "${ca_cert_file}" | awk -v folder="$tmp_dir" '/BEGIN CERTIFICATE/ { i++; } /BEGIN CERTIFICATE/, /END CERTIFICATE/ { print > folder "/cert-" i ".crt" }'
  crt_array=()
  while IFS='' read -r crt; do crt_array+=("$crt"); done <<< "$(ls -d "${tmp_dir}"/*.crt 2> /dev/null)"
  
  for path in "${crt_array[@]}"
  do
    ca_subject_hash=$(openssl x509 -in "${path}" -noout --subject_hash 2> /dev/null) 
    if [[ "${tls_issuer_hash}" == "${ca_subject_hash}" ]]; then
      echo ""  
      rm -rf "$tmp_dir"
      return
    fi
  done
  
  rm -rf "$tmp_dir"
  echo "${tls_cert_file}'s issuer_hash does not match with ${ca_cert_file}'s subject_hash"
}

function validate_certificate_rootca_match() {
  local cert_file=$1
  local ca_cert=$2
  
  cert_hash=$(openssl x509 -in "${cert_file}" -noout -issuer_hash 2> /dev/null ) 
  ca_hash=$(openssl x509 -in "${ca_cert}" -noout -issuer_hash 2> /dev/null) 

  if [[ "${cert_hash}" == "${ca_hash}" ]]; then
    echo "" 
  else 
    echo "Certificate and RootCA hash is mismatched."
  fi
}

function validate_certificate_expiry() {
  local cert_file=$1
  expiry=$(openssl x509 -in "${cert_file}" -noout -checkend 86400 2> /dev/null)

  if [[ "${expiry}" == "Certificate will not expire" ]]; then
    echo ""
  else
   echo "Ceritifcate already expired or will expire within 24hours"
  fi
}

function validate_certificate_key_length() {
  local cert_key=$1
  length=$(openssl rsa -in "${cert_key}" -text -noout 2> /dev/null | grep "Private-Key" | awk '{print $3}' | sed -e 's/(//g')

  if [[ ${length} -ge 2048 ]]; then
    echo ""
  else
    echo "Ceritifcate Key should be minimum of 2048, current length is ${length}"
  fi
}

function validate_certificate_san() {
  local cert_file=$1
  local fqdn=$2
  
  server_dns=$(openssl x509 -in "${cert_file}" -noout -text 2> /dev/null | grep "DNS:" | sed -e "s/DNS://g" | sed -e "s/[[:space:]]//g")
  if grep -q "${fqdn}" <<< "${server_dns}"; then
    echo ""
  else
    echo "Certificate ${cert_file} doesn't have ${fqdn} in SAN"
  fi
}

########
# $1 should be crt file path
########
function validate_pem_format() {
  pem_file_path=$1
  openssl x509 -in "${pem_file_path}" -noout -text >/dev/null 2>&1 && echo "" || echo "${pem_file_path} is not in pem format"
}

########
# $1 should be tls crt file path
########
function validate_tls_has_one_certificate() {
  pem_file_path=$1
  local count
  count="$(grep -o "BEGIN CERTIFICATE" "${pem_file_path}" | wc -l)"
  [[ "${count}" == 1 ]] && echo "" || echo "${pem_file_path} should have only one certificate, not chain of certificate"
}

function validate_server_certificate() {

  local ca_file_path=$1
  local cert_file_path=$2
  local key_file_path=$3
  local fqdn=$4
  local argo_fqdn=$5
  local objectstore_fqdn=$6
  local registry_fqdn=$7
  local monitoring_fqdn=$8
  local insights_fqdn=$9

  if [[ ! -f "${ca_file_path}" ]]; then
      server_cert_error+=("CA Bundle Certificate file not found: ${ca_file_path}")
  fi

  if [[ ! -f "${cert_file_path}" ]]; then
      server_cert_error+=("Server Certificate file not found: ${cert_file_path}")
  fi

  if [[ ! -f "${key_file_path}" ]]; then
      server_cert_error+=("Server Certificate Private Key file not found: ${key_file_path}")
  fi

  [[ ${#server_cert_error[@]} -ne 0 ]] && failure_output "Validating Certificate..... failed" "${server_cert_error[@]}"
  
  response=$(validate_pem_format "${cert_file_path}")
  [[ -n ${response} ]] && server_cert_error+=("${response}")

  response=$(validate_pem_format "${ca_file_path}")
  [[ -n ${response} ]] && server_cert_error+=("${response}")

  response=$(validate_certificate_key_match "${cert_file_path}" "${key_file_path}")
  [[ -n ${response} ]] && server_cert_error+=("${response}")
  
  # response=$(validate_certificate_rootca_match "${cert_file_path}" "${ca_file_path}")
  # [[ -n ${response} ]] && server_cert_error+=("${response}")
  
  response=$(validate_certificate_expiry "${cert_file_path}")
  [[ -n ${response} ]] && server_cert_error+=("${response}")
  
  response=$(validate_certificate_key_length "${key_file_path}")
  [[ -n ${response} ]] && server_cert_error+=("${response}")
  
  response=$(validate_tls_has_one_certificate "${cert_file_path}")
  [[ -n ${response} ]] && server_cert_error+=("${response}")
  
  response=$(validate_tls_cert_issuer_and_ca_subject_hash_match "${cert_file_path}" "${ca_file_path}")
  [[ -n "${response}" ]] && server_cert_error+=("${response}")
  
  response=$(validate_certificate_san "${cert_file_path}" "${fqdn}")
  [[ -n ${response} ]] && server_cert_error+=("${response}")

  response=$(validate_certificate_san "${cert_file_path}" "*.${fqdn}")
  [[ -n ${response} ]] && wild_card_san_error+=("${response}")
  
  if [[ -n "${response}" ]]; then
     response=$(validate_certificate_san "${cert_file_path}" "${argo_fqdn}")
     [[ -n ${response} ]] && subdomain_san_error+=("${response}")

     response=$(validate_certificate_san "${cert_file_path}" "${objectstore_fqdn}")
     [[ -n ${response} ]] && subdomain_san_error+=("${response}")

    [[ -n "${registry_fqdn}" ]] && response=$(validate_certificate_san "${cert_file_path}" "${registry_fqdn}") || response=""  
    [[ -n ${response} ]] && subdomain_san_error+=("${response}")

     response=$(validate_certificate_san "${cert_file_path}" "${monitoring_fqdn}")
     [[ -n ${response} ]] && subdomain_san_error+=("${response}")

     response=$(validate_certificate_san "${cert_file_path}" "${insights_fqdn}")
     [[ -n ${response} ]] && subdomain_san_error+=("${response}")
  fi

  if [[ ${#subdomain_san_error[@]} -ne 0 && ${#wild_card_san_error[@]} -ne 0 ]]; then
    server_cert_error+=("${subdomain_san_error[@]}")
    server_cert_error+=("${wild_card_san_error[@]}")
  fi

  if [[ -n $UPDATE_FQDN && "$UPDATE_FQDN" == "1" ]];
  then [[
     ${#server_cert_error[@]} -ne 0 ]] && warn_output "Validating Certificate..... failed" "${server_cert_error[@]}" && return
  else
    [[ ${#server_cert_error[@]} -ne 0 ]] && failure_output "Validating Certificate..... failed" "${server_cert_error[@]}"
  fi

  # donot remove this line, as calling function is dependent on this line to perform futher steps
  echo -n "success"
}

function validate_signing_certificate(){
    local cert_file=$1
    local pass=$2

    if [[ ! -f "${cert_file}" ]]; then
      token_cert_error+=("token signing certificate file not found: ${cert_file}")
    fi

    [[ ${#token_cert_error[@]} -ne 0 ]] && failure_output "Validating Token Signing Certificate..... failed" "${token_cert_error[@]}"

    if token_cert=$(openssl pkcs12 -in "$cert_file" -nodes --passin pass:"$pass" 2> /dev/null); then
        expiry=$(echo "${token_cert}" | openssl x509 -noout -checkend 86400)
        [[ "${expiry}" != "Certificate will not expire" ]] && token_cert_error+=("Ceritifcate already expired or will expire within 24hours")
    else
        token_cert_error+=("Password is invalid for certificate")
    fi

    [[ ${#token_cert_error[@]} -ne 0 ]] && failure_output "Validating Token Signing Certificate..... failed" "${token_cert_error[@]}"

    info "Validating Token Signing Certificate..... DONE"
}
