#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
ETCDCTL_CACERT="${ETCDCTL_CACERT:-/var/lib/rancher/rke2/server/tls/etcd/server-ca.crt}"
ETCDCTL_CERT="${ETCDCTL_CERT:-/var/lib/rancher/rke2/server/tls/etcd/client.crt}"
ETCDCTL_KEY="${ETCDCTL_KEY:-/var/lib/rancher/rke2/server/tls/etcd/client.key}"
ETCDCTL_BIN="${ETCDCTL_BIN_PATH}"
LONGHORN_NAMESPACE="longhorn-system"
LOCK_CREATED="false"

readonly SCRIPT_DIR
readonly ETCDCTL_CACERT
readonly ETCDCTL_CERT
readonly ETCDCTL_KEY
readonly LONGHORN_NAMESPACE

#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/utils.sh

export KUBECONFIG=/etc/rancher/rke2/rke2.yaml
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/bin:/usr/local/bin

trap on_exit 0

function on_exit() {
  [[ $LOCK_CREATED == "true" ]] && single_instance_lock "delete"
  [[ $display_usage_check == "true" ]] && return
  if [[ $exit_on_failure == "true" ]] ; then
    error "Node Removal script failed"
  else
    info "Node Removal script executed successfully"
  fi
}


function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Rotate kubernetes cluster nodes"
  echo
  echo "Arguments"
  echo "  -n|--name                          [Required] : Comma separated list of nodes to remove from the cluster"
  echo "  --ignore-missing-tainted            Ignore if a replacement tainted node is not available in the cluster"
  echo "  --ignore-missing-zone               Ignore if a replacement node is not available in the same zone as provided nodes"
  echo "  -d|--debug                          Increase logging verbosity to show all debug logs"
  echo "  -h|--help                           Show this help message and exit"
  echo
  echo "Examples"
  echo "  Remove node: server0 , server1 and server2 from k8s cluster"
  echo "  ${COMMAND_NAME} --nodes \"server0,server1,server2\""
  echo
  echo "***************************************************************************************"
  echo
  display_usage_check="true"
}

function parse_long_args() {
  # IGNORE_MISSING_NODES="false"
  IGNORE_MISSING_TAINTED_NODES="false"
  IGNORE_MISSING_ZONE_NODE="false"
  SKIP_NODE_DELETION="false"
  while (("$#")); do
    case "$1" in
    -n|--name)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        NODES_TO_REMOVE="$2"
        shift 2
      else
        error "Value for $1 is missing"
        shift 2
      fi
      ;;

    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    --ignore-missing-tainted)
      warn "Ignoring missing tainted nodes"
      IGNORE_MISSING_TAINTED_NODES="true"
      shift
      ;;

    --skip-node-deletion)
      warn "Skipping node deletion"
      SKIP_NODE_DELETION="true"
      shift
      ;;

    --ignore-missing-zone)
      warn "Ignoring missing zone nodes"
      IGNORE_MISSING_ZONE_NODE="true"
      shift
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      exit 1
      ;;
    esac
  done
  [[ -z $NODES_TO_REMOVE ]] && error "Error: Required argument  -n|--nodes missing"
  NODES_TO_REMOVE_JSON_ARRAY=$(jq -c -n --arg nodes "${NODES_TO_REMOVE}" '[$nodes | split(",") | .[] | gsub("^\\s+|\\s+$";"")]')
}


function get_node_type() {
  local node_name="$1"
  local all_nodes_data="$2"

  node_type=$(echo "${all_nodes_data}" | jq -r --arg name "${node_name}" '.items[] | select(.metadata.name==$name) | .metadata.labels["node-role.kubernetes.io/master"]')
  if [[ -n "${node_type}" ]] ; then
    echo "${node_type}"
  else
    error "Not able to determine if node: ${node_name} is server or agent. Please check if the node was removed"
  fi
}

function get_node_zone() {
  local node_name="$1"
  local all_nodes_data="$2"

  node_zone=$(echo "${all_nodes_data}" | jq -r --arg name "${node_name}" '.items[] | select(.metadata.name==$name) | .metadata.labels["topology.kubernetes.io/zone"]')
  if [[ -n "${node_zone}" ]] ; then
    echo "${node_zone}"
  else
    error "Not able to determine zone config of node: ${node_name}. Please check if the node was removed"
  fi
}

function validate_node_zone() {
  local type="$1"
  local node_name="$2"
  local all_nodes_data="$3"
  local remaining_nodes_data="$4"
  local node_type
  local node_zone


  node_type=$(get_node_type "${node_name}" "${all_nodes_data}")
  [[ $node_type == "null" ]] && return 0
  # [[ $type == "node" ]] && is_node_available=$(echo "$remaining_nodes_data" | jq --arg node_type "${node_type}"  '[.items[] | select((.metadata.labels["node-role.kubernetes.io/master"] | tostring) == $node_type).metadata.name] | length')
  [[ $type == "zone" ]] && node_zone=$(get_node_zone "${node_name}" "${all_nodes_data}")
  [[ $type == "zone" ]] && is_node_available=$(echo "$remaining_nodes_data" | jq --arg node_type "${node_type}"  --arg node_zone "${node_zone}"  '[.items[] | select((.metadata.labels["node-role.kubernetes.io/master"] | tostring) == $node_type and (.metadata.labels["topology.kubernetes.io/zone"] | tostring) == $node_zone).metadata.name] | length')
  if [[ "$is_node_available" -gt 0 ]] ; then
    return 0
  else
    # [[ $type == "node" ]] && echo "No node would be present with same role if we remove node: ${node_name}"
    [[ $type == "zone" ]] && echo "No node would be present with same role in the zone if we remove node: ${node_name}"
    return 1
  fi
}


function get_node_taint() {
  local node_name="$1"
  local all_nodes_data="$2"

  node_taints=$(echo "${all_nodes_data}" | jq -r --arg name "${node_name}" '.items[] | select(.metadata.name==$name) | (try ([.spec.taints[] | select(((.effect=="NoSchedule" and .key=="node.kubernetes.io/unschedulable") or (.effect=="NoSchedule" and .key=="node.kubernetes.io/unreachable") or (.effect=="NoExecute" and .key=="node.kubernetes.io/unreachable"))  |not)]) // []) as $a | if (($a | length) > 0) then ($a | sort_by(.key)) else "null"  end')
  if [[ -n "${node_taints}" ]] ; then
    echo "${node_taints}"
  else
    error "Not able determine taints of node: ${node_name}. Please check if the node was removed"
  fi
}


function validate_node_taints() {
  local node_name="$1"
  local all_nodes_data="$2"
  local remaining_nodes_data="$3"
  local node_with_same_taint_found

  node_taints=$(get_node_taint "${node_name}" "${all_nodes_data}")
  [[ $node_taints == "null" ]] && return 0

  node_array=$(echo "$remaining_nodes_data" | jq -r '.items[] | .metadata.name')
  node_with_same_taint_found="false"
  while IFS=$'\n' read -r node_name_iter
  do
    node_taints_r=$(get_node_taint "${node_name_iter}" "${all_nodes_data}")
    if [[ $node_taints_r != "null" ]] ; then
      if diff  <(jq '.' <<< "$node_taints") <(jq '.'  <<< "${node_taints_r}") >> /dev/null ; then
        node_with_same_taint_found="true"
        break
      fi
    fi
  done <<< "${node_array}"

  if [[ $node_with_same_taint_found == "false" ]] ; then
    error_without_exit "No node would be present with same taints configs if we remove ${node_name}"
    return 1
  fi
}

function validate_server_count() {
  local remaining_nodes_data="$1"
  local remaining_server_count
  local remainder

  remaining_server_count=$(echo "$remaining_nodes_data" | jq -r  '[.items[] | select(.metadata.labels["node-role.kubernetes.io/master"] == "true").metadata.name] | length')
  if [[ $remaining_server_count -ge 3 ]] ; then
    remainder=$((remaining_server_count%2))
    if [[ $remainder -eq 1 ]] ; then
      return 0
    else
      error_without_exit "Expecting odd number of server nodes after removal of provided nodes, Found even($remaining_server_count)"
      return 1
    fi
  else
    error_without_exit "Expecting server nodes >= 3 after removal of provided nodes, found only $remaining_server_count"
    return 1
  fi
}

function get_k8s_node_resource(){
  local resource="$1"
  local resource_type="$2"
  local node_name="$3"
  local node_stat
  local returnCode

  case "$resource_type" in
    allocated)
      case "$resource" in
        cpu)
          node_stat=$(kubectl describe node "${node_name}" | sed -n '/Allocated resources:/,/Events:/p' | grep "${resource}" | awk '{print $2}' | sed -r 's:m:/1000:g')
          node_stat=$((node_stat))
          ;;
        memory)
          node_stat=$(kubectl describe node "${node_name}" | sed -n '/Allocated resources:/,/Events:/p' | grep "${resource}" | awk '{print $2}' | tr 'k' 'K' | numfmt --from=auto)
          ;;
        *)
          error "Unsupported $resource for get_k8s_node_allocated_resource"
          ;;
      esac
    ;;
    allocatable)
      case "$resource" in
        cpu)
          node_stat=$(kubectl get node "${node_name}" -o json | jq -r '.status.allocatable.cpu' | sed -r 's:m:/1000:g')
          node_stat=$((node_stat))
          ;;
        memory)
          node_stat=$(kubectl get node "${node_name}" -o json | jq -r '.status.allocatable.memory' | tr 'k' 'K' | numfmt --from=auto)
          ;;
        *)
          error "Unsupported resource: $resource for get_k8s_node_allocated_resource"
          ;;
      esac
    ;;
    *)
      error "Unsupported resource_type: $resource_type for get_k8s_node_allocated_resource"
    ;;
  esac

  if [[ -n $node_stat ]] ; then
    echo "$node_stat"
    returnCode="0"
  else
    error_without_exit "Got empty value while extracting ${resource_type} ${resource} from node: ${node_name}"
    returnCode="1"
  fi
  return "${returnCode}"
}


function validate_cluster_capacity() {
  local remaining_nodes_data="$1"
  local node_cpu
  local node_memory
  local old_nodes_allocated_cpu=0
  local old_nodes_allocated_memory=0
  local new_nodes_allocated_cpu=0
  local new_nodes_allocated_memory=0
  local new_nodes_allocatable_cpu=0
  local new_nodes_allocatable_memory=0
  local new_nodes_remaining_cpu
  local new_nodes_remaining_memory
  local return_code=0

  node_array=$(echo "$NODES_TO_REMOVE_JSON_ARRAY" | jq -r '.[]')
  while IFS=$'\n' read -r node_name
  do
    kubectl get node "${node_name}"  2>&1 | grep -q 'Error from server (NotFound)' && continue
    node_cpu=$(get_k8s_node_resource "cpu" "allocated" "${node_name}")
    old_nodes_allocated_cpu=$((old_nodes_allocated_cpu+node_cpu))

    node_memory=$(get_k8s_node_resource "memory" "allocated" "${node_name}")
    old_nodes_allocated_memory=$((old_nodes_allocated_memory+node_memory))
  done <<< "${node_array}"


  remaining_nodes=$(echo "$remaining_nodes_data" | jq -r '.items[] | .metadata.name')
  while IFS=$'\n' read -r node_name
  do
    node_cpu=$(get_k8s_node_resource "cpu" "allocated" "${node_name}")
    new_nodes_allocated_cpu=$((new_nodes_allocated_cpu+node_cpu))

    node_memory=$(get_k8s_node_resource "memory" "allocated" "${node_name}")
    new_nodes_allocated_memory=$((new_nodes_allocated_memory+node_memory))

    node_cpu=$(get_k8s_node_resource "cpu" "allocatable" "${node_name}")
    new_nodes_allocatable_cpu=$((new_nodes_allocatable_cpu+node_cpu))

    node_memory=$(get_k8s_node_resource "memory" "allocatable" "${node_name}")
    new_nodes_allocatable_memory=$((new_nodes_allocatable_memory+node_memory))
  done <<< "${remaining_nodes}"

  new_nodes_remaining_cpu=$((new_nodes_allocatable_cpu-new_nodes_allocated_cpu))
  new_nodes_remaining_memory=$((new_nodes_allocatable_memory-new_nodes_allocated_memory))

  if [[ $new_nodes_remaining_cpu -lt $old_nodes_allocated_cpu ]] ; then
    echo "Cluster does not have enough CPU to run pods after node removal. [Required: ${old_nodes_allocated_cpu}, Available: ${new_nodes_remaining_cpu}]"
    return_code=1
  fi

  if [[ $new_nodes_remaining_memory -lt $old_nodes_allocated_memory ]] ; then
    echo "Cluster does not have enough RAM to run pods after node removal. [Required: ${old_nodes_allocated_memory}, Available: ${new_nodes_remaining_memory}]"
    return_code=1
  fi

  return "${return_code}"
}

function redis_ha_check(){
  local redis_cluster_namespace=redis-system
  local redis_cluster_controller_type=statefulset
  local redis_cluster_controller_name=redis-cluster
  local redis_cluster_pod_labels="redis.io/role=node"
  local redis_cluster_controller_replica_count
  local redis_cluster_k8s_nodes

  if ! kubectl -n "${redis_cluster_namespace}" get "${redis_cluster_controller_type}" "${redis_cluster_controller_name}" >> /dev/null 2>&1 ; then
    echo "Redis cluster not found,  tried get on  ${redis_cluster_controller_type}/${redis_cluster_controller_name} inside NS=${redis_cluster_namespace}"
    return 1
  fi

  redis_cluster_controller_replica_count=$(kubectl -n "${redis_cluster_namespace}" get "${redis_cluster_controller_type}" "${redis_cluster_controller_name}" -o json | jq -r '.spec.replicas')
  if [[ $redis_cluster_controller_replica_count -gt 2 ]] ; then
    echo "Redis cluster has HA enabled"
    return 0
  else
    redis_cluster_k8s_nodes=$(kubectl -n "${redis_cluster_namespace}" get pod -l "${redis_cluster_pod_labels}" -o json | jq -c '[.items[] | select(.spec.nodeName!=null).spec.nodeName]')
    echo "${redis_cluster_k8s_nodes}"
    return 2
  fi
}
#shellcheck disable=SC2120
function validate_pending_pods() {
  local cmd_timeout="${1:-5m}"
  if timeout "${cmd_timeout}"  bash -c "index=0 ; while kubectl get pod -A | grep -q 'Pending' ; do [[ \$index -eq 0  ]] && index=1 && echo 'Waiting for pending pods to schedule' > /dev/tty ; sleep 2 ; printf '.' > /dev/tty ; done" ; then
    return 0
  else
    echo "Pending Pods found in the cluster. Verify the cluster state and retry"
    return 1
  fi
}

#shellcheck disable=SC2120
function validate_ready_pods(){
  local retry_per_controller="${1:-8}"
  local timeout_per_controller_seconds="${2:-120}"
  local controller_data
  local k8s_controller_metadata
  local controller_ns
  local controller_type
  local controller_name
  

  controller_data=$(kubectl get deployment,statefulset -A -o json | jq -r '.items[] | {kind: .kind , selector: ([.spec.selector.matchLabels  | keys[] as $k| "\($k)=\(.[$k])"]| join(",")), name: .metadata.name, namespace: .metadata.namespace} | .namespace + ":" + .kind + ":" + .name')
  if [[ -z $controller_data ]] ; then
    echo "No deployment or statefulset found in the cluster"
    return 1
  fi
  while IFS=$'\n' read -r k8s_controller_metadata
  do
    [[ -z $k8s_controller_metadata ]] && continue
    controller_ns=$(echo "${k8s_controller_metadata}" | cut -d':' -f1)
    controller_type=$(echo "${k8s_controller_metadata}" | cut -d':' -f2)
    controller_name=$(echo "${k8s_controller_metadata}" | cut -d':' -f3)
    [[ -z $controller_ns ]] || [[ -z $controller_type ]] || [[ -z $controller_name ]] && continue
    ## Ignore rke2-coredns-rke2-coredns-autoscaler
    [[ $controller_type == "Deployment" ]] && [[ $controller_name == "rke2-coredns-rke2-coredns-autoscaler" ]] && [[ $controller_ns == "kube-system" ]] && continue
    failure="true"
    for index in $(seq "${retry_per_controller}")
    do
      ## Check if controller exists
      if ! kubectl -n "${controller_ns}" get "${controller_type}" "${controller_name}" >> /dev/null 2>&1 ; then
        failure="false"
        break
      fi
      if kubectl -n "${controller_ns}" rollout status "${controller_type}" "${controller_name}" -w --timeout="${timeout_per_controller_seconds}s" >> /dev/null ; then
        failure="false"
        break
      fi
    done
    if [[ $failure == "true" ]] ; then
      echo "Timeout: while waiting for pods of ${controller_type}/${controller_name} in namespace ${controller_ns} to become ready"
      return 1
    fi
  done <<< "${controller_data}"
}

function validate_remaining_nodes_longhorn_scheduling() {
  local remaining_nodes_data="$1"
  local all_nodes_data_lh="$2"
  local remaining_server
  local scheduling_disabled_nodes

  remaining_server=$(echo "$remaining_nodes_data" | jq -r -c  '[.items[] | select(.metadata.labels["node-role.kubernetes.io/master"] == "true").metadata.name]')
  scheduling_disabled_nodes=$(echo "$all_nodes_data_lh" | jq -c  --argjson  nodes "${remaining_server}"  '[.items[] | select(.metadata.name  as $s_nodes | $nodes | index($s_nodes)) | select(.spec.allowScheduling == true and  (.spec.disks[] | .allowScheduling==true) | not) | .metadata.name]')
  if [[ $scheduling_disabled_nodes != "[]" ]] ; then
    echo "Remaining node(s): ${scheduling_disabled_nodes} has scheduling disabled on longhorn. Please make sure all remaining servers have scheduling enabled on longhorn"
    return 1
  fi
}

function validate_volume_replication() {
  local error_msg=()
  local volumes
  local volume_replica_count

  volumes=$(kubectl get pvc -A -o json | jq -r '.items[] | select(.metadata.name | startswith("test-") | not) | .spec.volumeName')
  for volume in $volumes; do
    volume_replica_count=$(kubectl -n "${LONGHORN_NAMESPACE}" get volumes.longhorn.io "${volume}" -o jsonpath="{.spec.numberOfReplicas}")
    if [[ $volume_replica_count -lt 2 ]]; then 
        error_msg+=("$volume has only $volume_replica_count replica. Please upgrade to 2021.10.1 for volume replication to be enabled or delete the PVs if created for testing")
    fi
  done

  if [[ ${#error_msg[@]} != 0 ]] ; then
    echo "${error_msg[@]}"
    return 1
  fi
}

function pre_validate_check() {
  local all_nodes_data="$1"
  local all_nodes_data_lh="$2"
  local remaining_nodes_data="$3"
  local error_msg=()
  local warn_msg=()
  local node_array
  local remaining_healthy_nodes_data
  local index
  local validate_node_zone_status
  local validate_node_taints_status
  local ret_error_msg
  local exitCode
  local is_atlest_one_provided_node_exists="false"
  local node_name
  local is_node_present

  info "Running validation checks"
  set +e
  node_array=$(echo "$NODES_TO_REMOVE_JSON_ARRAY" | jq -r '.[]')
  index=0
  validate_node_zone_status="success"
  validate_node_taints_status="success"
  while IFS=$'\n' read -r node_name
  do
    is_node_present=$(echo "${all_nodes_data}" | jq --arg node "${node_name}" '[.items[] | select(.metadata.name==$node)] | length')
    case "$is_node_present" in
      0)
        warn_msg+=("Node: ${node_name} does not exist in the cluster")
        ;;
      1)
        is_atlest_one_provided_node_exists="true"
        if [[ $IGNORE_MISSING_ZONE_NODE != "true" ]] ; then
          [[ $index -eq 0 ]] && info "Checking if cluster has replacement server nodes in the same zone after removal of provided nodes" 
          ret_error_msg=$(validate_node_zone "zone" "${node_name}" "${all_nodes_data}" "${remaining_nodes_data}")
          exitCode="$?"
          [[ $exitCode -ne 0 ]] && error_msg+=("${ret_error_msg}") && validate_node_zone_status="fail"
        fi

        if [[ $IGNORE_MISSING_TAINTED_NODES != "true" ]] ; then
          [[ $index -eq 0 ]] && info "Checking if cluster has replacement tainted nodes after removal of provided nodes"
          ret_error_msg=$(validate_node_taints "${node_name}" "${all_nodes_data}" "${remaining_nodes_data}")
          exitCode="$?"
          [[ $exitCode -ne 0 ]] && error_msg+=("${ret_error_msg}") && validate_node_taints_status="fail"
        fi
        ;;
      *)
        error "More than 1 node found with the same name: ${node_name} . Please check the cluster state"
    esac
    index=$((index+1))
  done <<< "${node_array}"
  

  [[ $is_atlest_one_provided_node_exists == "false" ]] && exit_on_failure="false" && info "None of the provided node(s) exist in the cluster" && exit 0
  [[ $IGNORE_MISSING_ZONE_NODE != "true" ]] && [[ $validate_node_zone_status == "success" ]] && info "[PASS] Cluster has replacement server nodes in the same zone after removal of provided nodes"
  [[ $IGNORE_MISSING_TAINTED_NODES != "true" ]] && [[ $validate_node_taints_status == "success" ]] && info "[PASS] Cluster has replacement tainted nodes after removal of provided nodes"

  ## Unschedulable/NotReady check
  remaining_healthy_nodes_data=$(echo "$all_nodes_data" | jq --argjson  nodes "${NODES_TO_REMOVE_JSON_ARRAY}" '[.items[] | select(.metadata.name  as $s_nodes | $nodes | index($s_nodes) | not) | select(.spec.unschedulable==null and  (try (.status.conditions[] | select(.type=="Ready").status) // "False") == "True")] as $f_list | { "items" : $f_list }')
  info "Checking if cluster has Unschedulable/NotReady nodes after removal of provided nodes"
  if ! diff <(jq '.items | sort_by(.metadata.name)' <<< "${remaining_nodes_data}") <(jq '.items | sort_by(.metadata.name)' <<< "${remaining_healthy_nodes_data}") >> /dev/null ; then
    error "Cluster has Unschedulable/NotReady node(s) after removal of provided nodes. Please make sure all the remaining nodes are Schedulable and Ready"
  fi
  info "[PASS] Cluster has no Unschedulable/NotReady nodes after removal of provided nodes"

  remaining_nodes_data="${remaining_healthy_nodes_data}"
  info "Checking if cluster has required server nodes after removal of provided nodes"
  ret_error_msg=$(validate_server_count "${remaining_nodes_data}")
  exitCode="$?"
  [[ $exitCode -ne 0 ]] && error_msg+=("${ret_error_msg}")
  [[ $exitCode -eq 0 ]] && info "[PASS] Cluster has more than 2 server nodes and are odd in numbers after removal of provided nodes"

  info "Checking if cluster has enough capacity after removal of provided nodes"
  ret_error_msg=$(validate_cluster_capacity "${remaining_nodes_data}")
  exitCode="$?"
  [[ $exitCode -ne 0 ]] && error_msg+=("${ret_error_msg}")
  [[ $exitCode -eq 0 ]] && info "[PASS] Cluster has enough capacity after removal of provided nodes"

  info "Checking if cluster has any pending pod before starting node removal"
  ret_error_msg=$(validate_pending_pods)
  exitCode="$?"
  [[ $exitCode -ne 0 ]] && error_msg+=("${ret_error_msg}")
  [[ $exitCode -eq 0 ]] && info "[PASS] Cluster has no pending pods before starting node removal"

  info "Checking if cluster has any unhealthy pod before starting node removal"
  ret_error_msg=$(validate_ready_pods)
  exitCode="$?"
  [[ $exitCode -ne 0 ]] && error_msg+=("${ret_error_msg}")
  [[ $exitCode -eq 0 ]] && info "[PASS] Cluster has no unhealthy pods before starting node removal"

  info "Checking if replication is enabled on volumes" 
  ret_error_msg=$(validate_volume_replication)
  exitCode="$?"
  [[ $exitCode -ne 0 ]] && error_msg+=("${ret_error_msg}")
  [[ $exitCode -eq 0 ]] && info "[PASS] Cluster has volume replication enabled"

  info "Checking if cluster has any longhorn node out of remaining nodes with scheduling disabled"
  ret_error_msg=$(validate_remaining_nodes_longhorn_scheduling "${remaining_nodes_data}" "${all_nodes_data_lh}")
  exitCode="$?"
  [[ $exitCode -ne 0 ]] && error_msg+=("${ret_error_msg}")
  [[ $exitCode -eq 0 ]] && info "[PASS] Cluster does not have any longhorn node out of remaining nodes with scheduling disabled"


  info "Checking if cluster has redis HA enable"
  ## keep these global
  VERIFY_REDIS_NODE="false"
  REDIS_NODES=$(redis_ha_check)
  exitCode="$?"
  [[ $exitCode -eq 0 ]] && info "Redis cluster has HA enabled"
  [[ $exitCode -eq 1 ]] && error_msg+=("${error}")
  if [[ $exitCode -eq 2 ]] ; then
    info "Redis does not have HA enabled"
    [[ -n $REDIS_NODES ]] && [[ $REDIS_NODES != "" ]] && [[ $REDIS_NODES != "null" ]] && VERIFY_REDIS_NODE="true"
  fi

  if [[ ${#warn_msg[@]} != 0 ]] ; then
    warn_output "Node(s) not found in the cluster" "${warn_msg[@]}"
  fi

  if [[ ${#error_msg[@]} != 0 ]] ; then
    failure_output "Node(s) validation check failed" "${error_msg[@]}"
  fi
  set -e

  ## Keep validate_volume_status at last
  validate_volume_status
  info "Validation checks succeeded"
}

function cordon_nodes() {
  local node_array
  local node_name

  node_array=$(echo "$NODES_TO_REMOVE_JSON_ARRAY" | jq -r '.[]')
  while IFS=$'\n' read -r node_name
  do
    kubectl get node "${node_name}"  2>&1 | grep -q 'Error from server (NotFound)' && info "Node: ${node_name} not found" && continue
    info "Cordoning node: ${node_name}"
    kubectl cordon "${node_name}"
    info "Longhorn: Disabling scheduling on node: ${node_name}"
    kubectl -n "${LONGHORN_NAMESPACE}" patch nodes.longhorn.io "${node_name}" --type merge -p '{"spec": {"allowScheduling": false}}'
    info "Longhorn: Disabled scheduling on node: ${node_name} successfully"
  done <<< "${node_array}"
}

function get_rebuild_status() {
  local pv_name_json_array="$1"
  local lv_engines_status
  local lv_replica_rebuild_status
  local return_string

  lv_engines_status=$(kubectl -n "${LONGHORN_NAMESPACE}" get engines.longhorn.io -o json)
  lv_replica_rebuild_status=$(echo "${lv_engines_status}" |  jq --argjson pvNames "${pv_name_json_array}" '.items[] | select(.status.rebuildStatus!={} and (.spec.volumeName as $vN | $pvNames | index($vN)))')
  [[ -z ${lv_replica_rebuild_status} ]] || [[ ${lv_replica_rebuild_status} == "null" ]] && return 0
  return_string=$(echo "${lv_replica_rebuild_status}" | jq -r '. as $global | $global | .spec.replicaAddressMap | to_entries | map( {(.value) : .key } ) | add as $addReplicaMap | $global | .status.rebuildStatus | keys[] as $k | [$addReplicaMap[($k | split("//")[1])], (.[$k].progress)] | @csv' 2>&1 | awk -v FS="," 'BEGIN{print "PVC-Replica\t\t\t\t\t\tRebuild-Status(%)";}{printf "%s\t%s%s",$1,$2,ORS}')
  echo "${return_string}"
}

function display_rebuild_status() {
  local rebuild_status="$1"
  local rebuild_status_prev="$2"

  if ! diff <(echo "${rebuild_status}") <(echo "${rebuild_status_prev}") >> /dev/null 2>&1 ; 
  then
    if [[ -n ${rebuild_status} ]] && [[ $rebuild_status == *'"pvc-'* ]] ; then
      echo "${rebuild_status}"
    fi
  else
    printf "."
  fi
}

#shellcheck disable=SC2120
function validate_volume_status() {
  local timeout_seconds="${1:-1800}"
  local intermediate_check_timeout_seconds="${2:-300}"
  local robustness_unhealthy_volumes
  local newline_char=""
  local volCheckStartTime
  local currentTime
  local intermediate_check="false"
  local diffTime
  local rebuild_status
  local rebuild_status_prev

  volCheckStartTime=$(date +"%s")

  while true
  do
    robustness_unhealthy_volumes=$(kubectl -n "${LONGHORN_NAMESPACE}" get volumes.longhorn.io -o json | jq -c '[.items[] | select(.status.robustness!="healthy" and (.status.kubernetesStatus.pvcName | startswith("test-") | not)).metadata.name]')
    if [[ $robustness_unhealthy_volumes != "[]" ]] ; then
      if [[ -z $newline_char ]] ; then
        info "Longhorn: Waiting for volume robustness to become healthy"
        newline_char="\n"
      fi
      currentTime=$(date +"%s")
      diffTime=$((currentTime - volCheckStartTime))
      if [[ $diffTime -gt $intermediate_check_timeout_seconds ]] && [[ $diffTime -lt $timeout_seconds ]] && [[ $intermediate_check == "false" ]] ; then
        intermediate_check="true"
        echo ""
        warn "Longhorn: It is taking more time for volume${robustness_unhealthy_volumes} robustness to become healthy , will wait for another $((timeout_seconds - intermediate_check_timeout_seconds)) seconds before exiting"
        warn "In case, this operation results in timeout, you can safely rerun the script again"
      elif [[ $diffTime -gt $timeout_seconds ]] ; then
        echo ""
        error "Longhorn: Timeout while waiting volume${robustness_unhealthy_volumes} robustness to become healthy"
      fi
    else
      [[ -n ${newline_char} ]] && echo ""
      info "Longhorn: all volume robustness have become healthy"
      break
    fi
    sleep 60
    rebuild_status=$(get_rebuild_status "${robustness_unhealthy_volumes}")
    display_rebuild_status "${rebuild_status}" "${rebuild_status_prev}"
    rebuild_status_prev=${rebuild_status}
  done
}

function get_first_remaining_etcd_name() {
  local remaining_nodes_data="$1"
  local remaining_server_name

  remaining_server_name=$(echo "$remaining_nodes_data" | jq -r '[.items[] | select(.metadata.labels["node-role.kubernetes.io/master"] == "true").metadata.annotations["etcd.k3s.cattle.io/node-name"]] | sort | .[0]')
  if [[ -z ${remaining_server_name} ]] || [[ ${remaining_server_name} == "null" ]] ; then
    error "Error: Could not find etcd name of first remaining sever"
  else
    echo "${remaining_server_name}"
    return 0
  fi
}

function promote_etcd_leader() {
  local remaining_nodes_data="$1"
  local new_leader
  local etcdctl_bin
  local new_leader_id
  local etcd_endpoints
  local current_leader
  local retry=3
  local leader_changed="failed"

  new_leader=$(get_first_remaining_etcd_name "${remaining_nodes_data}")
  if [[ -z $ETCDCTL_BIN ]] ; then
    etcdctl_bin=$(find /var/lib/rancher/rke2/agent/containerd/ -name "etcdctl" -type f -perm -u+x  -print -quit)
    [[ -z ${etcdctl_bin} ]] && error "Error: Could not find etcdctl"
    ETCDCTL_BIN="${etcdctl_bin}"
  fi

  new_leader_id=$(${ETCDCTL_BIN} \
    --cacert="${ETCDCTL_CACERT}" \
    --cert="${ETCDCTL_CERT}" \
    --key="${ETCDCTL_KEY}" \
    member list | grep "${new_leader}" | awk '{print $1}' | cut -d',' -f1)
  [[ -z ${new_leader_id} ]] && error "Error: Could not find member id"

  etcd_endpoints=$(${ETCDCTL_BIN} \
    --cacert="${ETCDCTL_CACERT}" \
    --cert="${ETCDCTL_CERT}" \
    --key="${ETCDCTL_KEY}" \
    member list -w json | jq -r '[.members[].clientURLs[]] | join(",")')
  [[ -z ${etcd_endpoints} ]] && error "Error: Could not find etcd endpoints"

  current_leader=$(${ETCDCTL_BIN} --endpoints="${etcd_endpoints}" \
    --cacert="${ETCDCTL_CACERT}" \
    --cert="${ETCDCTL_CERT}" \
    --key="${ETCDCTL_KEY}" \
    endpoint status 2> /dev/null | grep 'true, false' | awk '{print $2}' | cut -d',' -f 1)

  [[ $current_leader == "$new_leader_id" ]] && return 0

  info "Promoting member: ${new_leader}[${new_leader_id}] as etcd leader"
  ${ETCDCTL_BIN} --endpoints="${etcd_endpoints}" \
    --cacert="${ETCDCTL_CACERT}" \
    --cert="${ETCDCTL_CERT}" \
    --key="${ETCDCTL_KEY}" \
    move-leader "${new_leader_id}"

  for index in $(seq "${retry}")
  do
    current_leader=$(${ETCDCTL_BIN} --endpoints="${etcd_endpoints}" \
      --cacert="${ETCDCTL_CACERT}" \
      --cert="${ETCDCTL_CERT}" \
      --key="${ETCDCTL_KEY}" \
      endpoint status  | grep 'true, false' | awk '{print $2}' | cut -d',' -f 1)

    if [[ $current_leader == "$new_leader_id" ]] ; then
      leader_changed="success"
      break
    fi
    sleep 5
  done

  if [[ $leader_changed == "failed" ]] ; then
    error "Failed to update etcd leader"
  fi
}

function evict_lh_replicas() {
  local node_name="$1"
  local lv_replica_evict_timeout_seconds="${2:-1800}"
  local lv_replica_evict_intermediate_check_timeout_seconds="${3:-300}"
  local evictionStartTime
  local newline_char=""
  local intermediate_check="false"
  local lh_replicas_rebuild
  local currentTime
  local diffTime
  local rebuild_status
  local rebuild_status_prev
  local returnCode


  ## NEW code
  info "Longhorn: Evicting replica from node ${node_name}"
  kubectl -n "${LONGHORN_NAMESPACE}" patch nodes.longhorn.io "${node_name}" --type merge -p '{"spec": {"evictionRequested": true}}'
  evictionStartTime=$(date +"%s")
  while true
  do
    lh_replicas_rebuild=$(kubectl -n "${LONGHORN_NAMESPACE}" get replicas.longhorn.io -o json | jq -r --arg nodeName "${node_name}" '[.items[] | select(.spec.nodeID==$nodeName and .status.currentState=="running").spec.volumeName]')
    if [[ $lh_replicas_rebuild != "[]" ]] ; then
      if [[ -z $newline_char  ]] ; then
        info "Longhorn: Waiting for replicas to move out of node: ${node_name}"
        newline_char="\n"
      fi
      currentTime=$(date +"%s")
      diffTime=$((currentTime - evictionStartTime))
      if [[ $diffTime -gt $lv_replica_evict_intermediate_check_timeout_seconds ]] && [[ $diffTime -lt $lv_replica_evict_timeout_seconds ]] && [[ $intermediate_check == "false" ]] ; then
        intermediate_check="true"
        echo ""
        warn "Longhorn: It is taking more time for the replicas to move out, will wait for another $((lv_replica_evict_timeout_seconds - lv_replica_evict_intermediate_check_timeout_seconds)) seconds before exiting"
        warn "In case, this operation results in timeout,  you can safely rerun the script again"
      elif [[ $diffTime -gt $lv_replica_evict_timeout_seconds ]] ; then
        echo ""
        error_without_exit "Longhorn: Timeout while waiting for replica to move out of node: ${node_name}"
        returnCode=1
        break
      fi
    else
      [[ -n ${newline_char} ]] && echo ""
      info "Longhorn: replicas evicted from node: ${node_name}"
      returnCode=0
      break
    fi
    sleep 60
    rebuild_status=$(get_rebuild_status "${lh_replicas_rebuild}")
    display_rebuild_status "${rebuild_status}" "${rebuild_status_prev}"
    rebuild_status_prev=${rebuild_status}
  done
  return "${returnCode}"
}

function drain_k8s_node() {
  local node_name="$1"
  local node_drain_retry="${2:-8}"
  local node_drain_timeout_seconds="${3:-120}"
  local node_drain_success="false"
  local index
  local dangaling_attachments
  local redis_node_drain_check

  info "Draining node: ${node_name}"
  for index in $(seq 0 "${node_drain_retry}")
  do
    if [[ $index -gt 0 ]] ; then
      warn "Retrying drain on node: ${node_name}"
      if kubectl drain "${node_name}" --delete-emptydir-data --force=true --grace-period=-1 --ignore-daemonsets=true --timeout="10s" --pod-selector 'longhorn.io/component!=instance-manager' >> /dev/null 2>&1 ; then
        dangaling_attachments=$(kubectl get volumeattachment -o json | jq -r --arg nodeName "${node_name}" '.items[] | select(.spec.nodeName==$nodeName).metadata.name')
        if [[ -n ${dangaling_attachments} ]] ; then
          while IFS=$'\n' read -r dangaling_attachment
          do
            [[ -z $dangaling_attachment ]] && continue
            kubectl delete volumeattachment "${dangaling_attachment}" || echo "Error deleting volumeattachment"
          done <<< "${dangaling_attachments}"
        fi
      fi
    fi
    if kubectl drain "${node_name}" --delete-emptydir-data --force=true --grace-period=-1 --ignore-daemonsets=true --skip-wait-for-delete-timeout=300 --timeout="${node_drain_timeout_seconds}s" >> "/tmp/${node_name}-drain.log" 2>&1 ; then
      node_drain_success="true"
      break
    fi
  done
  [[ $node_drain_success == "false" ]] && error "Error while draining node: ${node_name}. For more details, please check the logs /tmp/${node_name}-drain.log"
  [[ $node_drain_success == "true" ]] && info "Node: ${node_name} drained successfully"

  if [[ $VERIFY_REDIS_NODE == "true" ]] ; then
    redis_node_drain_check=$(jq -n --argjson redisNodes "${REDIS_NODES}" --arg nodeName "${node_name}"   '$redisNodes | index($nodeName)')
    if [[ $redis_node_drain_check != "null" ]] ; then
      kubectl -n redis-system create job "redis-cluster-recovery-job-$(date +%s)" --from=cronjob/redis-cluster-recovery-job
      warn "In case, the script fails with 'error: timed out' while running next 'Verifying if all pods managed by deployment/statefulset are ready'"
      warn "You can safely rerun the script again. This happens due to non-HA redis cluster"
    fi
  fi
}

function delete_nodes(){
  local node_array
  local node_name
  local index
  local retry=1
  local timeout_seconds=200
  local sleep_before_check=100
  local nodeDeletionStartTime
  local currentTime
  local node_status
  local warn_check
  local info_check
  local total_wait=$((timeout_seconds + sleep_before_check))
  local node_deleted


  node_array=$(echo "$NODES_TO_REMOVE_JSON_ARRAY" | jq -r '.[]')
  while IFS=$'\n' read -r node_name
  do
    info_check="false"
    warn_check="false"
    if kubectl get node "${node_name}"  2>&1 | grep -q 'Error from server (NotFound)' ; then
      info "Node: ${node_name} deleted successfully(NotFound)"
      continue
    fi
    
    while true
    do
      node_status=$(kubectl get node "${node_name}" -o json | jq -r '.status.conditions[] | select(.type=="Ready" and .status!="True" and (.message | contains("Kubelet stopped posting node status"))).status')
      if [[ -n $node_status ]] ; then
        [[ $info_check == "true" ]] || [[ $warn_check == "true" ]] && echo ""
        info "Deleting Node: ${node_name} from kubernetes cluster"
        node_deleted="false"
        for index in $(seq 0 "${retry}")
        do
          if kubectl delete node  --grace-period=100 "${node_name}" ; then
            info "Node: ${node_name} deleted successfully"
            node_deleted="true"
            break
          fi
        done
        if [[ $node_deleted == "true" ]] ; then
          break
        else
          error "Error: While deleting node: ${node_name}"
        fi
      else
        if [[ $info_check == "false" ]] ; then
          warn "Please shutdown/terminate node: ${node_name}. The script will not delete a healthy node as it may cause unexpected behavior"
          warn "Script will wait upto ${total_wait} seconds for the node to shutdown"
          info_check="true"
          nodeDeletionStartTime=$(date +"%s")
        fi
        currentTime=$(date +"%s")
        diffTime=$((currentTime - nodeDeletionStartTime))
        if [[ $diffTime -gt $total_wait ]] ; then
          echo ""
          error "Please rerun the script again when the node: ${node_name} is shutdown/terminated to remove it from kubernetes cluster"
          exit 1
        elif [[ $diffTime -gt $sleep_before_check ]] && [[ $warn_check == "false" ]] ; then
          warn_check="true"
          echo ""
          warn "Node: ${node_name} is still reporting healthy status"
        else
          printf "."
          sleep 10
        fi
      fi
    done
  done <<< "${node_array}"
}

#shellcheck disable=SC2120
function drain_nodes() {
  local node_drain_timeout_seconds="${1:-120}"
  local node_drain_retry="${2:-8}"
  local lv_replica_evict_timeout_seconds="${3:-1800}"
  local lv_replica_evict_intermediate_check_timeout_seconds="${4:-300}"
  local retry_evict_lh_on_failure="${5:-true}"
  local node_array
  local node_name
  local error
  local exitCode
  local retry_evict_lh="false"


  node_array=$(echo "$NODES_TO_REMOVE_JSON_ARRAY" | jq -r '.[]')
  while IFS=$'\n' read -r node_name
  do
    kubectl get node "${node_name}"  2>&1 | grep -q 'Error from server (NotFound)' && continue
    if ! evict_lh_replicas "${node_name}" "${lv_replica_evict_timeout_seconds}" "${lv_replica_evict_intermediate_check_timeout_seconds}" && [[ $retry_evict_lh_on_failure == "true" ]] ; then
      warn "Longhorn: replica eviction will be retried after node drain"
      retry_evict_lh="true"
    fi
    drain_k8s_node "${node_name}" "${node_drain_retry}" "${node_drain_timeout_seconds}"
    if [[ $retry_evict_lh == "true" ]] ; then
      warn "Longhorn: Retrying replica eviction after node drain"
      if ! evict_lh_replicas "${node_name}" "${lv_replica_evict_timeout_seconds}" "${lv_replica_evict_intermediate_check_timeout_seconds}" ; then
        error "Error: Longhorn replica eviction failed after drain on node: ${node_name}"
      fi
    fi
    #shellcheck disable=SC2119
    validate_volume_status
    info "Verifying if all pods managed by deployment/statefulset are ready after node: ${node_name} drained"
    error=$(validate_pending_pods)
    exitCode="$?"
    [[ $exitCode -ne 0 ]] && error "${error}"
    error=$(validate_ready_pods)
    exitCode="$?"
    [[ $exitCode -ne 0 ]] && error "${error}"
    info "All pods managed by deployment/statefulset are ready after node: ${node_name} drained"
  done <<< "${node_array}"
  info "All provided Nodes drained successfully"
}

function single_instance_lock(){
  set +e
  local op="$1"
  local cm_lock_name="graceful-node-removal-lock"
  local output
  case "$op" in
    create)
      output=$(kubectl -n kube-system get configmap "${cm_lock_name}" 2>&1)
      exit_code="$?"
      ## If lock already exists
      if [[ $exit_code -eq 0 ]] ; then
        error "Running multiple instances of the script is not allowed"
      fi
      ## If unexpected output
      if ! [[ $output == *"(NotFound)"* ]] ; then
        error "Failure while checking lock status"
      fi
      info "Creating instance lock"
      if kubectl -n kube-system create configmap graceful-node-removal-lock --dry-run=client -o json | kubectl apply -f - >> /dev/null 2>&1; then
        info "Instance lock created successfully"
      else
        error "Failure while creating instance lock"
      fi
      ;;
    delete)
      output=$(kubectl -n kube-system get configmap "${cm_lock_name}" 2>&1)
      exit_code="$?"
      if [[ $exit_code -eq 0 ]] ; then
        info "Deleting instance lock"
        if kubectl -n kube-system delete configmap "${cm_lock_name}" >> /dev/null 2>&1 ; then
          info "Instance lock deleted successfully"
        else
          error "Failure while deleting instance lock"
        fi
      else
        if ! [[ $output == *"(NotFound)"* ]] ; then
          error "Failure while checking lock status"
        fi
      fi
      ;;
    *)
      error "unsupported operation: ${op} while calling single_instance_lock"
  esac
  set -e
}

function main() {
  ## Keep these global
  display_usage_check="false"
  exit_on_failure=true
  ## local variables
  local all_nodes_data
  local all_nodes_data_lh
  local remaining_nodes_data
  ## Function call
  parse_long_args "$@"
  single_instance_lock "create" && LOCK_CREATED="true"
  all_nodes_data=$(kubectl get node -o json)
  all_nodes_data_lh=$(kubectl -n "${LONGHORN_NAMESPACE}" get nodes.longhorn.io  -o json)
  remaining_nodes_data=$(echo "$all_nodes_data" | jq --argjson  nodes "${NODES_TO_REMOVE_JSON_ARRAY}" '[.items[] | select(.metadata.name  as $s_nodes | $nodes | index($s_nodes) | not)] as $f_list | { "items" : $f_list }')
  pre_validate_check "${all_nodes_data}" "${all_nodes_data_lh}" "${remaining_nodes_data}"
  cordon_nodes
  #shellcheck disable=SC2119
  drain_nodes
  promote_etcd_leader "${remaining_nodes_data}"
  [[ $SKIP_NODE_DELETION != "true" ]] && delete_nodes
  exit_on_failure=false
}

main "$@"