#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

# Note: error_messages must be declared in prerequisites.sh

function parse_hardware_validation_args() {
  local machine_installation_type
  local product_installation_type
  local is_multi_node_installation

  machine_installation_type="single_node"
  product_installation_type="core_platform"
  is_multi_node_installation="$(get_optional_json_value ".multinode")"

  if [[ "${is_multi_node_installation}" == "true" ]]; then
      machine_installation_type="multi_node"
  fi

  if [[ "$(is_full_installation)" == "true" ]]; then
      product_installation_type="full_automation_suite"
  fi  
  
  MOUNT_POINT="$(get_mandatory_json_value ".infra.longhorn.mount")"
  MINIMUM_DATA_DISK_SIZE=$(get_mandatory_json_value ".infra.hardware_requirements.${machine_installation_type}.${product_installation_type}.data_disk")    
  if [[ "${INSTALLSTANDALONEAICENTER}" == "1" ]]; then
    [[ "${is_multi_node_installation}" == "true" ]] && MINIMUM_CPU_CORES=8 || MINIMUM_CPU_CORES=16
  else
    MINIMUM_CPU_CORES=$(get_mandatory_json_value ".infra.hardware_requirements.${machine_installation_type}.${product_installation_type}.cpu")
  fi
  MINIMUM_MEMORY_IN_GB=$(get_mandatory_json_value ".infra.hardware_requirements.${machine_installation_type}.${product_installation_type}.memory")
  MINIMUM_RANCHER_DIR_SIZE=$(get_mandatory_json_value ".infra.os_disk_requirements.rancher_dir_size")
  MINIMUM_KUBELET_DIR_SIZE=$(get_mandatory_json_value ".infra.os_disk_requirements.kubelet_dir_size")
  MINIMUM_ETCD_DIR_SIZE=$(get_mandatory_json_value ".infra.os_disk_requirements.etcd_dir_size")
  
  # Hardware requirement based on NodeType
  MINIMUM_CPU_CORES_TASKMINING=$(get_mandatory_json_value ".infra.hardware_requirements.task_mining.cpu")
  MINIMUM_CPU_CORES_GPU=$(get_mandatory_json_value ".infra.hardware_requirements.gpu.cpu")
  MINIMUM_CPU_CORES_JIT=$(get_mandatory_json_value ".infra.hardware_requirements.jit.cpu")
  MINIMUM_MEMORY_IN_GB_TASKMINING=$(get_mandatory_json_value ".infra.hardware_requirements.task_mining.memory")
  MINIMUM_MEMORY_IN_GB_GPU=$(get_mandatory_json_value ".infra.hardware_requirements.gpu.memory")
  MINIMUM_MEMORY_IN_GB_JIT=$(get_mandatory_json_value ".infra.hardware_requirements.jit.memory")  

  # Reduce 10% reserved size
  MINIMUM_MEMORY_IN_GB_EXCLUDING_RESERVED=$(((MINIMUM_MEMORY_IN_GB*9)/10))
  MINIMUM_RANCHER_DIR_SIZE_EXCLUDING_RESERVED=$(((MINIMUM_RANCHER_DIR_SIZE*9)/10))
  MINIMUM_KUBELET_DIR_SIZE_EXCLUDING_RESERVED=$(((MINIMUM_KUBELET_DIR_SIZE*9)/10))
  MINIMUM_ETCD_DIR_SIZE_EXCLUDING_RESERVED=$(((MINIMUM_ETCD_DIR_SIZE*9)/10))
  MINIMUM_DATA_DISK_SIZE_EXCLUDING_RESERVED=$(((MINIMUM_DATA_DISK_SIZE*9)/10))    
  MINIMUM_MEMORY_IN_GB_TASKMINING_EXCLUDING_RESERVED=$(((MINIMUM_MEMORY_IN_GB_TASKMINING*9)/10))
  MINIMUM_MEMORY_IN_GB_GPU_EXCLUDING_RESERVED=$(((MINIMUM_MEMORY_IN_GB_GPU*9)/10))
  MINIMUM_MEMORY_IN_GB_JIT_EXCLUDING_RESERVED=$(((MINIMUM_MEMORY_IN_GB_JIT*9)/10))
}

function cpu_validation() {
  local node_type=$1
  local cpu_count
  cpu_count=$(grep -c processor /proc/cpuinfo)
  if [[ -z $cpu_count ]]; then
    error_messages+=("Fetching system cpu info failed.")
  fi

  minimum_cpu_cores_required="${MINIMUM_CPU_CORES}"
  
  if [[ "${node_type}" == "task-mining" ]]; then
    minimum_cpu_cores_required="${MINIMUM_CPU_CORES_TASKMINING}"
  elif [[ "${node_type}" == "gpu" ]]; then
    minimum_cpu_cores_required="${MINIMUM_CPU_CORES_GPU}"
  elif [[ "${node_type}" == "jit" ]]; then
    minimum_cpu_cores_required="${MINIMUM_CPU_CORES_JIT}"
  fi

  # Stop installation if cpu count is less than 'minimum_cores' mentioned in defaults.json
  if (( cpu_count < minimum_cpu_cores_required )); then
    error_messages+=("Minimum CPU core requirement is $minimum_cpu_cores_required")
  fi
}

function memory_validation() {
  local node_type=$1
  local memory_info
  memory_info=$( grep MemTotal /proc/meminfo )
  if [[ -z "${memory_info}" ]]; then
    error_messages+=("Fetching system memory info failed.")
  fi

  # parse and get the memory in kb
  local memory_in_kb
  memory_in_kb=$(echo "${memory_info}" | xargs | cut -d' ' -f2)

  # convert memory info from kb to gb
  local memory_in_Gb=$(( "${memory_in_kb}"/1024/1024 ))

  minimum_memory_in_gb_required="${MINIMUM_MEMORY_IN_GB_EXCLUDING_RESERVED}"
  
  if [[ "${node_type}" == "task-mining" ]]; then
    minimum_memory_in_gb_required="${MINIMUM_MEMORY_IN_GB_TASKMINING_EXCLUDING_RESERVED}"
  elif [[ "${node_type}" == "gpu" ]]; then
    minimum_memory_in_gb_required="${MINIMUM_MEMORY_IN_GB_GPU_EXCLUDING_RESERVED}"
  elif [[ "${node_type}" == "jit" ]]; then
    minimum_memory_in_gb_required="${MINIMUM_MEMORY_IN_GB_JIT_EXCLUDING_RESERVED}"
  fi

  # Stop installation if memory size is less 'than minimum_memory_size' GB mentioned in defaults.json
  if (( memory_in_Gb < minimum_memory_in_gb_required )); then
    error_messages+=("Minimum memory requirement is $minimum_memory_in_gb_required GiB")
  fi
}

function validate_fstab_entry() {
  local MOUNT_POINT=$1
  # fstab validation
  local mount_available_in_fstab
  mount_available_in_fstab=$(grep -w "${MOUNT_POINT}" /etc/fstab) || true
  if [[ -z "${mount_available_in_fstab}" ]]; then
    error_messages+=("No entry found for ${MOUNT_POINT} in /etc/fstab.")
    return
  fi

  # If it's getting executed as non-root user, the check does not work
  [[ "$(id -un 2>/dev/null || true)" != "root" ]] && return

  local validata_disk_mount
  validata_disk_mount=$(mount -fav|grep -w "${MOUNT_POINT}") || true
  if [[ -z "${validata_disk_mount}" ]]; then
    error_messages+=("${MOUNT_POINT} entry in /etc/fstab not valid.")
  fi
}

function data_disk_validation() {
  [[ "$(systemctl is-enabled rke2-server 2>/dev/null)" == "enabled" ]] && info "Found RKE2 cluster already running, Skipping data disk check." && return
  
  if [[ ! -d "${MOUNT_POINT}" ]]; then
    error_messages+=("Data disk is not attached to path: ${MOUNT_POINT}")
    return
  fi

  local disk_info
  local disk_info_list
  disk_info=$(df -T "${MOUNT_POINT}" | tail -n +2 |  awk '{ print $2, $3 }')
  IFS=" " read -r -a disk_info_list <<< "${disk_info}"
  local type
  local size_in_Kb
  type="${disk_info_list[0]}"
  size_in_Kb="${disk_info_list[1]}"
  local size_in_GB=$((size_in_Kb/1024/1024))
  
  if [[ "${type}" != "ext4" && "${type}" != "xfs" ]]; then
    error_messages+=("Data disk of type \"${type}\" attached to path \"${MOUNT_POINT}\" is not supported. \n\tOnly ext4 and xfs disk type are supported")
  fi

  if [[ "${size_in_GB}" -lt "${MINIMUM_DATA_DISK_SIZE_EXCLUDING_RESERVED}" ]]; then
    error_messages+=("Minimum requirement for disk attached to \"${MOUNT_POINT}\" is ${MINIMUM_DATA_DISK_SIZE} GiB")
  fi

  validate_fstab_entry "${MOUNT_POINT}"
}

function rke2_disk_validation() {

  local rancher_partition
  local kubelet_partition
  local etcd_partition
  local rancher_path="/var/lib/rancher"
  local kubelet_path="/var/lib/kubelet"
  local etcd_path="/var/lib/rancher/rke2/server/db"
  local NODE_TYPE=$1

  if [[ "${NODE_TYPE}" == "none" || "${NODE_TYPE}" == "server" ]]; then 
    if [[ "$(systemctl is-enabled rke2-server 2>/dev/null)" == "enabled" ]]; then
      info "Found RKE2 cluster already running, Skipping rke2 disk check."
      return
    fi
  else
    if [[ "$(systemctl is-enabled rke2-agent 2>/dev/null)" == "enabled" ]]; then
      info "Found RKE2 cluster already running, Skipping rke2 disk check."
      return
    fi
  fi

  # Check if /var/lib/rancher is mount with a separate device
  if [[ -d "${rancher_path}" ]]; then
    rancher_partition=$(df "${rancher_path}" | sed -n '2 p' | cut -d' ' -f1)
    # Check that the corresponding device is indeed mounted at the desired location
    [[ $(df ${rancher_path} --output=target | grep ${rancher_path}) != "${rancher_path}" ]] && error_messages+=("${rancher_path} must be mounted with a separate partition or disk")
  else
    error_messages+=("${rancher_path} must be mounted with a separate partition or disk")
  fi

  # Check if /var/lib/kubelet is mount with a separate device
  if [[ -d "${kubelet_path}" ]]; then
    kubelet_partition=$(df "${kubelet_path}" | sed -n '2 p' | cut -d' ' -f1)
    # Check that the corresponding device is indeed mounted at the desired location
    [[ $(df ${kubelet_path} --output=target | grep ${kubelet_path}) != "${kubelet_path}" ]] && error_messages+=("${kubelet_path} must be mounted with a separate partition or disk")
  else
    error_messages+=("${kubelet_path} must be mounted with a separate partition or disk")
  fi

  if [[ -n "$rancher_partition" && "$(get_size_in_GB "$rancher_partition")" -lt "${MINIMUM_RANCHER_DIR_SIZE_EXCLUDING_RESERVED}" ]]; then
    error_messages+=("Minimum requirement for disk attached to device \"${rancher_partition}\" is ${MINIMUM_RANCHER_DIR_SIZE} GiB.")
  fi

  if [[ -n "$kubelet_partition" && "$(get_size_in_GB "$kubelet_partition")" -lt "${MINIMUM_KUBELET_DIR_SIZE_EXCLUDING_RESERVED}" ]]; then
    error_messages+=("Minimum requirement for disk attached to device \"${kubelet_partition}\" is ${MINIMUM_KUBELET_DIR_SIZE} GiB.")
  fi

  validate_fstab_entry "$rancher_path"
  validate_fstab_entry "$kubelet_path"
  
  if [[ "${NODE_TYPE}" == "none" || "${NODE_TYPE}" == "server" ]]; then 
    if [[ -d "${etcd_path}" ]]; then
      etcd_partition=$(df "${etcd_path}" | sed -n '2 p' | cut -d' ' -f1)
      if [[ "${etcd_partition}" == "${rancher_partition}" || "${etcd_partition}" == "${kubelet_partition}" ]]; then 
        error_message+=("\"${etcd_path}\" should be mounted on separate disk.")
      fi

      if [[ "$(get_size_in_GB "$etcd_partition")" -lt "${MINIMUM_ETCD_DIR_SIZE_EXCLUDING_RESERVED}" ]]; then
        error_messages+=("Minimum requirement for disk attached to device \"${etcd_partition}\" is ${MINIMUM_ETCD_DIR_SIZE} GiB.")
      fi
      
      validate_fstab_entry "$etcd_path"    
    else
      error_messages+=("\"${etcd_path}\" is not mounted.")
    fi
  fi 
}

# given the device name, returns the size in GB
function get_size_in_GB() {
  local size_in_bytes
  size_in_bytes=$(lsblk "$1" -b --nodeps --noheadings --paths --output size)
  local divisor=$((1024*1024*1024))
  local size_in_GB=$(((size_in_bytes+divisor-1)/divisor)) # this will round up value
  echo ${size_in_GB}
}