#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

function get_cmd_for_node_role() {
  cmd=$1
  node_role=$2

  if [[ "${node_role}" == "agent" ]]; then
    cmd="${cmd} --kubeconfig /var/lib/rancher/rke2/agent/kubelet.kubeconfig"
  fi
  export get_cmd_for_node_role_result="${cmd}"
}

########
# Validate Cluster Ready
########
function validate_cluster_ready() {
  local maxtry=$1
  local node_role=$2
  local try=0
  local result
  local node_name

  node_name=$(hostname)

  get_cmd_for_node_role "kubectl cluster-info" "${node_role}"
  cluster_cmd="${get_cmd_for_node_role_result}"
  unset get_cmd_for_node_role_result

  info "Check if cluster is ready to use..."
  while ! ${cluster_cmd} >/dev/null 2>&1  && (( try != maxtry )) ; do
    result=$(${cluster_cmd}) || true
    info "${result}"
    try=$((try+1))
    info "waiting for the cluster to be ready...${try}/${maxtry}" &&  sleep 5;
  done
  ${cluster_cmd} || error "Could not reach the cluster API"

  get_cmd_for_node_role "kubectl get node ${node_name} " "${node_role}"
  get_node_cmd="${get_cmd_for_node_role_result}"
  unset get_cmd_for_node_role_result

  try=0
  while ${get_node_cmd} | grep -w "NotReady" && (( try != maxtry )) ; do
    try=$((try+1))
    result=$(${get_node_cmd}) || true
    info "${result}"
    info "waiting for all nodes to be ready...${try}/${maxtry}" &&  sleep 5;
  done

  result=$(${get_node_cmd} | grep -w "NotReady") || true
  if [ -n "${result}" ] ; then
    error "Node is not ready... exiting"
  fi
}

########
# Stop RKE2 Server
########
function stop_rke2_server(){
  info "Stopping RKE2 Server..."
  systemctl stop rke2-server.service
  sleep 2

  local stopped=0
  local try=0
  local maxtry=30
  local status=""
  while (( try != maxtry )) ; do
    status="$(systemctl is-active rke2-server.service)" || true
    if { [ "${status}" == "active" ]; }; then
      try=$((try+1))
      info "waiting for the rk2 server to stop...${try}/${maxtry}" &&  sleep 2;
    else
      info "Stopped RKE2 Server successfully."
      stopped=1
      break
    fi
  done

  if [ ${stopped} -eq 0 ]; then
    error "RKE2 Server is still running."
  fi
}

function kill_kubelet_containerd(){
  KUBELET_PIDS=$(pgrep kubelet)
  CONTAINERD_PIDS=$(pgrep containerd)

  if [ -n "${KUBELET_PIDS}" ]; then
    for KUBELET_PID in ${KUBELET_PIDS}; do
      info "kill kubelet ${KUBELET_PID}"
      kill -9 "${KUBELET_PID}" || true
    done
  fi

  if [ -n "${CONTAINERD_PIDS}" ]; then
    for CONTAINERD_PID in ${CONTAINERD_PIDS}; do
      info "kill containerd ${CONTAINERD_PID}"
      kill -9 "${CONTAINERD_PID}" || true
    done
  fi
}

########
# Start RKE2 Server
########
function start_rke2_server(){
  info "Starting RKE2 Server..."
  systemctl start rke2-server.service
  sleep 2

  local started=0
  local try=0
  local maxtry=30
  local status=""
  while (( try != maxtry )) ; do
    systemctl start rke2-server.service
    status="$(systemctl is-active rke2-server.service)" || true
    if { [ "${status}" != "active" ]; }; then
      try=$((try+1))
      info "waiting for the rk2 server to be ready...${try}/${maxtry}" &&  sleep 2;
    else
      info "Started RKE2 Server successfully."
      started=1
      break
    fi
  done

  if [ ${started} -eq 0 ]; then
    error "RKE2 Server is still not running."
  fi
}

function delete_pv_forcefully() {
  local pv_name=$1

  kubectl delete pv "${pv_name}" --grace-period=0 --force &

  local success=0
  local try=0
  local maxtry=10
  while (( try < maxtry )); do
    # update finaluzer field to null
    result=$(kubectl get pv "${pv_name}" -o=json | jq '.metadata.finalizers = null' | kubectl apply -f -) || true
    info "${result}"

    result=$(kubectl get pv "${pv_name}")|| true
    info "${result}"
    if [[ -z "${result}" ]]; then
      success=1
      break;
    fi
    kubectl delete pv "${pv_name}" --grace-period=0 --force &
    info "Waiting to delete pv ${pv_name} ${try}/${maxtry}..."; sleep 10
    try=$(( try + 1 ))
  done

  if [ "${success}" -eq 0 ]; then
    error_without_exit "Failed to delete pv ${pv_name}."
  fi
}

function delete_longhornvolume_forcefully() {
  local pv_name=$1
  
  if ! kubectl -n longhorn-system get volumes.longhorn.io "${pv_name}" >/dev/null 2>&1 ; then
    info "Volume ${pv_name} not found, skip deletion."
    return
  fi

  kubectl -n longhorn-system delete volumes.longhorn.io "${pv_name}" --grace-period=0 --force &
  local success=0
  local try=0
  local maxtry=10
  while (( try < maxtry )); do
    local result=""
    # update finaluzer field to null
    result=$(kubectl -n longhorn-system get volumes.longhorn.io "${pv_name}" -o=json | jq '.metadata.finalizers = null' | kubectl apply -f -) || true
    info "${result}"

    result=$(kubectl -n longhorn-system get volumes.longhorn.io "${pv_name}")|| true
    info "${result}"
    if [[ -z "${result}" ]]; then
      success=1
      break;
    fi
    kubectl -n longhorn-system delete volumes.longhorn.io "${pv_name}" --grace-period=0 --force &
    info "Waiting to delete volume ${pv_name} ${try}/${maxtry}..."; sleep 10
    try=$(( try + 1 ))
  done

   if [ "${success}" -eq 0 ]; then
    error_without_exit "Failed to delete volume ${pv_name}."
  fi

}

function delete_volumeattachment_forcefully() {
  local volumeattachment_name=$1

  if ! kubectl get volumeattachments "${volumeattachment_name}" >/dev/null 2>&1 ; then
    info "Volumeattachment ${volumeattachment_name} not found, skip deletion."
    return
  fi

  kubectl delete volumeattachments "${volumeattachment_name}" --grace-period=0 --force &
  local success=0
  local try=0
  local maxtry=10
  while (( try < maxtry )); do
    local result=""
    # update finaluzer field to null
    result=$(kubectl get volumeattachments "${volumeattachment_name}" -o=json | jq '.metadata.finalizers = null' | kubectl apply -f -) || true
    info "${result}"

    result=$(kubectl get volumeattachments "${volumeattachment_name}") || true
    info "${result}"
    if [[ -z "${result}" ]]; then
      success=1
      break;
    fi
    kubectl delete volumeattachments "${volumeattachment_name}" --grace-period=0 --force &
    info "Waiting to delete volumeattachments ${volumeattachment_name} ${try}/${maxtry}..."; sleep 10
    try=$(( try + 1 ))
  done

   if [ "${success}" -eq 0 ]; then
    error_without_exit "Failed to delete volumeattachments ${volumeattachment_name}."
  fi
}

function scale_up_deployment() {
  local deployment_name=$1
  local namespace=$2
  local replica=$3

  # Scale up deployments using PVCs
  info "Start Scale Up deployment ${deployment_name}..."

  info "Waiting to scale up deployment..."

  local try=1
  local maxtry=15
  success=0
  while (( try != maxtry )) ; do
    result=$(kubectl scale deployment "${deployment_name}" --replicas="${replica}" -n "${namespace}") || true
    info "${result}"

    scaleup=$(kubectl get deployment "${deployment_name}" -n "${namespace}"|grep "${replica}"/"${replica}") || true
    if ! { [ -n "${scaleup}" ] && [ "${scaleup}" != " " ]; }; then
      try=$((try+1))
      info "waiting for the deployment ${deployment_name} to scale up...${try}/${maxtry}";sleep 30
    else
      info "Deployment scaled up successfully."
      success=1
      break
    fi
  done

  if [ ${success} -eq 0 ]; then
    warn "Deployment scaled up failed ${deployment_name}."
  fi
}

function scale_up_statefulset() {
  local statefulset_name=$1
  local namespace=$2
  local replica=$3
  local ownerReferences=$4

  # Scale up statefulsets using PVCs
  info "Start Scale Up statefulset ${statefulset_name}..."

  # validate and scale up ownerreference
  scale_ownerreferences "${ownerReferences}" "${namespace}" "${replica}"

  info "Waiting to scale up statefulset..."

  local try=1
  local maxtry=15
  local success=0
  while (( try != maxtry )) ; do

    kubectl scale statefulset "${statefulset_name}" --replicas="${replica}" -n "${namespace}"
    kubectl get statefulset "${statefulset_name}" -n "${namespace}"

    scaleup=$(kubectl get statefulset "${statefulset_name}" -n "${namespace}"|grep "${replica}"/"${replica}") || true
    if ! { [ -n "${scaleup}" ] && [ "${scaleup}" != " " ]; }; then
      try=$((try+1))
      info "waiting for the statefulset ${statefulset_name} to scale up...${try}/${maxtry}"; sleep 30
    else
      info "Statefulset scaled up successfully."
      success=1
      break
    fi
  done

  if [ ${success} -eq 0 ]; then
    warn "Statefulset scaled up failed ${statefulset_name}."
  fi
}

function scale_ownerreferences() {
  local ownerReferences=$1
  local namespace=$2
  local replicas=$3

  # no operation required
  if [[ -z "${ownerReferences}" || "${ownerReferences}" == "null" ]]; then
    return
  fi

  ownerReferences=$(echo "${ownerReferences}"| jq -c ".[]")
  for ownerReference in ${ownerReferences};
  do
    echo "Owner: ${ownerReference}"
    local resourceKind
    local resourceName
    resourceKind=$(echo "${ownerReference}"| jq -r ".kind")
    resourceName=$(echo "${ownerReference}"| jq -r ".name")

    if kubectl -n "${namespace}" get "${resourceKind}" "${resourceName}" >/dev/null 2>&1; then
      # scale replicas
      kubectl  -n "${namespace}" patch "${resourceKind}" "${resourceName}" --type json -p "[{\"op\": \"replace\", \"path\": \"/spec/replicas\", \"value\": ${replicas} }]"
    fi
  done
}

function scale_down_statefulset() {
  local statefulset_name=$1
  local namespace=$2
  local ownerReferences=$3

  info "Start Scale Down statefulset ${statefulset_name} under namespace ${namespace}..."

  # validate and scale down ownerreference
  scale_ownerreferences "${ownerReferences}" "${namespace}" 0

  local try=0
  local maxtry=30
  success=0
  while (( try != maxtry )) ; do
    result=$(kubectl scale statefulset "${statefulset_name}" --replicas=0 -n "${namespace}") || true
    info "${result}"
    scaledown=$(kubectl get statefulset "${statefulset_name}" -n "${namespace}"|grep 0/0) || true
    if { [ -n "${scaledown}" ] && [ "${scaledown}" != " " ]; }; then
      info "Statefulset scaled down successfully."
      success=1
      break
    else
      try=$((try+1))
      info "waiting for the statefulset ${statefulset_name} to scale down...${try}/${maxtry}";sleep 30
    fi
  done

  if [ ${success} -eq 0 ]; then
    warn "Statefulset ${statefulset_name} scaled down failed"
  fi
}

function scale_down_deployment(){
  local deployment_name=$1
  local namespace=$2

  info "Start Scale Down deployment ${deployment_name} under namespace ${namespace}..."
  info "Waiting to scale down deployment..."

  local try=0
  local maxtry=60
  success=0
  while (( try != maxtry )) ; do
    result=$(kubectl scale deployment "${deployment_name}" --replicas=0 -n "${namespace}") || true
    info "${result}"
    scaledown=$(kubectl get deployment "${deployment_name}" -n "${namespace}"|grep 0/0) || true
    if { [ -n "${scaledown}" ] && [ "${scaledown}" != " " ]; }; then
      info "Deployment scaled down successfully."
      success=1
      break
    else
      try=$((try+1))
      info "waiting for the deployment ${deployment_name} to scale down...${try}/${maxtry}";sleep 30
    fi
  done

  if [ ${success} -eq 0 ]; then
    warn "Deployment ${deployment_name} scaled down failed"
  fi
}

function scale_down_all_deployment(){
  local namespace=$1

  info "Starting to scale down all the deployments under namespace: $namespace..."

  local try=0
  local maxtry=180
  success=0

  deploy_count=$(kubectl --namespace "${namespace}" get deploy --no-headers | wc -l)
  kubectl scale deployment --replicas=0 --namespace "${namespace}" --all > /dev/null\
    || return 1

  while (( try != maxtry )) ; do
    scaledown_count=$(kubectl --namespace "${namespace}" get deploy --no-headers | grep -c 0/0) \
      || return 1
    if [[ -n "${scaledown_count}" && "${scaledown_count}" == "${deploy_count}" ]]; then
      info "Deployment scaled down successfully."
      return 0
    else
      info "waiting for all the deployments in namespace to scale down...${try}/${maxtry}"
      try=$((try+1))
      sleep 10
    fi
  done
  return 1
}

function scale_up_all_deployment() {
  local namespace=$1
  local replica=$2

  # Scale up deployments using PVCs
  info "Starting to scale down all the deployments under namespace: $namespace...."

  info "Waiting to scale up deployment..."

  kubectl --namespace "$namespace" scale deployment --replicas="$replica" --all > /dev/null \
    || return 1

  wait_for_all_pods_in_namespace_ready "$namespace" >/dev/null \
      || return 1

  return 0  
}

########
# Wait till the pods of an particular app are all in Ready status
# Arguments:
#   namespace
#   app_label
########
function wait_for_pods_ready_label_selector() {
  local namespace=$1
  local app_label=$2

  local try=0
  local maxtry=30
  while [[ $(kubectl -n "$namespace" get pods -l app="$app_label" -o 'jsonpath={..status.conditions[?(@.type=="Ready")].status}') =~ "False" ]] && (( try != maxtry )); do 
    try=$((try+1))
    info "Waiting for app: $app_label to be available...${try}/${maxtry}" &&  sleep 10;
  done
}

function __is_pod_ready() {
  local pod_status
  pod_status="$(kubectl -n "$namespace" get po "$1" -o 'jsonpath={.status.conditions[?(@.type=="Ready")].status}')"
  [[ "${pod_status}" == "True" ]] && echo "true" || echo "false"
}

function __pods_ready() {
  local pod

  if [[ "$#" == 0 ]]; then
    echo "true"
    return
  fi

  for pod in "${pods[@]}"; do
    if [[ "$(__is_pod_ready "$pod")" == "false" ]]; then
      echo "false"
      return
    fi
  done

  echo "true"
}

function are_all_pods_ready() {
  local all_pods
  all_pods=("$@")
  local pods_ready
  mapfile -t terminating_pods < <(kubectl -n "$namespace" get pods | grep "Terminating" | awk '{print $1}')

  pods=()
  for pod in "${all_pods[@]}"; do
    skip=
    for terminating_pod in "${terminating_pods[@]}"; do
        [[ "$pod" == "$terminating_pod" ]] && { skip=1; break; }
    done
    [[ -n $skip ]] || pods+=("$pod")
  done
  pods_ready="$(__pods_ready "${pods[@]}" "$namespace")"
  echo "${pods_ready}"
}

function wait_for_all_pods_in_namespace_ready() {
  local namespace=$1
  local start_time
  start_time=$(date +%s)
  local current_time
  current_time=$(date +%s)
  local all_deployments_ready
  
  mapfile -t deploy_list < <(kubectl get deploy -n "$namespace" | awk '{print $1}' | tail -n +2)
  while [[ $((current_time-start_time)) -le 1800 ]]; do
    all_deployments_ready="true"
    pending_deploy=()
    for deploy in "${deploy_list[@]}"; do
      if ! kubectl rollout status deployment "$deploy" -n "$namespace" --timeout 20s 2>/dev/null; then
        all_deployments_ready="false"
        pending_deploy+=("${deploy}")
      fi
    done
    
    pods=()
    for deploy in "${pending_deploy[@]}"; do
      selectors=$(kubectl get deploy "$deploy" -n "$namespace" --output=json | jq -j '.spec.selector.matchLabels | to_entries | map([.key, .value]|join("=")) | join(",")')
      mapfile -t pod_list < <(kubectl get pod -n "$namespace" -l "$selectors" | awk '{print $1}' | tail -n +2)
      pods+=("${pod_list[@]}")
    done

    if [[ "$all_deployments_ready" == "true" || "$(are_all_pods_ready "${pods[@]}")" == "true" ]]; then
      return 0
    fi
    
    current_time=$(date +%s)
    info "Waiting for pods to be ready..."
  done

  info "Waited for approx $((current_time-start_time)) seconds, but all pods are not ready yet."
  return 1
}

function get_tls_certs() {
  local namespace="istio-system"
  local secret_name="istio-ingressgateway-certs"
  if secret=$(kubectl get secret ${secret_name} --namespace ${namespace} -o json 2> /dev/null); then
    echo "${secret}"
  else
    error "Failed to get tls-certs"
  fi
}