#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

# shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/k8sutils.sh
# shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

########
# Set Longhorn Url
########
# shellcheck disable=SC2120
function set_longhorn_url(){
  local node_role=${1:-none}
  local longhorn_svc
  local longhorn_ip
  local longhorn_port

  get_cmd_for_node_role_result=""
  get_cmd_for_node_role "kubectl -n longhorn-system get svc longhorn-backend -o=json" "${node_role}"
  longhorn_svc_cmd="${get_cmd_for_node_role_result}"
  unset get_cmd_for_node_role_result

  # get longhorn backend ip & port
  longhorn_svc=$(${longhorn_svc_cmd} | jq -c '.spec| {ip: .clusterIP, port: .ports[0].port}')
  longhorn_ip=$(echo "${longhorn_svc}"| jq -r ".ip")
  longhorn_port=$(echo "${longhorn_svc}"| jq -r ".port")

  if [[ -z "${longhorn_ip}" || -z "${longhorn_port}" ]]; then
    error "Failed to set Longhorn RestApi Endpoint"
  fi
  LONGHORN_URL="http://${longhorn_ip}:${longhorn_port}"

  export LONGHORN_URL="${LONGHORN_URL}"
  info "Longhorn RestApi Endpoint: ${LONGHORN_URL}"
}

function wait_volume_healthy(){
  pv_name=$1
  local try=0
  local maxtry=30
  local success=0
  while (( try < maxtry ));do
    status=$(kubectl -n longhorn-system get volumes.longhorn.io "${pv_name}"  -o json |jq -r ".status.state") || true
    info "volume ${pv_name} status: ${status}"
    if [[ "${status}" == "detached" ]]; then
      success=1
      info "Volume ready to use"
      break
    fi
    info "waiting for volume to be ready to use${try}/${maxtry}..."; sleep 30
    try=$(( try + 1 ))
  done

  export WAIT_VOLUME_HEALTHY="${success}"
}

function check_longhorn_disks() {
    storage_classes=$(kubectl get storageclass -o json | jq -r ".items[].metadata|select((.name == \"longhorn\") or (.name == \"longhorn-backup\")).name")
    if [ -z "${storage_classes}" ]; then
      # return if longhorn storage not available in cluster
      return
    fi
    # set longhorn url
    set_longhorn_url
    # get nodes disks
    nodes_count=$(kubectl get nodes -o json| jq -r ".items[].metadata.name"|wc -l)
    if [[ "${nodes_count}" -ge 3 ]]; then
      longhorn_nodes=$(curl -X GET --noproxy "*" "${LONGHORN_URL}/v1/nodes") || true
      if [[ -n "${longhorn_nodes}" && "${longhorn_nodes}" != "null" && -n $(echo "${longhorn_nodes}"|jq -r ".data") ]]; then
        longhorn_nodes_count=$(echo "${longhorn_nodes}"|jq -r ".data[].name"|wc -l)
        if [[ "${longhorn_nodes_count}" -lt 3 ]]; then
          warn "multinode-cluster with >= 3 node requires atleast 3 node with disks attached"
          return
        fi
        longhorn_nodes_with_disk_sizes=$(echo "${longhorn_nodes}"|jq -c ".data[]|{name: .name, storageSize: .disks|to_entries[]|select(.key|test(\"^default-disk\")).value.storageMaximum}")
        # consider minimum disk of 500Gi per node
        disk_with_mimsize=0
        for node_with_disk_size in ${longhorn_nodes_with_disk_sizes}; do
          local node_name
          local disk_size
          node_name=$(echo "${node_with_disk_size}"| jq -r ".name")
          disk_size=$(echo "${node_with_disk_size}"| jq -r ".storageSize")
          local size_in_GB=$((disk_size/1024/1024/1024))
          if [[ "${size_in_GB}" -ge 500 ]]; then
            disk_with_mimsize=$(( disk_with_mimsize + 1 ))
            info "Disks sizes available in node ${node_name}: ${size_in_GB} Gib."
          else
            warn "Disks sizes available in node ${node_name}: ${size_in_GB} Gib."
          fi
        done
        if [[ "${disk_with_mimsize}" -lt 3 ]]; then
          warn "multinode-cluster with >= 3 node requires atleast 3 node with disks size >= to 500Gi."
        fi
      fi
    fi
}

function stop_disk_scheduling() {
  local node_name mount_path
  node_name=$1
  mount_path=$2
  node_role=${3:-none}
  # set longhorn url
  set_longhorn_url "${node_role}"
  # get nodes disks
  local try=0
  local maxtry=10
  local longhorn_node_data=""
  local ready=0
  while (( try != maxtry )) ; do
    longhorn_node_data=$(curl -X GET --noproxy "*" "${LONGHORN_URL}/v1/nodes/${node_name,,}" --connect-timeout 40) || true
    if [[ -n "${longhorn_node_data}" && "${longhorn_node_data}" != "null" && -n $(echo "${longhorn_node_data}"|jq -r ".disks[]") ]]; then
      info "longhorn node ${node_name} is ready"
      ready=1
      break;
    fi
    try=$((try+1))
    info "waiting for the longhorn node ${node_name} to be ready...${try}/${maxtry}" &&  sleep 40;
  done

  if (( ready == 1 )); then
    disks_metadata=$(echo "${longhorn_node_data}"|jq -r ".disks")
    data_disk_metadata=$(echo "$disks_metadata" | jq -r "to_entries | map(select(.value.path == \"${mount_path}\") | .key)[0]")
    if [[ -n "${data_disk_metadata}" && "${data_disk_metadata}" != "null" ]]; then
      updated_disk_metadata=$(echo "$disks_metadata" | jq -r ". | with_entries (if (.value.path == \"${mount_path}\") then (.value.allowScheduling = false) else . end)")
      # update disk metadata
      curl --noproxy "*" "${LONGHORN_URL}/v1/nodes/${node_name,,}?action=diskUpdate" --data-raw "{\"disks\": ${updated_disk_metadata}}" || true
      info "longhorn disk scheduling is disabled on node ${node_name}"
    fi
  else
    warn "longhorn disk scheduling is not disabled on node ${node_name}"
  fi
}

########
# Validates longhorn backend endpoint is reachable
########
function validate_longhorn_reachable {
  local try=1
  local maxtry=31
  local success=0
  while (( try != maxtry )) ; do
    local backup_resp=""
    local resp_message=""
    backup_resp=$(curl "${LONGHORN_URL}/v1?") || true
    if { [ -n "${backup_resp}" ] && [ "${backup_resp}" != " " ]; }; then
      # verify if backupvolumes is exposed
      resp_message=$(echo "${backup_resp}"| grep "/v1/backupvolumes")
      if [[ -n "${resp_message}" && "${resp_message}" != " " ]]; then
        info "Longhorn Backend is Accessible"
        success=1
        break;
      fi
      info "waiting for longhorn backend to comeup ...${try}/${maxtry}";
    fi
    try=$((try+1))
    sleep 10
  done

  if [ ${success} -eq 0 ]; then
    error "Longhorn backend not accessible, error: ${backup_resp}."
  fi
}