#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

required_infra_pkgs=(iscsi-initiator-utils jq nfs-utils rpcbind helm uuidgen nc openssl httpd-tools)
required_offline_pkgs=(podman)

NODE_TYPE=${NODE_TYPE:-none}
INSTALL_TYPE="online"
MACHINES=""
declare -a error_messages
declare -a warn_messages

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/hardware_validation.sh

function display_prereq_usage() {
    echo "usage: check_prerequisites -dhaofj"
    echo "  -d                              Run in debug"
    echo "  -h                              Display help"
    echo "  --install-type <online|offline> Specify installation type. To run specific pre-requisite checks"
    echo "  -m|--machines                   Comma seperated IP address / hostnames of all the machines"
    echo "  -j                              Type of node to execute pre-requisite checks: [server|agent|task-mining|gpu|jit], default:server"
}

function parse_prereq_args() {
    while (("$#")); do
        case "$1" in
        -d)
            info "Running in debug"
            shift
            set -x
            ;;

        -h | --help)
            display_prereq_usage
            shift
            exit 0
            ;;

        --install-type)
          if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
            INSTALL_TYPE="$2"
            [[ "${INSTALL_TYPE}" != "online" && "${INSTALL_TYPE}" != "offline" ]] && display_prereq_usage && error "Incorrect install type: ${INSTALL_TYPE} provided"
            shift 2
          else
            warn "Install type not provided, precheck will consider this as online setup."
            shift 2
          fi
          ;;

        -m | --machines)
          if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
            MACHINES="$2"
            shift 2
          else
            error "Value for $1 is missing"
            exit 1
          fi
          ;;

        -j)
          if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
            NODE_TYPE="$2"
            shift 2
          else
            warn "Node Type not provided, precheck will consider this as server node."
            shift 2
          fi
          ;;


        -*) # unsupported flags
            display_prereq_usage
            Error "Error: Unsupported flag $1"
            exit 1
            ;;
        esac
    done
}

parse_prereq_input_file() {
    PROXY_ENABLED="$(get_optional_json_value ".proxy.enabled")"
    HTTP_PROXY="$(get_optional_json_value ".proxy.http_proxy")"
    NO_PROXY="$(get_optional_json_value ".proxy.no_proxy")"
    CUSTOM_DNS_RESOLVER="$(get_optional_json_value ".infra.custom_dns_resolver")"
    parse_hardware_validation_args
}

# Checks if the url can be accessed from the VMs
validate_url_response() {
    local proxy
    proxy=""
    if [[ "${PROXY_ENABLED}" == "true" && "${NO_PROXY}" != *"${1}"* ]] ; then
        http_proxy_url=${HTTP_PROXY/http:\/\//}
        proxy="--proxy ${http_proxy_url}"
    fi

    # shellcheck disable=SC2086
    if nc ${proxy} -z -v -w5 "$1" "$2" &>/dev/null; then
        info "Successfully connected to $1 on port $2"
    else
        error_messages+=("Host $1:$2 is not reachable from this machine")
    fi
}

validate_telemetry_connectivity() {
  if [[ "${TELEMETRY_OPT_OUT:-false}" == "true" ]]; then
    info "Telemetry is off"
    return;
  fi

  validate_url_response "${TELEMETRY_FQDN}" 443

  local now
  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)

  function data() {
    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Prerequisites.ConnectivityCheck",
            "properties": {
                "ApplicationName": "UiPath.ServiceFabric.Installer"
            }
        }
    }
}
EOF
  }

  post_response=0
  send_telemetry_unsafe "$(data)" || post_response=$?

  if [[ ${post_response} -eq 0 ]]; then
      info "Successfully posted telemetry event to ${TELEMETRY_URL}."
  else
      error_messages+=("Cannot POST to ${TELEMETRY_URL} from this machine. Curl exit code: ${post_response}")
  fi
}

function check_infra_prerequisites() {
    create_missing_packages_and_commands_list "${required_infra_pkgs[@]}"
    if [ -n "$MISSING_PKGS" ]; then
        error_messages+=("The following packages are not installed on this machine: $MISSING_PKGS. \n\tPlease install them before running installer.")
    else
        info "All required packages are present: ${required_infra_pkgs[*]}"
    fi
}

function check_infra_prerequisites_and_exit() {
    check_required_packages_and_commands "${required_infra_pkgs[@]}"
}

function validate_hardware_requirements() {
    local gpu_info
    gpu_info=$(lspci | grep -i 'NVIDIA' | tr '\n' ',')
    [[ -n "${gpu_info}" ]] && info "Node is using GPU, skipping hardware validation checks for Infra platform." && return
    
    info "Starting hardware validation checks for Infra platform"

    cpu_validation "${NODE_TYPE}"
    memory_validation "${NODE_TYPE}"

    rke2_disk_validation "${NODE_TYPE}"
    [[ "${NODE_TYPE}" == "none" || "${NODE_TYPE}" == "server" ]] && data_disk_validation

    info "Successfully validated hardware requirements"
}

function check_offline_prerequisites() {
    create_missing_packages_and_commands_list "${required_offline_pkgs[@]}"
    if [ -n "$MISSING_PKGS" ]; then
        error_messages+=("Offline Install - The following packages are not installed on this machine: $MISSING_PKGS. \n\tPlease install them before running installer.")
    else
        info "Offline Install - All required packages are present: ${required_offline_pkgs[*]}"
    fi
}

function check_offline_prerequisites_and_exit() {
    check_required_packages_and_commands "${required_offline_pkgs[@]}"
}

# Checks if the port is already in use. $1 is the port to be checked
function validate_port_connection() {
    if bash -c "exec 6<>/dev/tcp/127.0.0.1/$1" &>/dev/null; then
        error_messages+=("Port $1 already in use")
    else
        info "Port $1 not in use"
    fi
}

function validate_required_ports() {
    [[ "$(systemctl is-enabled rke2-server 2>/dev/null)" == "enabled" ]] && info "RKE2 Server already installed, skipping ports checks." && return
    [[ "$(systemctl is-enabled rke2-agent 2>/dev/null)" == "enabled" ]] && info "RKE2 Agent already installed, skipping ports checks." && return
    validate_port_connection 80
    validate_port_connection 443
    #validate_port_connection 31443
    #validate_port_connection 31444
    #validate_port_connection 31445
}

function dns_resolve() {
    local domain="$1"
    local ips
    ips=$(getent ahosts  "${domain}" | awk '{print $1}' | sort | uniq)
    echo "${ips}"
}

function validate_ip() {
    local ip="$1"
    local domain="$2"
    local msg
    if ! echo "${ip}" | xargs -n 1 ipcalc -c -s
    then
        if [[ -n ${domain} ]]
        then
            msg="Invalid IP(s) Found: ${ips//$'\n'/,} for ${domain}"
        else
            msg="Invalid IP(s) Found: ${ips//$'\n'/,}"
        fi
        error_messages+=("${msg}")
    fi
}

function validate_subdomains() {
    local fqdn
    local fqdn_ips
    local subdomain_keys=(".subdomain.rook_ceph" ".subdomain.rancher_server" ".subdomain.argo_cd")
    local subdomain_key
    local subdomain_value
    local subdomain
    local subdomain_ips
    local insight_enabled
    local subdomain_resolve_map={}

    [[ ${INSTALL_TYPE} == "offline" ]] && subdomain_keys+=(".subdomain.docker_registry")
    insight_enabled="$(get_mandatory_json_value ".insights.enabled")"
    [[ ${insight_enabled} == "true" ]] && subdomain_keys+=(".subdomain.looker")
    fqdn="$(get_mandatory_json_value ".fqdn")"
    fqdn_ips=$(dns_resolve "${fqdn}")
    if [[ -z "${fqdn_ips}" ]]
    then
        error_messages+=("No IP Address found for ${fqdn}")
    else
        subdomain_resolve_map=$(echo "${subdomain_resolve_map}" |  jq --arg key "${fqdn}" --arg value "$(echo -n "${fqdn_ips}" | tr '\n' ',')" '.[$key]=($value)')
        validate_ip "${fqdn_ips}" "${fqdn}"
    fi
    for index in "${!subdomain_keys[@]}"
    do
        subdomain_key="${subdomain_keys[$index]}"
        subdomain_value="$(get_mandatory_json_value "${subdomain_key}")"
        subdomain="${subdomain_value}.${fqdn}"
        subdomain_ips=$(dns_resolve "${subdomain}")
        if [[ -z "${subdomain_ips}" ]]
        then
            error_messages+=("No IP Address found for ${subdomain}")
        else
            subdomain_resolve_map=$(echo "${subdomain_resolve_map}" |  jq --arg key "${subdomain}" --arg value "$(echo -n "${subdomain_ips}" | tr '\n' ',')" '.[$key]=($value)')
            validate_ip "${subdomain_ips}" "${subdomain}"
            if [[ $subdomain_ips != "${fqdn_ips}" ]]
            then
                error_messages+=("Incorrect DNS config found. DNS config for \"${subdomain}\" must be exactly same as \"${fqdn}\"")
            else
                info "Successfully validated resolution of ${subdomain}"
            fi
        fi
    done
    echo "${subdomain_resolve_map}" | jq -r 'keys[] as $k | "\($k)=\(.[$k])"' > /tmp/subdomain_resolve_map.txt
}

function validate_external_connectivity() {
    validate_url_response "get.helm.sh" 443
    validate_url_response "registry.uipath.com" 443
    validate_url_response "rpm.rancher.io" 443
    validate_url_response "index.docker.io" 443
    validate_url_response "github.com" 443
    validate_url_response "github-releases.githubusercontent.com" 443
    validate_url_response "docker.io" 443
    validate_url_response "www.uipath.com" 443
    validate_url_response "activate.uipath.com" 443
    validate_url_response "license.uipath.com" 443
    validate_url_response "du-metering.uipath.com" 443
    validate_url_response "taskminingartifacts.blob.core.windows.net" 443
    validate_url_response "registry-data.uipath.com" 443
    validate_url_response "production.cloudflare.docker.com" 443
    validate_telemetry_connectivity
}

function check_networkmanager_services() {
    if systemctl is-enabled nm-cloud-setup.service &>/dev/null ;
    then
        error_messages+=("Please disable nm-cloud-setup.service by running 'sudo systemctl disable nm-cloud-setup.service'")
    fi

    if systemctl is-enabled nm-cloud-setup.timer &>/dev/null ;
    then
        error_messages+=("Please disable nm-cloud-setup.timer by running 'sudo systemctl disable nm-cloud-setup.timer'")
    fi
}

function check_services_enabled() {
    if ! systemctl is-active iscsid &>/dev/null ;
    then
        error_messages+=("Please install iscsi-initiator-utils package and enable iscsid service by running 'systemctl enable iscsid && systemctl start iscsid'")
    fi
}

function validate_nameserver() {
    # If it's getting executed as non-root user, the check does not work
    [[ "$(id -un 2>/dev/null || true)" != "root" ]] && return 

    DNS_RESOLVER_FILE="/etc/resolv.conf"
    [[ -n "${CUSTOM_DNS_RESOLVER}" ]] && DNS_RESOLVER_FILE=${CUSTOM_DNS_RESOLVER}

    if [[ ! -s "${DNS_RESOLVER_FILE}" ]]; then
      error_messages+=("nameserver validation failed, ${DNS_RESOLVER_FILE} missing or empty")
      return
    fi
    local ip_addresses
    ip_addresses=$(< "${DNS_RESOLVER_FILE}" grep "^nameserver"| awk '{print $2}' )
    if [[ ${ip_addresses} == "" ]]; then
      error_messages+=("nameserver validation failed, nameserver not configured correctly in ${DNS_RESOLVER_FILE}.")
      return
    fi

    # check if the address is not of 127.0.0.0/8 CIDR
    for ip_address in ${ip_addresses}; do
      [[ ! "${ip_address}" =~ ^127.* ]] || error_messages+=("Loopback ip address ${ip_address} found in ${DNS_RESOLVER_FILE}. Local dns resolver can not be used.")
    done
}

function validate_auditd_state() {
    if systemctl is-enabled auditd &>/dev/null;
    then
        # enabled set to 2 means, it is immutable
        if [[ "$(get_auditctl_parameter_value "enabled")" == "2" ]] && [[ "$(get_auditctl_parameter_value "backlog_wait_time")" != "1" ]]; then
            error_messages+=("Automation Suite require backlog_wait_time to be set 1. Please follow our troubleshooting guide to set backlog_wait_time to 1.\n Troubleshooting Guide Link - https://docs.uipath.com/automation-suite/docs/troubleshooting")
        fi
    fi
}

function grep_keyword_in_logrotate_dir() {
    local grep_keyword="$1"
    local logrotate_dir_path
    logrotate_dir_path="/etc/logrotate.d/"
    if [[ -d "${logrotate_dir_path}" ]]; then
        mapfile -t files < <(grep -rl "${logrotate_dir_path}" -e "${grep_keyword}")
        for file in "${files[@]}"; do
            warn_messages+=("log rotation configuration is set to ${grep_keyword} instead of 'daily' or 'hourly' for ${file}")
        done
    fi
}

function validate_logrotated_config() {
    grep_keyword_in_logrotate_dir "weekly"
    grep_keyword_in_logrotate_dir "monthly"
    grep_keyword_in_logrotate_dir "yearly"
}

function run_bundled_check(){
    local bundle_args=()
    # [[ $CHECK_PRE_REQUISITES == "none" ]] && return
    # [[ ! -f "${SCRIPT_DIR}/Prerequisite_Checks/connectivity/main.sh" ]] && return
    # [[ ! -f "${SCRIPT_DIR}/validateUiPathASReadiness.sh" ]] && return

    [[ -n "${MACHINES}" ]] && bundle_args+=(--machines "${MACHINES}")
    [[ ${NODE_TYPE} != "none" ]] && bundle_args+=(--node-type "${NODE_TYPE}")
    bundle_args+=(--install-type "${INSTALL_TYPE}")
    bundle_args+=(--input "${INFILE}")
    ## confirm this
    [[ ${NODE_TYPE} == "none" ]] && [[ $INSTALL_TYPE == "offline" ]] && bundle_args+=(--primary-server)
    [[ ${INSTALLSTANDALONEAICENTER} -eq 1 ]] && bundle_args+=(--install-standalone-aicenter)
    "${SCRIPT_DIR}/validateUiPathASReadiness.sh" "${bundle_args[@]}"
}

function check_prerequisites() {
    
    export INFILE="${MERGEDINPUTFILE}"
    parse_prereq_args "$@"

    parse_prereq_input_file

    check_infra_prerequisites

    [[ $INSTALL_TYPE == "offline" ]] && check_offline_prerequisites

    validate_hardware_requirements

    validate_required_ports

    validate_subdomains
    
    validate_logrotated_config
    
    check_networkmanager_services

    check_services_enabled

    validate_auditd_state

    validate_nameserver


    if ! run_bundled_check ; then
        error_messages+=("Failure running prerequisites validation checks")
    fi

    [[ ${#warn_messages[@]} != 0 ]] && warn_output "Warnings" "${warn_messages[@]}"
    
    echo
    
    [[ ${#error_messages[@]} != 0 ]] && failure_output "Validation Failed" "${error_messages[@]}"
    
    info "Prerequisites validation completed"
}
