#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi

BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

PVC_COMPONENT_TO_OPERATOR_DEPLOY_MAP='{
  "monitoring": "rancher-monitoring-operator",
  "rabbitmq": "rabbitmq-cluster-operator",
  "mongodb": "mongodb-kubernetes-operator"
}'

SUB_COMPONENT_MAP='{
  "objectstore": ["mon", "data"],
  "registry":    ["data"],
  "monitoring":  ["alertmanager","prometheus"],
  "rabbitmq":    ["server"],
  "mongodb":     ["data", "logs"]
}'

GLOBAL_RETRY=15
LONGHORN_NAMESPACE="longhorn-system"
LONGHORN_MANGER_LABELS="app=longhorn-manager"
LONGHORN_NODES_STORAGE_SCHEDULED={}

readonly PVC_COMPONENT_TO_OPERATOR_DEPLOY_MAP
readonly SCRIPT_DIR
readonly GLOBAL_RETRY
readonly SUB_COMPONENT_MAP
readonly LONGHORN_NAMESPACE
readonly LONGHORN_MANGER_LABELS

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/Fabric_Installer/bin:/usr/local/bin

#shellcheck disable=SC1090,SC1091
source "${SCRIPT_DIR}"/utils.sh

trap on_exit 0


function on_exit() {
  local local_remove_pid
  local pvc_owner_pid
  local pvc_op_pid
  trap '' 2
  (single_instance_resize_lock "remove" "true") &
  local_remove_pid=$!
  (scale_up_pvc_owners) &
  pvc_owner_pid=$!
  if [[ $operator_scale_up == "yes" ]] && [[ !  $COMPONENT =~ (objectstore|registry) ]]
  then
    (scale_operator "up") &
    pvc_op_pid=$!
    wait "$pvc_op_pid"
  fi
  wait "$local_remove_pid" "$pvc_owner_pid"
  trap 2
}


function display_usage() {
  local suported_subcomponents
  suported_subcomponents=$(echo "${SUB_COMPONENT_MAP}" | jq -r --arg component "${COMPONENT}" '.[$component] | join("|")')
  local suported_components
  if [[ ${suported_subcomponents} == "null" ]]
  then
    suported_components=$(echo "${SUB_COMPONENT_MAP}" | jq -r '. | keys | join("|")')
    error "Error: Unsupported component: ${COMPONENT}. Possible component values: [${suported_components}]"
    exit 1
  fi
  local first_suported_subcomponent
  first_suported_subcomponent=$(echo "${SUB_COMPONENT_MAP}" | jq -r --arg component "${COMPONENT}" '.[$component][0]')
  echo
  echo "***************************************************************************************"
  echo
  echo "Resize PVC provided by different CSI drivers"
  echo 
  echo "Arguments"
  echo "  -s|--size            [Required] : Size of PVC. Allowed formats are"
  echo "                       - Plain integer"
  echo "                       - A fixed-point number using one of these suffixes: E, P, T, G, M, k"
  echo "                       - A fixed-point number using one of power-of-two equivalents suffixes: Ei, Pi, Ti, Gi, Mi, Ki" 
  echo "  -b|--sub-component   [Required] : Sub-component to resize. Possible Values [${suported_subcomponents}]"
  echo "  -d|--debug           Increase logging verbosity to show all debug logs"
  echo "  -h|--help            Show this help message and exit"
  echo "  -f|--force           Allow deletion of pods to resize the PVC. This may cause downtime"
  echo
  echo "Examples"
  echo "  Update PVC size for ${COMPONENT} ${first_suported_subcomponent} to 105 Gi"
  echo "  ${COMMAND_NAME} --size 105Gi  --sub-component \"${first_suported_subcomponent}\""
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -f | --force)
      info "Running with force=true"
      shift
      FORCE="yes"
      ;;

    -s | --size)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        ORG_SIZE="$2"
        SIZE=$(k8s_convert_fix_point_number "${ORG_SIZE}")
        shift 2
      else
        error "Value for $1 is missing"
        shift 2
      fi
      ;;

    -b | --sub-component)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        SUB_COMPONENT="$2"
        shift 2
      else
        error "Value for $1 is missing"
        shift 2
      fi
      ;;           

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      exit 1
      ;;
    esac
  done
  [[ -z $SUB_COMPONENT ]] && error "Error: Required argument  -b | --sub-component missing"
  [[ -z $SIZE ]] && error "Error: Required argument  -s | --size missing"
  [ -n "${SIZE##*[!0-9]*}" ] || error "Error: Incorrect value for -s | --size, expecting a number"
}

function get_pvc_labels() {
  case "$COMPONENT" in
    objectstore)
      NS="rook-ceph"
      case "$SUB_COMPONENT" in
        data)
          PVC_LABELS="ceph.rook.io/DeviceSet"
          ;;
        mon)
          PVC_LABELS="app=rook-ceph-mon"
          ;;
        *)
          error "Error: Unsupported sub component=${SUB_COMPONENT} for Component=${COMPONENT}"
          exit 1
          ;;
      esac
      ;;
    registry)
      NS="docker-registry"
      case "$SUB_COMPONENT" in
        data)
          PVC_LABELS="app=docker-registry"
          ;;
        *)
          error "Error: Unsupported sub component=${SUB_COMPONENT} for Component=${COMPONENT}"
          exit 1
          ;;
      esac
      ;;
    monitoring)
      NS="cattle-monitoring-system"
      case "$SUB_COMPONENT" in
        alertmanager)
          PVC_LABELS="app.kubernetes.io/name=alertmanager"
          ;;
        prometheus)
          PVC_LABELS="app.kubernetes.io/name=prometheus"
          ;;          
        *)
          error "Error: Unsupported sub component=${SUB_COMPONENT} for Component=${COMPONENT}"
          exit 1
          ;;
      esac
      ;;
    mongodb)
      NS="mongodb"
      case "$SUB_COMPONENT" in
        data)
          PVC_LABELS="volume-type=data"
          ;;
        logs)
          PVC_LABELS="volume-type=logs"
          ;;
        *)
          error "Error: Unsupported sub component=${SUB_COMPONENT} for Component=${COMPONENT}"
          exit 1
          ;;
      esac
      ;;
    rabbitmq)
      NS="rabbitmq"
      OP_NS="rabbitmq-system"
      case "$SUB_COMPONENT" in
        server)
          PVC_LABELS="app.kubernetes.io/name=rabbitmq"
          ;;          
        *)
          error "Error: Unsupported sub component=${SUB_COMPONENT} for Component=${COMPONENT}"
          exit 1
          ;;
      esac
      ;;                      
    *)
      error "Error: Unsupported component=${COMPONENT}"
      exit 1
      ;;
  esac
}

function get_pvc_provisioner() {
  PVC_DATA=$(kubectl -n "${NS}" get pvc -l "${PVC_LABELS}" -o json)
  PVC_PROVISIONER=$(echo "${PVC_DATA}" | jq -r '.items[] |  .metadata.name + "=" + .metadata.annotations["volume.beta.kubernetes.io/storage-provisioner"] + "=" + .spec.volumeName ')
  if [[ $PVC_PROVISIONER == "" ]] || [[ $PVC_PROVISIONER == "null" ]]
  then
    error "Error: Could not determine pvc provisioner"
    exit 1
  fi
}

function wait_until_pvc_attached() {
  local  retry="${1:-10}"
  retry=$((retry-1))
  [[ $retry -lt 0 ]] && error "Error: timeout waiting for pvc: ${PVC_NAME} to attach" && exit 1
  kubectl -n "${NS}" describe pvc "${PVC_NAME}" | sed -n '/Used By:/,/:/p' | sed  '$ d' | awk '{print $NF}' | while IFS=$'\n' read -r PVC_POD
  do
    if [[ $PVC_POD == "<none>" ]]
    then
      sleep 20
      info "try:$((GLOBAL_RETRY - retry))/${GLOBAL_RETRY} Waiting for PVC: ${PVC_NAME} in namespace: ${NS} to attach"
      wait_until_pvc_attached "${retry}"
    elif [[ $PVC_POD == "" ]]
    then
      error "Error: EMPTY 'Used By' config for PVC: ${PVC_NAME} in namespace: ${NS}"
      exit 1
    else 
      kubectl -n "${NS}" get po "${PVC_POD}" | grep -q 'Completed' && continue
      if kubectl -n "${NS}" wait pod "${PVC_POD}" --for=condition=Ready --timeout=1m 2>/dev/null
      then
        info "Successfully attached PVC: ${PVC_NAME} in namespace: ${NS}"
      else
        sleep 15
        info "try:$((GLOBAL_RETRY - retry))/${GLOBAL_RETRY} Waiting for PVC: ${PVC_NAME} in namespace: ${NS} to attach"
        wait_until_pvc_attached "${retry}"
      fi
    fi
  done
}


function wait_until_pvc_detached() {
  local  retry="${1:-10}"
  retry=$((retry-1))
  [[ $retry -lt 0 ]] && rollback_pvc_owners && error "Error: timeout waiting for pvc: ${PVC_NAME} to detached" && exit 1
  kubectl -n "${NS}" describe pvc "${PVC_NAME}" | sed -n '/Used By:/,/:/p' | sed  '$ d' | awk '{print $NF}' | while IFS=$'\n' read -r PVC_POD
  do
    if [[ $PVC_POD != "" ]] && [[ $PVC_POD != "<none>" ]]
    then
      kubectl -n "${NS}" get po "${PVC_POD}" | grep -q 'Completed' && continue
      if kubectl -n "${NS}" wait pod "${PVC_POD}" --for=delete --timeout=1m 2>/dev/null
      then
        info "Successfully detached PVC: ${PVC_NAME} in namespace: ${NS}"
      else
        sleep 15
        info "try:$((GLOBAL_RETRY - retry))/${GLOBAL_RETRY} Waiting for PVC: ${PVC_NAME} in namespace: ${NS} to detached"
        wait_until_pvc_detached "${retry}"
      fi
    else
      info "Successfully detached PVC: ${PVC_NAME} in namespace: ${NS}"
    fi
  done
}

function check_current_pvc_size() {
  local notify="$1"
  local force
  CURRENT_PVC_SIZE=$(kubectl -n "${NS}" get pvc "${PVC_NAME}" -o json | jq -r '.status.capacity.storage')
  CURRENT_PVC_SIZE=$(k8s_convert_fix_point_number "${CURRENT_PVC_SIZE}")
  PVC_DIFF=$((SIZE - CURRENT_PVC_SIZE))

  [[ $notify != "yes" ]] && return

  DO_RESIZE="no"

  if [[ ${SIZE} -gt ${CURRENT_PVC_SIZE} ]]
  then
    DO_RESIZE="yes"
  elif [[ ${SIZE} -lt "${CURRENT_PVC_SIZE}" ]]
  then
    warn "Provided size for PVC: ${PVC_NAME} is less than current size. Your CSI may or may not support reducing PVC size"
    if [[ $FORCE != "yes" ]]
    then
      warn "If you wish to continue PVC size reduction operation, type 'yes' and hit Enter for confirmation. For non-interactive run use --force or -f"
      read -r force </dev/tty
      if [[ $force == 'yes' ]] || [[ $force == "'yes'" ]]
      then
        warn "Trying to reduce PVC size"
        DO_RESIZE="yes"
      else
        info "Unexpected user input received: ${force}, aborting PVC: ${PVC_NAME} size reduction"
        exit 2
      fi
    else
      warn "Trying to reduce PVC size with force"
      DO_RESIZE="yes"
    fi
  else
    info "Provided size for PVC: ${PVC_NAME} is same as current size"
  fi
}

function wait_until_pvc_resized() {
  local  retry="${1:-10}"
  retry=$((retry-1))
  [[ $retry -lt 0 ]] && rollback_pvc_owners && error "Error: timeout waiting for pvc: ${PVC_NAME} to resize" && exit 1
  CURRENT_PVC_SIZE=$(kubectl -n "${NS}" get pvc "${PVC_NAME}" -o json | jq -r '.status.capacity.storage')
  if [[ ${CURRENT_PVC_SIZE} != "${ORG_SIZE}" ]]
  then
    sleep 15
    info "try:$((GLOBAL_RETRY - retry))/${GLOBAL_RETRY} Waiting for PVC: ${PVC_NAME} in namespace: ${NS} to resize"
    wait_until_pvc_resized "${retry}"
  else
    info "Successfully resized PVC: ${PVC_NAME} in namespace: ${NS}"
  fi
}

function scale_operator() {
  op="$1"
  local retry="${2:-10}"
  [[ $retry -le 0 ]] && error "Error: timeout waiting for operator to scale ${op}"
  local op_ns="${NS}"
  ##support operator in different NS than actual pod
  [[ -n $OP_NS ]] && op_ns="${OP_NS}"
  case "$op" in
    "down")
      operator_replicas=0
      OPERATOR_DEPLOY_REPLICAS=$(kubectl -n ${op_ns} get deploy "${OPERATOR_DEPLOY}" -o json | jq -r '.spec.replicas')
      POD_STATE="delete"
      ;;
    "up")
      if [[ $OPERATOR_DEPLOY_REPLICAS != "" ]] && [[ $OPERATOR_DEPLOY_REPLICAS -gt 0 ]]
      then
        operator_replicas="${OPERATOR_DEPLOY_REPLICAS}"
      else
        operator_replicas=1
      fi
      POD_STATE="condition=Ready"
      ;;
    *)
      error "Error: Unsupported operation for scale_operator"
      exit 1
      ;;
  esac  

  if [[ $OPERATOR_DEPLOY == "null" ]]
  then
    error "Error: cannot scale down operator deployment ${OPERATOR_DEPLOY} in namespace: ${NS} for Component: ${COMPONENT}"
    exit 1
  fi
  kubectl -n "${op_ns}" patch deploy "${OPERATOR_DEPLOY}" -p "{\"spec\":{\"replicas\":${operator_replicas}}}"
  kubectl -n "${op_ns}" rollout status deploy "${OPERATOR_DEPLOY}" --timeout=10m
  OP_PODS=$(kubectl -n "${op_ns}" get po -o json | jq -r  --arg deploy "${OPERATOR_DEPLOY}" '.items[] | select(.metadata.name | startswith($deploy)) | .metadata.name')
  [[ -z "${OP_PODS}" ]] && return
  while IFS=$'\n' read -r OP_POD
  do
    if ! kubectl -n "${op_ns}" wait --for=${POD_STATE} pod/"${OP_POD}" --timeout=30s
    then
      sleep 5
      scale_operator "${op}" "${retry}"
    fi
  done <<< "${OP_PODS}"
}

function update_scale_down_pod_controllers() {
  local key="${1}"
  local value="${2}"
  SCALED_DOWN_POD_CONTROLLERS=$(echo "${SCALED_DOWN_POD_CONTROLLERS}" | jq --arg key "${key}" --arg value "${value}" '.[$key] = ($value| tonumber)')
}

function tried_resize_pvc() {
  local key="${1}"
  local value="${2}"
  TRIED_RESIZED_PVCS=$(echo "${TRIED_RESIZED_PVCS}" | jq --arg key "${key}" --arg value "${value}" '.[$key] = ($value| tonumber)')  
}

function scale_pods_using_pvc() {
  op="$1"

  case "$op" in
    "down")
      owner_replicas=0
      ;;
    "up")
      owner_replicas=${OWNER_REPLICAS}
      ;;
    *)
      error "Error: Unsupported operation for scale_pods_using_pvc"
      exit 1
      ;;
  esac

  if [[ $CONTROLLER_KIND != "" ]] && [[ $CONTROLLER_NAME != "" ]]
  then
    if [[ $op == "down" ]]
    then 
      CURRENT_REPLICAS=$(kubectl -n "${NS}" get "${CONTROLLER_KIND}" "${CONTROLLER_NAME}" -o jsonpath='{.spec.replicas}')
      validate "integer" "${CURRENT_REPLICAS}" "CURRENT_REPLICAS"
      [[ $CURRENT_REPLICAS -eq 0 ]] && warn "Current replica count for ${CONTROLLER_KIND}/${CONTROLLER_NAME} is zero"
      [[ $CURRENT_REPLICAS -gt 0 ]] && update_scale_down_pod_controllers "${CONTROLLER_KIND}=${CONTROLLER_NAME}" "${CURRENT_REPLICAS}"
    fi
    kubectl -n "${NS}" patch "${CONTROLLER_KIND}" "${CONTROLLER_NAME}" -p "{\"spec\":{\"replicas\":${owner_replicas}}}"
    kubectl -n "${NS}" rollout status "${CONTROLLER_KIND}" "${CONTROLLER_NAME}" --timeout=10m
  else
    if [[ $POD_NAME != "" ]]
    then
      error "Error: Could not find owner resource for pod: ${POD_NAME} in namespace: ${NS} for Component: ${COMPONENT}"
      exit 1
    fi
  fi
}

function scale_up_pvc_owners() {
  PVC_POD_CONTROLLERS_COUNT=$(echo "${SCALED_DOWN_POD_CONTROLLERS}" | jq -r '. | keys | length')
  if [[ $PVC_POD_CONTROLLERS_COUNT -gt 0 ]]
  then
    PVC_POD_CONTROLLERS=$(echo "$SCALED_DOWN_POD_CONTROLLERS" | jq -r '. | keys[]')
    while IFS=$'\n' read -r PVC_POD_CONTROLLER
    do
      CONTROLLER_KIND=$(echo "${PVC_POD_CONTROLLER}" | cut -d'=' -f1)
      CONTROLLER_NAME=$(echo "${PVC_POD_CONTROLLER}" | cut -d'=' -f2)
      OWNER_REPLICAS=$(echo "${SCALED_DOWN_POD_CONTROLLERS}" | jq -r --arg key "${PVC_POD_CONTROLLER}" '.[$key]')
      scale_pods_using_pvc "up"
    done <<< "${PVC_POD_CONTROLLERS}"
  fi

  TRIED_RESIZED_PVC_COUNT=$(echo "${TRIED_RESIZED_PVCS}" | jq -r '. | keys | length')
  if [[ ${TRIED_RESIZED_PVC_COUNT} -gt 0 ]]
  then
    PVC_PROVISIONER=$(echo "${TRIED_RESIZED_PVCS}" | jq -r '. | keys[]')
    while IFS=$'\n' read -r PVC_PROVISIONER_DATA
    do
      PVC_NAME=$(echo "${PVC_PROVISIONER_DATA}" | cut -d'=' -f 1)
      PROVISIONER_NAME=$(echo "${PVC_PROVISIONER_DATA}" | cut -d'=' -f 2)
      case "${PROVISIONER_NAME}" in
        "driver.longhorn.io")
          skip_attach_check=$(echo "${SKIP_PVC_ATTACHMENT_CHECK}" | jq --arg key "${PVC_NAME}" 'if .[$key] != null then true else false end')
          if [[ $skip_attach_check == "true" ]]
          then
            warn "Skipping attach check for PVC: ${PVC_NAME} as no pod was using it"
          else
            wait_until_pvc_attached "${GLOBAL_RETRY}"
          fi
          ;;
        *)
          error "Error: Unsupported provisioner: ${PROVISIONER_NAME}"
          exit 1
          ;;
      esac
    done <<< "${PVC_PROVISIONER}"    
  fi
}


function rollback_pvc_owners() { 
  warn "Failed resizing the PVC: ${PVC_NAME} in namespace: ${NS}, ROLLING BACK"
  scale_up_pvc_owners
  if [[ $operator_scale_up == "yes" ]] && [[ !  $COMPONENT =~ (objectstore|registry) ]]
  then
    scale_operator "up"
  fi
  single_instance_resize_lock "remove"
  error "Error: Failed to resize pvc: ${PVC_NAME}. Exiting"
  exit 1
}

function resize_longhorn_pvc() {
  tried_resize_pvc "${PVC_NAME}=${PROVISIONER_NAME}=${PV_NAME}" "1"
  if ! kubectl -n ${NS} patch pvc "${PVC_NAME}" -p "{\"spec\": {\"resources\": {\"requests\": {\"storage\": \"${ORG_SIZE}\" }}}}"
  then
    rollback_pvc_owners
  fi
}

function single_instance_resize_lock() {
  op="${1}"
  ignore_warning="${2:-false}"
  case "${op}" in
    "create")
      if test -f "/tmp/${NS}_${PVC_NAME}.lock"
      then
        error "Error: Running more than 1 resizing script at a given time is not supported."
        exit 1
      fi
      touch "/tmp/${NS}_${PVC_NAME}.lock"
      ;;      
    "remove")
      if test -f "/tmp/${NS}_${PVC_NAME}.lock"
      then
        rm -f "/tmp/${NS}_${PVC_NAME}.lock"
      else
        [[ "${ignore_warning}" == "true" ]] && return
        warn "Single instance resizing lock not found. Please verify if resizing worked as expected"
      fi 
      ;;
    "check")
      if test -f "/tmp/${NS}_${PVC_NAME}.lock"
      then
        error "Error: Running more than 1 resizing script at a given time is not supported."
        exit 1
      fi
      ;;      
    *)
      error "Error: Unsupported operation for single_instance_resize_lock"
  esac
}

function get_pod_owner_using_pvc() {
  PVC_POD_DATA=$(kubectl -n "${NS}" describe pvc "${PVC_NAME}" | sed -n '/Used By:/,/:/p' | sed  '$ d' | awk '{print $NF}')
  while IFS=$'\n' read -r PVC_POD
  do
    pod_owner_none="false"
    if [[ $PVC_POD == "<none>" ]]
    then
      pod_owner_none="true"
      return
    fi
    kubectl -n "${NS}" get po "${PVC_POD}" | grep -q 'Completed' && pod_owner_none="true" && continue
    if [[ $PVC_POD != "" ]]
    then
      POD_NAME="${PVC_POD}"
      controller_data=$(kubectl -n "${NS}" get po "${POD_NAME}" -o json | jq -r '[.metadata.ownerReferences[] | select(.controller==true)][0] | .kind + "=" + .name')
      [[ $controller_data == "" ]] && error "Error: Could not determine owner for pod: ${POD_NAME}" && exit 1
      CONTROLLER_KIND=$(echo "${controller_data}" | cut -d'=' -f1)
      CONTROLLER_NAME=$(echo "${controller_data}" | cut -d'=' -f2)
      if [[ $CONTROLLER_KIND == "ReplicaSet" ]]
      then
        controller_data=$(kubectl  -n "${NS}" get "${CONTROLLER_KIND}" "${CONTROLLER_NAME}" -o json | jq -r '[.metadata.ownerReferences[] | select(.controller==true)][0] | .kind + "=" + .name')
        CONTROLLER_KIND=$(echo "${controller_data}" | cut -d'=' -f1)
        CONTROLLER_NAME=$(echo "${controller_data}" | cut -d'=' -f2)
      fi
      break
    else
      POD_NAME=""
      CONTROLLER_KIND=""
      CONTROLLER_NAME=""
    fi
  done <<< "${PVC_POD_DATA}"
}

function check_confirmation() {
  local force
  if [[ $FORCE != "yes" ]]
  then
    warn "Please be aware resizing a PVC will cause downtime, we are going to scale down the pods using the PVC(s) as resizing needs PVC in detached state"
    warn "If you wish to continue, type 'yes' and hit Enter for confirmation. For non-interactive run use --force or -f"
    read -r force </dev/tty
    if [[ $force == 'yes' ]] || [[ $force == "'yes'" ]]
    then
      info "Proceeding to resize PVC(s)"
    else
      info "Unexpected user input received: ${force}, exiting resize PVC(s) operation"
      exit 2
    fi
  else
    info "Proceeding to resize PVC(s) with force"
  fi  
}

function validate() {
  local type="$1"
  local input="$2"
  local variable_name="$3"
  case "${type}" in
    integer)
      echo "${input}" | grep -q -E '^\-[0-9]+$' && input="${input/-/}"
      if [[ -z $input ]] || [ -z "${input##*[!0-9]*}" ]
      then
        error "Invalid value: ${input} for integer validation in variable ${variable_name}"
        exit 1
      fi
      ;;
    *)
      error "Error: Unsupported type: ${type}"
      exit 
  esac
}


function get_longhorn_node_capacity() {
  local pvc_path_hosts_hard_limit
  pvc_path_hosts_hard_limit=$(echo "${longhorn_nodes}" | jq -r --arg pvc_host "${pvc_host}" --arg pvc_path "${pvc_path}" --arg over_provisioning_percentage "${over_provisioning_percentage}" '.data[] | select(.id==$pvc_host) | .disks | to_entries[] | select(.value.path==$pvc_path) | ((.value.storageMaximum - .value.storageReserved) * (($over_provisioning_percentage | tonumber)/100))')
  validate "integer" "${pvc_path_hosts_hard_limit}" "pvc_path_hosts_hard_limit"

  pvc_path_hosts_storage_scheduled=$(echo "${longhorn_nodes}" | jq -r --arg pvc_host "${pvc_host}" --arg pvc_path "${pvc_path}" '.data[] | select(.id==$pvc_host) | .disks | to_entries[] | select(.value.path==$pvc_path) | .value.storageScheduled')
  validate "integer" "${pvc_path_hosts_storage_scheduled}" "pvc_path_hosts_storage_scheduled"

  ## Set hard limit
  LONGHORN_NODES_STORAGE_SCHEDULED=$(echo "${LONGHORN_NODES_STORAGE_SCHEDULED}" | jq --arg key "${pvc_host}=${pvc_path}" --arg value "${pvc_path_hosts_hard_limit}" '.[$key]["hard_limit"] = ($value | tonumber)')
  ## Set calculated storage scheduled
  LONGHORN_NODES_STORAGE_SCHEDULED=$(echo "${LONGHORN_NODES_STORAGE_SCHEDULED}" | jq --arg key "${pvc_host}=${pvc_path}" --arg value "${pvc_path_hosts_storage_scheduled}" --arg pvc_diff "${PVC_DIFF}"  'if .[$key]["calculated_storage"] == null then .[$key]["calculated_storage"] = (($value | tonumber) + ($pvc_diff | tonumber)) else .[$key]["calculated_storage"] = (.[$key]["calculated_storage"] + ($pvc_diff | tonumber)) end')

}

#shellcheck disable=SC2046
function get_longhorn_storage_capacity() {
  [[ $PVC_DIFF -le 0 ]] && return
  validate "integer" "${PVC_DIFF}" "PVC_DIFF"

  over_provisioning_percentage=$(kubectl -n ${LONGHORN_NAMESPACE}  exec $(kubectl -n ${LONGHORN_NAMESPACE}  get po  -l "${LONGHORN_MANGER_LABELS}"  --field-selector status.phase=Running -o jsonpath='{.items[0].metadata.name}') -- curl -s  "http://longhorn-backend.${LONGHORN_NAMESPACE}:9500/v1/settings/storage-over-provisioning-percentage" | jq -r '.value')
  validate "integer" "${over_provisioning_percentage}" "over_provisioning_percentage"

  ## Get PVC volumes details
  longhorn_volumes=$(kubectl -n ${LONGHORN_NAMESPACE}  exec $(kubectl -n ${LONGHORN_NAMESPACE}  get po  -l "${LONGHORN_MANGER_LABELS}"  --field-selector status.phase=Running -o jsonpath='{.items[0].metadata.name}') -- curl -s "http://longhorn-backend.${LONGHORN_NAMESPACE}:9500/v1/volumes")
  pvc_path_hosts=$(echo "${longhorn_volumes}" | jq -r --arg pv_id "${PV_NAME}" '.data[] | select(.id==$pv_id) | .replicas[] | select(.running==true) | .diskPath + "=" + .hostId')
  [[ -z ${pvc_path_hosts} ]] && error "Error: Could not fetch PVC hosts from longhorn"

  ## Get PVC hosts details
  longhorn_nodes=$(kubectl -n ${LONGHORN_NAMESPACE}  exec $(kubectl -n ${LONGHORN_NAMESPACE}  get po  -l "${LONGHORN_MANGER_LABELS}"  --field-selector status.phase=Running -o jsonpath='{.items[0].metadata.name}') -- curl -s "http://longhorn-backend.${LONGHORN_NAMESPACE}:9500/v1/nodes")
  [[ -z ${longhorn_nodes} ]] && error "Error: Could not fetch nodes from longhorn"

  while IFS=$'\n' read -r pvc_path_host
  do
    pvc_path=$(echo "${pvc_path_host}" | cut -d'=' -f 1)
    pvc_host=$(echo "${pvc_path_host}" | cut -d'=' -f 2)
    get_longhorn_node_capacity
  done <<< "${pvc_path_hosts}"
}

function is_longhorn_storage_capacity_available() {
  local hard_limit_crossed_nodes
  local longhorn_node
  local longhorn_disk

  hard_limit_crossed_nodes=$(echo "${LONGHORN_NODES_STORAGE_SCHEDULED}" | jq -r '. | to_entries[] | select(.value.calculated_storage > .value.hard_limit).key')
  if [[ -n "${hard_limit_crossed_nodes}" ]]
  then
    while IFS=$'\n' read -r hard_limit_crossed_node
    do
      hard_limit=$(echo "${LONGHORN_NODES_STORAGE_SCHEDULED}" | jq -r --arg key "${hard_limit_crossed_node}" '.[$key]["hard_limit"]')
      calculated_storage=$(echo "${LONGHORN_NODES_STORAGE_SCHEDULED}" | jq -r --arg key "${hard_limit_crossed_node}" '.[$key]["calculated_storage"]')
      longhorn_node=$(echo "${hard_limit_crossed_node}" | cut -d'=' -f 1)
      longhorn_disk=$(echo "${hard_limit_crossed_node}" | cut -d'=' -f 2)
      warn "Node: ${longhorn_node} does not have enough storage capacity for disk: ${longhorn_disk} to expand the requested PVC(s). Storage Capacity(max): $(numfmt --to=iec-i "${hard_limit}")  Requested Capacity: $(numfmt --to=iec-i "${calculated_storage}")"
    done <<< "${hard_limit_crossed_nodes}"
    error "Error: Aborting expansion, Node(s) do not have enough storage capacity to expand the requested PVC(s)"
    exit 1
  fi
}

function k8s_convert_fix_point_number() {
  local k8s_number="${1}"
  local k8s_integer
  k8s_integer=$(echo "${k8s_number}" | numfmt --from=auto)
  echo "${k8s_integer}"
}

function main() {
  parse_long_args "$@"
  get_pvc_labels
  get_pvc_provisioner

  local index=0

  OPERATOR_DEPLOY=$(echo "${PVC_COMPONENT_TO_OPERATOR_DEPLOY_MAP}" | jq -r --arg component "${COMPONENT}" '.[$component]')

  SCALED_DOWN_POD_CONTROLLERS={}
  TRIED_RESIZED_PVCS={}
  SKIP_PVC_ATTACHMENT_CHECK={}

  ## validate overall capacity availability
  while IFS=$'\n' read -r PVC_PROVISIONER_DATA
  do
    PVC_NAME=$(echo "${PVC_PROVISIONER_DATA}" | cut -d'=' -f 1)
    PROVISIONER_NAME=$(echo "${PVC_PROVISIONER_DATA}" | cut -d'=' -f 2)
    PV_NAME=$(echo "${PVC_PROVISIONER_DATA}" | cut -d'=' -f 3)
    case "${PROVISIONER_NAME}" in
      "driver.longhorn.io")
        check_current_pvc_size
        get_longhorn_storage_capacity
        get_pod_owner_using_pvc
        [[ $pod_owner_none == "true" ]] && SKIP_PVC_ATTACHMENT_CHECK=$(echo "${SKIP_PVC_ATTACHMENT_CHECK}" | jq --arg key "${PVC_NAME}" '.[$key]=1')
        ;;
      *)
        error "Error: Unsupported provisioner: ${PROVISIONER_NAME}"
        exit 1
        ;;
    esac      
  done <<< "${PVC_PROVISIONER}"
  is_longhorn_storage_capacity_available

  ## Actual Resize
  while IFS=$'\n' read -r PVC_PROVISIONER_DATA
  do
    PVC_NAME=$(echo "${PVC_PROVISIONER_DATA}" | cut -d'=' -f 1)
    PROVISIONER_NAME=$(echo "${PVC_PROVISIONER_DATA}" | cut -d'=' -f 2)
    PV_NAME=$(echo "${PVC_PROVISIONER_DATA}" | cut -d'=' -f 3)
    case "${PROVISIONER_NAME}" in
      "driver.longhorn.io")
        check_current_pvc_size "yes"
        if [[ $DO_RESIZE == "yes" ]]
        then
          get_pod_owner_using_pvc
          single_instance_resize_lock "check"
          if [[ $index -eq 0 ]]
          then
            if [[ $pod_owner_none != "true" ]] && [[ $POD_NAME != "" ]]
            then 
              check_confirmation
              if [[ ! $COMPONENT =~ (objectstore|registry) ]]
              then
                operator_scale_up="yes"
                scale_operator "down"
              fi
              index=$((index+1))
            fi
          fi
          single_instance_resize_lock "create"
          if [[ $pod_owner_none != "true" ]]
          then
            scale_pods_using_pvc "down"
            wait_until_pvc_detached "${GLOBAL_RETRY}"
          fi
          resize_longhorn_pvc
          wait_until_pvc_resized "${GLOBAL_RETRY}"
          single_instance_resize_lock "remove"
        else
          info "Resizing not required for PVC: ${PVC_NAME}"
        fi
        ;;
      *)
        error "Error: Unsupported provisioner: ${PROVISIONER_NAME}"
        exit 1
        ;;
    esac
  done <<< "${PVC_PROVISIONER}"
}

main "$@"