#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

# returns 1 when A greater than B
# returns 0 when A equals B
# returns -1 when A lower than B

set -eu

declare -A weights
weights=( ["rc"]=3 ["preview"]=2 ["private"]=1 )
regex='^([0-9]+)\.([0-9]+)\.([0-9]+)(-(([a-z]+)(([0-9]*)?))?\.?([a-z0-9+]*))?$'

outcome() {
  result=$1
  printf "%s$result\n"
  exit 0
}

number() {
  number=$1
  re='^[0-9]+$'
  if ! [[ $number =~ $re ]] ; then
    echo "error: ${number} Not a number" >&2; exit 1
  fi
}

compare() {
    firstVersion=$1 #2021.10.0-rc11.metadata
    secondVersion=$2 #2021.10.1-preview2.metadata
    
    # check if equal
    [[ "${firstVersion}" = "${secondVersion}" ]] && outcome "0"

    [[ $firstVersion =~ $regex ]]
    [[ -z ${BASH_REMATCH[0]+x} ]] && echo "error: $firstVersion is not a valid version" && exit 1
    first_major="${BASH_REMATCH[1]}"
    first_minor="${BASH_REMATCH[2]}"
    first_patch="${BASH_REMATCH[3]}"
    first_prerelease="${BASH_REMATCH[5]}"
    first_type="${BASH_REMATCH[6]}"
    first_iteration="${BASH_REMATCH[7]}"
    first_meta="${BASH_REMATCH[9]}"
    
    [[ $secondVersion =~ $regex ]]
    [[ -z ${BASH_REMATCH[0]+x} ]] && echo "error: $secondVersion is not a valid version" && exit 1
    second_major="${BASH_REMATCH[1]}"
    second_minor="${BASH_REMATCH[2]}"
    second_patch="${BASH_REMATCH[3]}"
    second_prerelease="${BASH_REMATCH[5]}"
    second_type="${BASH_REMATCH[6]}"
    second_iteration="${BASH_REMATCH[7]}"
    second_meta="${BASH_REMATCH[9]}"

    number "${first_major}"
    number "${second_major}"
    (( first_major > second_major )) && outcome "1"
    (( first_major < second_major )) && outcome "-1"

    number "${first_minor}"
    number "${second_minor}"
    (( first_minor > second_minor )) && outcome "1"
    (( first_minor < second_minor )) && outcome "-1"

    number "${first_patch}"
    number "${second_patch}"
    (( first_patch > second_patch )) && outcome "1"
    (( first_patch < second_patch )) && outcome "-1"

    # Prerelease is older than release
    # Check if only first version is prerelease
    [[ -n $first_prerelease && -z $second_prerelease ]] && outcome "-1"

    # Check if only second version is prerelease
    [[ -z $first_prerelease && -n $second_prerelease ]] && outcome "1"

    # If there are no prerelease tags, compare metas
    if [[ -z $first_prerelease && -z $second_prerelease ]]; then
        # The version with meta is newer than the one without
        [[ -z $first_meta && -n $second_meta ]] && outcome "1"
        [[ -n $first_meta && -z $second_meta ]] && outcome "-1"
        if [[ $first_meta > $second_meta ]]; then outcome "1"; else outcome "-1"; fi
    fi

    # If prerelease tags are present, compare types, iterations and meta - in that order
    [[ "${weights[$first_type]:-0}" > "${weights[$second_type]:-0}" ]] && outcome "1"
    [[ "${weights[$first_type]:-0}" < "${weights[$second_type]:-0}" ]] && outcome "-1"

    # If weights undefined, compare alphabetically
    if [[ -z "${weights[$first_type]+x}" && -z "${weights[$second_type]+x}" ]]; then
        [[ $first_prerelease > $second_prerelease ]] && outcome "1"
        [[ $first_prerelease < $second_prerelease ]] && outcome "-1"
    fi
    
    # If type or prerelease is equal, check iteration
    (( first_iteration > second_iteration )) && outcome "1"
    (( first_iteration < second_iteration )) && outcome "-1"

    # Finally, check meta
    if [[ $first_meta > $second_meta ]]; then outcome "1"; else outcome "-1"; fi
}

printf "%s$(compare "$@")\n"
