#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

function parse_step() {
    STEP=$1
    COMPONENT="$(get_mandatory_inline_json_value ".component" "${STEP}")"
    export COMPONENT
    COMPONENT_SOURCE_VERSION="$(get_optional_inline_json_value ".sourceVersion" "${STEP}")"
    if [[ -n "${COMPONENT_SOURCE_VERSION}" ]]; then
        export COMPONENT_SOURCE_VERSION
    fi
    COMPONENT_TARGET_VERSION="$(get_mandatory_inline_json_value ".targetVersion" "${STEP}")"
    export COMPONENT_TARGET_VERSION
    UPGRADE_TYPE="$(get_mandatory_inline_json_value ".type" "${STEP}")"
    export UPGRADE_TYPE
}

function get_node_type() {
  local NODE_NAME
  local VALUE
  NODE_NAME=$1
  VALUE=$(kubectl get nodes/"${NODE_NAME}" -o json | jq -r '.metadata.labels."node-role.kubernetes.io/control-plane"')
  [[ ${VALUE} == "true" ]] && echo "server" || echo "agent"
}

####
# RKE2 version is obtained from kubectl get nodes output. Adding --no-header to suppress the headers for the table format
# parsing the multiline output, adding the server name (1st position in the table output) and version (5th position)
####
function get_rke2_versions() {
  local kubectl_nodes
  local server_name
  local version
  kubectl_nodes=$(kubectl get nodes --no-headers)
  while read -r crt_node; do
    server_name=$(awk '{print $1}' <<<"${crt_node}")
    version=$(awk '{print $5}' <<<"${crt_node}")
    infraVersion=$(get_node_annotation_value "${server_name}" "uipath/infra-version")
    node_type=$(get_node_type "${server_name}")
    jq --arg server "$server_name" --arg version "$version" --arg infraVersion "${infraVersion:-0}" --arg nodeType "$node_type" '.nodes += [{name : $server, version: $version, infraVersion: $infraVersion, nodeType: $nodeType}]' "${VERSION_FILE_PATH}" >versions.json.tmp && mv versions.json.tmp "${VERSION_FILE_PATH}"
  done <<<"${kubectl_nodes}"
  info "Determined RKE2 versions"
}

####
# docker-registry is installed using helm in its own namespace.
# helm outputs warnings which cannot be suppressed, so they are redirected to /dev/null
# assumes a single release is present. if more are present, the first one is taken into consideration
####
function get_docker_registry_version() {
  local docker_releases
  docker_releases=$(helm list -n docker-registry -o json 2>/dev/null)
  local docker_version
  docker_version=$(jq -er '. | map(select(.name=="docker-registry")) | .[0].app_version' <<<"${docker_releases}") || error "No docker-registry release was found"
  jq --arg docker_version "$docker_version" '."docker-registry" = $docker_version' "${VERSION_FILE_PATH}" >versions.json.tmp && mv versions.json.tmp "${VERSION_FILE_PATH}"
  info "Detected docker-registry app_version=${docker_version}"
}

####
# longhorn is installed using helm in its own namespace.
# helm outputs warnings which cannot be suppressed, so they are redirected to /dev/null
# assumes a single release is present. if more are present, the first one is taken into consideration
####
function get_longhorn_version() {
  local longhorn_releases
  longhorn_releases=$(helm list -n longhorn-system -o json 2>/dev/null)
  local longhorn_version
  longhorn_version=$(jq -er '. | map(select(.name=="longhorn")) | .[0].app_version' <<<"${longhorn_releases}") || error "No longhorn release was found"
  jq --arg longhorn_version "$longhorn_version" '.longhorn = $longhorn_version' "${VERSION_FILE_PATH}" >versions.json.tmp && mv versions.json.tmp "${VERSION_FILE_PATH}"
  info "Detected longhorn app_version=${longhorn_version}"
}

####
# argocd outputs both agent and server version
# the server version is in semver format, with metadata separated by "+"
# metadata is removed and only the version is added to the json
####
function get_argocd_version() {
  local argocd_version
  # the script has set -o pipefail, and any argocd error will trigger a failure
  argocd_version=$(kubectl get deployments/argocd-server -n argocd -o json | jq -erc '.metadata.labels.version') || error "Could not determine ArgoCD version"
  jq --arg argocd_version "$argocd_version" '.argocd = $argocd_version' "${VERSION_FILE_PATH}" >versions.json.tmp && mv versions.json.tmp "${VERSION_FILE_PATH}"
  info "Detected ArgoCD server version=${argocd_version}"
 }

function get_apps_versions() {
  local argocd_app_list
  local crt_app app_version
  argocd_app_list=$(kubectl get applications -n argocd -o json 2>/dev/null| jq -erc '.items[].metadata.name') || error "Could not obtain the list of ArgoCD applications"
  while read -r crt_app; do
    app_version=$(kubectl get applications/"${crt_app}" -n argocd -o json 2>/dev/null | jq -erc '.spec.source.targetRevision') || error "Could not obtain the version for ${crt_app}"
    info "ArgoCD app ${crt_app} version=${app_version}"
    jq --arg crt_app "$crt_app" --arg app_version "$app_version" '. + {($crt_app) : $app_version}' "${VERSION_FILE_PATH}" >versions.json.tmp && mv versions.json.tmp "${VERSION_FILE_PATH}"
  done <<<"${argocd_app_list}"
}

function get_version_for_component() {
  local COMPONENT=$1
  TEMP_INFILE="${INFILE}"
  INFILE="${VERSION_FILE_PATH}"
  version="$(get_mandatory_json_value ".${COMPONENT}")"
  echo "${version}"
  INFILE="${TEMP_INFILE}"
}

function validate_step() {
  info "Validating upgrade step for $COMPONENT"
  if [ "${UPGRADE_TYPE}" = "versionless" ]; then
    info "Step for $COMPONENT is versionless. Executing..."
    exit 0
  fi
  info "Checking if $COMPONENT is already at version $COMPONENT_TARGET_VERSION"
  COMPONENT_CURRENT_VERSION="$(get_version_for_component "${COMPONENT}")"
  if [ "${COMPONENT_CURRENT_VERSION}" = "${COMPONENT_TARGET_VERSION}" ]; then
    info "Component ${COMPONENT} already running at ${COMPONENT_TARGET_VERSION}. Skipping step"
    exit 0
  fi
  if [[ -n "${COMPONENT_SOURCE_VERSION}" ]]; then
    info "The $COMPONENT component has a source version restriction, checking if upgrade is supported."
    if [ "${COMPONENT_CURRENT_VERSION}" != "${COMPONENT_SOURCE_VERSION}" ]; then
      error "Component ${COMPONENT} is not running a supported version. Exiting upgrade"
    fi
  fi
}

####
# Wait for specific CRD group in a namespace to become available
####
function wait_for_crd() {
  info "Waiting for plans.upgrade.cattle.io to become available."
  local crd=$1
  local namespace=$2
  local try=0
  local maxtry=10
  local status="notready"
  while [[ ${status} == "notready" ]]  && (( try != maxtry )) ; do
        try=$((try+1))
        kubectl -n "$namespace" wait --for condition=established --timeout=60s "$crd" && status="ready"
        info "Waiting for $crd CRDs to become available...${try}/${maxtry}" && sleep 20;
  done
  [[ ${status} == "ready" ]] || error "$crd CRDs did not become available in the expected time."
}

####
# Checks if ceph cluster is in error state and exits the installer
####
function ceph_health_exit_on_error() {
  local rook_cluster_namespace=rook-ceph
  toolbox_pod=$(kubectl -n "${rook_cluster_namespace}" get pod --field-selector=status.phase=Running -l app=rook-ceph-tools -o jsonpath='{.items[0].metadata.name}')
  if [[ -z "${toolbox_pod}" ]];
  then
      error "Ceph toolbox pod is not up. Please check and retry upgrade."
  fi
  if kubectl -n ${rook_cluster_namespace} exec -it "${toolbox_pod}" -- ceph status | grep -q HEALTH_ERROR;
  then
      kubectl -n ${rook_cluster_namespace} exec -it "${toolbox_pod}" -- ceph status
      error "Ceph cluster is in error state, please recover it before proceeding with upgrade."
  fi
}
