#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

########
# Utils module for different script functions
########

# TODO: this needs to be based on target env type so we don't have to patch the release branch to repoint to customer instance
TELEMETRY_INSTRUMENTATION_KEY=5da85bdb-6422-49b5-a6fa-8d3a0907a0f0
TELEMETRY_FQDN=dc.services.visualstudio.com
TELEMETRY_URL=https://${TELEMETRY_FQDN}/v2/track
function send_telemetry_unsafe() {
  curl -sf "${TELEMETRY_URL}" -k -X POST \
    -H 'Content-Type: application/json' -m 30 -d "${1}" > /dev/null 2>&1
}

function send_telemetry() {
  send_telemetry_unsafe "$@" || true
}

function ensure_machine_hash() {
  if [[ "${MACHINE_HASH:-none}" == "none" ]]; then
    MACHINE_HASH="$(uname -a | sha256sum -t - | cut -d' ' -f1)"
  fi

  export MACHINE_HASH
}

function service_exists() {
    local service_name=$1
    if [[ $(systemctl list-units --all -t service --full --no-legend "$service_name.service" | cut -f1 -d' ') == $service_name.service ]]; then
      echo "true"
    else
      echo "false"
    fi
}

function send_checkpoint_telemetry() {
  local telemetry_content

  if [[ "${TELEMETRY_OPT_OUT:-false}" == "true" ]]; then
    return;
  fi

  local now
  local checkpoint

  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)
  checkpoint=${1}

  ensure_machine_hash

  function data() {
    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Checkpoint",
            "properties": {
                "ApplicationName": "UiPath.ServiceFabric.Installer",
                "InstallId": "${INSTALL_ID:-none}",
                "MachineHash": "${MACHINE_HASH}",
                "ClusterToken" : "${RKE_TOKEN:-none}",
                "ClusterId": "${CLUSTER_ID:-none}",
                "Checkpoint": "${checkpoint}"
            }
        }
    }
}
EOF
  }

  telemetry_content="$(data)"
  send_telemetry "${telemetry_content}"
}

function send_error_telemetry() {
  local telemetry_content
  
  if [[ "${TELEMETRY_OPT_OUT:-false}" == "true" ]]; then
    return;
  fi

  local now
  local origin
  local message

  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)
  origin=${1}
  message=${*:2}

  ensure_machine_hash

  function data() {
    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Error",
            "properties": {
                "ApplicationName": "UiPath.ServiceFabric.Installer",
                "InstallId": "${INSTALL_ID:-none}",
                "MachineHash": "${MACHINE_HASH}",
                "ClusterToken" : "${RKE_TOKEN:-none}",
                "ClusterId": "${CLUSTER_ID:-none}",
                "Origin": "${origin}",
                "Message": "${message}"
            }
        }
    }
}
EOF
  }

  telemetry_content="$(data)"
  send_telemetry "${telemetry_content}"
}

function get_message_with_errorcode() {
  local err_code=$1
  err_message=$(jq -r ".error|select( has (\"${err_code}\")).\"${err_code}\"" "${MODULES_DIR}"/errcode_message_mapping.json)
  # update positional args
  local i=0
  #shellcheck disable=SC2048
  for arg in ${*:2}; do
    err_message=${err_message//\{$i\}/${arg}}
    i=$((i+1))
  done

  echo "${err_message}"
}

function error_with_code() {
  local err_code=$1
  local error_message
  error_message="$(get_message_with_errorcode "${err_code}" "${*:2}")"

  echo -e "\e[0;31m[ERROR][$(date +'%Y-%m-%dT%H:%M:%S%z')]:\e[0m ${err_code} : ${error_message}" >&2
  send_error_telemetry "$0" "${error_message}"
  exit 1
}

function error() {
  echo -e "\e[0;31m[ERROR][$(date +'%Y-%m-%dT%H:%M:%S%z')]:\e[0m $*" >&2
  send_error_telemetry "$0" "$1"
  exit 1
}

function error_without_exit() {
  echo -e "\e[0;31m[ERROR][$(date +'%Y-%m-%dT%H:%M:%S%z')]:\e[0m $*" >&2
}

function info() {
  echo "[INFO] [$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*"
}

function warn() {
  echo -e "\e[0;33m[WARN] [$(date +'%Y-%m-%dT%H:%M:%S%z')]:\e[0m $*" >&2
}

function display_usage() {
  echo "usage: $(basename "$0") [-hd] -i infile -o outfile"
  echo "  -d          Run in debug"
  echo "  -h          Display help"
  echo "  -i infile   Specify input json file"
  echo "  -o outfile  Specify output json file"
}

########
# Parse command line arguments. Exit if illegal arguments are passed
# Arguments:
#   1: String to check
#   2: Error message
# Outputs:
#   INFILE and OUTFILE are initialized with their respective values
########
function exit_if_string_empty() {
  if [[ -z "$1" ]]; then
    error "$2"
  fi
}

########
# Parse command line arguments. Exit if illegal arguments are passed
# Globals:
#   INFILE
#   OUTFILE
# Outputs:
#   INFILE and OUTFILE are initialized with their respective values
########
function parse_args() {
  INFILE=""
  OUTFILE=""
  
  OFFLINE_INSTALLER=""
  while getopts 'hdi:o:f:t:' flag "$@"; do
    case "${flag}" in
    d)
      info "Running in debug"
      set -x
      ;;
    i)
      info "Using input parameters file: ${OPTARG}"
      INFILE=${OPTARG}
      ;;
    f)
      info "Using defaults parameters file: ${OPTARG}"
      DEFAULTSFILE=${OPTARG}
      ;;
    t)
      info "Using offlineInstaller parmeter: ${OPTARG}"
      # shellcheck disable=SC2034
      OFFLINE_INSTALLER=${OPTARG}
      ;;      
    h)
      display_usage
      exit 0
      ;;
    o)
      info "Using output parameters file: ${OPTARG}"
      OUTFILE=${OPTARG}
      ;;
    :)
      error "Invalid option: ${OPTARG} requires an argument."
      ;;
    *)
      error "Unexpected option ${flag}"
      ;;
    esac
  done
  if [[ -z "${INFILE}" || -z "${OUTFILE}" ]]; then
    display_usage
    error "Mandatory parameters not found"
  fi
  if [[ ! -f "${INFILE}" ]]; then
    error "File ${INFILE} not found"
  fi
  shift $((OPTIND - 1))
}

########
# Merge defaults.json and ${INFILE} into merged_input.json
# Globals:
#   DEFAULTSFILE
# Outputs:
#   Generates "merged_input.json"
########
function merge_default_values() {
  if [[ ! -f "${DEFAULTSFILE}" ]]; then
    error "File ${DEFAULTSFILE} not found"
    exit 1
  fi
  if [ -n "${INFILE}" ];
  then
    jq '.' "${INFILE}" &> /dev/null || error "Input file ${INFILE} is not a valid json  and cannot be parsed"
    jq '.' "${DEFAULTSFILE}" &> /dev/null || error "Defaults file ${DEFAULTSFILE} is not a valid json and cannot be parsed"
    jq -s '.[0] * .[1]' "${DEFAULTSFILE}" "${INFILE}" > "${MERGEDINPUTFILE}"
  else
    cat "${DEFAULTSFILE}" > "${MERGEDINPUTFILE}"
  fi
  INFILE="${MERGEDINPUTFILE}"
}

#######
# Override HA flag to false if 3 master nodes are not available
#######
function check_available_masters() {
    AVAILABLE_MASTER_NODE_COUNT=$(kubectl get nodes |grep master|grep -wc "Ready")

    if [ "${PROFILE}" == "ha" ] && [ "${AVAILABLE_MASTER_NODE_COUNT}" -lt 3 ]; then
      error_with_code "13005"
    fi
}

function get_fabric_install_paths() {
  local BASE_DIR=$1
  readonly BINARY_DIR="${BASE_DIR}/bin"
  export BINARY_DIR
  readonly VENDOR_DIR="${BASE_DIR}/vendor"
  export VENDOR_DIR
}

function detect_os() {
  # shellcheck disable=SC1091
  OS_ID=$(source /etc/os-release ; echo "$ID")
  # shellcheck disable=SC1091
  OS_VERSION=$(source /etc/os-release ; echo "$VERSION_ID")
  OS_MAJOR_VERSION=$(echo "$OS_VERSION" | cut -d. -f1)
  OS_MINOR_VERSION=$(echo "$OS_VERSION" | cut -d. -f2)
  export OS_ID
  export OS_VERSION
  export OS_MAJOR_VERSION
  export OS_MINOR_VERSION
}

########
# Load OS-specific utils.sh
########
function load_os_utils() {
  detect_os
  local modules_directory
  modules_directory=$1
  if [ "$OS_ID" == "ubuntu" ] || [ "$OS_ID" == "debian" ]; then
    #shellcheck disable=SC1090,SC1091
    source "${modules_directory}/debian-utils.sh"
  elif [ "$OS_ID" == "rhel" ]; then
    #shellcheck disable=SC1090,SC1091
    source "${modules_directory}/rhel-utils.sh"
  else
    warn "OS not supported. Skipping OS-specific module loading."
  fi
  info "Detected OS: $OS_ID MajorVersion: $OS_MAJOR_VERSION MinorVersion: $OS_MINOR_VERSION"
}

########
# Check if package is installed or command is present
# Uses check_package_installed from os-specific modules loaded above
########
function check_package_or_command() {
  local COMMAND_EXISTS=1
  local PACKAGE_EXISTS=1
  local PATHS
  local IFS=':'
  PATHS="$(whereis "$1")"
  # shellcheck disable=SC2162
  read -a PATHARR <<<"$PATHS"
  if [ ${#PATHARR[@]} -gt "1" ]; then
    COMMAND_EXISTS=0
  fi

  check_package_installed "$1" && PACKAGE_EXISTS=0 || PACKAGE_EXISTS=1

  if [ $PACKAGE_EXISTS == 0 ] || [ $COMMAND_EXISTS == 0 ]; then
    return 0
  else
    return 1
  fi
}

########
# Check whether a list of packages and/or commands are available
# Usage:
# required_pkgs=(jq curl)
# check_required_packages_and_commands ${required_pkgs[@]}
########
function check_required_packages_and_commands() {
  create_missing_packages_and_commands_list "$@"
  if [ -n "$MISSING_PKGS" ]; then
    error "The following packages and/or commands are not available:$MISSING_PKGS. Please install them and try again."
  fi
}


########
# Check whether a list of packages and/or commands are available and puts them in MISSING_PKGS env variable
# Usage:
# required_pkgs=(jq curl)
# create_missing_packages_and_commands_list ${required_pkgs[@]}
########
function create_missing_packages_and_commands_list() {
  export MISSING_PKGS
  local required_pkgs=("$@")
  local missing_pkgs=""
  for pkg in "${required_pkgs[@]}"; do
    local package_or_command_exists
    check_package_or_command "$pkg" && package_or_command_exists=0 || package_or_command_exists=1
    if [ $package_or_command_exists == 1 ]; then
      missing_pkgs+=" $pkg"
    fi
  done
  MISSING_PKGS="$missing_pkgs"
}

########
# Get value from config json
# Globals:
#   INFILE
# Arguments:
#   json_path
# Outputs:
#   Returns the value or exits if not found
########
function get_mandatory_json_value() {
  local json_path="$1"
  local value

  value="$(jq -r "${json_path}" < "${INFILE}")"
  if [[ ${value} != null ]]; then
    echo "${value}"
  else
    error "${json_path} not found in input file"
  fi
}

########
# Get value from inline json
# Arguments:
#   json_path
#   json_string
# Outputs:
#   Returns the value or exits if not found
########
function get_mandatory_inline_json_value() {
  local json_string="$2"
  local json_path="$1"
  local value
  #shellcheck disable=SC2015
  value="$(echo "${json_string}" | jq -r "${json_path}")" \
    && echo "${value}" \
    || error "${json_path} not found in input json"
}

########
# Get optional value from inline json
# Arguments:
#   json_path
#   json_string
# Outputs:
#   Returns the value or an empty string if not found
########
function get_optional_inline_json_value() {
  local json_string="$2"
  local json_path="$1"
  local value
  #shellcheck disable=SC2015
  value="$(echo "${json_string}" | jq -r "${json_path}")"
  if [[ ${value} != null ]]; then
    echo "${value}"
  fi
}

########
# Get value from config json and base64 encode
# Globals:
#   INFILE
# Arguments:
#   json_path
# Outputs:
#   Returns the base64 encoded value or exits if not found
########
function get_mandatory_b64enc_json_value() {
  local json_path="$1"
  local value

  value="$(jq -r "${json_path}" < "${INFILE}")"
  if [[ ${value} != null ]]; then
    echo -n "${value}" | base64 -w 0 | sed -e 's/[[:space:]]//g'
  else
    error "${json_path} not found in input file"
  fi
}

########
# Get value from config json
# Globals:
#   INFILE
# Arguments:
#   json_path
# Outputs:
#   Returns the value or exits if not found
########
function get_optional_json_value() {
  local json_path="$1"
  local value

  value="$(jq -r "${json_path}" < "${INFILE}")"
  if [[ ${value} != null ]]; then
    echo "${value}"
  fi
}

########
# Get value from config json and base64 encode
# Globals:
#   INFILE
# Arguments:
#   json_path
# Outputs:
#   Returns the base64 encoded value or exits if not found
########
function get_optional_b64enc_json_value() {
  local json_path="$1"
  local value

  value="$(jq -r "${json_path}" < "${INFILE}")"
  if [[ ${value} != null ]]; then
    echo -n "${value}" | base64 -w 0 | sed -e 's/[[:space:]]//g'
  else
    echo "" | base64 -w 0 | sed -e 's/[[:space:]]//g'
  fi
}


#####
# Open Port Forward to service or pod
# To use:
# openport argocd service/argocd-server 443
#####
function openport() {
    local try=0
    local maxtry=30
    local PORTFORWARD_PORT=""
    local NAMESPACE=$1
    local ENDPOINT=$2
    local PORT=$3
    kubectl port-forward -n "$NAMESPACE" "$ENDPOINT" ":$PORT" >/dev/null &
    while ((try != maxtry)) && [[ -z $PORTFORWARD_PORT ]] ; do
        try=$((try + 1))
        PORTFORWARD_PORT=$(ss -lp4 | grep "pid=$!" | awk '{print $5}' | awk -F ":" '{print $2}')
        sleep 1
    done
    if [[ -z $PORTFORWARD_PORT ]]; then
        error "Could not open port $PORT to $ENDPOINT in namespace $NAMESPACE"
    else
        echo "$PORTFORWARD_PORT"
    fi
}

#####
# Kill existing kubectl port-forward process based on port
# To use:
# closeport 31842
#####
function closeport() {
    local PORTFORWARD_PORT=$1
    local PID
    PID=$(ss -lp4 | grep "kubectl" | grep "127.0.0.1:$PORTFORWARD_PORT" | awk '{print $7}' | awk -F "pid=" '{print $2}' | awk -F "," '{print $1}')
    kill "$PID"
}

####
# Wait for job to become completed, filter by label
####
function wait_for_job_complete_by_label() {
  local label=$1
  local namespace=$2
  local try=0
  local maxtry=50
  local status="notready"
  while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
    try=$((try+1))
    kubectl -n "$namespace" wait --for=condition=complete job -l "$label" && status="ready"
    info "Waiting for the $label jobs to complete...${try}/${maxtry}" && sleep 30;
  done
  [[ ${status} == "ready" ]] || error "Failed to complete $label jobs."
}

########
# Install Packages depending on OS
########
function load_package_utils() {
  detect_os
  local INSTALL_TYPE
  local MODULES_DIRECTORY
  MODULES_DIRECTORY=$1
  INSTALL_TYPE=$2
  if [ "$OS_ID" == "rhel" ]; then
    #shellcheck disable=SC1090,SC1091
    source "$MODULES_DIRECTORY/install_packages/$INSTALL_TYPE-$OS_ID.sh"
  else
    warn "Installation of packages is not supported for $OS_ID"
  fi
}

########
# Check if the script is getting triggered from root or not
########
function require_root_user() {
    local user=""
    user="$(id -un 2>/dev/null || true)"
    if [ "$user" != "root" ]; then
        error "Infra installer need to be run as root."
    fi
}

########
# Get value from config json or return default value
# Globals:
#   INFILE
# Arguments:
#   json_path
#   default_value
# Outputs:
#   Returns the value or return default value if not found
########
function get_json_value_or_return_default_value() {
  local json_path="$1"
  local default_value="$2"
  local value
  value="$(get_optional_json_value "${json_path}")"
  if [[ -n ${value} ]]; then
    echo "${value}"
  else
    echo "${default_value}"
  fi
}

function failure_output() {
  local title=$1
  shift
  local failures=("$@")
  printf '=%.0s' {1..100}
  echo
  error_without_exit "$title"
  for error in "${failures[@]}"
  do
    [[ -n ${error} ]] && error_without_exit ">>> $error"
  done
  printf '=%.0s' {1..100}
  echo
  exit 1
}

function warn_output() {
  local title=$1
  shift
  local warnings=("$@")
  printf '=%.0s' {1..100} >&2
  echo >&2
  warn "$title"
  for warn in "${warnings[@]}"
  do
    [[ -n ${warn} ]] && warn ">>> $warn"
  done
  printf '=%.0s' {1..100} >&2
  echo >&2
}

########
# Validates whether the argument is empty or not
########
function check_if_empty() {
    local arg_name=$1
    local arg_val=$2

    if [ -z "$arg_val" ]; then
        error "$arg_name argument is missing"
    fi
}


# Validate if any non core service is enabled
function is_full_installation(){
  local aicenter
  local business_apps
  local du
  local taskmining
  #aicenter
  aicenter=$(get_optional_json_value ".aicenter.enabled")
  #business-apps
  business_apps="$(get_optional_json_value ".apps.enabled")"
  #DU
  du="$(get_optional_json_value ".documentunderstanding.enabled")"
  #taskmining
  taskmining="$(get_optional_json_value ".task_mining.enabled")"

  if [[ ( -n "${aicenter}" && "${aicenter}" == "true" ) ||
    ( -n "${taskmining}" && "${taskmining}" == "true" ) ||
    ( -n "${du}" && "${du}" == "true" ) ||
    ( -n "${business_apps}" && "${business_apps}" == "true" ) ]]; then
    echo "true"
  else
    echo "false"
  fi
}

function get_existing_trap() {
  local signal=$1
  (( $# == 1 )) ||  error "${FUNCNAME[0]} needs a signal parameter"
  trap_cmd() { printf '%s\n' "$3"; }
  eval trap_cmd "$(trap -p "${signal}")"
}

# call option with parameters: $1=name $2=value $3=file
function add_or_replace_option() {
    name=${1//\//\\/}
    value=${2//\//\\/}
    sed -i \
        -e '/^#\?\(\s*'"${name}"'\s*=\s*\).*/{s//\1'"${value}"'/;:a;n;ba;q}' \
        -e '$a'"${name}"'='"${value}" "$3"
}

# It will fetch value associated with parameter for auditctl -s command
# Make sure you check `systemctl is-enabled auditd` before calling this function
function get_auditctl_parameter_value() {
  local parameter_name=$1
  local output
  output=$(auditctl -s | grep "${parameter_name}" | tail -1)
  IFS=' ' read -r -a output_value <<<"${output}"
  echo "${output_value[1]}"
}

function stop_incoming_connections_istio() {
  # Disable incoming calls
  kubectl -n istio-system patch daemonset istio-ingressgateway -p '{"spec": {"template": {"spec": {"nodeSelector": {"non-existing": "true"}}}}}' &>/dev/null || true
}

function start_incoming_connections_istio() {
  # enable gateway, as we need it during upgrade
  kubectl -n istio-system patch daemonset istio-ingressgateway --type json -p='[{"op": "remove", "path": "/spec/template/spec/nodeSelector/non-existing"}]' &>/dev/null || true
}

function wait_for_istio_pods() {
  kubectl -n istio-system rollout status daemonset istio-ingressgateway
}


export BACKUP_MARKER_FILE="/etc/rancher/as-backup-marker.txt"
export CONSENT_FILE="/etc/rancher/as-upgrade-consent.txt"
export RECURRING_JOB_BACKUP_LOCATION="/var/lib/uipath/uipath-backup-recurringjob.yaml"
export BACKUP_CRONTAB_LOCATION="/var/lib/uipath/crontab.bakup"
export MAINTENANCE_MODE_DEPLOY_REPLICA_CM="uipath-deployment-replicas"

function delete_recurring_job() {
  backup_job_name="${1}"

  # validate pvc cronjob for backup created
  local try=0
  local maxtry=10
  local exists=""
  while (( try < maxtry ));
  do
    # delete recurring job
    kubectl -n longhorn-system delete recurringjob "${backup_job_name}" --force || true
    kubectl -n longhorn-system delete cronjobs "${backup_job_name}" --force || true

    exists=$(kubectl -n longhorn-system get cronjobs "${backup_job_name}" &>/dev/null) || true
    if [[ -z "${exists}" ]]; then
      info "Recurring cronjob ${backup_job_name} deleted successfully"
      break;
    fi

    warn "Recurring cronjob ${backup_job_name} not deleted yet, ${try}/${maxtry}..."; sleep 5
    try=$((try + 1))
    exists=""
  done

  if [[ -n "${exists}" ]];then
    error "Failed to delete recurring job for ${backup_job_name}"
  fi
}

function pause_backups_temp() {
  # Disable etcd backup
  info "Disabling regular backups for now."
  mkdir -p /var/lib/uipath/
  ([[ -f "${BACKUP_CRONTAB_LOCATION}" ]] && grep -q backupjob "${BACKUP_CRONTAB_LOCATION}") || crontab -l > ${BACKUP_CRONTAB_LOCATION}
  crontab -l > backupcron
  sed -i '/backupjob/d' backupcron
  crontab backupcron
  rm backupcron

  if kubectl -n longhorn-system get recurringjob uipath-backup &>/dev/null;
  then
    kubectl -n longhorn-system get recurringjob uipath-backup -o yaml > "${RECURRING_JOB_BACKUP_LOCATION}"
    delete_recurring_job uipath-backup
  fi

  info "Disabled regular backups during upgrade Successfully."
}

function resume_backup_schedule() {
  delete_tmp_files
  [[ -f "${RECURRING_JOB_BACKUP_LOCATION}" ]] && kubectl -n longhorn-system apply -f "${RECURRING_JOB_BACKUP_LOCATION}" && rm -f "${RECURRING_JOB_BACKUP_LOCATION}"
  [[ -f "${BACKUP_CRONTAB_LOCATION}" ]] && crontab "${BACKUP_CRONTAB_LOCATION}" && rm -f "${BACKUP_CRONTAB_LOCATION}"
  info "Resumed scheduled backups on the node."
}

function save_latest_etcd_snapshots() {
  export PATH=$PATH:/usr/local/bin/
  # Take 2 etcd snapshots, --name is the prefix used for snapshot
  rke2 etcd-snapshot --name etcd-snapshot-ondemand
  rke2 etcd-snapshot --name etcd-snapshot-ondemand
}

function trigger_on_demand_backup() {
  info "Triggering etcd backup"
  save_latest_etcd_snapshots
  [[ -f /usr/bin/backupjob ]] && (/usr/bin/backupjob || error "Failed to run backup job")
  info "Triggering volumes backup"
  if kubectl -n longhorn-system get cronjob uipath-backup &>/dev/null;
  then
    kubectl -n longhorn-system delete job uipath-backup-manual 2>/dev/null || true
    kubectl -n longhorn-system create job --from=cronjob/uipath-backup "uipath-backup-manual"
    # wait for jobs completion
    info "Waiting for longhorn backup job to get completed"
    kubectl -n longhorn-system wait --for=condition=complete job/uipath-backup-manual --timeout=1800s
  fi
  info "On demand backup completed"
}

function cluster_backup() {
  if [[ ! -f ${CONSENT_FILE} ]] && [[ "${UPGRADE_SILENT_MODE}" == "false" ]];
  then
    # If we have not taken consent yet
    if crontab -l | grep -q backupjob;
    then
      ## Backup is enabled in the Cluster, check for SQL Server Backup as well
      printf "\n"
      printf "Now the setup will back up the cluster.\n"
      printf "The backup is used to restore the cluster to a previous working state in case of an upgrade failure.\n"
      printf "Warning: It is highly recommended to back up all the SQL Databases as well. Without it the data can be inconsistent in the cluster in the event of a restore.\n"
      printf "\n\n"
      printf "Enter your choice (y to continue and q to quit the upgrade) [y/q]:"
      read -r input

      case "${input}" in
          y) printf "Continuing the upgrade process with backup\n";
            echo "Upgrade Consent" > ${CONSENT_FILE};;
          q) exit 0;;
          *) printf "\nInvalid choice\n";
              exit 0;;
      esac
    else
      printf "\n"
      printf "Warning: Backup is not enabled in the Cluster.\n"
      printf "It is highly recommended to have the cluster backed up before upgrade.\n"
      printf "Without a backup, the cluster cannot be restored to a previous working state in case of an upgrade failure.\n"
      printf "\n\n"
      printf "Enter your choice (y to continue and q to quit the upgrade) [y/q]: "
      read -r input

      case "${input}" in
          y) printf "Continuing the upgrade process without backup\n";
            echo "Upgrade Consent" > ${CONSENT_FILE};;
          q) exit 0;;
          *) printf "\nInvalid choice\n";
              exit 0;;
      esac
    fi
  fi

  if [[ ! -f "${BACKUP_MARKER_FILE}" ]] && crontab -l | grep -q backupjob;
  then
    ## Trigger backup of the cluster.
    trigger_on_demand_backup
    ## Pause further backups
    pause_backups_temp
    ## Create a file to indicate Backup was completed
    echo "Backup Complete" > ${BACKUP_MARKER_FILE}
  fi
}

function delete_tmp_files(){
  rm -f ${BACKUP_MARKER_FILE}
  rm -f ${CONSENT_FILE}
}

function argocd_login_internal() {
  local BASE_DIR
  BASE_DIR=$(cd "${MODULES_DIR}/../Fabric_Installer" && pwd)
  [[ -n "${BINARY_DIR:-}" ]] || get_fabric_install_paths "${BASE_DIR}"
  [[ -n "${ARGOCD_NAMESPACE:-}" ]] || get_argocd_install_params &>/dev/null
  ARGOCD_PORT=443
  ARGOCD_FQDN=$(kubectl -n argocd get svc argocd-server -o jsonpath="{.spec.clusterIP}")
  ARGOCD_PASSWORD_SECRET="argocd-admin-password"
  export ARGO_PLAINTEXT="true"
  (login_argocd "${ARGOCD_PORT}" "${ARGOCD_FQDN}" "${ARGOCD_PASSWORD_SECRET}" "argocd" "argocd" &> /dev/null) || true
  unset ARGO_PLAINTEXT
  argocd app list &>/dev/null || info "Could not log in to ArgoCD"
}

function update_configmap_with_replicas(){
  echo "$1" > deploy-replicas.json
  kubectl -n uipath create cm "${MAINTENANCE_MODE_DEPLOY_REPLICA_CM}" --from-file=deploy-replicas.json --dry-run=client -o yaml | kubectl apply -f -
}

function get_stored_replicas(){
  if kubectl -n uipath get cm "${MAINTENANCE_MODE_DEPLOY_REPLICA_CM}" &>/dev/null;
  then
    kubectl -n uipath get cm "${MAINTENANCE_MODE_DEPLOY_REPLICA_CM}" -o jsonpath="{.data.deploy-replicas\.json}"
  else
    echo "[]"
  fi
}

function enable_maintenance_mode() {
  info "Puting the cluster in maintenance mode. This will take some time."
  
  WAIT_TIME="5m"
  if [[ -n "${1:-}" && "$1" == "wait" ]];
  then
    [[ -z "$2" ]] || WAIT_TIME="$2"
  fi

  stop_incoming_connections_istio
  sleep "$WAIT_TIME"
  # Scale down all uipath deployments
  UIPATH_DEPLOYMENTS=$(kubectl -n uipath get deploy --no-headers -o custom-columns=NAME:.metadata.name)
  ORIGINAL_REPLICAS=$(get_stored_replicas)
  for deploy in $UIPATH_DEPLOYMENTS;
  do
    stored_replica=$(echo "$ORIGINAL_REPLICAS" | jq --arg deploy "${deploy}" '.[] | select(.name==$deploy)' | jq '.replicas' | tr -d '"')
    # If there exists already an entry with this deployment's replica count, skip updating it
    if [[ -z "${stored_replica}" ]];
    then
      orig_replica=$(kubectl -n uipath get deploy "$deploy" -o jsonpath="{.spec.replicas}")
      ORIGINAL_REPLICAS=$(jq --arg replica "${orig_replica}" --arg deploy "${deploy}" '. += [{ name: $deploy, replicas : $replica }]' <<< "${ORIGINAL_REPLICAS}")
      update_configmap_with_replicas "${ORIGINAL_REPLICAS}"
    fi
    kubectl -n uipath scale deploy "$deploy" --replicas=0
    kubectl -n uipath rollout status deploy "$deploy" || true
  done

  terminatingpodscount=-1
  try=0
  maxtry=30
  while(( try < maxtry )); do
    try=$(( try + 1 ))
    terminatingpodscount=$(kubectl -n uipath get pods --no-headers | grep -c "Terminating") || true
    if [[ "${terminatingpodscount}" -le 1 ]]; then
      info "All pods in uipath namespace terminated"
      break
    fi
    info "Total pods in Terminating state ${terminatingpodscount}, Waiting for all pods to terminate... ${try}/${maxtry}"; sleep 10
  done

  # Suspend all cronjobs in uipath namespace
  UIPATH_CRONJOBS=$(kubectl -n uipath get cronjobs --no-headers -o custom-columns=NAME:.metadata.name)
  for cronjob in $UIPATH_CRONJOBS;
  do
    kubectl -n uipath patch cronjob "$cronjob" -p '{"spec" : {"suspend" : true }}'
  done

  # Add a label to the node indicating maintenance mode
  info "Adding maintenance label on the nodes"
  ALL_NODES=$(kubectl get nodes --no-headers -o custom-columns=NAME:.metadata.name)
  for node in $ALL_NODES ;
  do
    kubectl label --overwrite "nodes/$node" "uipath-maintenance=true" &>/dev/null
  done
}

function disable_maintenance_mode() {
  info "Removing cluster maintenance mode"
  start_incoming_connections_istio && wait_for_istio_pods
  [[ -n "${1:-}" && "$1" == "--no-argo-login" ]] || argocd_login_internal

  # Scale UP all uipath deployments
  UIPATH_DEPLOYMENTS=$(kubectl -n uipath get deploy --no-headers -o custom-columns=NAME:.metadata.name)
  ORIGINAL_REPLICAS=$(get_stored_replicas)
  if [[ "$ORIGINAL_REPLICAS" != "[]" ]];
  then
    for deploy in $UIPATH_DEPLOYMENTS;
    do
      orig_replica=$(echo "$ORIGINAL_REPLICAS" | jq --arg deploy "${deploy}" '.[] | select(.name==$deploy)' | jq '.replicas' | tr -d '"')
      if [[ -z "${orig_replica}" ]];
      then 
        # If stored replica not found, set it to 1
        orig_replica=1
      fi
      kubectl -n uipath scale deploy "$deploy" --replicas="${orig_replica}"
    done
  fi
  kubectl -n uipath delete cm "${MAINTENANCE_MODE_DEPLOY_REPLICA_CM}" --ignore-not-found

  # Resume all cronjobs in uipath namespace
  UIPATH_CRONJOBS=$(kubectl -n uipath get cronjobs --no-headers -o custom-columns=NAME:.metadata.name)
  for cronjob in $UIPATH_CRONJOBS;
  do
    kubectl -n uipath patch cronjob "$cronjob" -p '{"spec" : {"suspend" : false }}'
  done

  ALL_APPS=$(argocd app list -o name 2>/dev/null) || true
  for app in $ALL_APPS;
  do
    info "Syncing deployments in app ${app}"
    app_deployments=$(argocd app resources "$app" 2>/dev/null | grep Deployment | awk '{print $4}') || true
    for deploy in $app_deployments;
    do
      argocd app sync "$app" --resource "apps:Deployment:$deploy" --force &>/dev/null || true
    done

    info "Syncing cronjobs in app ${app}"
    app_cronjobs=$(argocd app resources "$app" 2>/dev/null | grep CronJob | awk '{print $4}') || true
    for cronjob in $app_cronjobs;
    do
      argocd app sync "$app" --resource "batch:CronJob:$cronjob" --force &>/dev/null || true
    done
  done

  # Scale down all uipath deployments
  UIPATH_DEPLOYMENTS=$(kubectl -n uipath get deploy --no-headers -o custom-columns=NAME:.metadata.name)
  for deploy in $UIPATH_DEPLOYMENTS;
  do
    replicas=$(kubectl -n uipath get deploy "${deploy}" -o jsonpath="{.spec.replicas}")
    if [[ $replicas == 0 ]];
    then
      warn "Could not scale back up $deploy deployments"
    fi
  done

  info "Removing maintenance label from the nodes"
  ALL_NODES=$(kubectl get nodes --no-headers -o custom-columns=NAME:.metadata.name)
  for node in $ALL_NODES ;
  do
    kubectl label --overwrite "nodes/$node" "uipath-maintenance"- &>/dev/null
  done
}

function is_maintenance_mode_enabled() {
  MAINTENANCE_NODES_COUNT=$(kubectl get nodes -o name --selector="uipath-maintenance=true" | wc -l)
  ALL_NODES_COUNT=$(kubectl get nodes -o name | wc -l)
  if [[ "${MAINTENANCE_NODES_COUNT}" == "${ALL_NODES_COUNT}" ]];
  then
    info "Yes. Maintenance mode is enabled on the Cluster"
  else
    info "No. Maintenance mode is not enabled in the Cluster"
  fi
}


function get_node_annotation_value() {
  local NODE_NAME
  local ANNOTATION_NAME
  NODE_NAME=$1
  ANNOTATION_NAME=$2
  VALUE=$(kubectl get nodes/"${NODE_NAME}" -o json | jq -er --arg ANNOTATION "${ANNOTATION_NAME}" '.metadata.annotations[$ANNOTATION]')
  [[ ${VALUE} == null ]] || echo "${VALUE}"
}

function update_node_annotation_value() {
  local NODE_NAME
  local ANNOTATION_NAME
  local ANNOTATION_VALUE
  NODE_NAME=$1
  ANNOTATION_NAME=$2
  ANNOTATION_VALUE=$3
  kubectl annotate --overwrite nodes/"${NODE_NAME}" "${ANNOTATION_NAME}=${ANNOTATION_VALUE}"
}

function determine_first_node_upgrade() {
  local kubectl_nodes
  local version
  local result
  version=$1
  result="true"
  [[ "$(systemctl is-enabled rke2-server 2>/dev/null)" == "enabled" ]] || {
    # If it's an agent node, return false
    echo "false"
    return
  }
  kubectl_nodes=$(kubectl get nodes --no-headers)
  while read -r crt_node; do
    server_name=$(awk '{print $1}' <<<"${crt_node}")
    current_version=$(get_node_annotation_value "${server_name}" "uipath/infra-version")
    [[ "${current_version}" == "${version}" ]] && result="false"
  done <<<"${kubectl_nodes}"
  echo "${result}"
}

function get_enabled_services() {
  local enabled_sevices=()
  local service_iter
  local is_enabled
  ## Get this list from defaults.json
  local all_service_name=("platform" "orchestrator" "test_manager" "automation_hub" "automation_ops" "insights" "action_center" "aicenter" "apps" "documentunderstanding" "task_mining" "serverlesscontrolplane")

  if [[ ${INSTALLSTANDALONEAICENTER} == "1" ]] ; then
    enabled_sevices=("aicenter" "documentunderstanding")
  else
    for service_iter in "${all_service_name[@]}"
    do
      if [[ $service_iter == "platform" ]] || [[ $service_iter == "orchestrator" ]] || [[ $service_iter == "serverlesscontrolplane" ]] ; then
        enabled_sevices+=("${service_iter}")
      else
        is_enabled="$(get_optional_json_value ".${service_iter}.enabled")"
        if [[ -n "${is_enabled}" ]] && [[ $is_enabled == "true" ]] ; then
          enabled_sevices+=("${service_iter}")
        fi
      fi
    done
  fi
  echo "${enabled_sevices[@]}" | tr ' ' ','
}

function add_test_result() {
  local test_status="$1"
  local test_details="$2"
  local test_mitigation="$3"
  local test_color
  local color_reset='\e[0m'

    case "$test_status" in
    PASSED)
      ## Green
      test_color='\u001b[32m'
      ;;
    WARN)
      ## yellow
      test_color='\u001b[33m'
      ;;
    FAILED)
      ## red
      test_color='\u001b[31m'
      ;;
    INFO)
      ##
      test_color='\u001b[36m'
      ;;
    *)
      error "Unsupported test status: ${test_status} while calling add_test_result"
      ;;
  esac

  if [[ -n "${test_mitigation}" ]] ; then
    echo -e ">>> [${test_color}${test_status}${color_reset}] ${test_details}\n    ${test_mitigation}"
  else
    echo -e ">>> [${test_color}${test_status}${color_reset}] ${test_details}"
  fi
}

########
# Label configMap/secret to enable secret/configmap copy to other namespaces
# Arguments:
#   namespace
#   object_type
#   object_name
########
function label_object() {
    local namespace=$1
    local object_type=$2
    local object_name=$3

    check_if_empty namespace "$namespace"
    check_if_empty object_type "$object_type"
    check_if_empty object_name "$object_name"

    kubectl -n "$namespace" label "$object_type" "$object_name" config-discovery=yes 2>/dev/null || true
}