#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS
export PATH="$PATH:/usr/local/sbin:/sbin:/bin:/usr/sbin:/usr/bin:/root/bin:/usr/local/bin"

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  echo "Could not determine script path"
  exit 1
fi

MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  echo "Could not determine Modules path"
  exit 1
fi

export MODULES_DIR

if [[ -f "${MODULES_DIR}/utils.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${MODULES_DIR}/utils.sh"
else
  echo "Common utils not found at ${MODULES_DIR}/utils.sh"
  exit 1
fi

##deployment_type online|offline. Default to online 
DEPLOYMENT_TYPE="online"
##Comma separated list of node's private/internal ips passed via argument
NODES=""
##Config file PATH
CONFIG_FILE_NAME="config.json"
CONFIG="${SCRIPT_DIR}/${CONFIG_FILE_NAME}"
##enabled services
ENABLED_SERVICES_JSON_ARRAY='[]'
##exit code
EXIT_CODE=0

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Run connectivity checks between nodes, outbound connection test for online setup"
  echo
  echo "Arguments"
  echo "  -m|--machines                      Comma seperated IP address / hostnames of all the machines to run connectivity check against"
  echo "  --install-type                     Possible values will be online or offline. Defaults to online"
  echo "  -s|--enabled-services              Comma separated list of enabled services"
  echo "  -d|--debug                         Increase logging verbosity to show all debug logs"
  echo "  -h|--help                          Show this help message and exit"
  echo  
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      echo "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -m|--machines)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        NODES="$2"
        shift 2
      else
        echo "Value for $1 is missing"
        exit 1
      fi
      ;;     

    -s|--enabled-services)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        ENABLED_SERVICES="$2"
        ENABLED_SERVICES_JSON_ARRAY=$(jq -c -n --arg services "${ENABLED_SERVICES}" '[$services | split(",") | .[] | gsub("^\\s+|\\s+$";"")]')
        shift 2
      else
        echo "Value for $1 is missing"
        exit 1
      fi
      ;;   

    --install-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        DEPLOYMENT_TYPE="$2"
        shift 2
      else
        echo "Value for $1 is missing"
        exit 1
      fi
      ;;

    -*|*) # unsupported flags
      # display_usage
      # echo "Error: Unsupported flag $1"
      # exit 1
      shift
      ;;
    esac
  done

  # [[ -z $NODES ]] && echo "Error: Required argument -m|--machines missing while running connectivity check" && exit 1
  NODES_JSON_ARRAY=$(jq -c -n --arg nodes "${NODES}" '[$nodes | split(",") | .[] | gsub("^\\s+|\\s+$";"")]')
  [[ ! -f ${CONFIG} ]] && echo "Error: config file '${CONFIG}' not found" && exit 1
  
}

function check_if_required_packages_available() {
  local package
  local required_packages=("jq" "nc")
  for package in "${required_packages[@]}"
  do
    if ! which "${package}" >> /dev/null 2>&1 ; then
      echo "'${package}' not found in PATH or is not installed. Make sure '${package}' is installed and is available in the PATH"
      exit 1
    fi
  done
}

function transport_connectivity_check() {
  local proto="$1"
  local ports="$2"
  local extra_arg=""
  local node_array
  local timeout=5
  local output
  local connection_status
  local node
  local port
  
  [[ $proto == "udp" ]] &&  extra_arg="--udp"
  node_array=$(echo "$NODES_JSON_ARRAY" | jq -r '.[]')
  while IFS=$'\n' read -r node
  do
    node=$(printf '%s' "$node" | tr -d '[:space:]')
    [[ -z "${node}" ]] && continue
    while IFS=$'\n' read -r port
    do
      port=$(printf '%s' "$port" | tr -d '[:space:]')
      [[ -z "${port}" ]] && continue
      output=$(nc -z --verbose ${extra_arg} --wait ${timeout} "${node}"  "${port}" 2>&1)
      case "$output" in
        *"bytes sent"*)
          connection_status="connected"
          ;;
        *"Connection refused"*)
          connection_status="refused"
          ;;
        *"Connection timed out"*)
          connection_status="timeout"
          ;;
        *)
          connection_status="unknown"
          ;;
      esac

      if [[ $proto == "tcp" ]] ; then
        if [[ $connection_status == "connected" ]] || [[ $connection_status == "refused" ]] ; then
          add_test_result "PASSED" "Successfully connected to ${node} on ${proto} port: ${port}"
        else
          add_test_result "FAILED" "Connection to ${node} on ${proto} port: ${port} failed" "* Ensure firewalld service on destination is stopped and disabled${MITIGATION_DELIMITER}* Ensure TCP port: ${port} is allowed via firewall"
          EXIT_CODE=51
        fi
      else
        if [[ $connection_status == "refused" ]] ; then
          add_test_result "PASSED" "Successfully connected to ${node} on ${proto} port: ${port}"
        elif [[ $connection_status == "connected" ]] ; then
          add_test_result "WARN" "If the check ran when no process was listening on ${proto} port: ${port} on Host: ${node} and ICMP is allowed on the network, then connection to ${node} on ${proto} port: ${port} failed" "* Ensure firewalld service on destination is stopped and disabled${MITIGATION_DELIMITER}* Ensure UDP traffic on port: ${port} is allowed via firewall"
        else
          add_test_result "FAILED" "Connection to ${node} on ${proto} port: ${port} failed" "* Ensure firewalld service on destination is stopped and disabled${MITIGATION_DELIMITER}* Ensure UDP port: ${port} is allowed via firewall"
          EXIT_CODE=51
        fi
      fi
    done <<< "${ports}"
  done <<< "${node_array}"
}

function outbound_connectivity_checks() {
  local url_list
  local url
  local output

  url_list=$(jq -r  --argjson enabledServices "${ENABLED_SERVICES_JSON_ARRAY}"  '.outbound_checks | to_entries[]  | select((.key as $k | $enabledServices | index($k)) or .key=="common").value[]' "${CONFIG}")
  while IFS=$'\n' read -r url
  do
    url=$(printf '%s' "$url" | tr -d '[:space:]')
    [[ -z "${url}" ]] && continue
    output=$(curl -m 10  -w "%{http_code}" --silent --output /dev/null "${url}")
    if [[ $output == "000" ]] ; then
      add_test_result "FAILED" "Connection to ${url} failed" "* Make sure outbound connection is allowed via firewall/proxy server"
      EXIT_CODE=51
    else
      add_test_result "PASSED" "Successfully connected to ${url}"
    fi
  done <<< "${url_list}"
}

function run_validation() {
  local tcp_ports
  local udp_ports

  tcp_ports=$(jq -r '.node_checks.tcp_ports[]' "${CONFIG}")
  udp_ports=$(jq -r '.node_checks.udp_ports[]' "${CONFIG}")
  if [[ -n $NODES ]] ; then
    echo "Validating cluster connections... STARTED"
    transport_connectivity_check "tcp" "${tcp_ports}"
    transport_connectivity_check "udp" "${udp_ports}"
    echo "Validating cluster connections... DONE"
    echo ""
  fi
  [[ $DEPLOYMENT_TYPE == "offline" ]] && return 0
  echo "Validating external connections... STARTED"
  outbound_connectivity_checks
  echo "Validating external connections... DONE"
}


function main() {
  parse_long_args "$@"
  check_if_required_packages_available
  echo "=====================Connectivity Validation======================="
  run_validation
  echo "==================================================================="
}

main "$@"
exit "${EXIT_CODE}"