#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

#GLOBALS:
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

readonly SCRIPT_DIR
readonly MODULES_DIR
export MISSING_DEGRADED_APP_LIST=""

declare -a ADDITIONAL_PARAMETERS
declare -a MANDATORY_PARAMETERS
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"
#shellcheck disable=SC1090,SC1091,SC2034
source "${MODULES_DIR}/argocd_utils.sh"
#shellcheck disable=SC1090,SC1091,SC2034
source "${MODULES_DIR}/cluster_utils.sh"
#shellcheck disable=SC1090,SC1091,SC2034
source "${MODULES_DIR}/k8sutils.sh"
#shellcheck disable=SC1090,SC1091,SC2034
source "${SCRIPT_DIR}/common-services-installer.sh"
#FUNCTIONS

function parse_standalone_aicenter_input_file() {
  ORCHESTRATOR_URL="$(get_mandatory_json_value ".aicenter.orchestrator_url")"
  [[ "${ORCHESTRATOR_URL: -1}" != "/" ]] && ORCHESTRATOR_URL+="/"

  IDENTITY_SERVER_URL="$(get_mandatory_json_value ".aicenter.identity_server_url")"
  ORCHESTRATOR_CERT_FILE_PATH="$(get_mandatory_json_value ".aicenter.orchestrator_cert_file_path")"
  ORCHESTRATOR_CERT_FILE_PATH="${ORCHESTRATOR_CERT_FILE_PATH/#\~/$HOME}"
  IDENTITY_CERT_FILE_PATH="$(get_mandatory_json_value ".aicenter.identity_cert_file_path")"
  IDENTITY_CERT_FILE_PATH="${IDENTITY_CERT_FILE_PATH/#\~/$HOME}"
  IDENTITY_ACCESS_TOKEN="$(get_mandatory_json_value ".aicenter.identity_access_token")"
}

function get_additional_parameters() {
  ADDITIONAL_PARAMETERS=()
    [ -z "${DOCKER_REGISTRY_PULL_SECRET}" ] || ADDITIONAL_PARAMETERS+=("-p global.imagePullSecrets[0]='uipathpullsecret'" )

  ADDITIONAL_PARAMETERS+=(-p global.applications.apps.aicenter.standalone.enabled="true" )
  ADDITIONAL_PARAMETERS+=(-p global.applications.apps.platform.enabled="false" )
}

function read_orch_certificate_files() {
  if [[ -f ${ORCHESTRATOR_CERT_FILE_PATH} ]]; then
    ORCHESTRATOR_CERT_FILE_CONTENT=$(base64 -i "${ORCHESTRATOR_CERT_FILE_PATH}")
    # shellcheck disable=SC2001,2086
    ORCHESTRATOR_CERT_FILE_CONTENT=$(echo ${ORCHESTRATOR_CERT_FILE_CONTENT} | sed -e 's/[[:space:]]//g')
  else
    error "${ORCHESTRATOR_CERT_FILE_PATH} not found"
  fi

  if [[ -f ${IDENTITY_CERT_FILE_PATH} ]]; then
    IDENTITY_CERT_FILE_CONTENT=$(base64 -i "${IDENTITY_CERT_FILE_PATH}")
    # shellcheck disable=SC2001,2086
    IDENTITY_CERT_FILE_CONTENT=$(echo ${IDENTITY_CERT_FILE_CONTENT} | sed -e 's/[[:space:]]//g')
  else
    IDENTITY_CERT_FILE_CONTENT=${ORCHESTRATOR_CERT_FILE_CONTENT}
  fi
}

function install_app_of_apps() {  
  get_additional_parameters

  get_mandatory_app_create_parameters

  argocd app create uipath \
    --repo "${APP_OF_APPS_REPO}" \
    --revision "${APP_OF_APPS_VERSION}" \
    --helm-chart "helm/${APP_OF_APPS_CHART}" \
    --dest-server https://kubernetes.default.svc \
    --upsert \
    --sync-policy "automated" \
    --sync-retry-backoff-duration 10s \
    --sync-retry-backoff-factor 2 \
    --sync-retry-backoff-max-duration 20m \
    --sync-retry-limit 4 \
    "${ADDITIONAL_PARAMETERS[@]}" \
    "${MANDATORY_PARAMETERS[@]}" \
    --values values-aicenter.yaml \
    -p global.userInputs.aicenter.sqlConnectionString="${AICENTER_SQL_CONNECTION:-PLACEHOLDER}" \
    -p global.userInputs.aicenter.orchestratorUrl="${ORCHESTRATOR_URL}" \
    -p global.userInputs.aicenter.identityServerUrl="${IDENTITY_SERVER_URL}" \
    -p global.userInputs.aicenter.orchestratorCert="${ORCHESTRATOR_CERT_FILE_CONTENT}" \
    -p global.userInputs.aicenter.identityCert="${IDENTITY_CERT_FILE_CONTENT}" \
    -p global.userInputs.aicenter.identityAccessToken="${IDENTITY_ACCESS_TOKEN}" \
    -p global.userInputs.documentunderstanding.datamanager.sqlConnectionString="${DU_DATAMANAGER_SQL_CONNECTION:-PLACEHOLDER}" \
    || error "Failed to create app of apps"

  info "Application uipath created"
  
  sync_with_wait "uipath" "argocd" "app.kubernetes.io/instance=uipath" "service-fabric=service-presync-checks" "uipath"
}

function set_output() {
  cat > "${OUTFILE}" <<EOF
{ 
  "aicenter_url": "https://${CLUSTER_FQDN}/aifabric_",
  "documentunderstanding_url": "https://${CLUSTER_FQDN}/du_"
}
EOF
}

#MAIN
function main() {

  cd "${SCRIPT_DIR}"
  parse_args "$@"

  load_os_utils "$SCRIPT_DIR/../Modules"
  
  check_requirements
  
  parse_sf_input_file

  parse_aicenter_input_file

  parse_documentunderstanding_input_file

  parse_standalone_aicenter_input_file

  send_checkpoint_telemetry "AiCenter Services: Input Parsed"

  [[ "$(is_update_fqdn)" == "true" ]] && update_service_start_fqdn_state

  get_fabric_install_paths "${SCRIPT_DIR}"
  get_argocd_install_params
  login_argocd "${ARGOCD_NODEPORT_HTTP}" "${ARGOCD_FQDN}" "${ARGOCD_PASSWORD_SECRET_NAME}" "${ARGOCD_NAMESPACE}" "argocd"

  validate_platform_aicenter_installation_option "aicenter"

  parse_offline_input_parameters

  local try=0
  local maxtry=30
  local status="notready"
  while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
          try=$((try+1))
          kubectl cluster-info >/dev/null 2>&1  && status="ready"
          info "waiting for cluster to be ready...${try}/${maxtry}" &&  sleep 10;
  done
  kubectl cluster-info || error "Could not reach the cluster API"

  declare -a namespaces=("uipath")

  kubectl cluster-info || error "KubeAPI connection failed"

  argocd cluster list || error "ArgoCD connection failed"

  send_checkpoint_telemetry "AiCenter Services: Cluster Ready"

  create_namespaces "${namespaces[@]}"

  read_orch_certificate_files

  check_available_masters
  
  create_preinstall_override_config

  ensure_cluster_id

  send_checkpoint_telemetry "AiCenter Services: ClusterId Created"
  send_input_telemetry

  install_app_of_apps

  [[ "$(is_update_fqdn)" == "true" ]] && update_fqdn "${namespaces[@]}"

  set_output

  send_checkpoint_telemetry "AiCenter Services: Success"
  info "UiPath Services installed successfully."
}

main "$@"