#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#shellcheck disable=SC1090,SC1091,SC2034

########
# Get sql connection string from config json and base64 encode
# Use global template if string not found
# Globals:
#   INFILE
# Arguments:
#   json_path
#   sql_template_type
#   default_db_name
# Outputs:
#   Returns the base64 encoded value or exits if not found
########
function get_sql_connection_string_b64() {
  local connection_string connection_string_decoded
  connection_string=$(get_optional_sql_connection_string_b64 "$@")
  connection_string_decoded=$(echo "${connection_string}" | base64 -d)

  if [[ -n ${connection_string_decoded} ]];
  then
    echo "${connection_string}"
  else
    local json_path="$1"
    local sql_template_type="$2"
    error "Missing sql param ${json_path} and no sql_connection_string_template of ${sql_template_type} was specified."
  fi
}

########
# Get sql connection string from config json and base64 encode
# Globals:
#   INFILE
# Arguments:
#   json_path
#   sql_template_type
#   default_db_name
# Outputs:
#   Returns the base64 encoded value or default b64 value if not found
########
function get_optional_sql_connection_string_b64() {
  local json_path="$1"
  local sql_template_type="$2"
  local default_db_name="$3"
  local value
  local default_value
  local found
  
  local create_db_string
  create_db_string="$(get_json_value_or_return_default_value ".sql.create_db" "false")"
  value="$(jq -er "${json_path}" < "${INFILE}")"
  found=$?
  if [[ ${create_db_string} == "true" ||  ${found} != 0 ]]; then
    default_value="$(create_sql_connection_string_from_template "${sql_template_type}" "${default_db_name}")"
    echo "${default_value}"
  else
    echo -n "${value}" | base64 -w 0 | sed -e 's/[[:space:]]//g'
  fi
}

function create_sql_connection_string_from_template() {
  local sql_template_type="$1"
  local default_db_name="$2"
  local sql_connection_string_template sql_connection_string_template_jdbc sql_connection_string_template_odbc
  sql_connection_string_template="$(get_optional_json_value ".sql_connection_string_template")"
  sql_connection_string_template_jdbc="$(get_optional_json_value ".sql_connection_string_template_jdbc")"
  sql_connection_string_template_odbc="$(get_optional_json_value ".sql_connection_string_template_odbc")"

  if [[ -n ${sql_connection_string_template} && ${sql_template_type} == "dotnet" ]]; then
    default_value="${sql_connection_string_template//DB_NAME_PLACEHOLDER/$default_db_name}"
    echo -n "${default_value}" | base64 -w 0 | sed -e 's/[[:space:]]//g'
  elif [[ -n ${sql_connection_string_template_jdbc} && ${sql_template_type} == "jdbc" ]]; then
    default_value="${sql_connection_string_template_jdbc//DB_NAME_PLACEHOLDER/$default_db_name}"
    echo -n "${default_value}" | base64 -w 0 | sed -e 's/[[:space:]]//g'
  elif [[ -n ${sql_connection_string_template_odbc} && ${sql_template_type} == "odbc" ]]; then
    default_value="${sql_connection_string_template_odbc//DB_NAME_PLACEHOLDER/$default_db_name}"
    echo -n "${default_value}" | base64 -w 0 | sed -e 's/[[:space:]]//g'
  else
    # leave newline to prevent empty secret strings causing malformed base64 secrets
    echo "" | base64 -w 0 | sed -e 's/[[:space:]]//g'
  fi
}

function check_sql_connection_str(){
  # empty value 64 encoding Cg==
  if [[ "$1" == "Cg==" && "$2" == "" ]]; then
    echo "0"
  else
    echo "1"
  fi
}

########
# Parse input json file to set the config values
# Globals:
#   INFILE
# Outputs:
#   Sets global config variables
########
function parse_input_file() {

  parse_sf_input_file

  ORCHESTRATOR_SQL_CONNECTION="$(get_sql_connection_string_b64 ".orchestrator.sql_connection_str" "dotnet" "AutomationSuite_Orchestrator")"
  ORCHESTRATOR_TA_SQL_CONNECTION="${ORCHESTRATOR_SQL_CONNECTION}"
  ORCHESTRATOR_TA_SQL_CONNECTION_OVERRIDE="$(get_optional_json_value ".orchestrator.testautomation.sql_connection_str")"
  [[ ${CREATE_DB} == "false" && ${ORCHESTRATOR_TA_SQL_CONNECTION_OVERRIDE} != "" ]] && ORCHESTRATOR_TA_SQL_CONNECTION="$(get_mandatory_b64enc_json_value ".orchestrator.testautomation.sql_connection_str" "dotnet" "AutomationSuite_Orchestrator")"
  ORCHESTRATOR_TESTAUTOMATION_ENABLED="$(get_optional_json_value ".orchestrator.testautomation.enabled")"
  ORCHESTRATOR_ROBOT_LOGS_ELASTIC_URI="$(get_optional_json_value ".orchestrator.orchestrator_robot_logs_elastic.elastic_uri")"
  ORCHESTRATOR_ROBOT_LOGS_ELASTIC_AUTH_USERNAME="$(get_optional_json_value ".orchestrator.orchestrator_robot_logs_elastic.elastic_auth_username")"
  ORCHESTRATOR_ROBOT_LOGS_ELASTIC_AUTH_PASSWORD="$(get_optional_json_value ".orchestrator.orchestrator_robot_logs_elastic.elastic_auth_password")"

  PLATFORM_SQL_CONNECTION="$(get_optional_sql_connection_string_b64 ".platform.sql_connection_str" "dotnet" "AutomationSuite_Platform")"
  LOCATION_SQL_CONNECTION="${PLATFORM_SQL_CONNECTION}"
  LOCATION_SQL_CONNECTION_OVERRIDE="$(get_optional_json_value ".platform.location.sql_connection_str")"
  (( $(check_sql_connection_str "${PLATFORM_SQL_CONNECTION}" "${LOCATION_SQL_CONNECTION_OVERRIDE}") == 0 )) && error_with_code "13010" ".platform.location.sql_connection_str" ".platform.sql_connection_str"
  [[ ${CREATE_DB} == "false" && ${LOCATION_SQL_CONNECTION_OVERRIDE} != "" ]] && LOCATION_SQL_CONNECTION="$(get_mandatory_b64enc_json_value ".platform.location.sql_connection_str" "dotnet" "AutomationSuite_Platform")"
  OMS_SQL_CONNECTION="${PLATFORM_SQL_CONNECTION}"
  OMS_SQL_CONNECTION_OVERRIDE="$(get_optional_json_value ".platform.oms.sql_connection_str")"
  (( $(check_sql_connection_str "${PLATFORM_SQL_CONNECTION}" "${OMS_SQL_CONNECTION_OVERRIDE}") == 0 )) && error_with_code "13010" ".platform.oms.sql_connection_str" ".platform.sql_connection_str"
  [[ ${CREATE_DB} == "false" && ${OMS_SQL_CONNECTION_OVERRIDE} != "" ]] && OMS_SQL_CONNECTION="$(get_mandatory_b64enc_json_value ".platform.oms.sql_connection_str" "dotnet" "AutomationSuite_Platform")"
  IDENTITY_SQL_CONNECTION="${PLATFORM_SQL_CONNECTION}"
  IDENTITY_SQL_CONNECTION_OVERRIDE="$(get_optional_json_value ".platform.identity.sql_connection_str")"
  (( $(check_sql_connection_str "${PLATFORM_SQL_CONNECTION}" "${IDENTITY_SQL_CONNECTION_OVERRIDE}") == 0 )) && error_with_code "13010" ".platform.identity.sql_connection_str" ".platform.sql_connection_str"
  [[ ${CREATE_DB} == "false" && ${IDENTITY_SQL_CONNECTION_OVERRIDE} != "" ]] && IDENTITY_SQL_CONNECTION="$(get_mandatory_b64enc_json_value ".platform.identity.sql_connection_str" "dotnet" "AutomationSuite_Platform")"
  LRM_SQL_CONNECTION="${PLATFORM_SQL_CONNECTION}"
  LRM_SQL_CONNECTION_OVERRIDE="$(get_optional_json_value ".platform.lrm.sql_connection_str")"
  (( $(check_sql_connection_str "${PLATFORM_SQL_CONNECTION}" "${LRM_SQL_CONNECTION_OVERRIDE}") == 0 )) && error_with_code "13010" ".platform.lrm.sql_connection_str" ".platform.sql_connection_str"
  [[ ${CREATE_DB} == "false" && ${LRM_SQL_CONNECTION_OVERRIDE} != "" ]] && LRM_SQL_CONNECTION="$(get_mandatory_b64enc_json_value ".platform.lrm.sql_connection_str" "dotnet" "AutomationSuite_Platform")"
  LA_SQL_CONNECTION="${PLATFORM_SQL_CONNECTION}"
  LA_SQL_CONNECTION_OVERRIDE="$(get_optional_json_value ".platform.la.sql_connection_str")"
  (( $(check_sql_connection_str "${PLATFORM_SQL_CONNECTION}" "${LA_SQL_CONNECTION_OVERRIDE}") == 0 )) && error_with_code "13010" ".platform.la.sql_connection_str" ".platform.sql_connection_str"
  [[ ${CREATE_DB} == "false" && ${LA_SQL_CONNECTION_OVERRIDE} != "" ]] && LA_SQL_CONNECTION="$(get_mandatory_b64enc_json_value ".platform.la.sql_connection_str" "dotnet" "AutomationSuite_Platform")"
  AUDIT_SQL_CONNECTION="${PLATFORM_SQL_CONNECTION}"
  AUDIT_SQL_CONNECTION_OVERRIDE="$(get_optional_json_value ".platform.audit.sql_connection_str")"
  (( $(check_sql_connection_str "${PLATFORM_SQL_CONNECTION}" "${AUDIT_SQL_CONNECTION_OVERRIDE}") == 0 )) && error_with_code "13010" ".platform.audit.sql_connection_str" ".platform.sql_connection_str"
  [[ ${CREATE_DB} == "false" && ${AUDIT_SQL_CONNECTION_OVERRIDE} != "" ]] && AUDIT_SQL_CONNECTION="$(get_mandatory_b64enc_json_value ".platform.audit.sql_connection_str" "dotnet" "AutomationSuite_Platform")"
  ORCHESTRATOR_UPDATESERVER_ENABLED="$(get_optional_json_value ".orchestrator.updateserver.enabled")"
  ORCHESTRATOR_UPD_SQL_CONNECTION="${PLATFORM_SQL_CONNECTION}"
  ORCHESTRATOR_UPD_SQL_CONNECTION_OVERRIDE="$(get_optional_json_value ".orchestrator.updateserver.sql_connection_str")"
  (( $(check_sql_connection_str "${PLATFORM_SQL_CONNECTION}" "${ORCHESTRATOR_UPD_SQL_CONNECTION_OVERRIDE}") == 0 )) && error_with_code "13010" ".orchestrator.updateserver.sql_connection_str" ".platform.sql_connection_str"
  [[ ${CREATE_DB} == "false" && ${ORCHESTRATOR_UPD_SQL_CONNECTION_OVERRIDE} != "" ]] && ORCHESTRATOR_UPD_SQL_CONNECTION="$(get_mandatory_b64enc_json_value ".orchestrator.updateserver.sql_connection_str" "dotnet" "AutomationSuite_Platform")"

  TESTMANAGER_SQL_CONNECTION="$(get_optional_sql_connection_string_b64 ".test_manager.sql_connection_str" "dotnet" "AutomationSuite_Test_Manager")"

  AUTOMATIONOPS_SQL_CONNECTION="${PLATFORM_SQL_CONNECTION}"
  AUTOMATIONOPS_SQL_CONNECTION_OVERRIDE="$(get_optional_json_value ".automation_ops.sql_connection_str")"
  (( $(check_sql_connection_str "${PLATFORM_SQL_CONNECTION}" "${AUTOMATIONOPS_SQL_CONNECTION_OVERRIDE}") == 0 )) && error_with_code "13010" ".automation_ops.sql_connection_str" ".platform.sql_connection_str"
  [[ ${CREATE_DB} == "false" && ${AUTOMATIONOPS_SQL_CONNECTION_OVERRIDE} != "" ]] && AUTOMATIONOPS_SQL_CONNECTION="$(get_mandatory_b64enc_json_value ".automation_ops.sql_connection_str" "dotnet" "AutomationSuite_Platform")"

  AUTOMATIONHUB_SQL_CONNECTION="$(get_optional_sql_connection_string_b64 ".automation_hub.sql_connection_str" "dotnet" "AutomationSuite_Automation_Hub")"
  
  INSIGHTS_SQL_CONNECTION="$(get_optional_sql_connection_string_b64 ".insights.sql_connection_str" "dotnet" "AutomationSuite_Insights")"
  INSIGHTS_SMTP_TLS_VERSION="$(get_optional_json_value ".insights.smtp_configuration.tls_version")"
  INSIGHTS_SMTP_FROM_EMAIL="$(get_optional_json_value ".insights.smtp_configuration.from_email")"
  INSIGHTS_SMTP_HOST="$(get_optional_json_value ".insights.smtp_configuration.host")"
  INSIGHTS_SMTP_PORT="$(get_optional_json_value ".insights.smtp_configuration.port")"
  INSIGHTS_SMTP_USERNAME="$(get_optional_json_value ".insights.smtp_configuration.username")"
  INSIGHTS_SMTP_PASSWORD="$(get_optional_b64enc_json_value ".insights.smtp_configuration.password")"

  TASKMINING_SQL_CONNECTION="$(get_optional_sql_connection_string_b64 ".task_mining.sql_connection_str" "dotnet" "AutomationSuite_Task_Mining")"

  IDENTITY_HOST_ADMIN_USERNAME="$(get_mandatory_json_value ".admin_username")"
  IDENTITY_HOST_ADMIN_PASSWORD="$(get_mandatory_b64enc_json_value ".admin_password")"
  IDENTITY_SAML_CURRENT_SERVICE_CERT_FILE="$(get_optional_json_value ".identity_certificate.saml_current_service_cert_file")"
  IDENTITY_SAML_CURRENT_SERVICE_CERT_FILE="${IDENTITY_SAML_CURRENT_SERVICE_CERT_FILE/#\~/$HOME}"
  IDENTITY_SAML_CURRENT_SERVICE_CERT_PASS="$(get_optional_b64enc_json_value ".identity_certificate.saml_current_service_cert_pass")"
  IDENTITY_SAML_FUTURE_SERVICE_CERT_FILE="$(get_optional_json_value ".identity_certificate.saml_future_service_cert_file")"
  IDENTITY_SAML_FUTURE_SERVICE_CERT_FILE="${IDENTITY_SAML_FUTURE_SERVICE_CERT_FILE/#\~/$HOME}"
  IDENTITY_SAML_FUTURE_SERVICE_CERT_PASS="$(get_optional_b64enc_json_value ".identity_certificate.saml_future_service_cert_pass")"
  IDENTITY_LDAPS_CUSTOM_ROOTCA_FILE="$(get_optional_json_value ".identity_certificate.ldap_cert_authority_file")"
  IDENTITY_LDAPS_CUSTOM_ROOTCA_FILE="${IDENTITY_LDAPS_CUSTOM_ROOTCA_FILE/#\~/$HOME}"
  IDENTITY_KRB5_KEYTAB="$(get_optional_json_value ".identity_krb5_keytab_secret")"
  
  ORCHESTRATOR_KERBEROS_AD_USER="$(get_optional_json_value ".orchestrator.kerberos_auth_config.ad_username")"
  ORCHESTRATOR_KERBEROS_KEYTAB="$(get_optional_json_value ".orchestrator.kerberos_auth_config.user_keytab")"
  PLATFORM_KERBEROS_AD_USER="$(get_optional_json_value ".platform.kerberos_auth_config.ad_username")"
  PLATFORM_KERBEROS_KEYTAB="$(get_optional_json_value ".platform.kerberos_auth_config.user_keytab")"
  DISCOVERYGROUP_KERBEROS_AD_USER="$(get_optional_json_value ".discovery_group.kerberos_auth_config.ad_username")"
  DISCOVERYGROUP_KERBEROS_KEYTAB="$(get_optional_json_value ".discovery_group.kerberos_auth_config.user_keytab")"
  TESTMANAGER_KERBEROS_AD_USER="$(get_optional_json_value ".test_manager.kerberos_auth_config.ad_username")"
  TESTMANAGER_KERBEROS_KEYTAB="$(get_optional_json_value ".test_manager.kerberos_auth_config.user_keytab")"
  AUTOMATIONOPS_KERBEROS_AD_USER="$(get_optional_json_value ".automation_ops.kerberos_auth_config.ad_username")"
  AUTOMATIONOPS_KERBEROS_KEYTAB="$(get_optional_json_value ".automation_ops.kerberos_auth_config.user_keytab")"
  AICENTER_KERBEROS_AD_USER="$(get_optional_json_value ".aicenter.kerberos_auth_config.ad_username")"
  AICENTER_KERBEROS_KEYTAB="$(get_optional_json_value ".aicenter.kerberos_auth_config.user_keytab")"
  DOCUMENTUNDERSTANDING_KERBEROS_AD_USER="$(get_optional_json_value ".documentunderstanding.kerberos_auth_config.ad_username")"
  DOCUMENTUNDERSTANDING_KERBEROS_KEYTAB="$(get_optional_json_value ".documentunderstanding.kerberos_auth_config.user_keytab")"
  INSIGHTS_KERBEROS_AD_USER="$(get_optional_json_value ".insights.kerberos_auth_config.ad_username")"
  INSIGHTS_KERBEROS_KEYTAB="$(get_optional_json_value ".insights.kerberos_auth_config.user_keytab")"

  parse_aicenter_input_file

  parse_documentunderstanding_input_file
}

########
# Parse sf core platform input json file to set the config values
# Globals:
#   INFILE
# Outputs:
#   Sets global config variables
########
function parse_sf_input_file() {
  VERSION="$(get_optional_json_value ".version")"
  CLUSTER_FQDN="$(get_mandatory_json_value ".fqdn")"
  LOOKER_FQDN="$(get_mandatory_json_value ".subdomain.looker").${CLUSTER_FQDN}"
  CEPH_FQDN="$(get_mandatory_json_value ".subdomain.rook_ceph").${CLUSTER_FQDN}"
  CEPH_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
  ARGOCD_FQDN="$(get_mandatory_json_value ".subdomain.argo_cd").${CLUSTER_FQDN}"
  DOCKER_REGISTRY_FQDN="$(get_mandatory_json_value ".subdomain.docker_registry").${CLUSTER_FQDN}"
  DOCKER_REGISTRY_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
  LOOKER_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"

  PROFILE="$(get_mandatory_json_value ".profile")"
  STORAGE_CLASS="$(get_mandatory_json_value ".storage_class")"
  STORAGE_CLASS_SINGLE_REPLICA="$(get_mandatory_json_value ".storage_class_single_replica")"
  DOCKER_REGISTRY_URL="$(get_mandatory_json_value ".registries.docker.url")"
  DOCKER_REGISTRY_NODEPORT="$(get_mandatory_json_value ".infra.docker_registry.node_port")"
  DOCKER_REGISTRY_PULL_SECRET="$(get_optional_json_value ".registries.docker.pull_secret_value")"
  TELEMETRY_OPT_OUT="$(get_optional_json_value ".telemetry_optout")"

  APP_OF_APPS_REPO="$(get_mandatory_json_value ".registries.helm.url")"
  APP_OF_APPS_VERSION="$(get_mandatory_json_value ".version")"
  APP_OF_APPS_CHART="$(get_mandatory_json_value ".services.chart")"

  PROXY_ENABLED="$(get_optional_json_value ".proxy.enabled")"
  HTTP_PROXY="$(get_optional_json_value ".proxy.http_proxy")"
  HTTPS_PROXY="$(get_optional_json_value ".proxy.https_proxy")"
  NO_PROXY="$(get_optional_json_value ".proxy.no_proxy")"

  ENABLESQLINTEGRATEDAUTH="$(get_optional_json_value ".kerberos_auth_config.enabled")"
  ADDOMAIN="$(get_optional_json_value ".kerberos_auth_config.ad_domain")"
  ADUSERNAME="$(get_optional_json_value ".kerberos_auth_config.default_ad_username")"
  USERKEYTAB="$(get_optional_json_value ".kerberos_auth_config.default_user_keytab")"
  TICKETLIFETIME="$(get_optional_json_value ".kerberos_auth_config.ticket_lifetime_in_hour")"

  CREATE_DB="$(get_json_value_or_return_default_value ".sql.create_db" "false")"
  # We used default connection string to create services database
  SQL_DEFAULT_CONNECTION_STRING="$(get_optional_json_value ".sql_connection_string_template")"
  if [[ -n ${SQL_DEFAULT_CONNECTION_STRING} ]]; then
    SQL_DEFAULT_CONNECTION_STRING="${SQL_DEFAULT_CONNECTION_STRING//DB_NAME_PLACEHOLDER/master}"
    SQL_DEFAULT_CONNECTION_STRING="$(echo -n "${SQL_DEFAULT_CONNECTION_STRING}" | base64 -w 0 | sed -e 's/[[:space:]]//g')"
  fi

  RKE_TOKEN="$(get_mandatory_json_value ".rke_token")"
  ZONE_RESILIENCE="$(get_json_value_or_return_default_value ".zone_resilience" "false")"

  # Disable telemetry in case of offline
  if [[ "${INSTALL_TYPE}" == "offline" ]];
  then
    TELEMETRY_OPT_OUT="true"
  fi
}

function parse_aicenter_input_file() {
  AICENTER_SQL_CONNECTION="$(get_optional_sql_connection_string_b64 ".aicenter.sql_connection_str" "jdbc" "AutomationSuite_AICenter")"
}

function parse_documentunderstanding_input_file() {
  DU_DATAMANAGER_SQL_CONNECTION="$(get_optional_sql_connection_string_b64 ".documentunderstanding.datamanager.sql_connection_str" "odbc" "AutomationSuite_DU_Datamanager")"
  DU_SERVICES_HANDWRITING_ENABLED="$(get_optional_json_value ".documentunderstanding.handwriting.enabled")"
  DU_SERVICES_HANDWRITING_MAX_CPU_PER_POD="$(get_optional_json_value ".documentunderstanding.handwriting.max_cpu_per_pod")"
}

function parse_offline_input_parameters() {  
  if [[ ${INSTALL_TYPE} == 'offline' ]];then
    APP_OF_APPS_REPO="${DOCKER_REGISTRY_FQDN}:${DOCKER_REGISTRY_EXTERNAL_PORT}"
    DOCKER_REGISTRY_URL="localhost:${DOCKER_REGISTRY_NODEPORT}"
  fi
}

function get_enabled_services() {
  SERVICES_ENABLED=()
  SERVICE_NAMES_ENABLED=()
  ALL_SERVICE_NAMES=("platform" "orchestrator" "testmanager" "automationhub" "automationops" "insights" "actioncenter" "aicenter" "ba" "documentunderstanding" "taskmining" )
  local val
  
  # platform and orchestrator are enabled by default
  SERVICES_ENABLED+=(-p global.applications.apps.platform.enabled="true" )
  SERVICE_NAMES_ENABLED+=("platform" "orchestrator")
 
 # actioncenter
  val="$(get_optional_json_value ".action_center.enabled")"
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.actioncenter.enabled="${val}")
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.actions.enabled="${val}")
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.processes.enabled="${val}")
  [ "$val" != "true" ] || SERVICE_NAMES_ENABLED+=("actioncenter")

  #aicenter
  val="$(get_optional_json_value ".aicenter.enabled")"
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.aicenter.enabled="${val}")
  [ "$val" != "true" ] || SERVICE_NAMES_ENABLED+=("aicenter")

  #automationhub
  val="$(get_optional_json_value ".automation_hub.enabled")"
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.automationhub.enabled="${val}")
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.automationstore.enabled="${val}")
  [ "$val" != "true" ] || SERVICE_NAMES_ENABLED+=("automationhub")

  #automationops
  val="$(get_optional_json_value ".automation_ops.enabled")"
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.automationops.enabled="${val}")
  [ "$val" != "true" ] || SERVICE_NAMES_ENABLED+=("automationops")

  #business-apps
  val="$(get_optional_json_value ".apps.enabled")"
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.ba.enabled="${val}")
  [ "$val" != "true" ] || SERVICE_NAMES_ENABLED+=("ba")

  #DU
  val="$(get_optional_json_value ".documentunderstanding.enabled")"
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.documentunderstanding.enabled="${val}")
  [ "$val" != "true" ] || SERVICE_NAMES_ENABLED+=("documentunderstanding")
  
  #Insights
  val="$(get_optional_json_value ".insights.enabled")"
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.insights.enabled="${val}")
  [ "$val" != "true" ] || SERVICE_NAMES_ENABLED+=("insights")

  #taskmining
  val="$(get_optional_json_value ".task_mining.enabled")"
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.taskmining.enabled="${val}")
  [ "$val" != "true" ] || SERVICE_NAMES_ENABLED+=("taskmining")

  #testmanager
  val="$(get_optional_json_value ".test_manager.enabled")"
  [ -z "$val" ] || SERVICES_ENABLED+=(-p global.applications.apps.testmanager.enabled="${val}")
  [ "$val" != "true" ] || SERVICE_NAMES_ENABLED+=("testmanager")
}

function check_requirements() {
    local required_pkgs=(kubectl helm uuidgen argocd jq)
    check_required_packages_and_commands "${required_pkgs[@]}"
}

function create_namespaces() {
  local namespaces=("$@")

  for namespace in "${namespaces[@]}"
  do
    kubectl get namespace "${namespace}" \
      || kubectl create namespace "${namespace}" \
      || error "Failed to create namespace ${namespace}"

    # Don't fail if the label is already set
    kubectl label namespace "${namespace}" istio-injection=enabled || true
    kubectl label namespace "${namespace}" uipath-injection=enabled || true
  done
}

function read_identity_certificate_files() {

  # load optional SAML service certificates
  if [[ -f ${IDENTITY_SAML_CURRENT_SERVICE_CERT_FILE} ]]; then
    IDENTITY_SAML_CURRENT_SERVICE_CERT_CONTENT=$(base64 -i "${IDENTITY_SAML_CURRENT_SERVICE_CERT_FILE}")
    # shellcheck disable=SC2001,2086
    IDENTITY_SAML_CURRENT_SERVICE_CERT_CONTENT=$(echo ${IDENTITY_SAML_CURRENT_SERVICE_CERT_CONTENT} | sed -e 's/[[:space:]]//g')
  else
    # have empty placeholder value
    IDENTITY_SAML_CURRENT_SERVICE_CERT_CONTENT=""
  fi
  if [[ -f ${IDENTITY_SAML_FUTURE_SERVICE_CERT_FILE} ]]; then
    IDENTITY_SAML_FUTURE_SERVICE_CERT_CONTENT=$(base64 -i "${IDENTITY_SAML_FUTURE_SERVICE_CERT_FILE}")
    # shellcheck disable=SC2001,2086
    IDENTITY_SAML_FUTURE_SERVICE_CERT_CONTENT=$(echo ${IDENTITY_SAML_FUTURE_SERVICE_CERT_CONTENT} | sed -e 's/[[:space:]]//g')
  else
    # have empty placeholder value
    IDENTITY_SAML_FUTURE_SERVICE_CERT_CONTENT=""
  fi
    # load optional LDAPS custom rootCA
  if [[ -f ${IDENTITY_LDAPS_CUSTOM_ROOTCA_FILE} ]]; then
    IDENTITY_LDAPS_CUSTOM_ROOTCA_CONTENT=$(base64 -i "${IDENTITY_LDAPS_CUSTOM_ROOTCA_FILE}")
    # shellcheck disable=SC2001,2086
    IDENTITY_LDAPS_CUSTOM_ROOTCA_CONTENT=$(echo ${IDENTITY_LDAPS_CUSTOM_ROOTCA_CONTENT} | sed -e 's/[[:space:]]//g')
  else
    # have empty placeholder value
    IDENTITY_LDAPS_CUSTOM_ROOTCA_CONTENT=""
  fi
}

# generate unique identifier to send with telemetry so customer can be identified in AppInsights.
# don't regenerate it if it is already there.
function ensure_cluster_id() {
  local app_details

  local try=0
  local maxtry=3
  local status="notready"
  while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
          try=$((try+1))
          app_details=$(argocd app get uipath --show-params 2>/dev/null) && status="ready"
          info "waiting for the uipath argo app details ...${try}/${maxtry}" &&  sleep 5;
  done
  
  [[ ${status} == "ready" ]] && CLUSTER_ID=$(echo "${app_details}" | grep telemetry.clusterId | awk '{print $2}')
  [[ ${status} == "notready" ]] && CLUSTER_ID="$(uuidgen)"

  if [[ -z ${CLUSTER_ID} ]]; then
    error "Failed to generate a telemetry clusterId"
  else
    info "[Telemetry] Using telemetry clusterId: ${CLUSTER_ID}"
  fi
}

function send_resource_telemetry() {
  local telemetry_content

  if [[ "${TELEMETRY_OPT_OUT:-false}" == "true" ]]; then
    info "Telemetry is off"
    return;
  fi

  local now
  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)

  ensure_machine_hash

  REQUESTED_MEM=$(kubectl describe nodes | grep "memory  " |  awk '{sum += $2} END {print sum}')
  REQUESTED_CPU=$(kubectl describe nodes | grep "cpu  " |  awk '{sum += $2} END {print sum}')
  NODE_COUNT=$(($(kubectl get nodes | wc -l) - 1))
  CORE_COUNT=$(kubectl get nodes -o=jsonpath='{.items[0].status.capacity.cpu}')

  info "Requested CPU: ${REQUESTED_CPU}, Requested Memory: ${REQUESTED_MEM}, Node Count: ${NODE_COUNT}, Core Count: ${CORE_COUNT}"

  function data() {
    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Resource.Requests",
            "properties": {
                "ApplicationName": "UiPath.ServiceFabric.Installer",
                "ProductVersion": "${VERSION}",
                "ClusterToken" : "${RKE_TOKEN}",
                "MachineHash" : "${MACHINE_HASH}",
                "ClusterId": "${CLUSTER_ID}",
                "Profile" : "${PROFILE}",
                "OptionalServices" : "${SERVICES_ENABLED[@]}",
                "RequestedCpu": "${REQUESTED_CPU}",
                "RequestedMemory": "${REQUESTED_MEM}",
                "NodeCount": "${NODE_COUNT}",
                "CoreCount": "${CORE_COUNT}",
            }
        }
    }
}
EOF
  }

  telemetry_content="$(data)"
  send_telemetry "${telemetry_content}"
}

function send_input_telemetry() {
  local telemetry_content

  if [[ "${TELEMETRY_OPT_OUT:-false}" == "true" ]]; then
    info "Telemetry is off"
    return;
  fi

  local now
  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)

  local appOfAppsRepo
  local dockerRegistryUrl
  if [[ "${INSTALL_TYPE}" == "online" ]]; then
    appOfAppsRepo=${APP_OF_APPS_REPO}
    dockerRegistryUrl=${DOCKER_REGISTRY_URL}
  else
    appOfAppsRepo="(redacted)"
    dockerRegistryUrl="(redacted)"
  fi

  ensure_machine_hash

  function data() {
    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Services.Input",
            "properties": {
                "ApplicationName": "UiPath.ServiceFabric.Installer",
                "ProductVersion": "${VERSION}",
                "InstallType": "${INSTALL_TYPE}",
                "InstallId": "${INSTALL_ID:-none}",
                "ClusterToken" : "${RKE_TOKEN}",
                "MachineHash" : "${MACHINE_HASH}",
                "ClusterId": "${CLUSTER_ID}",
                "AppOfAppsVersion" : "${APP_OF_APPS_VERSION}",
                "AppOfAppsRepo" : "${appOfAppsRepo}",
                "Profile" : "${PROFILE}",
                "StorageClass" : "${STORAGE_CLASS}",
                "StorageClassSingleReplica" : "${STORAGE_CLASS_SINGLE_REPLICA}",
                "DockerRegistryUrl" : "${dockerRegistryUrl}",
                "OptionalServices" : "${SERVICES_ENABLED[@]}"
            }
        }
    }
}
EOF
  }

  telemetry_content="$(data)"
  send_telemetry "${telemetry_content}"
}

function create_custom_configmaps() 
{
    local applications=("$@")
    local configmaps=()
    read -r -a configmaps <<< "$(kubectl -n uipath get configmaps -o jsonpath='{.items[*].metadata.name}')"

    for application in "${applications[@]}"; do
        local custom_configmap_name="$application-customconfig"
        
        if [[ ! ${configmaps[*]} =~ ${custom_configmap_name} ]]; then
            local custom_configmap="{
                \"apiVersion\": \"v1\",
                \"data\": {
                    \"nlog.json\": \"{}\",
                    \"values.json\": \"{}\"
                },
                \"kind\": \"ConfigMap\",
                \"metadata\": {
                    \"name\": \"$custom_configmap_name\",
                    \"namespace\": \"uipath\",
                    \"creationTimestamp\": null
                }
            }"
            echo "$custom_configmap" | kubectl apply -f -
        fi

    done
}

function create_preinstall_override_config(){
  kubectl apply -f "${SCRIPT_DIR}/k8s/preinstall-override-config.yaml"
}

# return mandatory parameters required to create service apps of apps
function get_mandatory_app_create_parameters() {
  MANDATORY_PARAMETERS=()

  MANDATORY_PARAMETERS+=(-p global.userInputs.hostname="${CLUSTER_FQDN}" )
  MANDATORY_PARAMETERS+=(-p global.userInputs.lookerHostname="${LOOKER_FQDN}" )
  MANDATORY_PARAMETERS+=(-p global.userInputs.lookerExternalPort="${LOOKER_EXTERNAL_PORT}")
  MANDATORY_PARAMETERS+=(-p global.userInputs.cephHostname="${CEPH_FQDN}" )
  MANDATORY_PARAMETERS+=(-p global.userInputs.cephExternalPort="${CEPH_EXTERNAL_PORT}")
  
  MANDATORY_PARAMETERS+=(-p global.userInputs.telemetry.clusterId="${CLUSTER_ID}" )
  MANDATORY_PARAMETERS+=(-p global.userInputs.telemetry.optOut="${TELEMETRY_OPT_OUT:-false}" )
  MANDATORY_PARAMETERS+=(-p global.repoURL="${APP_OF_APPS_REPO}" )
  MANDATORY_PARAMETERS+=(-p global.registryUrl="${DOCKER_REGISTRY_URL}" )
  MANDATORY_PARAMETERS+=(-p global.installType="${INSTALL_TYPE}" )
  MANDATORY_PARAMETERS+=(-p global.profile="${PROFILE}" )
  MANDATORY_PARAMETERS+=(-p global.zoneResilience="${ZONE_RESILIENCE}" )
  MANDATORY_PARAMETERS+=(-p global.storageClass="${STORAGE_CLASS}" )
  MANDATORY_PARAMETERS+=(-p global.storageClassSingleReplica="${STORAGE_CLASS_SINGLE_REPLICA}" )
  MANDATORY_PARAMETERS+=(-p global.proxy.enabled="${PROXY_ENABLED:-false}" )
  MANDATORY_PARAMETERS+=(-p global.proxy.httpproxy="${HTTP_PROXY:-PLACEHOLDER}" )
  MANDATORY_PARAMETERS+=(-p global.proxy.httpsproxy="${HTTPS_PROXY:-PLACEHOLDER}" )
  MANDATORY_PARAMETERS+=(-p global.proxy.noproxy="${NO_PROXY:-PLACEHOLDER},${CLUSTER_FQDN},.${CLUSTER_FQDN},127.0.0.0/8,localhost,kubernetes,.local,.svc,.cluster,rook-ceph-rgw-rook-ceph.rook-ceph.svc.cluster.local,actioncenter-actions,actioncenter-bupproxyservice,actioncenter-processes,ai-app-svc,ai-appmanager-svc,ai-deployer-svc,ai-helper-svc,ai-pkgmanager-svc,ai-trainer-svc,apps-designer,apps-runtime,apps-server,apps-wsserver,automationhub-ah-frontdoor-service,automationhub-ah-open-api-service,automationhub-ah-tenant-service,automationhub-ah-web-client,identity-service-api,identity-service-web,insights-insightslooker,insights-insightsportal,insights-insightsprovisioning,orchestrator,platform-audit-service,platform-license-accountant,platform-license-resource-manager,platform-location-service,platform-organization-management-service,platform-portal,studio-governance-api,studio-governance-web,testmanager,webhook-webhook-service,.uipath" )
  MANDATORY_PARAMETERS+=(-p global.kerberosAuthConfig.enabled="${ENABLESQLINTEGRATEDAUTH:-false}" )
  MANDATORY_PARAMETERS+=(-p global.kerberosAuthConfig.adDomain="${ADDOMAIN:-PLACEHOLDER}" )
  MANDATORY_PARAMETERS+=(-p global.kerberosAuthConfig.adUserName="${ADUSERNAME:-PLACEHOLDER}" )
  MANDATORY_PARAMETERS+=(-p global.kerberosAuthConfig.userKeytab="${USERKEYTAB:-PLACEHOLDER}" )
  MANDATORY_PARAMETERS+=(-p global.kerberosAuthConfig.ticketLifetimeInHours="${TICKETLIFETIME:-PLACEHOLDER}" )

  MANDATORY_PARAMETERS+=(-p global.userInputs.sql.createDB="${CREATE_DB}" ) 
  if [[ "${CREATE_DB}" == "true" ]]; then
    MANDATORY_PARAMETERS+=(-p global.userInputs.sql.defaultConnectionString="${SQL_DEFAULT_CONNECTION_STRING}" )
  fi
}

######
# check and throw error when user is trying to install standalone aicenter amd platform in same cluster
######
function validate_platform_aicenter_installation_option() {
  local namespace service_install_type
  local trigger_service_install_type=$1
  
  service_install_type=$(get_cluster_config_default "SERVICE_INSTALL_TYPE" "all")

  if [[ "${service_install_type}" != "${trigger_service_install_type}" ]]; then
    error "Platform service and standalone aicenter service installation in same cluster is not supported"
  fi
}

function update_service_start_fqdn_state () {
  local fqdn_update_state
  fqdn_update_state=$(get_cluster_config "update_fqdn_state")
  [[ ${fqdn_update_state} == "" ]] \
    && error_with_code "13007"
  update_cluster_config "update_fqdn_state" "SERVICE_STARTED"
}

function update_fqdn() {
  local namespaces=("$@")

  # Standalone ai center does not use redis cache, so clean redis only when install type is not standalone ai center
  if [[ ${INSTALLSTANDALONEAICENTER} == "0" ]]; then
    if ! wait_for_sync_and_health "argocd" "platform" ;
    then
      error_with_code "13008"
    fi

    info "Clearing redis cache."
    
    redis_password=$(kubectl get secrets/redb-redis-cluster-db -n uipath --template '{{ .data.password }}' | base64 -d)
    redis_port=$(kubectl get secrets/redb-redis-cluster-db -n uipath --template '{{ .data.port }}' | base64 -d)
    kubectl --namespace redis-system \
      exec -it redis-cluster-0 \
      -c redis-enterprise-node "--" sh \
      -c "redis-cli -p $redis_port -a $redis_password --scan --pattern is:* | xargs redis-cli -p $redis_port -a $redis_password del" >/dev/null
    
    info "Redis cache cleared."
  fi
  
  info "Restarting all the pods"
  # If updating FQDN then also restart the pods to update the certificate in trust store of all the services

  profile=$(get_cluster_config_default "profile" "ha")

  if [[ "$profile" == "ha"  ]]; 
  then
    kubectl rollout restart deployment --namespace "${namespaces[0]}" >/dev/null
    wait_for_all_pods_in_namespace_ready "${namespaces[0]}" >/dev/null \
      || error "FQDN update failed - timeout waiting for all pods to restart and ready in namespace ${namespaces[0]}"
  else
    scale_down_all_deployment "${namespaces[0]}" \
      || error "FQDN update failed - Scalling down all deployment in namespace ${namespaces[0]} failed"
    scale_up_all_deployment "${namespaces[0]}" "1" \
      || error "FQDN update failed - Scalling up all deployment in namespace ${namespaces[0]} failed" 
  fi

  info "Successfully updated FQDN..."

  update_cluster_config "update_fqdn_status" "SERVICE_DONE"
  update_cluster_config "update_fqdn" ""
}

function is_update_fqdn() {
  local is_update_fqdn_operation
  is_update_fqdn_operation=$(get_cluster_config "update_fqdn" )
  
  if [[ "${is_update_fqdn_operation}" == "true" ]]; then
    echo "true"
  else
    echo "false"
  fi
}
