#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

# kill subprocesses whem the caller script terminates
trap "trap - SIGTERM && jobs -p | xargs -r kill" INT TERM EXIT

#GLOBALS:
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error_with_code "10001"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error_with_code "10002"
fi

## Locale
LANG_SERVICES_PREFIX="[Products]"
LANG_SERVICES_STARTED_PREFIX="[Products] [Started]"
LANG_SERVICES_FAILED_PREFIX="[Products] [Failed]"
LANG_SERVICES_COMPLETED_PREFIX="[Products] [Completed]"
LANG_SERVICES_CONFIG_MESSAGE="Generating the config resources"

readonly SCRIPT_DIR
readonly MODULES_DIR
export MISSING_DEGRADED_APP_LIST=""
declare -a SERVICES_ENABLED
declare -a SERVICE_NAMES_ENABLED
declare -a MANDATORY_PARAMETERS
declare -A APPS_STATUS=()
declare -i APPS_PROCESSED=0
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/argocd_utils.sh"
#shellcheck disable=SC1090,SC1091,SC2034
source "${MODULES_DIR}/cluster_utils.sh"
#shellcheck disable=SC1090,SC1091,SC2034
source "${MODULES_DIR}/k8sutils.sh"
#shellcheck disable=SC1090,SC1091,SC2034
source "${SCRIPT_DIR}/common-services-installer.sh"
#FUNCTIONS

function install_app_of_apps() {
  local pid

  additional_parameters=()
    [ -z "${DOCKER_REGISTRY_PULL_SECRET}" ] || additional_parameters+=(-p "global.imagePullSecrets[0]=uipathpullsecret")

  certificate_parameters=()
    [ -z "$IDENTITY_SAML_CURRENT_SERVICE_CERT_CONTENT" ] || certificate_parameters+=(-p "global.userInputs.certificate.identity.saml.currentServiceCert=\"${IDENTITY_SAML_CURRENT_SERVICE_CERT_CONTENT}\"")
    [ -z "$(base64 -d <<< "$IDENTITY_SAML_CURRENT_SERVICE_CERT_PASS")" ] || certificate_parameters+=(-p "global.userInputs.certificate.identity.saml.currentServiceCertPassword=\"${IDENTITY_SAML_CURRENT_SERVICE_CERT_PASS}\"")
    [ -z "$IDENTITY_SAML_FUTURE_SERVICE_CERT_CONTENT" ] || certificate_parameters+=(-p "global.userInputs.certificate.identity.saml.futureServiceCert=\"${IDENTITY_SAML_FUTURE_SERVICE_CERT_CONTENT}\"")
    [ -z "$(base64 -d <<< "$IDENTITY_SAML_FUTURE_SERVICE_CERT_PASS")" ] || certificate_parameters+=(-p "global.userInputs.certificate.identity.saml.futureServiceCertPassword=\"${IDENTITY_SAML_FUTURE_SERVICE_CERT_PASS}\"")
    [ -z "$IDENTITY_LDAPS_CUSTOM_ROOTCA_CONTENT" ] || certificate_parameters+=(-p "global.userInputs.certificate.identity.ldaps.customRootCA=\"${IDENTITY_LDAPS_CUSTOM_ROOTCA_CONTENT}\"")

  get_mandatory_app_create_parameters

  local try=0
  local maxtry=5
  local status="notready"
  while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
    try=$((try+1))
    argocd app create uipath \
      --repo "${APP_OF_APPS_REPO}" \
      --revision "${APP_OF_APPS_VERSION}" \
      --helm-chart "helm/${APP_OF_APPS_CHART}" \
      --dest-server https://kubernetes.default.svc \
      --upsert \
      --sync-policy "automated" \
      --sync-retry-backoff-duration 10s \
      --sync-retry-backoff-factor 2 \
      --sync-retry-backoff-max-duration 20m \
      --sync-retry-limit 4 \
      "${SERVICES_ENABLED[@]}" \
      "${additional_parameters[@]}" \
      "${MANDATORY_PARAMETERS[@]}" \
      "${certificate_parameters[@]}" \
      -p global.userInputs.orchestrator.sqlConnectionString="${ORCHESTRATOR_SQL_CONNECTION}" \
      -p global.userInputs.orchestrator.testautomation.enabled="${ORCHESTRATOR_TESTAUTOMATION_ENABLED}" \
      -p global.userInputs.orchestrator.testautomation.sqlConnectionString="${ORCHESTRATOR_TA_SQL_CONNECTION}" \
      -p global.userInputs.orchestrator.updateserver.enabled="${ORCHESTRATOR_UPDATESERVER_ENABLED}" \
      -p global.userInputs.orchestrator.updateserver.sqlConnectionString="${ORCHESTRATOR_UPD_SQL_CONNECTION}" \
      -p global.userInputs.orchestrator.orchestratorRobotLogsElasticUri="${ORCHESTRATOR_ROBOT_LOGS_ELASTIC_URI}" \
      -p global.userInputs.orchestrator.orchestratorRobotLogsElasticAuthUsername="${ORCHESTRATOR_ROBOT_LOGS_ELASTIC_AUTH_USERNAME}" \
      -p global.userInputs.orchestrator.orchestratorRobotLogsElasticAuthPassword="${ORCHESTRATOR_ROBOT_LOGS_ELASTIC_AUTH_PASSWORD}" \
      -p global.userInputs.identity.sqlConnectionString="${IDENTITY_SQL_CONNECTION}" \
      -p global.userInputs.identity.hostAdminUsername="${IDENTITY_HOST_ADMIN_USERNAME}" \
      -p global.userInputs.identity.hostAdminPassword="${IDENTITY_HOST_ADMIN_PASSWORD}" \
      -p global.userInputs.identity.krb5KeytabSecret="${IDENTITY_KRB5_KEYTAB:-UExBQ0VIT0xERVI=}" \
      -p global.userInputs.oms.sqlConnectionString="${OMS_SQL_CONNECTION}" \
      -p global.userInputs.location.sqlConnectionString="${LOCATION_SQL_CONNECTION}" \
      -p global.userInputs.lrm.sqlConnectionString="${LRM_SQL_CONNECTION}" \
      -p global.userInputs.la.sqlConnectionString="${LA_SQL_CONNECTION}" \
      -p global.userInputs.audit.sqlConnectionString="${AUDIT_SQL_CONNECTION}" \
      -p global.userInputs.automationops.sqlConnectionString="${AUTOMATIONOPS_SQL_CONNECTION:-PLACEHOLDER}" \
      -p global.userInputs.taskmining.sqlConnectionString="${TASKMINING_SQL_CONNECTION:-PLACEHOLDER}" \
      -p global.userInputs.testmanager.sqlConnectionString="${TESTMANAGER_SQL_CONNECTION:-PLACEHOLDER}" \
      -p global.userInputs.automationhub.sqlConnectionString="${AUTOMATIONHUB_SQL_CONNECTION:-PLACEHOLDER}" \
      -p global.userInputs.aicenter.sqlConnectionString="${AICENTER_SQL_CONNECTION:-PLACEHOLDER}" \
      -p global.userInputs.documentunderstanding.datamanager.sqlConnectionString="${DU_DATAMANAGER_SQL_CONNECTION:-PLACEHOLDER}" \
      -p global.userInputs.documentunderstanding.datamanager.sqlConnectionString="${DU_DATAMANAGER_SQL_CONNECTION:-PLACEHOLDER}" \
      -p global.userInputs.documentunderstanding.handwriting.enabled="${DU_SERVICES_HANDWRITING_ENABLED:-PLACEHOLDER}" \
      -p global.userInputs.documentunderstanding.handwriting.max_cpu_per_pod="${DU_SERVICES_HANDWRITING_MAX_CPU_PER_POD:-PLACEHOLDER}" \
      -p global.userInputs.insights.sqlConnectionString="${INSIGHTS_SQL_CONNECTION:-PLACEHOLDER}" \
      -p global.userInputs.insights.smtpPassword="${INSIGHTS_SMTP_PASSWORD:-PLACEHOLDER}" \
      -p global.userInputs.insights.smtpTlsVersion="${INSIGHTS_SMTP_TLS_VERSION:-PLACEHOLDER}" \
      -p global.userInputs.insights.smtpFromEmail="${INSIGHTS_SMTP_FROM_EMAIL:-PLACEHOLDER}" \
      -p global.userInputs.insights.smtpHost="${INSIGHTS_SMTP_HOST:-PLACEHOLDER}" \
      -p global.userInputs.insights.smtpPort="${INSIGHTS_SMTP_PORT:-PLACEHOLDER}" \
      -p global.userInputs.insights.smtpUsername="${INSIGHTS_SMTP_USERNAME:-PLACEHOLDER}" \
      -p global.userInputs.orchestrator.kerberosAuthConfig.adUserName="${ORCHESTRATOR_KERBEROS_AD_USER:-PLACEHOLDER}" \
      -p global.userInputs.orchestrator.kerberosAuthConfig.userKeytab="${ORCHESTRATOR_KERBEROS_KEYTAB:-UExBQ0VIT0xERVI=}" \
      -p global.userInputs.platform.kerberosAuthConfig.adUserName="${PLATFORM_KERBEROS_AD_USER:-PLACEHOLDER}" \
      -p global.userInputs.platform.kerberosAuthConfig.userKeytab="${PLATFORM_KERBEROS_KEYTAB:-UExBQ0VIT0xERVI=}" \
      -p global.userInputs.discoverygroup.kerberosAuthConfig.adUserName="${DISCOVERYGROUP_KERBEROS_AD_USER:-PLACEHOLDER}" \
      -p global.userInputs.discoverygroup.kerberosAuthConfig.userKeytab="${DISCOVERYGROUP_KERBEROS_KEYTAB:-UExBQ0VIT0xERVI=}" \
      -p global.userInputs.testmanager.kerberosAuthConfig.adUserName="${TESTMANAGER_KERBEROS_AD_USER:-PLACEHOLDER}" \
      -p global.userInputs.testmanager.kerberosAuthConfig.userKeytab="${TESTMANAGER_KERBEROS_KEYTAB:-UExBQ0VIT0xERVI=}" \
      -p global.userInputs.automationops.kerberosAuthConfig.adUserName="${AUTOMATIONOPS_KERBEROS_AD_USER:-PLACEHOLDER}" \
      -p global.userInputs.automationops.kerberosAuthConfig.userKeytab="${AUTOMATIONOPS_KERBEROS_KEYTAB:-UExBQ0VIT0xERVI=}" \
      -p global.userInputs.aicenter.kerberosAuthConfig.adUserName="${AICENTER_KERBEROS_AD_USER:-PLACEHOLDER}" \
      -p global.userInputs.aicenter.kerberosAuthConfig.userKeytab="${AICENTER_KERBEROS_KEYTAB:-UExBQ0VIT0xERVI=}" \
      -p global.userInputs.documentunderstanding.kerberosAuthConfig.adUserName="${DOCUMENTUNDERSTANDING_KERBEROS_AD_USER:-PLACEHOLDER}" \
      -p global.userInputs.documentunderstanding.kerberosAuthConfig.userKeytab="${DOCUMENTUNDERSTANDING_KERBEROS_KEYTAB:-UExBQ0VIT0xERVI=}" \
      -p global.userInputs.insights.kerberosAuthConfig.adUserName="${INSIGHTS_KERBEROS_AD_USER:-PLACEHOLDER}" \
      -p global.userInputs.insights.kerberosAuthConfig.userKeytab="${INSIGHTS_KERBEROS_KEYTAB:-UExBQ0VIT0xERVI=}" \
      >/dev/null 2>&1 \
      && status="ready"

      [[ ${status} == "ready" ]] || { info "Creating uipath app...try ${try}/${maxtry}"; sleep 5; }
  done
  [[ ${status} == "ready" ]] || error_with_code "13001"

  info "Application uipath created"
  
  # sent this to background so we can poll for the apps status
  sync_with_wait "uipath" "argocd" "app.kubernetes.io/instance=uipath" "service-fabric=service-presync-checks" "uipath" &

  # keep printing the apps status as it changes
  pid="$!"
  while [ -d "/proc/${pid}" ];do
    get_argo_apps_status
    sleep 5
  done
 
  # get the last status once the process finished
  get_argo_apps_status
  wait "${pid}" || error_with_code "13002"
}

function set_output() {
  cat > "${OUTFILE}" <<EOF
{ 
  "orchestrator_url": "https://${CLUSTER_FQDN}/orchestrator_",
  "identity_url": "https://${CLUSTER_FQDN}/identity_",
  "oms_url": "https://${CLUSTER_FQDN}/organization",
  "lrm_url": "https://${CLUSTER_FQDN}/lrm",
  "la_url": "https://${CLUSTER_FQDN}/la",
  "testmanager_url": "https://${CLUSTER_FQDN}/testmanager_",
  "automationhub_url": "https://${CLUSTER_FQDN}/automationhub_",
  "automationops_url": "https://${CLUSTER_FQDN}/roboticsops_",
  "webhook_url": "https://${CLUSTER_FQDN}/webhook_",
  "taskmining_url": "https://${CLUSTER_FQDN}/taskmining_",
  "actioncenter_url": "https://${CLUSTER_FQDN}/bupproxyservice_",
  "ba_url": "https://${CLUSTER_FQDN}/apps_",
  "aicenter_url": "https://${CLUSTER_FQDN}/aifabric_",
  "documentunderstanding_url": "https://${CLUSTER_FQDN}/du_",
  "insights_url": "https://${CLUSTER_FQDN}/insights_",
  "businessapps_url": "https://${CLUSTER_FQDN}/apps_"
}
EOF
}

#MAIN
function main() {
  info "${LANG_SERVICES_STARTED_PREFIX} ${LANG_SERVICES_CONFIG_MESSAGE}"
  # Use trap to append extra error messages to functions imported in this script
  old_trap="$(get_existing_trap "EXIT")"
  #shellcheck disable=SC2064
  trap "error_without_exit \"${LANG_SERVICES_FAILED_PREFIX} ${LANG_SERVICES_CONFIG_MESSAGE}\"; ${old_trap}" EXIT

  cd "${SCRIPT_DIR}"

  [[ "$(is_update_fqdn)" == "true" ]] && update_service_start_fqdn_state

  parse_args "$@"

  load_os_utils "$SCRIPT_DIR/../Modules"

  check_requirements

  parse_input_file

  send_checkpoint_telemetry "Services: Input Parsed"

  get_fabric_install_paths "${SCRIPT_DIR}"
  get_argocd_install_params
  login_argocd "${ARGOCD_NODEPORT_HTTP}" "${ARGOCD_FQDN}" "${ARGOCD_PASSWORD_SECRET_NAME}" "${ARGOCD_NAMESPACE}" "argocd"

  validate_platform_aicenter_installation_option "all"

  parse_offline_input_parameters

  get_enabled_services

  local try=0
  local maxtry=30
  local status="notready"
  while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
          try=$((try+1))
          kubectl cluster-info >/dev/null 2>&1  && status="ready"
          info "waiting for cluster to be ready...${try}/${maxtry}" &&  sleep 10;
  done
  kubectl cluster-info || error_with_code "10004"

  declare -a namespaces=("uipath")
  declare -a custom_configmaps_applications=("orchestrator" "identity-service")

  kubectl cluster-info || error_with_code "13003"

  argocd cluster list || error_with_code "13004"

  send_checkpoint_telemetry "Services: Cluster Ready"

  create_namespaces "${namespaces[@]}"

  read_identity_certificate_files  

  check_available_masters
  
  create_custom_configmaps "${custom_configmaps_applications[@]}"

  create_preinstall_override_config

  ensure_cluster_id
  send_checkpoint_telemetry "Services: ClusterId Created"
  send_input_telemetry


  #shellcheck disable=SC2064
  trap "${old_trap}" EXIT

  info "${LANG_SERVICES_COMPLETED_PREFIX} ${LANG_SERVICES_CONFIG_MESSAGE}"

  install_app_of_apps

  [[ "$(is_update_fqdn)" == "true" ]] && update_fqdn "${namespaces[@]}"

  set_output

  send_checkpoint_telemetry "Services: Success"

  info "UiPath Services installed successfully."

  send_resource_telemetry
}



# Get status for all argo apps declared by SERVICE_NAMES_ENABLED
function get_argo_apps_status() {
  local status
  
  # get and print app status only if it changed
  # shellcheck disable=SC2076
  while (( APPS_PROCESSED < ${#ALL_SERVICE_NAMES[@]} )); do
    for app in "${ALL_SERVICE_NAMES[@]}"; do
      if [[ " ${SERVICE_NAMES_ENABLED[*]} " =~ " ${app} " ]]; then
        [[ -z "${APPS_STATUS[$app]-}" ]] && { 
          APPS_STATUS[$app]="Started";
          info "${LANG_SERVICES_PREFIX} [Started] Installing $app";
        }
        status="$(get_argo_app_status "${app}")"
        status="${status:-"Started"}"
        status="${status/Healthy/Completed}"
        status="${status/Missing/Started}"
        [[ "${APPS_STATUS[$app]-}" == "${status}" ]] || {
          APPS_STATUS[$app]="${status}"
          info "${LANG_SERVICES_PREFIX} [${status}] Installing $app"
        }
      else
        status="Skipped"
        [[ "${APPS_STATUS[$app]-}" == "${status}" ]] || {
          APPS_STATUS[$app]="${status}"
          info "${LANG_SERVICES_PREFIX} [${status}] Installing $app"
        }
      fi

      # Start from the beginning when we manage to get the status for a new app.
      # This way the first apps processed are updated sooner instead of waiting for a full first pass
      # through the array.
      [[ ${APPS_STATUS["${app}_processed"]-} ]] || {
          APPS_STATUS["${app}_processed"]=true;
          APPS_PROCESSED+=1
          break;
        }
    done
  done
}

main "$@"
