#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

########
# Install/Update ArgoCD
# Input arguments:
# $1 - Upgrade Step JSON
# $2 - Validate Step FLAG
########

#GLOBALS:
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/upgrade_utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/argocd_utils.sh"

function validate_rke2_step {
  local TEMP_INFILE="${INFILE}"
  local NEEDS_UPGRADE=0
  export INFILE="${VERSION_FILE_PATH}"
  info "Checking if RKE2 was already upgraded to ${COMPONENT_TARGET_VERSION}."
  NODES="$(get_mandatory_json_value ".nodes")"
  for row in $(echo "${NODES}" | jq -c '.[]'); do
      node="$(get_mandatory_inline_json_value ".name" "${row}")"
      version="$(get_mandatory_inline_json_value ".version" "${row}")"
      info "Found ${node} running ${version}"
      if [ "${version}" != "${COMPONENT_TARGET_VERSION}" ]; then
        NEEDS_UPGRADE=1
      fi
  done
  if [ "${NEEDS_UPGRADE}" = 0 ]; then
    info "RKE2 already upgraded, skipping"
    exit 0
  fi
  INFILE="${TEMP_INFILE}"
}

function whitelist_gatekeeper() {
  if ! grep 'system-upgrade' <<< "$(kubectl get K8sPSPAllowedUsers/psp-pods-allowed-user-ranges --template '{{.spec.match.excludedNamespaces}}')"; then
    kubectl patch K8sPSPAllowedUsers/psp-pods-allowed-user-ranges --type "json" -p '[{"op":"add","path":"/spec/match/excludedNamespaces/-","value":"system-upgrade"}]'
  fi
  if ! grep 'system-upgrade' <<< "$(kubectl get K8sPSPAllowPrivilegeEscalationContainer/psp-allow-privilege-escalation-container --template '{{.spec.match.excludedNamespaces}}')"; then
    kubectl patch K8sPSPAllowPrivilegeEscalationContainer/psp-allow-privilege-escalation-container --type "json" -p '[{"op":"add","path":"/spec/match/excludedNamespaces/-","value":"system-upgrade"}]'
  fi
  if ! grep 'system-upgrade' <<< "$(kubectl get K8sPSPCapabilities/psp-capabilities --template '{{.spec.match.excludedNamespaces}}')"; then
    kubectl patch K8sPSPCapabilities/psp-capabilities --type "json" -p '[{"op":"add","path":"/spec/match/excludedNamespaces/-","value":"system-upgrade"}]'
  fi
  if ! grep 'system-upgrade' <<< "$(kubectl get K8sPSPPrivilegedContainer/psp-privileged-container --template '{{.spec.match.excludedNamespaces}}')"; then
    kubectl patch K8sPSPPrivilegedContainer/psp-privileged-container --type "json" -p '[{"op":"add","path":"/spec/match/excludedNamespaces/-","value":"system-upgrade"}]'
  fi
  if ! grep 'system-upgrade' <<< "$(kubectl get K8sPSPReadOnlyRootFilesystem/psp-readonlyrootfilesystem --template '{{.spec.match.excludedNamespaces}}')"; then
    kubectl patch K8sPSPReadOnlyRootFilesystem/psp-readonlyrootfilesystem --type "json" -p '[{"op":"add","path":"/spec/match/excludedNamespaces/-","value":"system-upgrade"}]'
  fi
  if ! grep 'system-upgrade' <<< "$(kubectl get K8sPSPVolumeTypes/psp-volume-types --template '{{.spec.match.excludedNamespaces}}')"; then
    kubectl patch K8sPSPVolumeTypes/psp-volume-types --type "json" -p '[{"op":"add","path":"/spec/match/excludedNamespaces/-","value":"system-upgrade"}]'
  fi
}

function execute_step() {
  info "Executing $UPGRADE_TYPE upgrade for $COMPONENT"
  sed -i "s;##REGISTRY_PLACEHOLDER##;${DOCKER_REGISTRY_URL}/;" "${UPGRADE_VENDOR_DIR}/rancher/system-upgrade-controller.yaml"
  sed -i "s;##REGISTRY_PLACEHOLDER##;${DOCKER_REGISTRY_URL}/;" "${UPGRADE_VENDOR_DIR}/rancher/upgrade-plans.yaml"
  sed -i "s;##RKE2_TARGET_VERSION##;${COMPONENT_TARGET_VERSION};" "${UPGRADE_VENDOR_DIR}/rancher/upgrade-plans.yaml"
  kubectl apply -f "${UPGRADE_VENDOR_DIR}/rancher/system-upgrade-controller.yaml"
  wait_for_crd "crd/plans.upgrade.cattle.io" "system-upgrade"
  kubectl apply -f "${UPGRADE_VENDOR_DIR}/rancher/upgrade-plans.yaml"
  wait_for_job_complete_by_label "upgrade.cattle.io/plan=server-upgrade-plan" "system-upgrade"
  #Check if agents exist
  local NUM_AGENTS
  NUM_AGENTS=$(kubectl get node --selector='!node-role.kubernetes.io/control-plane' -o json | jq -r '.items | length')
  if [ "${NUM_AGENTS}" -gt 0 ]; then
    wait_for_job_complete_by_label "upgrade.cattle.io/plan=agent-upgrade-plan" "system-upgrade"
  fi
  info "Done!"
}

function main() {
  parse_step "${STEP}"
  whitelist_gatekeeper
  validate_rke2_step
  execute_step
}

export STEP=$1
# Pass additional arguments from upgrade path to main function.
main "${@:2}"