#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS:
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/upgrade_utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/argocd_utils.sh"

####
# Sets current rook-version and ceph-version from all pods
####
function set_rook_ceph_version() {
    local rook_cluster_namespace=rook-ceph
    current_rook_version=$(kubectl -n $rook_cluster_namespace get deployment -l rook_cluster=$rook_cluster_namespace -o jsonpath='{range .items[*]}{"rook-version="}{.metadata.labels.rook-version}{"\n"}{end}' | sort | uniq)
    current_ceph_version=$(kubectl -n $rook_cluster_namespace get deployment -l rook_cluster=$rook_cluster_namespace -o jsonpath='{range .items[*]}{"ceph-version="}{.metadata.labels.ceph-version}{"\n"}{end}' | sort | uniq)
}

####
# Waits for all rook-ceph pods to be updated to latest version and checks ceph health in the end
####
function check_latest_ceph_health() {
    osd_deploy_orig=$(kubectl -n rook-ceph get cephcluster rook-ceph -o json | jq '.spec.mon.count')
    kubectl -n rook-ceph scale deploy rook-ceph-operator --replicas=0
    kubectl -n rook-ceph delete deploy -l app=rook-ceph-osd
    kubectl -n rook-ceph scale deploy rook-ceph-operator --replicas=1
    
    local rook_cluster_namespace=rook-ceph
    local try=0
    local maxtry=40
    info "Waiting for rook-ceph to be updated to latest version.. ${try}/${maxtry}"
    while ((try != maxtry));
    do
        osd_count=$(kubectl -n rook-ceph get deploy -l app=rook-ceph-osd --no-headers 2>/dev/null | wc -l)
        if [[ "${osd_count}" == "${osd_deploy_orig}" ]];
        then
            set_rook_ceph_version
            if [[ "${current_rook_version}" == "rook-version=v1.7.9" && "${current_ceph_version}" == "ceph-version=16.2.6-0" ]];
            then
                info "Rook-Ceph updated to latest version"
                break
            fi
        fi
        try=$((try+1))
        info "Waiting for rook-ceph to be updated to latest version.. ${try}/${maxtry}"
        sleep 30
    done

    osd_deployments=$(kubectl -n rook-ceph get deploy -l app=rook-ceph-osd -o name)
    for deploy in ${osd_deployments};
    do
        kubectl -n rook-ceph rollout status "$deploy" --timeout=3m || true
    done

    kubectl -n rook-ceph delete pods -l app=rook-ceph-rgw
    gateway_deploy=$(kubectl -n rook-ceph get deploy -l app=rook-ceph-rgw -o name)
    for deploy in ${gateway_deploy};
    do
        kubectl -n rook-ceph rollout status "$deploy" --timeout=3m || true
    done

    set_rook_ceph_version
    if [[ "${current_rook_version}" == "rook-version=v1.7.9" && "${current_ceph_version}" == "ceph-version=16.2.6-0" ]];
    then
        # If all pods are successfully updated, check ceph health
        ceph_health_exit_on_error
        info "Rook ceph upgrade successful"
    else
        # As all pods could not be updated to latest version, exit
        kubectl -n $rook_cluster_namespace get deployments -l rook_cluster=$rook_cluster_namespace -o jsonpath='{range .items[*]}{.metadata.name}{"  \treq/upd/avl: "}{.spec.replicas}{"/"}{.status.updatedReplicas}{"/"}{.status.readyReplicas}{"  \trook-version="}{.metadata.labels.rook-version}{"\n"}{end}'
        error "Rook-Ceph could not be updated to latest version"
    fi
}

check_latest_ceph_health