#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

COMMAND_NAME=$(basename "$0")
readonly SCRIPT_DIR
readonly MODULES_DIR
  
export KUBECONFIG=/etc/rancher/rke2/rke2.yaml
export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/bin:/usr/local/bin

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/argocd_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/certificate_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cluster_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/../Infra_Installer/disaster_recovery/longhorn_helper.sh"

#FUNCTIONS
function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "configureUiPathAS.sh controls and manage UiPath Automation Suites"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [command]"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Available Commands:"
  echo "  config                               Manage cluster configuration"
  echo "  tls-cert                             Manage tls and server certificate"
  echo "  additional-ca-certs                  Manage additional ca certificates"
  echo "  identity                             Manage identity service"
  echo "  objectstore                          Manage objectstore"
  echo "  registry                             Manage registry"
  echo "  monitoring                           Manage monitoring"
  echo "  rabbitmq                             Manage rabbitmq"
  echo "  mongodb                              Manage mongo"
  echo "  node                                 Manage k8s nodes"
  echo "  enable-maintenance-mode              Enables maintenance mode on the Cluster"
  echo "  disable-maintenance-mode             Disables maintenance mode on the Cluster"
  echo "  is-maintenance-enabled               Checks if maintenance mode is enabled on the Cluster"
  echo "  resume-scheduled-backups             Resumes the paused scheduled backups"
  echo "  verify-volumes-backup                Verify if all volumes are backed up"
  echo "Flags:"
  echo "  -h|--help                           Display help"
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  local PARAMS=""
  while (("$#")); do
    export COMPONENT=$1
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    config)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_cluster_config "$@"
    ;;

    tls-cert)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_tls_cert "$@"
    ;;

    additional-ca-certs)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_additional_ca_certs "$@"
    ;;

    identity)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_identity "$@"
    ;;

    objectstore)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_objectstore "$@"
    ;;

    registry)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_registry "$@"
    ;;

    monitoring)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_monitoring "$@"
    ;;

    rabbitmq)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_rabbitmq "$@"
    ;;

    mongodb)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_mongodb "$@"
    ;;

    node)
      shift
      export COMMAND_NAME="${COMMAND_NAME} ${COMPONENT}"
      configure_node "$@"
    ;;

    enable-maintenance-mode)
      shift
      enable_maintenance_mode "$@"
      info "Enabled maintenance mode successfully." && exit 0
      ;;

    disable-maintenance-mode)
      shift
      disable_maintenance_mode
      info "Disabled maintenance mode successfully." && exit 0
      ;;
    
    is-maintenance-enabled)
      shift
      is_maintenance_mode_enabled
      exit 0
    ;;

    resume-scheduled-backups)
      shift
      resume_backup_schedule
      exit 0
    ;;

    verify-volumes-backup)
      shift
      check_if_all_volumes_backed_up
      exit 0
    ;;

    --kubeconfig)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        [[ -f "${2}" ]] || error "$2 not found"
        KUBECONFIG_PATH="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported command or flag $1"
      exit 1
    ;;
    esac
  done
  # set positional arguments in their proper place
  eval set -- "$PARAMS"
  display_usage
}

function configure_tls_cert() {
  login_argocd_using_cluster_config
  if "${SCRIPT_DIR}"/Cluster_Configure/tls-cert/main.sh "$@"; then
    exit 0
  fi
  exit 1
}

function configure_objectstore() {
  if "${SCRIPT_DIR}"/Cluster_Configure/objectstore/main.sh "$@"; then
    exit 0
  fi
  exit 1
}

function configure_registry() {
  if "${SCRIPT_DIR}"/Cluster_Configure/registry/main.sh "$@"; then
    exit 0
  fi
  exit 1
}

function configure_monitoring() {
  if "${SCRIPT_DIR}"/Cluster_Configure/monitoring/main.sh "$@"; then
    exit 0
  fi
  exit 1
}

function configure_rabbitmq() {
  if "${SCRIPT_DIR}"/Cluster_Configure/rabbitmq/main.sh "$@"; then
    exit 0
  fi
  exit 1
}

function configure_mongodb() {
  if "${SCRIPT_DIR}"/Cluster_Configure/mongodb/main.sh "$@"; then
    exit 0
  fi
  exit 1
}

function configure_node() {
  if "${SCRIPT_DIR}"/Cluster_Configure/node/main.sh "$@"; then
    exit 0
  fi
  exit 1
}


function configure_identity() {
  login_argocd_using_cluster_config
  if "${SCRIPT_DIR}"/Cluster_Configure/identity/identity.sh "$@"; then
    exit 0
  fi
    exit 1
}

function configure_additional_ca_certs() {
  if "${SCRIPT_DIR}"/Cluster_Configure/additional-ca-certs/main.sh "$@"; then
    exit 0
  fi
    exit 1
}

function configure_cluster_config() {
  login_argocd_using_cluster_config
  if "${SCRIPT_DIR}"/Cluster_Configure/config/main.sh "$@"; then
    exit 0
  fi
    exit 1
}

function main() {
  parse_long_args "$@"
  export KUBECONFIG="${KUBECONFIG_PATH:-/etc/rancher/rke2/rke2.yaml}"
}

main "$@"
