#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

YELLOW='\033[0;33m'
RED='\033[0;31m'
NC='\033[0m'
CYAN='\033[0;36m'
GREEN='\033[0;32m'

function error() {
  echo -e "${RED}[ERROR][$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*${NC}\n" >&2
  exit 1
}

function info() {
  echo -e "${CYAN}[INFO] [$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*${NC}"
}

function warn() {
  echo -e "${YELLOW}[WARN] [$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*${NC}" >&2
}

function success() {
  echo -e "${GREEN}[SUCCESS] [$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*${NC}" >&2
}

function check_running_as_root() {
  if [ "$EUID" -ne 0 ]; then
    error "Please run script as root"
  fi
}

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Utility to configure the disk for UiPath Automation Suite Installation."
  echo "Run this script to configure the disks on new machine or to extend the size of datadisk"
  echo
  echo "Arguments"
  echo "  -n|--node-type                  NodeType, Possible values: agent, server. Default to server."
  echo "  -i|--install-type               Installation mode, Possible values: online, offline. Default to online."
  echo "  -c|--cluster-disk-name          Device to host rancher and  kubelet. Ex: /dev/sdb."
  echo "  -e|--etcd-disk-name             Device to host etcd, Not required for agent node. Ex: /dev/sdb."
  echo "  -l|--data-disk-name             Device to host datadisk, Not required for agent node. Ex: /dev/sdc."
  echo "  -b|--bundle-disk-name           Device to host the uipath bundle. "
  echo "                                    Only required for offline installation on 1st server node. "
  echo "  -f|--complete-suite             Installing complete product suite or any of these products: "
  echo "                                    aicenter, apps, taskmining, documentunderstanding."
  echo "                                    This will configure the datadisk volume to be 2TiB instead of 512Gi."
  echo "  -p|--primary-server             Is this machine is first server machine? Applicable only for airgap install."
  echo "                                    This is the machine on which UiPath AutomationSuite bundle will be installed."
  echo "                                    Default to false."
  echo "  -x|--extend-data-disk           Extend the datadisk. Either attach new disk or resize the exiting datadisk."
  echo "  -r|--resize                     Used in conjunction of with --extend-data-disk to resize the exiting volume,"
  echo "                                    instead of adding new volume. "
  echo "  -d|--debug                      Run in debug."
  echo "  -h|--help                       Display help."
  echo
  echo "ExampleUsage:"
  echo
  echo "  $(basename "$0") --node-type server --install-type online \\"
  echo "    --cluster-disk-name /dev/sdb --etcd-disk-name /dev/sdc \\"
  echo "    --data-disk-name /dev/sdd"
  echo
  echo "  $(basename "$0") --data-disk-name /dev/sdh \\"
  echo "    --extend-data-disk"
  echo
  echo "***************************************************************************************"
  echo
}

UIPATH_VG="uipathvg"
DATADISK_VG="uipathdatavg"
ETCD_VG="uipathetcdvg"
BUNDLE_VG="uipathbundlevg"

RANCHER_LV="rancherlv"
KUBELET_LV="kubeletlv"
INSTALLER_LV="installerlv"
DATADISK_LV="longhornlv"
ETCD_LV="etcdlv"
BUNDLE_LV="bundlelv"

RANCHER_LV_SIZE=${RANCHER_LV_SIZE:-"190"}
KUBELET_LV_SIZE=${KUBELET_LV_SIZE:-"56"}
INSTALLER_LV_SIZE=${INSTALLER_LV_SIZE:-"10"}

FULL_SUITE=0
PRIMARY_SERVER=0
EXTEND_DATA_DISK=0
RESIZE=0
CLUSTER_DISK_SIZE=${CLUSTER_DISK_SIZE:-"$((RANCHER_LV_SIZE + KUBELET_LV_SIZE + INSTALLER_LV_SIZE))"}
FULL_SUITE_DATADISK_SIZE=${FULL_SUITE_DATADISK_SIZE:-"2048"}
BASIC_SUITE_DATADISK_SIZE=${BASIC_SUITE_DATADISK_SIZE:-"512"}
ETCD_DISK_SIZE=${ETCD_DISK_SIZE:-"16"}
BUNDLE_DISK_SIZE=${BUNDLE_DISK_SIZE:-"512"}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -n | --node-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        NODE_TYPE="$2"
        shift 2
      else
        error "Value for $1 is missing"
        shift 2
      fi
      ;;

    -i | --install-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        INSTALL_TYPE="$2"
        shift 2
      else
        error "Value for $1 is missing"
        shift 2
      fi
      ;;

    -c | --cluster-disk-name)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        CLUSTER_DEV_NAME="$2"
        shift 2
      else
        error "Value for $1 is missing"
        shift 2
      fi
      ;;

    -e | --etcd-disk-name)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        ETCD_DEV_NAME="$2"
        shift 2
      else
        error "Value for $1 is missing"
        shift 2
      fi
      ;;

    -l | --data-disk-name)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        DATADISK_DEV_NAME="$2"
        shift 2
      else
        error "Value for $1 is missing"
        shift 2
      fi
      ;;

    -b | --bundle-disk-name)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        BUNDLE_DEV_NAME="$2"
        shift 2
      else
        error "Value for $1 is missing"
        shift 2
      fi
      ;;

    -p | --primary-server)
      PRIMARY_SERVER=1
      shift
      ;;

    -f | --complete-suite)
      FULL_SUITE=1
      shift
      ;;

    -x | --extend-data-disk)
      EXTEND_DATA_DISK=1
      shift
      ;;

    -r | --resize)
      RESIZE=1
      shift
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -* | *) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      exit 1
      ;;
    esac
  done

  [[ -z $NODE_TYPE ]] && warn "-n|--node-type parameter is missing. Setting it as server" && NODE_TYPE="server"
  [[ -z $INSTALL_TYPE ]] && warn "-i|--install-type parameter is missing. Setting it as online" && INSTALL_TYPE="online"

  if [[ "$EXTEND_DATA_DISK" == 0 ]]; then
    [[ -z $CLUSTER_DEV_NAME ]] && error "Error: Required argument -c|--cluster-disk-name missing" && is_partition_exist "$CLUSTER_DEV_NAME"

    [[ $NODE_TYPE == "server" ]] && [[ -z $ETCD_DEV_NAME ]] &&
      error "Error: Required argument -e|--etcd-disk-name missing" &&
      is_disk_exist "$ETCD_DEV_NAME"

    [[ $NODE_TYPE == "server" ]] && [[ -z $DATADISK_DEV_NAME ]] &&
      error "Error: Required argument -l|--data-disk-name missing" &&
      is_partition_exist "$DATADISK_DEV_NAME"

    [[ "${NODE_TYPE}" != "server" && "${NODE_TYPE}" != "agent" ]] &&
      error "Error: Argument -n|--node-type possible values are agent, server"

    [[ "${INSTALL_TYPE}" != "online" && "${INSTALL_TYPE}" != "offline" ]] &&
      error "Error: Argument -i|--install-type possible values are online, offline"

    disk_size=$(get_size_in_GB "$CLUSTER_DEV_NAME")
    [[ "$disk_size" -lt "${CLUSTER_DISK_SIZE}" ]] &&
      error "Size of cluster disk should be atleast ${CLUSTER_DISK_SIZE}GiB. Current provisioned disk size is ${disk_size}GiB."

    if [[ "${NODE_TYPE}" == "server" ]]; then
      data_disk_size=$(get_size_in_GB "$DATADISK_DEV_NAME")
      [[ "${FULL_SUITE}" == "1" && "$data_disk_size" -lt "${FULL_SUITE_DATADISK_SIZE}" ]] &&
        error "Minimum requirement for complete suite datadisk is ${FULL_SUITE_DATADISK_SIZE}GiB. Current provision disk size is ${data_disk_size}GiB."

      [[ "${FULL_SUITE}" == "0" && "$data_disk_size" -lt "${BASIC_SUITE_DATADISK_SIZE}" ]] &&
        error "Minimum requirement for basic suite datadisk is ${BASIC_SUITE_DATADISK_SIZE}GiB. Current provision disk size is ${data_disk_size}GiB."

      etcd_disk_size=$(get_size_in_GB "$ETCD_DEV_NAME")
      [[ "$etcd_disk_size" -lt "${ETCD_DISK_SIZE}" ]] &&
        error "Size of etcd disk should be ${ETCD_DISK_SIZE}GiB. Current provision disk size is ${etcd_disk_size}GiB."

      if [[ "${ETCD_DEV_NAME}" == "$DATADISK_DEV_NAME" ]] || [[ "${ETCD_DEV_NAME}" == "$CLUSTER_DEV_NAME" ]]; then
        error "etcd cannot be co-hosted on datadisk or clusterdisk as etcd is very sensitive to IO"
      fi

      [[ "${INSTALL_TYPE}" == "offline" && "${PRIMARY_SERVER}" == "1" && -z $BUNDLE_DEV_NAME ]] &&
      error "Error: Required argument -b|--bundle-disk-name missing."

      [[ "${INSTALL_TYPE}" == "offline" && "${PRIMARY_SERVER}" == "1" && -n $BUNDLE_DEV_NAME ]] &&
        is_partition_exist "$BUNDLE_DEV_NAME"
    fi
  else
    [[ -z $DATADISK_DEV_NAME ]] && [[ $RESIZE == "0" ]] &&
      error "Error: Required argument -l|--data-disk-name missing" &&
      is_partition_exist "$DATADISK_DEV_NAME"
  fi
}

function get_size_in_GB() {
  local dev_name=$1
  local size_in_bytes
  size_in_bytes=$(lsblk "$dev_name" -b --nodeps --noheadings --paths --output size)
  local divisor=$((1024 * 1024 * 1024))
  local size_in_GB=$(((size_in_bytes + divisor - 1) / divisor)) # this will round up value
  echo ${size_in_GB}
}

function is_disk_exist() {
  local dev_name=$1
  lsblk --nodeps --noheadings --paths |
    grep -q "${dev_name}" ||
    error "Disk $dev_name does not exist"
}

function is_partition_exist() {
  local dev_name=$1
  lsblk  --noheadings --paths |
    grep -q "${dev_name}" ||
    error "Partition  $dev_name does not exist"
}

function create_lv() {
  local lv_name=$1
  local lv_size=$2
  local vg_name=$3

  dev_name="/dev/$vg_name/$lv_name"

  part_size=$(vgdisplay "$vg_name" -c | cut -d ":" -f13)

  parts=$((lv_size * 1024 * 1024 / part_size - 1))

  lvdisplay -c | cut -d ':' -f1 | grep -q "$dev_name" ||
    lvcreate -n "$lv_name" -l "$parts" "$vg_name" -y >/dev/null ||
    error "Unable to create logical volume for host partition: $dev_name"

  echo "$dev_name"
}

function create_vg() {
  local vg_name=$1
  local dev_name=$2

  vgdisplay -c | cut -d ':' -f1 | grep -q "$vg_name" ||
    vgcreate "$vg_name" "$dev_name" ||
    error "Unable to create volume group \"$vg_name\" from device \"$dev_name\""
}

function extend_vg() {
  local vg_name=$1
  local dev_name=$2

  if vgdisplay -c | cut -d ':' -f1 | grep -q "$vg_name"; then
    vgextend "$vg_name" "$dev_name"
  else
    error "Volume group \"$vg_name\" not found. Unable to extend volume from device \"$dev_name\""
  fi

}

function format_device() {
  local device_name=$1
  current_fs=$(lsblk -no FSTYPE "$device_name")
  [[ $current_fs == "xfs" ]] && info "$device_name is already formatted in xfs format, skipping"
  if [[ -z "${current_fs}" ]]; then
    mkfs.xfs -f "${device_name}" || error "Unable to format $device_name"
    info "$device_name formatted in xfs format"
  fi
}

function mount_partition() {
  local device_name=$1
  local mount_path=$2

  format_device "$device_name"

  if entry=$(grep -sE "(^|\s)$mount_path($|\s)" /proc/mounts); then
    mounted_dev_name=$(echo "$entry" | cut -d " " -f1)
    info "Partition: $mount_path is already mounted using Device: $mounted_dev_name, skipping"
  else
    mkdir -p "$mount_path"
    mount "${device_name}" "${mount_path}" || error "Unable to mount partition $mount_path under Device $device_name"
    success "Successfully mounted partition $mount_path using device $device_name"
  fi
}

function add_fstab_entry() {
  local dev_name=$1
  local mount_path=$2
  local part_name=$3

  uuid=$(blkid -o value -s UUID "${dev_name}")
  if [[ -z "${uuid}" ]]; then error "Unable to fetch UUID. Invalid DeviceName: $uuid"; fi

  grep -q "${mount_path}" /etc/fstab ||
    printf "# uipath-$part_name\nUUID=%s    %s    xfs    defaults,nofail    0    0\n" "${uuid}" "${mount_path}" >>/etc/fstab
}

function configure_cluster_disk() {
  local dev_name=$1
  local mount_path_rancher="/var/lib/rancher"
  local mount_path_kubelet="/var/lib/kubelet"
  local mount_path_installer="/opt/UiPathAutomationSuite"

  create_vg $UIPATH_VG "$dev_name"

  rancher_dev_name=$(create_lv $RANCHER_LV "${RANCHER_LV_SIZE}" $UIPATH_VG)
  mount_partition "$rancher_dev_name" $mount_path_rancher
  add_fstab_entry "$rancher_dev_name" $mount_path_rancher "rancher"
  info "$mount_path_rancher is configured"

  kubelet_dev_name=$(create_lv $KUBELET_LV "${KUBELET_LV_SIZE}" $UIPATH_VG)
  mount_partition "$kubelet_dev_name" $mount_path_kubelet
  add_fstab_entry "$kubelet_dev_name" $mount_path_kubelet "kubelet"
  info "$mount_path_kubelet is configured"

  installer_dev_name=$(create_lv $INSTALLER_LV "${INSTALLER_LV_SIZE}" $UIPATH_VG)
  mount_partition "$installer_dev_name" $mount_path_installer
  add_fstab_entry "$installer_dev_name" $mount_path_installer "installer"
  info "$mount_path_installer is configured"

  success "Cluster disk is configured successfully at dev location $dev_name"
}

function configure_data_disk() {
  local dev_name=$1
  local mount_path_datadisk="/datadisk"

  create_vg $DATADISK_VG "$dev_name"

  disk_size=$BASIC_SUITE_DATADISK_SIZE
  [[ "$FULL_SUITE" == 1 ]] && disk_size=$FULL_SUITE_DATADISK_SIZE

  datadisk_dev_name=$(create_lv $DATADISK_LV "${disk_size}" $DATADISK_VG)
  mount_partition "$datadisk_dev_name" $mount_path_datadisk
  add_fstab_entry "$datadisk_dev_name" $mount_path_datadisk "datadisk"
  info "$mount_path_datadisk is configured"

  success "Data disk is configured successfully at dev location $dev_name"
}

function configure_etcd_disk() {
  local dev_name=$1
  local mount_path_etcd="/var/lib/rancher/rke2/server/db"

  create_vg "$ETCD_VG" "$dev_name"

  etcd_dev_name=$(create_lv $ETCD_LV "${ETCD_DISK_SIZE}" $ETCD_VG)
  mount_partition "$etcd_dev_name" $mount_path_etcd
  add_fstab_entry "$etcd_dev_name" $mount_path_etcd "etcd"
  info "$mount_path_etcd is configured"

  success "etcd disk is configured successfully at dev location $dev_name"
}

function configure_bundle_disk() {
  local dev_name=$1
  local mount_path_bundle="/uipath"

  create_vg "$BUNDLE_VG" "$dev_name"

  bundle_dev_name=$(create_lv $BUNDLE_LV "${BUNDLE_DISK_SIZE}" $BUNDLE_VG)
  mount_partition "$bundle_dev_name" $mount_path_bundle
  add_fstab_entry "$bundle_dev_name" $mount_path_bundle "bundle"
  info "$mount_path_bundle is configured"

  success "UiPath bundle disk is configured successfully at dev location $dev_name"
}

function extend_data_disk() {
  local dev_name=$1

  [[ "$RESIZE" == "0" ]] && extend_vg $DATADISK_VG "$dev_name"
  lv_dev_name="/dev/$DATADISK_VG/$DATADISK_LV"

  if lvextend -r -l +100%FREE "$lv_dev_name"; then
    success "Successfully extended partition /dev/$DATADISK_VG/$DATADISK_LV"
  else
    error "unable to extend the Datadisk volumne: $lv_dev_name"
  fi
}

function detect_os() {
  # shellcheck disable=SC1091
  echo "$(source /etc/os-release ; echo "$ID")"
}

function check_package_installed_in_debian() {
    status="$(dpkg-query -W --showformat='${db:Status-Status}' "$1" &>/dev/null)" && return 1
    [[ ! "$status" == installed ]] && return 1
    
    return 0
}

function check_package_installed_in_rhel() {
    rpm -q "lvm2" &>/dev/null && return 0 || return 1
}

function is_lvm_exported_in_path() {
  lvm version &> /dev/null || error "lvm is not exported in environment variable"
}

function is_lvm_installed() {
  OS_ID="$(detect_os)"
  if [[ "${OS_ID}" == "rhel" ]]; then
    check_package_installed_in_rhel "lvm2" || error "lvm is not installed on machine"
  else
    check_package_installed_in_debian "lvm2" || error "lvm is not installed on machine"
  fi
}

function main() {
  is_lvm_installed && is_lvm_exported_in_path

  [[ "$EXTEND_DATA_DISK" == "1" ]] && extend_data_disk "$DATADISK_DEV_NAME" && exit

  [[ -n ${CLUSTER_DEV_NAME} ]] && configure_cluster_disk "$CLUSTER_DEV_NAME"

  [[ -n ${DATADISK_DEV_NAME} && $NODE_TYPE == "server" ]] && configure_data_disk "$DATADISK_DEV_NAME"

  [[ -n "$ETCD_DEV_NAME" && "$NODE_TYPE" == "server" ]] && configure_etcd_disk "$ETCD_DEV_NAME"

  [[ -n "$BUNDLE_DEV_NAME" && "$NODE_TYPE" == "server" && "$PRIMARY_SERVER" == "1" && "$INSTALL_TYPE" ]] && configure_bundle_disk "$BUNDLE_DEV_NAME"
}

check_running_as_root
parse_long_args "$@"
main
