#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e
#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

readonly SCRIPT_DIR
readonly MODULES_DIR

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

function parse_args() {
  INFILE=""
  while getopts 'di:' flag "$@"; do
    case "${flag}" in
    d)
      info "Running in debug"
      set -x
      ;;
    i)
      info "Using input parameters file: ${OPTARG}"
      INFILE=${OPTARG}
      ;;
    :)
      error "Invalid option: ${OPTARG} requires an argument."
      ;;
    *)
      error "Unexpected option ${flag}"
      ;;
    esac
  done
  if [[ -z "${INFILE}" ]]; then
    error "Mandatory parameters not found"
    display_usage
    exit 1
  fi
  if [[ ! -f "${INFILE}" ]]; then
    error "File ${INFILE} not found"
    exit 1
  fi
  shift $((OPTIND - 1))
}

function get_json_value() {
  local json_path="$1"
  local value

  if ! value="$(jq -er "${json_path}" < "${INFILE}")"; then
    error "${json_path} not found in input file"
  fi

  echo "$value"
}

function parse_input_file() {
  set -e
  MOUNT_POINT="/datadisk"
  SERVERS="$(get_json_value ".servers_ips[0:]" | jq "select(length > 0) | .[]")"
  AGENTS="$(get_json_value ".agents_ips[0:]" | jq "select(length > 0) | .[]")"
  # NFS_SERVER="$(get_json_value ".nfs_ip[0]" | jq "select(length > 0) | .[]")"
  RESTORE_SERVERS="$(get_json_value ".restore_servers_ips[0:]" | jq "select(length > 0) | .[]")"
  RESTORE_AGENTS="$(get_json_value ".restore_agents_ips[0:]" | jq "select(length > 0) | .[]")"

  set +e
}

function install_nfs_server() {
  dnf install nfs-utils -y

  info "Starting NFS Server..."
  systemctl start nfs-server.service
  systemctl enable nfs-server.service

  nfs_server_status
}

function nfs_server_status() {
  started=0
  local try=0
  local maxtry=30
  while (( try != maxtry )) ; do
    active=$(systemctl status nfs-server.service|grep -w active)
    if ! { [ -n "${active}" ] && [ "${active}" != " " ]; }; then
      try=$((try+1))
      info "waiting for the nfs server to be ready...${try}/${maxtry}" &&  sleep 2;
    else
      info "Started NFS Server successfully."
      started=1
      break
    fi
  done

  if [ ${started} -eq 0 ]; then
    error "NFS Server is still not running... exiting restore"
  fi
}

function setup_permissions_on_mountpath() {
  chown -R nobody: "${MOUNT_POINT}"
  chmod -R 777 "${MOUNT_POINT}"

  systemctl restart nfs-utils.service

  nfs_server_status
}

function expose_mountpath_to_nodes() {
  local nfs_client_ips_array=()
  while IFS='' read -r line; do nfs_client_ips_array+=("$line"); done <<< "${SERVERS}"
  while IFS='' read -r line; do nfs_client_ips_array+=("$line"); done <<< "${AGENTS}"
  while IFS='' read -r line; do nfs_client_ips_array+=("$line"); done <<< "${RESTORE_SERVERS}"
  while IFS='' read -r line; do nfs_client_ips_array+=("$line"); done <<< "${RESTORE_AGENTS}"
  for nfs_client_ip in "${nfs_client_ips_array[@]}";
  do
    if [[ -n "${nfs_client_ip}" && "${nfs_client_ip}" != " " ]]; then
      echo "${MOUNT_POINT}    ${nfs_client_ip}(rw,sync,no_all_squash,root_squash)" >> /etc/exports
    fi
  done

  exportfs -arv
  exportfs -s
}

function disable_firewalld() {
  systemctl stop firewalld
  systemctl disable firewalld
}

function main() {
  parse_args "$@"
  parse_input_file

  disable_firewalld

  install_nfs_server

  setup_permissions_on_mountpath

  expose_mountpath_to_nodes
}

main "$@"