#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error_with_code "10001"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error_with_code "10002"
fi

readonly SCRIPT_DIR
readonly MODULES_DIR

export PATH=$PATH:/var/lib/rancher/rke2/bin:${SCRIPT_DIR}/bin:/usr/local/bin
export HELM_EXPERIMENTAL_OCI=1
export OFFLINE_TMP_FOLDER="/tmp"
export DOT_FOLDER=~/.uipath

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/prerequisites.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/certificate_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cluster_utils.sh

export DEFAULTSFILE="${SCRIPT_DIR}/defaults.json"
export MERGEDINPUTFILE="${SCRIPT_DIR}/merged_input.json"

export ETCD_RESTORE=${ETCD_RESTORE:-false}

export UPGRADE_SILENT_MODE=${UPGRADE_SILENT_MODE:-false}

INSTALL_ID="$(uuidgen)"
export INSTALL_ID
readonly INSTALL_ID

readonly lenghty_info_message="This operation may be lengthy. Please allow for it to complete..."

INSTALLINFRA=0
INSTALLFABRIC=0
INSTALLSERVICES=0
# Exporting as it's required in infra-install.sh
export DO_UPGRADE=0
SETUPBACKUP=0
SETUPRESTORE=0
CHECK_PRE_REQUISITES="false"
INSTALLSTANDALONEAICENTER=0
INSTALL_OFFLINE_PREREQS=0
SKIP_PACKAGES_INSTALL=0
SKIP_PRE_REQS=0
INSTALL_TYPE=""
MACHINES=""


#FUNCTIONS
function display_usage() {
  echo "usage: $(basename "$0") [-hd] -i infile"
  echo "  -d                                                  Run in debug"
  echo "  -h                                                  Display help"
  echo "  -k|--install-kubernetes                             Install Kubernetes and infrastructure modules"
  echo "  -j|--join-node <server|agent|task-mining|gpu>       Type of node to be joined to an existing cluster"
  echo "  --kubeconfig                                        Specify path to the kubeconfig file"
  echo "  -f|--install-fabric                                 Install Fabric"
  echo "  -s|--install-services                               Install UiPath Services"
  echo "  -a|--install-all                                    Run all installers in order.If no installation is explicitly selected, --install-all will be implied"
  echo "  -b|--backup                                         Setup Backup"
  echo "  -r|--restore                                        Setup Restore"
  echo "  --restore-volume                                    Restore Volume"
  echo "  --disable-backup                                    Disable Backup"
  echo "  -i|--input <input.json>                             Specify input json file"
  echo "  -o|--output <output.json>                           Specify output json file"
  echo "  --install-type <online|offline>                     Specify installation type. Defaults to online. Cannot be set to online together with --offline-bundle."
  echo "  --offline-bundle </tmp/offline_bundle.tar.gz>       Specify the path to the offline bundle"
  echo "  --optional-offline-bundle </tmp/aic_bundle.tar.gz>  Specify the path to the optional offline bundle"
  echo "  --offline-tmp-folder </uipath>                      Specify parent directory for the temporary offline install folder"
  echo "  --install-offline-prereqs                           Install offline prerequisites"
  echo "  -p|--pre-reqs                                       Check if the VM meets all the pre-requisites for Automation Suite, this has to be passed alone to the installer"
  echo "  -u|--upgrade                                        Upgrade an existing Automation Suite cluster"
  echo "  --skip-packages-install                             Disable installation of package dependencies(helm, jq, etc.)"
  echo "  --skip-pre-reqs                                     Skips pre-requisite checks during install"
  echo "  --install-standalone-aicenter                       Install standalone aicenter service only"
  echo "  --one-time-manual-backup                            Trigger Manual Backup and disable it after completion."
  echo "  -m|--machines <comma seperated machine's IP>        Comma seperated IP address / hostnames of all the machines"
}

function contains() {
  if [[ $1 =~ (^|[[:space:]])$2($|[[:space:]]) ]]; then
    echo "0"
  else
    echo "1"
  fi
}

#######
# Parse long input arguments
# Either -a or --argument or --argument-with-dashes
# To pass a value to an argument, separate with space
# --argument value
######
function parse_long_args() {
  local PARAMS=""
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      DEBUG=1
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -k | --install-kubernetes)
      INSTALLINFRA=1
      shift
      ;;

    -j | --join-node)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        NODE_TYPE="$2"
        if [[ "${NODE_TYPE}" == "server" ]]; then
          NODE_ROLE="server"
        else
          NODE_ROLE="agent"
        fi
        result=$(contains "server agent task-mining gpu" "${NODE_TYPE}")
        [[ "${result}" == "1" ]] && display_usage && error_with_code "10020" "${NODE_TYPE}"
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    --kubeconfig)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        [[ -f "${2}" ]] || error_with_code "10006" "$2"
        KUBECONFIG_PATH="$2"
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    -f | --install-fabric)
      INSTALLFABRIC=1
      shift
      ;;

    -s | --install-services)
      INSTALLSERVICES=1
      shift
      ;;

    -a | --install-all)
      INSTALLINFRA=1
      INSTALLFABRIC=1
      INSTALLSERVICES=1
      shift
      ;;

    -b|--backup)
      SETUPBACKUP=1
      shift
      ;;

    -r|--restore)
      SETUPRESTORE=1
      shift
      ;;

    --volume-restore)
      RESTORE_VOLUME="true"
      shift
      ;;

    --disable-backup)
      DISABLE_BACKUP="true"
      shift
      ;;

    -p|--pre-reqs)
      CHECK_PRE_REQUISITES="true"
      shift
      ;;

    -m|--machines)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        MACHINES="$2"
        shift 2
      else
        display_usage
        error_with_code "10005" "${1}"
        shift 2
      fi
      ;;

    --skip-pre-reqs)
      SKIP_PRE_REQS=1
      shift
      ;;

    -i | --input)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        INFILE=$ABSOLUTEPATH
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    -o | --output)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        OUTFILE=$ABSOLUTEPATH
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    --offline-bundle)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        OFFLINE_BUNDLE_PATH=$ABSOLUTEPATH
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    --offline-tmp-folder)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        OFFLINE_TMP_FOLDER=$ABSOLUTEPATH
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    --optional-offline-bundle)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        OPTIONAL_OFFLINE_BUNDLE_PATH=$ABSOLUTEPATH
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    --install-offline-prereqs)
      INSTALL_OFFLINE_PREREQS=1
      shift
      ;;

    -u|--upgrade)
      export DO_UPGRADE=1
      shift
      ;;

    --accept-license-agreement)
      export LICENSE_AGREEMENT="accept"
      shift
      ;;

    --install-standalone-aicenter)
      INSTALLSTANDALONEAICENTER=1
      shift
      ;;

    --skip-packages-install)
      SKIP_PACKAGES_INSTALL=1
      shift
      ;;

    --non-interactive-upgrade)
      UPGRADE_SILENT_MODE=true
      shift
      ;;

    --one-time-manual-backup)
      ONE_TIME_MANUAL_BACKUP="true"
      SETUPBACKUP=1
      shift
      ;;
    --install-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        INSTALL_TYPE="$2"
        [[ "${INSTALL_TYPE}" != "online" && "${INSTALL_TYPE}" != "offline" ]] && display_usage && error_with_code "10028" "${INSTALL_TYPE}"
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    -*) # unsupported flags
      display_usage
      error_with_code "10007" "$1"
      exit 1
      ;;
    *) # preserve positional arguments
      PARAMS="$PARAMS $1"
      shift
      ;;
    esac
  done

  [[ "${INSTALL_TYPE}" == "online" ]] && [[ -n ${OFFLINE_BUNDLE_PATH} ]] && error_with_code "10029"
  [[ -n "${OFFLINE_BUNDLE_PATH}" ]] && INSTALL_TYPE="offline"
  [[ -z "${INSTALL_TYPE}" ]] && INSTALL_TYPE="online"
  [[ "${INSTALL_TYPE}" == "offline" ]] && [[ -z ${OFFLINE_BUNDLE_PATH} ]] && ((INSTALLINFRA == 1)) && error_with_code "10030"

  [[ "${INSTALL_OFFLINE_PREREQS}" == "1" ]] && [[ -z ${OFFLINE_BUNDLE_PATH} ]] &&
      error_with_code "10008"

  [[ "${CHECK_PRE_REQUISITES}" != "false" ]] &&
    [[ "${INSTALLINFRA}|${INSTALLFABRIC}|${INSTALLSERVICES}|${SETUPBACKUP}|${SETUPRESTORE}" != "0|0|0|0|0" ]] &&
      error_with_code "10009"

  # check if any installation will be executed
  if [[ "${INSTALLINFRA}|${INSTALLFABRIC}|${INSTALLSERVICES}|${SETUPBACKUP}|${SETUPRESTORE}" == "0|0|0|0|0|0" ]] &&
        [[ -z ${OPTIONAL_OFFLINE_BUNDLE_PATH} ]] &&
        [[ ${CHECK_PRE_REQUISITES} == "false" ]] ; then
    info "At least one installer or pre-requisites checks needs to be executed."
    display_usage
    exit 1
  fi
  # validate that fabric is not skipped
  if [[ "${INSTALLINFRA}|${INSTALLFABRIC}|${INSTALLSERVICES}" == "1|0|1" ]]; then
    error_with_code "12008"
  fi

  # don't install fabric and services when joining nodes
  if ((INSTALLFABRIC + INSTALLSERVICES != 0)) && [[ "${NODE_ROLE}" == "agent" ]]; then
    error_with_code "10010"
  fi

  if (( INSTALLINFRA + INSTALLFABRIC + INSTALLSERVICES > 0 )) && [[ -n ${OPTIONAL_OFFLINE_BUNDLE_PATH} ]] ; then
    error_with_code "10021"
  fi

  # validate that no two backup & restore were selected
  if (( SETUPBACKUP + SETUPRESTORE == 2 )) ; then
    error_with_code "10012"
  fi

  # check if provided input file was provided and exists
  if [[ -z "${INFILE}" ]]; then
    error_with_code "10013"
  fi
  if [[ ! -f "${INFILE}" ]]; then
    error_with_code "10014" "${INFILE}"
  fi
  # check if output file was provided
  if [[ -z "${OUTFILE}" && ${CHECK_PRE_REQUISITES} == "false" ]]; then
    error_with_code "10015"
  fi

  # set positional arguments in their proper place
  eval set -- "$PARAMS"
}

########
# Parse input json file to set the config values
# Globals:
#   INFILE
# Outputs:
#   Sets global config variables
########
function parse_input_file() {
  VERSION="$(get_mandatory_json_value ".version")"
  CLUSTER_FQDN="$(get_mandatory_json_value ".fqdn")"
  DOCKER_REGISTRY_FQDN="$(get_mandatory_json_value ".subdomain.docker_registry").${CLUSTER_FQDN}"
  ARGO_FQDN="$(get_mandatory_json_value ".subdomain.argo_cd").${CLUSTER_FQDN}"
  MONITORING_FQDN="$(get_mandatory_json_value ".subdomain.rancher_server").${CLUSTER_FQDN}"
  INSIGHTS_FQDN="$(get_mandatory_json_value ".subdomain.looker").${CLUSTER_FQDN}"
  OBJECTSTORE_FQDN="$(get_mandatory_json_value ".subdomain.rook_ceph").${CLUSTER_FQDN}"
  DOCKER_REGISTRY_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
  # support for online private installation from OCI Helm registry
  HELM_REGISTRY_URL="$(get_mandatory_json_value ".registries.helm.url")"
  HELM_REGISTRY_USERNAME="$(get_optional_json_value ".registries.helm.username")"
  HELM_REGISTRY_PASSWORD="$(get_optional_json_value ".registries.helm.password")"
  PODMAN_FOLDER="$(get_mandatory_json_value ".infra.offline.dir")/lib/containers"
  export HELM_REGISTRY_URL
  export HELM_REGISTRY_USERNAME
  export HELM_REGISTRY_PASSWORD

  # # infra specific versions
  DOCKER_REGISTRY_VERSION="$(get_mandatory_json_value ".infra.docker_registry.version")"
  LONGHORN_VERSION="$(get_mandatory_json_value ".infra.longhorn.version")"

  DOCKER_REGISTRY_NODEPORT="$(get_mandatory_json_value ".infra.docker_registry.node_port")"

  RANCHER_ISTIO_VERSION="$(get_mandatory_json_value ".infra.rancher_istio.version")"

  TELEMETRY_OPT_OUT="$(get_optional_json_value ".telemetry_optout")"

  INSTALL_MULTINODE="$(get_optional_json_value ".multinode")"

  if [[ "${INSTALL_TYPE}" == "offline" ]];
  then
    TELEMETRY_OPT_OUT="true"
  fi

  export INSTALL_MULTINODE
  export TELEMETRY_OPT_OUT
  export CLUSTER_FQDN
  export INSTALLSTANDALONEAICENTER
  export VERSION
}

# setup local helm cache and charts for infra installation
# Helm charts are pulled from the OCI compliant registry and stored locally
# if HELM_REGISTRY_USERNAME and HELM_REGISTRY_PASSWORD are set, then the helm registry is private and authentication is required
function download_infra_charts() {
  info "Using local helm folder ${TMP_HELM_CHARTS_DIR}"
  if [[ -n "${HELM_REGISTRY_USERNAME}" ]] && [[ -n "${HELM_REGISTRY_PASSWORD}" ]]; then
    info "Using private Helm registry located at ${HELM_REGISTRY_URL}"
    echo "${HELM_REGISTRY_PASSWORD}" |
      helm registry login "${HELM_REGISTRY_URL}" \
        --username "${HELM_REGISTRY_USERNAME}" \
        --password-stdin &> /dev/null
  fi

  declare -a charts=(
    "helm/docker-registry ${DOCKER_REGISTRY_VERSION}"
    "vendor/longhorn ${LONGHORN_VERSION}"
    "vendor/rancher-istio ${RANCHER_ISTIO_VERSION}"
    "helm/istio-configure ${DOCKER_REGISTRY_VERSION}"
    "helm/self-heal-operator ${DOCKER_REGISTRY_VERSION}"
  )

  for crt_chart in "${charts[@]}"; do
    IFS=' ' read -ra crt_chart_arr <<<"${crt_chart}"

    local try=0
    local maxtry=10
    local status="notready"
    while [[ ${status} == "notready" ]] && ((try != maxtry)); do
      try=$((try + 1))
      info "Pulling chart ${HELM_REGISTRY_URL}/${crt_chart_arr[0]}:${crt_chart_arr[1]} ...try ${try}/${maxtry}"
      helm chart pull "${HELM_REGISTRY_URL}"/"${crt_chart_arr[0]}":"${crt_chart_arr[1]}" &> /dev/null && status="ready"

      [[ ${status} == "ready" ]] || sleep 2
    done
    [[ ${status} == "ready" ]] || error_with_code "10022" "${HELM_REGISTRY_URL}/${crt_chart_arr[0]}:${crt_chart_arr[1]}"
    helm chart export "${HELM_REGISTRY_URL}"/"${crt_chart_arr[0]}":"${crt_chart_arr[1]}" --destination "${TMP_HELM_CHARTS_DIR}" &> /dev/null
  done
}

# Verify fabric installer bundle against the expected SHA sum
function check_bundle() {
  local bundle_path=${1}
  local bundle_sha=${2}

  info "Verifying ${bundle_path}"

  [[ -f "${bundle_path}" ]] || error_with_code "10006" "${bundle_path}"
  [[ -f "${bundle_sha}" ]] || error_with_code "10006" "${bundle_sha}"

  SHA_ACTUAL=$(sha256sum "${bundle_path}" | awk '{print $1}')
  SHA_EXPECTED=$(<"${bundle_sha}")

  if [ "${SHA_ACTUAL}" != "${SHA_EXPECTED}" ]; then
    error_with_code "10016" "${bundle_path}"
  fi

  info "${bundle_path} passed the checks"
}

# Populate docker registry with images from the offline bundle
function import_sf_images() {
  local sf_folder="${1}"
  local image_inventory="${sf_folder}/images/image-tags.txt"
  local image_tar="${sf_folder}/images/images.tar"

  info "Populating Docker Registry at https://localhost:${DOCKER_REGISTRY_NODEPORT}"

  [[ -f "${image_inventory}" ]] || error_with_code "10006" "${image_inventory}"
  local all_images_present=1

  while read -r image_name; do
    podman image exists "localhost:${DOCKER_REGISTRY_NODEPORT}/${image_name}" || all_images_present=0
  done <"${image_inventory}"

  if (( all_images_present != 1 ));
    then
      info "Loading image tar ${image_tar}"
      info "${lenghty_info_message}"
      export TMPDIR="${PODMAN_FOLDER}/tmp"
      mkdir -p "${TMPDIR}"
      # inline ENV for https://github.com/containers/podman/pull/5412 and https://github.com/containers/podman/issues/5960
      podman load -i "${image_tar}" || error_with_code "10017" "${image_tar}"
    else
      info "All images present. Skipping podman load"
  fi
  while read -r image_name; do
    local try=0
    local maxtry=10
    local status="notready"
    while [[ ${status} == "notready" ]] && ((try != maxtry)); do
      try=$((try + 1))
      info "Pushing image ${image_name}...try ${try}/${maxtry}" && sleep 5
      podman push "localhost:${DOCKER_REGISTRY_NODEPORT}/${image_name}" \
        --tls-verify=false >/dev/null &&
        status="ready"
      [[ ${status} == "ready" ]] || sleep 60
    done
    [[ ${status} == "ready" ]] || error_with_code "10018" "${image_name}"
  done <"${image_inventory}"

  info "Populated Docker Registry"
}

# Populate docker registry with charts from the offline bundle
function import_sf_charts() {
  local sf_folder="${1}"
  local chart_inventory="${sf_folder}/helm/helm-tags.txt"

  info "Populating Helm Registry at ${HELM_REGISTRY_URL}"

  [[ -f "${chart_inventory}" ]] || error_with_code "10006" "${chart_inventory}"
  while read -r line; do
    local chart_name file_name
    chart_name=$(awk '{print $1}' <<< "${line}")
    file_name=$(awk '{print $2}' <<< "${line}")
    helm chart save "${sf_folder}/helm/${file_name}" "${HELM_REGISTRY_URL}/${chart_name}"
    local try=0
    local maxtry=10
    local status="notready"
    while [[ ${status} == "notready" ]]  && (( try != maxtry )) ; do
      try=$((try+1))
      info "Pushing local chart ${sf_folder}/helm/${file_name} to ${chart_name}...try ${try}/${maxtry}" && sleep 2;
      helm chart push "${HELM_REGISTRY_URL}/${chart_name}" && status="ready"
    done
    [[ ${status} == "ready" ]] || error_with_code "10019" "${sf_folder}/helm/${file_name}"
  done < "${chart_inventory}"

  info "Populated Helm Registry"
}

function import_offline_sf_artifacts() {
  check_offline_prerequisites_and_exit
  local sf_bundle="${OFFLINE_INSTALL_FOLDER}/sf-bundle.tar.gz"
  local sf_sha="${OFFLINE_INSTALL_FOLDER}/sf-bundle-sha256sum.txt"
  local offline_sf_dir="${OFFLINE_INSTALL_FOLDER}/sf"

  configure_docker_tools

  check_bundle "${sf_bundle}" "${sf_sha}"
  info "Extracting ${sf_bundle} to ${offline_sf_dir}"
  info "${lenghty_info_message}"
  mkdir -p "${offline_sf_dir}" && tar -xzf "${sf_bundle}" -C "${offline_sf_dir}"
  rm -f "${sf_bundle}"

  import_sf_images "${offline_sf_dir}"
  import_sf_charts "${offline_sf_dir}"
}

function import_on_demand_artifacts() {
  check_offline_prerequisites_and_exit
  local offline_bundle_dir
  local bundle
  local filename
  local bundle_root

  configure_docker_tools

  for bundle in "${OFFLINE_BUNDLE_FOLDER}"/*.tar.gz; do
      filename=$(basename -- "${bundle}")
      echo "${filename}"
      bundle_root="${filename%%.*}"
      offline_bundle_dir="${OFFLINE_BUNDLE_FOLDER}/${bundle_root}"
      check_bundle "${OFFLINE_BUNDLE_FOLDER}/${bundle_root}.tar.gz" "${OFFLINE_BUNDLE_FOLDER}/${bundle_root}-sha256sum.txt"
      info "Extracting ${bundle} to ${offline_bundle_dir}"
      mkdir -p "${offline_bundle_dir}" && tar -xzf "${bundle}" -C "${offline_bundle_dir}"

      import_sf_images "${offline_bundle_dir}"
  done
}

function import_offline_infra_artifacts() {
  local rke2_bundle="${OFFLINE_INSTALL_FOLDER}/rke2-bundle.tar.gz"
  local rke2_sha="${OFFLINE_INSTALL_FOLDER}/rke2-bundle-sha256sum.txt"
  local infra_bundle="${OFFLINE_INSTALL_FOLDER}/infra-bundle.tar.gz"
  local infra_sha="${OFFLINE_INSTALL_FOLDER}/infra-bundle-sha256sum.txt"
  local offline_rke2_dir="${OFFLINE_INSTALL_FOLDER}/rke2"
  local rke2_agent_images_dir="/var/lib/rancher/rke2/agent/images"
  local offline_infra_dir="${OFFLINE_INSTALL_FOLDER}/infra"
  local infra_chart_inventory="${OFFLINE_INSTALL_FOLDER}/infra/helm/helm-tags.txt"

  check_bundle "${rke2_bundle}" "${rke2_sha}"
  info "Extracting ${rke2_bundle} to ${offline_rke2_dir}"
  mkdir -p "${offline_rke2_dir}" && tar -xzf "${rke2_bundle}" -C "${offline_rke2_dir}"
  rm -f "${rke2_bundle}"

  export INSTALL_RKE2_ARTIFACT_PATH="${offline_rke2_dir}"
  info "Set INSTALL_RKE2_ARTIFACT_PATH to ${offline_rke2_dir}"

  check_bundle "${infra_bundle}" "${infra_sha}"
  info "Extracting ${infra_bundle} to ${offline_infra_dir}"
  mkdir -p "${offline_infra_dir}" && tar -xzf "${infra_bundle}" -C "${offline_infra_dir}"
  rm -f "${infra_bundle}"

  mkdir -p "${rke2_agent_images_dir}"
  cp -r "${offline_infra_dir}/images"/* "${rke2_agent_images_dir}"
  info "Infra images copied to ${rke2_agent_images_dir}"

  info "Preparing the infra helm charts folder"
  [[ -f "${infra_chart_inventory}" ]] || error_with_code "10006" "${infra_chart_inventory}"
  TMP_HELM_CHARTS_DIR="${offline_infra_dir}/helm"
  while read -r line; do
    local chart_name file_name
    chart_name=$(awk '{print $1}' <<< "${line}")
    file_name=$(awk '{print $2}' <<< "${line}")
    info "Expanding ${file_name} to ${TMP_HELM_CHARTS_DIR}/${chart_name}"
    helm chart save "${TMP_HELM_CHARTS_DIR}/${file_name}" "localhost:${DOCKER_REGISTRY_NODEPORT}/${chart_name}" > /dev/null
    helm chart export "localhost:${DOCKER_REGISTRY_NODEPORT}/${chart_name}" --destination "${TMP_HELM_CHARTS_DIR}" > /dev/null
  done < "${infra_chart_inventory}"
}

function install_offline_prereqs() {
  local prereqs_bundle="${OFFLINE_INSTALL_FOLDER}/prereqs.tar.gz"
  local prereqs_sha="${OFFLINE_INSTALL_FOLDER}/prereqs-sha256sum.txt"
  local offline_prereqs_dir="${OFFLINE_INSTALL_FOLDER}/prereqs"
  local container_tools_enabled="false"
  local pkgs_list=(iscsi-initiator-utils jq nfs-utils rpcbind util-linux nc podman openssl httpd-tools)

  info "Installing offline prerequisites from ${offline_prereqs_dir}"
  check_bundle "${prereqs_bundle}" "${prereqs_sha}"
  mkdir -p "${offline_prereqs_dir}" && tar -xzf "${prereqs_bundle}" -C "${offline_prereqs_dir}"

  cat > /etc/yum.repos.d/uipath_prereqs.repo <<EOF
[uipath_prereqs]
name=uipath_prereqs
baseurl=file://${offline_prereqs_dir}
enabled=0
gpgcheck=0
EOF

  dnf module list --disablerepo="*" --enablerepo="uipath_prereqs" --enabled container-tools && container_tools_enabled="true"
  [[ ${container_tools_enabled} == "true" ]] && dnf -y module disable container-tools --disablerepo="*" --enablerepo="uipath_prereqs"
  for pkg in "${pkgs_list[@]}"; do
    dnf -y install --disablerepo=* --enablerepo="uipath_prereqs" "${pkg}" --allowerasing
  done
  systemctl enable iscsid
  systemctl start iscsid
}

# configure the podman installation
# create both rootless and rootfull paths
# update the config file (storage.conf) to point to these paths
# update the config file (storage.conf) to uncomment the overlayfs config
function configure_docker_tools() {
  local podman_config_path="/etc/containers/storage.conf"
  local podman_rootless_path="${PODMAN_FOLDER}/podman_rootless"
  local podman_rootfull_path="${PODMAN_FOLDER}/podman_rootfull"

  if [[ -n "${PODMAN_FOLDER}" ]]; then
    info "Configuring podman to use ${PODMAN_FOLDER}"
    mkdir -p "${podman_rootless_path}" || error_with_code "10031" "${podman_rootless_path}"
    mkdir -p "${podman_rootfull_path}" || error_with_code "10031" "${podman_rootfull_path}"
    # update the graphroot config line
    sed -i "s|^\(graphroot[[:space:]]=\).*|\1\"${podman_rootfull_path}\"|" "${podman_config_path}"

    # uncomment and update the rootless path config line
    sed -i "s|^#[[:space:]]*\(rootless_storage_path[[:space:]]*=\).*|\1\"${podman_rootless_path}\"|" "${podman_config_path}"
  fi

  podman login "localhost:${DOCKER_REGISTRY_NODEPORT}" --username "${HELM_REGISTRY_USERNAME}" --password "${HELM_REGISTRY_PASSWORD}" --tls-verify=false || error_with_code "10026" "${DOCKER_REGISTRY_NODEPORT}"

  local try=0
  local maxtry=10
  local status="notready"
  while [[ ${status} == "notready" ]] && ((try != maxtry)); do
    try=$((try + 1))
    info "Helm login to registry at ${HELM_REGISTRY_URL}...try ${try}/${maxtry}"
    helm registry login "${HELM_REGISTRY_URL}" --username "${HELM_REGISTRY_USERNAME}" --password "${HELM_REGISTRY_PASSWORD}" && status="ready"
    [[ ${status} == "ready" ]] || sleep 30
  done
  [[ ${status} == "ready" ]] || error_with_code "10023" "${HELM_REGISTRY_URL}"
  # uncomment the mount program
  sed -i "/^#mount_program[[:space:]]*=/s|^#||g" "${podman_config_path}"
}

function setup_offline_infra() {
  if [[ ${INSTALL_TYPE} == "offline" ]]; then
    import_offline_infra_artifacts
  else
    TMP_HELM_CHARTS_DIR=$(mktemp -d -p "${OFFLINE_TMP_FOLDER}" -t sf-infra-helm-charts.XXXXXXXXXX)
    download_infra_charts
  fi
  info "Using infra helm charts from ${TMP_HELM_CHARTS_DIR}"
  export TMP_HELM_CHARTS_DIR
}

function install_infra() {
  setup_offline_infra
  "${SCRIPT_DIR}"/Infra_Installer/infra-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/infra_output.json"
}

function install_fabric() {
  "${SCRIPT_DIR}"/Fabric_Installer/fabric-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/fabric_output.json"
}

function install_services() {
  "${SCRIPT_DIR}"/Services_Installer/services-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/services_output.json"
}

function do_upgrade() {
  (( "${INSTALLINFRA}" == "1" )) && setup_offline_infra
  "${SCRIPT_DIR}"/Upgrade/upgrade.sh -i "${INFILE}" -o "${SCRIPT_DIR}/upgrade_output.json" -k "${INSTALLINFRA}" -f "${INSTALLFABRIC}" -s "${INSTALLSERVICES}"
}

function install_aicenter_services() {
  "${SCRIPT_DIR}"/Services_Installer/aicenter-service-installer.sh -i "${INFILE}" -o "${SCRIPT_DIR}/services_output.json"
}

function setup_backup() {
    "${SCRIPT_DIR}"/Infra_Installer/disaster_recovery/backup.sh -i "${INFILE}" -o "${SCRIPT_DIR}/backup_output.json"
}

function setup_restore() {
    if [[ -z "${RESTORE_VOLUME}" ]]; then
      setup_offline_infra
    fi
    "${SCRIPT_DIR}"/Infra_Installer/disaster_recovery/restore.sh -i "${INFILE}" -o "${SCRIPT_DIR}/restore_output.json"
}

function check_eula() {
    [[ -z "${LICENSE_AGREEMENT}" ]] || EULA="${LICENSE_AGREEMENT}"
    if [[ "${EULA}" != "accept" ]]; then
      echo "Use of paid UiPath products and services is subject to the licensing agreement executed between you and UiPath."
      echo "Unless otherwise indicated by UiPath, use of free UiPath products is subject to the associated licensing agreement available here: https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website)."
      echo "By setting the environment variable LICENSE_AGREEMENT=accept or by using the --accept-license-agreement argument, you agree to the applicable licensing agreement."
      exit 0
    fi
}

function validate_cluster_requirement() {
  local allow_installation_value
  allow_installation_value=$("${MODULES_DIR}"/cluster-validation.sh)

  echo "${allow_installation_value}"
}

function set_offline_variables() {
  local docker_registry_password_name="docker-registry-credentials"
  HELM_REGISTRY_URL="${DOCKER_REGISTRY_FQDN}:${DOCKER_REGISTRY_EXTERNAL_PORT}"
  HELM_REGISTRY_PASSWORD=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.password\}\} | base64 -d)
  HELM_REGISTRY_USERNAME=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.username\}\} | base64 -d)
}

function parse_input(){
    merge_default_values && export INFILE
    parse_input_file
    info "Provided input file: ${INFILE}"
}

function read_certificate_files() {
  if [[ (-z "${NODE_ROLE}" || "${NODE_ROLE}" == "none") && "${ETCD_RESTORE}" != "true" ]]; then
    CERT_CA_FILE_PATH="$(get_optional_json_value ".server_certificate.ca_cert_file")"
    CERT_CA_FILE_PATH="${CERT_CA_FILE_PATH/#\~/$HOME}"

    CERT_TLS_FILE_PATH="$(get_optional_json_value ".server_certificate.tls_cert_file")"
    CERT_TLS_FILE_PATH="${CERT_TLS_FILE_PATH/#\~/$HOME}" #allow paths relative to the home directory

    CERT_KEY_FILE_PATH="$(get_optional_json_value ".server_certificate.tls_key_file")"
    CERT_KEY_FILE_PATH="${CERT_KEY_FILE_PATH/#\~/$HOME}"

    if [[ ! -f "${CERT_CA_FILE_PATH}" || ! -f "${CERT_TLS_FILE_PATH}" || ! -f "${CERT_KEY_FILE_PATH}" ]]; then
      info "One or more certificate files not found..."
      info "Generating TLS certificate..."

      create_self_issued_certificate "${CLUSTER_FQDN}" "${CERT_CA_FILE_PATH}" "${CERT_TLS_FILE_PATH}" "${CERT_KEY_FILE_PATH}"

      # adding certificate in the trust store of node
      cp "${CERT_CA_FILE_PATH}" "/etc/pki/ca-trust/source/anchors/"
      update-ca-trust
    fi
    cert_validation_status=$(validate_server_certificate "${CERT_CA_FILE_PATH}" "${CERT_TLS_FILE_PATH}" "${CERT_KEY_FILE_PATH}" "${CLUSTER_FQDN}" "${ARGO_FQDN}" "${OBJECTSTORE_FQDN}" "${DOCKER_REGISTRY_FQDN}" "${MONITORING_FQDN}" "${INSIGHTS_FQDN}")
    [[ "${cert_validation_status}" == "success" ]] && info "Validating Server Certificate... Done"

    update_fqdn=$(get_cluster_config "update_fqdn")
    if [[ "${cert_validation_status}" != "success" && "${update_fqdn}" == "true" ]];
    then
      cat << EOF

Certificate provided is invalid for new FQDN, would you like us to configure new self signed certificate?
If you wish to continue, type 'yes' and hit Enter for confirmation.

EOF
    read -r force </dev/tty
    if [[ $force == 'yes' ]] || [[ $force == "'yes'" ]]
      then
        info "Proceeding to configure new certificate"
      else
        info "Unexpected user input received: ${force}, exiting cluster fqdn update operation"
        exit 2
      fi
      create_self_issued_certificate "${CLUSTER_FQDN}" "${CERT_CA_FILE_PATH}" "${CERT_TLS_FILE_PATH}" "${CERT_KEY_FILE_PATH}"

      # adding certificate in the trust store of node
      cp "${CERT_CA_FILE_PATH}" "/etc/pki/ca-trust/source/anchors/"
      update-ca-trust
    fi


    export CERT_CA_FILE_PATH
    export CERT_TLS_FILE_PATH
    export CERT_KEY_FILE_PATH

    if [[ ${INSTALLSTANDALONEAICENTER} == 0 ]]; then
      TOKEN_SIGNING_CERT="$(get_optional_json_value ".identity_certificate.token_signing_cert_file")"
      TOKEN_SIGNING_CERT="${TOKEN_SIGNING_CERT/#\~/$HOME}"

      TOKEN_SIGNING_PASS="$(get_optional_json_value ".identity_certificate.token_signing_cert_pass")"

      if [[ ! -f "${TOKEN_SIGNING_CERT}" || -z "${TOKEN_SIGNING_PASS}" ]]; then
        info "Token signing certificate files not found..."
        info "Generating token signing certificate..."

        TOKEN_SIGNING_PASS="$(tr </dev/urandom -dc _A-Z-a-z-0-9 | head -c 16)"
        create_token_signing_certificate "${CERT_TLS_FILE_PATH}" "${CERT_KEY_FILE_PATH}" "${TOKEN_SIGNING_CERT}" "${TOKEN_SIGNING_PASS}"
      fi
      validate_signing_certificate "${TOKEN_SIGNING_CERT}" "${TOKEN_SIGNING_PASS}"

      export TOKEN_SIGNING_CERT
      export TOKEN_SIGNING_PASS
    fi
  fi
}

# Checks if previous install type is in sync with the current one, offline or online
function verify_install_type() {
  local existing_install_type
  if systemctl is-enabled rke2-server &>/dev/null || systemctl is-enabled rke2-agent &>/dev/null;
  then
    if rpm -q rke2-common &>/dev/null;
    then
      existing_install_type="online"
    else
      existing_install_type="offline"
    fi

    # Error out if the existing install type does not match the current one
    if [[ "${existing_install_type}" != "${1}" ]];
    then
      error_with_code "10024" "${existing_install_type}"
    fi
  fi
}

function is_cluster_installed() {
  [[ "$(systemctl is-enabled rke2-server 2>/dev/null)" == "enabled" ]] && echo -n "true" && return
  [[ "$(systemctl is-enabled rke2-agent 2>/dev/null)" == "enabled" ]] && echo -n "true" && return
  echo "false"
}

function is_updating_fqdn() {
  if [[ $(is_cluster_installed) == "true" ]];
  then
    current_cluster_fqdn=$(get_cluster_config "cluster_fqdn")
    update_fqdn=$(get_cluster_config "update_fqdn")
    {
      [[ ${update_fqdn} == "true" ]] || [[ -n "${current_cluster_fqdn}" && "${current_cluster_fqdn}" != "${CLUSTER_FQDN}" ]];
    } && echo -n "true" && return
  fi
  echo -n "false"
}

function check_and_warn_fqdn_update() {
  if [[ $(is_updating_fqdn) == "true" ]];
  then

  cat << EOF

Changing Cluster FQDN may result in disruption of Automation Suite usage.
It may disconnect your Robots, Mobile Orchestrator users, ML Activities, ML Skills,
and it may invalidate any pending confirmation emails.

If you wish to continue, type 'yes' and hit Enter for confirmation.
EOF
  read -r force </dev/tty
   if [[ $force == 'yes' ]] || [[ $force == "'yes'" ]]
    then
      info "Proceeding to update cluster fqdn"
    else
      info "Unexpected user input received: ${force}, exiting cluster fqdn update operation"
      exit 2
    fi

    local service_install_type
    service_install_type=$(get_cluster_config "SERVICE_INSTALL_TYPE")
    if [[ "${service_install_type}" == "aicenter" ]]; then
      export INSTALLSTANDALONEAICENTER=1
    fi
    update_cluster_config "update_fqdn" "true"
    export UPDATE_FQDN=1
  fi
}

function get_agent_type() {
  local taints
  taints=$(< "/etc/rancher/rke2/config.yaml" grep "node-taint")
  local agent_type="agent"
  if [[ -n "${taints}" ]]; then
    if [[ "${taints}" == *"task.mining"* ]]; then
      agent_type="task-mining"
    elif [[ "${taints}" == *"gpu"* ]]; then
      agent_type="gpu"
    elif [[ "${taints}" == *"jit"* ]]; then
      agent_type="jit"
    fi
  fi
  echo "${agent_type}"
}

function main() {

  local allow_installation_of_services="true"
  if (( SETUPBACKUP == 0 )); then
    parse_input
  fi

  check_eula
  load_os_utils "${MODULES_DIR}"
  export KUBECONFIG="${KUBECONFIG_PATH:-/etc/rancher/rke2/rke2.yaml}"
  # If upgrading, determine current node type.
  if [[ "${DO_UPGRADE}" == "1" ]]; then
    [[ "$(systemctl is-enabled rke2-server 2>/dev/null)" == "enabled" ]] && NODE_ROLE="server" && NODE_TYPE="server"
    [[ "$(systemctl is-enabled rke2-agent 2>/dev/null)" == "enabled" ]] && NODE_ROLE="agent" && NODE_TYPE=$(get_agent_type)
    [[ -z ${NODE_ROLE} ]] && error_with_code "11028"
    if [[ ${NODE_ROLE} == "server" && $(determine_first_node_upgrade "${VERSION}") == "true" ]]; then
      NODE_ROLE="none"
    fi
  fi
  [[ ${NODE_ROLE} == "agent" ]] && export KUBECONFIG="/var/lib/rancher/rke2/agent/kubelet.kubeconfig"
  export INSTALLSTANDALONEAICENTER
  export NODE_ROLE=${NODE_ROLE:-none}
  export NODE_TYPE=${NODE_TYPE:-none}


  local prerequisites_args=()
  prerequisites_args+=(--install-type "${INSTALL_TYPE}")
  [[ -n "${MACHINES}" ]] && prerequisites_args+=(--machines "${MACHINES}")
  [[ "${CHECK_PRE_REQUISITES}" == "true" ]] && check_prerequisites "${prerequisites_args[@]}" -j "${NODE_TYPE}" && exit 0

  [[ "${INSTALLINFRA}" == "1" ]] && require_root_user

  [[ -d "${DOT_FOLDER}" ]] || mkdir -p "${DOT_FOLDER}"

  if [[ "${DO_UPGRADE}" == "1" && "${INSTALLINFRA}" == "1" && $(determine_first_node_upgrade "${VERSION}") == "true" ]];
  then
    cluster_backup
  fi

  if (( INSTALLINFRA + INSTALLFABRIC + INSTALLSERVICES + DO_UPGRADE + SETUPRESTORE != 0 )); then
    if [[ ${INSTALL_TYPE} == "offline" ]]; then
      if [[ -z "${RESTORE_VOLUME}" ]]; then
        verify_install_type "offline"
        if [[ -n ${OFFLINE_BUNDLE_PATH} ]]; then
          info "Running the offline install using ${OFFLINE_BUNDLE_PATH}"
          OFFLINE_INSTALL_FOLDER="${OFFLINE_TMP_FOLDER}/sf_installer"
          mkdir -p "${OFFLINE_INSTALL_FOLDER}"
          info "Extracting ${OFFLINE_BUNDLE_PATH} to ${OFFLINE_INSTALL_FOLDER}"
          info "${lenghty_info_message}"
          tar -xzf "${OFFLINE_BUNDLE_PATH}" -C "${OFFLINE_INSTALL_FOLDER}"
          info "Files used for the offline installer have been extracted to ${OFFLINE_INSTALL_FOLDER}"
          if (( INSTALL_OFFLINE_PREREQS != 0 )); then
            install_offline_prereqs
          fi
        fi
        [[ "${SKIP_PRE_REQS}" == "0" ]] && check_prerequisites "${prerequisites_args[@]}" -j "${NODE_TYPE}"
      fi
    else
      verify_install_type "online"
      load_package_utils "${MODULES_DIR}" "${INSTALL_TYPE}"
      [[ "${SKIP_PACKAGES_INSTALL}" == "0" ]] && install_package_dependencies "${MODULES_DIR}"
      [[ "${SKIP_PRE_REQS}" == "0" ]] && check_prerequisites "${prerequisites_args[@]}" -j "${NODE_TYPE}"
      OFFLINE_INSTALL_FOLDER=""
      info "Running the online install"
    fi
  fi

  if (( SETUPRESTORE + SETUPBACKUP == 0 )); then
    check_and_warn_fqdn_update
  fi

  info "###############################################"
  info ""
  info "   Running UiPath Automation Suite Installer"
  info "             Version ${VERSION}"
  info ""
  info "###############################################"

  if [[ -n ${OPTIONAL_OFFLINE_BUNDLE_PATH} ]]; then
    info "Importing optional bundle at ${OPTIONAL_OFFLINE_BUNDLE_PATH}"
    OFFLINE_BUNDLE_FOLDER="${OFFLINE_TMP_FOLDER}/sf_on_demand"
    mkdir -p "${OFFLINE_BUNDLE_FOLDER}"
    info "Extracting ${OPTIONAL_OFFLINE_BUNDLE_PATH} to ${OFFLINE_BUNDLE_FOLDER}"
    tar -xzf "${OPTIONAL_OFFLINE_BUNDLE_PATH}" -C "${OFFLINE_BUNDLE_FOLDER}"
    info "Files used for the on demand bundle have been extracted to ${OFFLINE_BUNDLE_FOLDER}"
    set_offline_variables
    import_on_demand_artifacts
    exit 0
  fi

  export INSTALL_TYPE OFFLINE_INSTALL_FOLDER

  if (( SETUPRESTORE + SETUPBACKUP == 0 )); then
    read_certificate_files
  fi

  [[ -n ${NODE_ROLE} ]] && export NODE_ROLE=${NODE_ROLE}
  [[ -n ${NODE_TYPE} ]] && export NODE_TYPE=${NODE_TYPE}

  [[ -n "${RESTORE_VOLUME}" ]] && export RESTORE_VOLUME=${RESTORE_VOLUME}
  [[ -n "${DISABLE_BACKUP}" ]] && export DISABLE_BACKUP=${DISABLE_BACKUP}
  [[ "${DO_UPGRADE}" == "0" && "${INSTALLINFRA}" == "1" ]] && install_infra
  [[ -n "${ONE_TIME_MANUAL_BACKUP}" ]] && export ONE_TIME_MANUAL_BACKUP=${ONE_TIME_MANUAL_BACKUP}
  # get secrets access is not available in agent nodes kubeconfig so disabling this for agent node
  if [[ ${INSTALL_TYPE} == "offline" && "${NODE_ROLE}" != "agent" ]]; then
    set_offline_variables
  fi
  [[ -n ${OFFLINE_BUNDLE_PATH} ]] && [[ "${INSTALLFABRIC}" == "1" ]] && import_offline_sf_artifacts
  [[ "${INSTALLINFRA}|${INSTALLFABRIC}|${INSTALLSERVICES}|${INSTALLSTANDALONEAICENTER}" == "1|1|1|0" && "${INSTALL_MULTINODE}" == "true" ]] && allow_installation_of_services="$(validate_cluster_requirement)"
  [[ "${allow_installation_of_services}" == "false" ]] && warn "Skipping fabric and service installation because minimum cluster resource requirements are not met, Please check above warnings!"
  [[ "${DO_UPGRADE}" == "0" && "${INSTALLFABRIC}" == "1" && "${allow_installation_of_services}" == "true" ]] && install_fabric
  [[ "${DO_UPGRADE}" == "0" && "${INSTALLSERVICES}" == "1" && "${INSTALLSTANDALONEAICENTER}" == "0" && "${allow_installation_of_services}" == "true" ]] && install_services
  [[ "${DO_UPGRADE}" == "0" && "${INSTALLSERVICES}" == "1" && "${INSTALLSTANDALONEAICENTER}" == "1" ]] && install_aicenter_services
  [[ "${SETUPBACKUP}" == "1" ]] && setup_backup
  [[ "${SETUPRESTORE}" == "1" ]] && setup_restore
  [[ "${DO_UPGRADE}" == "1" ]] && do_upgrade

  echo "{}" >"${OUTFILE}"
  if [[ -f "${SCRIPT_DIR}/infra_output.json" ]]; then
    jq -s '.[0] * .[1]' "${OUTFILE}" "${SCRIPT_DIR}/infra_output.json" >"tempout.json"
    cat "tempout.json" >"${OUTFILE}"
    rm "tempout.json"
  fi
  if [[ -f "${SCRIPT_DIR}/fabric_output.json" ]]; then
    jq -s '.[0] * .[1]' "${OUTFILE}" "${SCRIPT_DIR}/fabric_output.json" >"tempout.json"
    cat "tempout.json" >"${OUTFILE}"
    rm "tempout.json"
  fi
  if [[ -f "${SCRIPT_DIR}/services_output.json" ]]; then
    jq -s '.[0] * .[1]' "${OUTFILE}" "${SCRIPT_DIR}/services_output.json" >"tempout.json"
    cat "tempout.json" >"${OUTFILE}"
    rm "tempout.json"
  fi
  if [[ -f "${SCRIPT_DIR}/upgrade_output.json" ]]; then
    jq -s '.[0] * .[1]' "${OUTFILE}" "${SCRIPT_DIR}/upgrade_output.json" >"tempout.json"
    cat "tempout.json" >"${OUTFILE}"
    rm "tempout.json"
  fi

  if (( DEBUG != 1)); then
    rm -rf "${OFFLINE_INSTALL_FOLDER}"
  fi
}

parse_long_args "$@"
main
