#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  echo "Could not determine script path"
  exit 1
fi

MODULES_DIR=$(cd "${SCRIPT_DIR}/Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  echo "Could not determine Modules path"
  exit 1
fi

export MODULES_DIR

if [[ -f "${MODULES_DIR}/utils.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${MODULES_DIR}/utils.sh"
else
  echo "Common utils not found at ${MODULES_DIR}/utils.sh"
  exit 1
fi

if [[ -f "${SCRIPT_DIR}/defaults.json" ]] ; then
  export DEFAULTSFILE="${SCRIPT_DIR}/defaults.json"
else
  error "Default json file not found at ${SCRIPT_DIR}/defaults.json"
fi

export MERGEDINPUTFILE="${SCRIPT_DIR}/merged_input.json"


CHECKS_DIR=$(cd "${SCRIPT_DIR}/Prerequisite_Checks" && pwd)
if [[ -z "${CHECKS_DIR}" ]]; then
  error "Could not determine Prerequisite_Checks path"
fi

export PATH="$PATH:/usr/local/sbin:/sbin:/bin:/usr/sbin:/usr/bin:/root/bin:/usr/local/bin"
export KUBECONFIG="${KUBECONFIG_PATH:-/etc/rancher/rke2/rke2.yaml}"

ENABLED_SEVICES=""
export MITIGATION_DELIMITER="\n    "
UNEXPECTED_FAILURE="false"
PREREQUISITE_CHECK_FAILED="false"
INFILE=""
NODES=""
NODE_TYPE=""
INSTALL_TYPE=""
PRIMARY_SERVER="false"
INSTALL_MISSING_PACKAGES="false"
CONFIGURE_SYSTEM_SERVICES="false"
INSTALLSTANDALONEAICENTER="0"


function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "validateUiPathASReadiness.sh Runs pre-install validation checks"
  echo 
  echo "Arguments:"
  echo " -i|--input                           [Required] Specify cluster config json file"
  echo " -m|--machines                        Comma seperated IP address / hostnames of all the machines"
  echo " -n|--node-type                       Possible values [server|agent|taskmining|gpu]. Defaults to server"
  echo " --install-type                       Possible values [online|offline]. Defaults to online"
  echo " -p|--primary-server                  First machine, and only applicable in case of offline installation"
  echo " --install-missing-packages           Install missing tools and system libraries"
  echo " --configure-system-services          Configure the system services"
  echo " --install-standalone-aicenter        Run checks for standalone aicenter service only"
  echo " -h|--help                            Display help"
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  local iter
  while (("$#")); do
    # export COMPONENT=$1
    case "$1" in
    -d | --debug)
      shift
      set -x
    ;;

    -h | --help)
      display_usage
      shift
      exit 0
    ;;

    -i | --input)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        local ABSOLUTEPATH
        ABSOLUTEPATH=$(readlink -f "$2") || error_with_code "10032" "$2"
        INFILE=$ABSOLUTEPATH
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;    

    -m | --machines)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        NODES="$2"
        shift 2
      else
        error "Value for $1 is missing"
        exit 1
      fi
      ;;

    -n | --node-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        NODE_TYPE="$2"
        shift 2
      else
        error "Value for $1 is missing"
        exit 1
      fi
      ;; 

    --install-type)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        INSTALL_TYPE="$2"
        shift 2
      else
        error "Value for $1 is missing"
        exit 1
      fi
      ;;

    -p | --primary-server)
      PRIMARY_SERVER="true"
      shift
    ;;  

    --install-missing-packages)
      INSTALL_MISSING_PACKAGES="true"
      shift
    ;;
    --configure-system-services)
      CONFIGURE_SYSTEM_SERVICES="true"
      shift
    ;;  

    --install-standalone-aicenter)
      shift
      INSTALLSTANDALONEAICENTER="1"
    ;;

    -*|*) # unsupported flags
      display_usage
      echo "Error: Unsupported flag $1"
      exit 1
      ;;
    esac
  done
  # display_usage
  # check if provided input file was provided and exists
  if [[ -z "${INFILE}" ]]; then
    error_with_code "10013"
  fi
  if [[ ! -f "${INFILE}" ]]; then
    error_with_code "10014" "${INFILE}"
  fi

  ##Validate Ips
  local invalid_ips=()
  if [[ -n "${NODES}" ]]; then
    NODES=$(echo "${NODES}" | tr ',' '\n')
    while IFS=$'\n' read -r iter
    do
      iter=$(printf '%s' "$iter" | tr -d '[:space:]')
      [[ -z "${iter}" ]] && continue
      iter=$(echo "${iter}" | tr -d '[:space:]')
      if ! ipcalc -c "${iter}" ; then
        invalid_ips+=("Invalid IP: ${iter}")
      fi
    done <<< "${NODES}"

    if [[ ${#invalid_ips[@]} != 0 ]] ; then
      failure_output "Invalid IP(s) provided in input" "${invalid_ips[@]}"
    fi
  fi

  # Validate node type
  if [[ -z "${NODE_TYPE}" ]] ; then
    NODE_TYPE="server"
  else
    case "${NODE_TYPE}" in
      server)
      ;;
      agent)
      ;;
      task-mining)
      ;;
      gpu)
      ;;
      *)
      error "Unsupported node type: ${NODE_TYPE} provided"
      ;;
    esac
  fi

  # Validate install type
  if [[ -z "${INSTALL_TYPE}" ]] ; then
    INSTALL_TYPE="online"
  else
    case "$INSTALL_TYPE" in
      online)
      ;;
      offline)
      ;;
      *)
      error "Unsupported install type: ${INSTALL_TYPE} provided"
      ;;
    esac
  fi

  if [[ -z "${PRIMARY_SERVER}" ]] ; then
    PRIMARY_SERVER="false"
  fi

  if [[ -z "${INSTALL_MISSING_PACKAGES}" ]] ; then
    INSTALL_MISSING_PACKAGES="false"
  fi

  if [[ -z "${CONFIGURE_SYSTEM_SERVICES}" ]] ; then
    CONFIGURE_SYSTEM_SERVICES="false"
  fi  

  if [[ -z "${INSTALLSTANDALONEAICENTER}" ]] ; then
    INSTALLSTANDALONEAICENTER="0"
  fi
}

function is_failure() {
  local exitCode="$1"
  case "$exitCode" in
    0)
    ;;
    51)
      PREREQUISITE_CHECK_FAILED="true"
    ;;
    *)
      UNEXPECTED_FAILURE="true"
      error_without_exit "Unexpected error occurred while executing connectivity check"
    ;;
  esac
}


function run_all() {
  local exitCode
  ## Run connectivity check
  "${CHECKS_DIR}/connectivity/main.sh" "$@" --enabled-services "${ENABLED_SEVICES}"
  is_failure "$?"

  ## Run other check

  ## Keep this at last
  if [[ $PREREQUISITE_CHECK_FAILED == "true" ]] ; then
    echo -e "\u001b[31mOne or more installation readiness checks are failed\e[0m"
  fi

  if [[ $UNEXPECTED_FAILURE == "false" ]] && [[ $PREREQUISITE_CHECK_FAILED == "false" ]] ; then
    echo -e "\u001b[32mInstallation readiness checks are passed. You can now proceed to install Automation Suite\e[0m"
    exitCode=0
  else
    exitCode=1
  fi
  exit $exitCode
}


function main() {
  parse_long_args "$@"
  diff "$INFILE" "${MERGEDINPUTFILE}" >> /dev/null 2>&1 || merge_default_values
  export INFILE
  ENABLED_SEVICES=$(get_enabled_services)
  run_all "$@"
}

main "$@"